Package: tidypredict
Title: Run Predictions Inside the Database
Version: 1.1.1
Authors@R: c(
    person("Emil", "Hvitfeldt", , "emil.hvitfeldt@posit.co", role = c("aut", "cre")),
    person("Edgar", "Ruiz", , "edgar@posit.co", role = "aut"),
    person("Max", "Kuhn", , "max@posit.co", role = "aut")
  )
Description: It parses a fitted 'R' model object, and returns a formula in
    'Tidy Eval' code that calculates the predictions.  It works with
    several databases back-ends because it leverages 'dplyr' and 'dbplyr'
    for the final 'SQL' translation of the algorithm. It currently
    supports lm(), glm(), randomForest(), ranger(), rpart(), earth(),
    xgb.Booster.complete(), lgb.Booster(), catboost.Model(), cubist(), and
    ctree() models.
License: MIT + file LICENSE
URL: https://tidypredict.tidymodels.org,
        https://github.com/tidymodels/tidypredict
BugReports: https://github.com/tidymodels/tidypredict/issues
Depends: R (>= 3.6)
Imports: cli, dplyr (>= 0.7), generics, jsonlite, knitr, purrr, rlang
        (>= 1.1.1), tibble, tidyr
Suggests: bonsai, covr, Cubist (>= 0.5.1), data.table, DBI, dbplyr,
        earth (>= 5.1.2), glmnet, lightgbm, methods, mlbench,
        modeldata, nycflights13, parsnip, partykit, randomForest,
        ranger (>= 0.14.1), RhpcBLASctl, rpart (>= 4.1.0), rmarkdown,
        RSQLite, survival, testthat (>= 3.2.0), withr, xgboost, yaml
VignetteBuilder: knitr
Config/Needs/website: tidyverse/tidytemplate
Config/testthat/edition: 3
Encoding: UTF-8
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-08-23 23:07:51 UTC; emilhvitfeldt
Author: Emil Hvitfeldt [aut, cre],
  Edgar Ruiz [aut],
  Max Kuhn [aut]
Maintainer: Emil Hvitfeldt <emil.hvitfeldt@posit.co>
Repository: CRAN
Date/Publication: 2026-08-24 05:20:02 UTC

# Generated by roxygen2: do not edit by hand

S3method(acceptable_formula,glm)
S3method(acceptable_formula,lm)
S3method(as_parsed_model,list)
S3method(knit_print,tidypredict_test)
S3method(parse_model,catboost.Model)
S3method(parse_model,cubist)
S3method(parse_model,earth)
S3method(parse_model,glm)
S3method(parse_model,glmnet)
S3method(parse_model,lgb.Booster)
S3method(parse_model,lm)
S3method(parse_model,model_fit)
S3method(parse_model,party)
S3method(parse_model,randomForest)
S3method(parse_model,ranger)
S3method(parse_model,rpart)
S3method(parse_model,xgb.Booster)
S3method(print,tidypredict_test)
S3method(tidy,pm_regression)
S3method(tidypredict_fit,"_xgb.Booster")
S3method(tidypredict_fit,catboost.Model)
S3method(tidypredict_fit,cubist)
S3method(tidypredict_fit,earth)
S3method(tidypredict_fit,glm)
S3method(tidypredict_fit,glmnet)
S3method(tidypredict_fit,lgb.Booster)
S3method(tidypredict_fit,lm)
S3method(tidypredict_fit,model_fit)
S3method(tidypredict_fit,party)
S3method(tidypredict_fit,pm_catboost)
S3method(tidypredict_fit,pm_lgb)
S3method(tidypredict_fit,pm_regression)
S3method(tidypredict_fit,pm_tree)
S3method(tidypredict_fit,pm_xgb)
S3method(tidypredict_fit,randomForest)
S3method(tidypredict_fit,ranger)
S3method(tidypredict_fit,rpart)
S3method(tidypredict_fit,xgb.Booster)
S3method(tidypredict_interval,data.frame)
S3method(tidypredict_interval,glm)
S3method(tidypredict_interval,list)
S3method(tidypredict_interval,lm)
S3method(tidypredict_test,"_xgb.Booster")
S3method(tidypredict_test,catboost.Model)
S3method(tidypredict_test,default)
S3method(tidypredict_test,glmnet)
S3method(tidypredict_test,lgb.Booster)
S3method(tidypredict_test,model_fit)
S3method(tidypredict_test,party)
S3method(tidypredict_test,rpart)
S3method(tidypredict_test,xgb.Booster)
export(.build_case_when_tree)
export(.build_linear_pred)
export(.build_nested_case_when_tree)
export(.extract_catboost_trees)
export(.extract_earth_multiclass)
export(.extract_glmnet_multiclass)
export(.extract_lgb_trees)
export(.extract_partykit_classprob)
export(.extract_ranger_classprob)
export(.extract_ranger_trees)
export(.extract_rf_classprob)
export(.extract_rf_trees)
export(.extract_rpart_classprob)
export(.extract_xgb_trees)
export(.partykit_tree_info_full)
export(.rpart_tree_info_full)
export(acceptable_formula)
export(as_parsed_model)
export(parse_model)
export(set_catboost_categories)
export(tidy)
export(tidypredict_fit)
export(tidypredict_interval)
export(tidypredict_sql)
export(tidypredict_sql_interval)
export(tidypredict_test)
export(tidypredict_to_column)
import(rlang)
importFrom(dplyr,"%>%")
importFrom(dplyr,case_when)
importFrom(dplyr,mutate)
importFrom(generics,tidy)
importFrom(knitr,knit_print)
importFrom(purrr,imap)
importFrom(purrr,map)
importFrom(purrr,map2)
importFrom(purrr,map_chr)
importFrom(purrr,map_dbl)
importFrom(purrr,map_dfr)
importFrom(purrr,map_int)
importFrom(purrr,map_lgl)
importFrom(purrr,reduce)
importFrom(purrr,transpose)
importFrom(stats,binomial)
importFrom(stats,predict)
importFrom(stats,qt)
importFrom(stats,setNames)
importFrom(tibble,tibble)
importFrom(utils,head)

# tidypredict <a href="https://tidypredict.tidymodels.org"><img src="man/figures/logo.png" align="right" height="138" alt="tidypredict website" /></a>

[![R-CMD-check](https://github.com/tidymodels/tidypredict/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/tidymodels/tidypredict/actions/workflows/R-CMD-check.yaml)
[![CRAN
status](https://www.r-pkg.org/badges/version/tidypredict)](https://CRAN.R-project.org/package=tidypredict)
[![Downloads](https://cranlogs.r-pkg.org/badges/tidypredict)](https://CRAN.R-project.org/package=tidypredict)
[![Codecov test
coverage](https://codecov.io/gh/tidymodels/tidypredict/graph/badge.svg)](https://app.codecov.io/gh/tidymodels/tidypredict)
[![lifecycle](https://img.shields.io/badge/lifecycle-stable-brightgreen.svg)](https://lifecycle.r-lib.org/articles/stages.html)

The main goal of `tidypredict` is to enable running predictions inside
databases. It reads the model, extracts the components needed to
calculate the prediction, and then creates an R formula that can be
translated into SQL. In other words, it is able to parse a model such as
this one:

``` r
model <- lm(mpg ~ wt + cyl, data = mtcars)
```

`tidypredict` can return a SQL statement that is ready to run inside the
database. Because it uses `dplyr`’s database interface, it works with
several databases back-ends, such as MS SQL:

``` r
tidypredict_sql(model, dbplyr::simulate_mssql())
```

    ## <SQL> (39.6862614802529 + (`wt` * -3.19097213898374)) + (`cyl` * -1.5077949682598)

## Installation

Install `tidypredict` from CRAN using:

``` r
install.packages("tidypredict")
```

Or install the **development version** using `devtools` as follows:

``` r
install.packages("remotes")
remotes::install_github("tidymodels/tidypredict")
```

## Functions

`tidypredict` has only a few functions, and it is not expected that
number to grow much. The main focus at this time is to add more models
to support.

| Function | Description |
|----|----|
| `tidypredict_fit()` | Returns an R formula that calculates the prediction |
| `tidypredict_sql()` | Returns a SQL query based on the formula from `tidypredict_fit()` |
| `tidypredict_to_column()` | Adds a new column using the formula from `tidypredict_fit()` |
| `tidypredict_test()` | Tests `tidypredict` predictions against the model’s native `predict()` function |
| `tidypredict_interval()` | Same as `tidypredict_fit()` but for intervals (only works with `lm` and `glm`) |
| `tidypredict_sql_interval()` | Same as `tidypredict_sql()` but for intervals (only works with `lm` and `glm`) |
| `parse_model()` | Creates a list spec based on the R model |
| `as_parsed_model()` | Prepares an object to be recognized as a parsed model |

## How it works

<img src="man/figures/howitworks.png">

Instead of translating directly to a SQL statement, `tidypredict`
creates an R formula. That formula can then be used inside `dplyr`. The
overall workflow would be as illustrated in the image above, and
described here:

1.  Fit the model using a base R model, or one from the packages listed
    in [Supported Models](#supported-models)
2.  `tidypredict` reads model, and creates a list object with the
    necessary components to run predictions
3.  `tidypredict` builds an R formula based on the list object
4.  `dplyr` evaluates the formula created by `tidypredict`
5.  `dplyr` translates the formula into a SQL statement, or any other
    interfaces.
6.  The database executes the SQL statement(s) created by `dplyr`

### Parsed model spec

`tidypredict` writes and reads a spec based on a model. Instead of
simply writing the R formula directly, splitting the spec from the
formula adds the following capabilities:

1.  No more saving models as `.rds` - Specifically for cases when the
    model needs to be used for predictions in a Shiny app.
2.  Beyond R models - Technically, anything that can write a proper
    spec, can be read into `tidypredict`. It also means, that the parsed
    model spec can become a good alternative to using *PMML.*

## Supported models

The following models are supported by `tidypredict`:

- Linear Regression - `lm()`
- Generalized Linear model - `glm()`
- Elastic net models - `glmnet::glmnet()`
- Random Forest models - `randomForest::randomForest()`
- Random Forest models, via `ranger` - `ranger::ranger()`
- MARS models - `earth::earth()`
- Decision tree models - `rpart::rpart()`
- XGBoost models - `xgboost::xgb.Booster`
- LightGBM models - `lightgbm::lgb.Booster`
- CatBoost models - `catboost::catboost.Model`
- Cubist models - `Cubist::cubist()`
- Tree models, via `partykit` - `partykit::ctree()`

### `parsnip`

`tidypredict` supports models fitted via the `parsnip` interface. The
ones confirmed currently work in `tidypredict` are:

- `lm()` - `parsnip`: `linear_reg()` with *“lm”* as the engine.
- `glmnet::glmnet()` - `parsnip`: `linear_reg()` or `logistic_reg()`
  with *“glmnet”* as the engine.
- `randomForest::randomForest()` - `parsnip`: `rand_forest()` with
  *“randomForest”* as the engine.
- `ranger::ranger()` - `parsnip`: `rand_forest()` with *“ranger”* as the
  engine.
- `earth::earth()` - `parsnip`: `mars()` with *“earth”* as the engine.
- `rpart::rpart()` - `parsnip`: `decision_tree()` with *“rpart”* as the
  engine.
- `xgboost::xgb.Booster` - `parsnip`: `boost_tree()` with *“xgboost”* as
  the engine.
- `lightgbm::lgb.Booster` - `parsnip`: `boost_tree()` with *“lightgbm”*
  as the engine (via `bonsai`).
- `catboost::catboost.Model` - `parsnip`: `boost_tree()` with
  *“catboost”* as the engine (via `bonsai`).

### `broom`

The `tidy()` function from broom works with linear models parsed via
`tidypredict`

``` r
pm <- parse_model(lm(wt ~ ., mtcars))
tidy(pm)
```

    ## # A tibble: 11 × 2
    ##    term        estimate
    ##    <chr>          <dbl>
    ##  1 (Intercept) -0.231  
    ##  2 mpg         -0.0417 
    ##  3 cyl         -0.0573 
    ##  4 disp         0.00669
    ##  5 hp          -0.00323
    ##  6 drat        -0.0901 
    ##  7 qsec         0.200  
    ##  8 vs          -0.0664 
    ##  9 am           0.0184 
    ## 10 gear        -0.0935 
    ## 11 carb         0.249

## Contributing

This project is released with a [Contributor Code of
Conduct](https://contributor-covenant.org/version/2/0/CODE_OF_CONDUCT.html).
By contributing to this project, you agree to abide by its terms.

- For questions and discussions about tidymodels packages, modeling, and
  machine learning, please [post on Posit
  Community](https://forum.posit.co/new-topic?category_id=15&tags=tidymodels,question).

- If you think you have encountered a bug, please [submit an
  issue](https://github.com/tidymodels/tidypredict/issues).

- Either way, learn how to create and share a
  [reprex](https://reprex.tidyverse.org/articles/articles/learn-reprex.html)
  (a minimal, reproducible example), to clearly communicate about your
  code.

- Check out further details on [contributing guidelines for tidymodels
  packages](https://www.tidymodels.org/contribute/) and [how to get
  help](https://www.tidymodels.org/help/).

## ----setup, include = FALSE---------------------------------------------------
if (requireNamespace("lightgbm", quietly = TRUE)) {
  library(tidypredict)
  library(lightgbm)
  library(dplyr)
  eval_code <- TRUE
} else {
  eval_code <- FALSE
}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = eval_code
)

## -----------------------------------------------------------------------------
library(lightgbm)

# Prepare data
X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
y <- mtcars$hp

dtrain <- lgb.Dataset(X, label = y, colnames = c("mpg", "cyl", "disp"))

model <- lgb.train(
  params = list(
    num_leaves = 4L,
    learning_rate = 0.5,
    objective = "regression",
    min_data_in_leaf = 1L
  ),
  data = dtrain,
  nrounds = 10L,
  verbose = -1L
)

## -----------------------------------------------------------------------------
tidypredict_fit(model)

## -----------------------------------------------------------------------------
library(dplyr)

mtcars %>%
  tidypredict_to_column(model) %>%
  glimpse()

## -----------------------------------------------------------------------------
tidypredict_test(model, xg_df = X)

## -----------------------------------------------------------------------------
X_bin <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
y_bin <- mtcars$am

dtrain_bin <- lgb.Dataset(X_bin, label = y_bin, colnames = c("mpg", "cyl", "disp"))

model_bin <- lgb.train(
  params = list(
    num_leaves = 4L,
    learning_rate = 0.5,
    objective = "binary",
    min_data_in_leaf = 1L
  ),
  data = dtrain_bin,
  nrounds = 10L,
  verbose = -1L
)

tidypredict_test(model_bin, xg_df = X_bin)

## -----------------------------------------------------------------------------
X_iris <- data.matrix(iris[, 1:4])
colnames(X_iris) <- c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width")
y_iris <- as.integer(iris$Species) - 1L

dtrain_iris <- lgb.Dataset(X_iris, label = y_iris, colnames = colnames(X_iris))

model_multi <- lgb.train(
  params = list(
    num_leaves = 4L,
    learning_rate = 0.5,
    objective = "multiclass",
    num_class = 3L,
    min_data_in_leaf = 1L
  ),
  data = dtrain_iris,
  nrounds = 5L,
  verbose = -1L
)

fit_formulas <- tidypredict_fit(model_multi)
names(fit_formulas)

## -----------------------------------------------------------------------------
iris %>%
  mutate(
    prob_setosa = !!fit_formulas$class_0,
    prob_versicolor = !!fit_formulas$class_1,
    prob_virginica = !!fit_formulas$class_2
  ) %>%
  select(Species, starts_with("prob_")) %>%
  head()

## -----------------------------------------------------------------------------
set.seed(123)
n <- 200
cat_data <- data.frame(
  cat_feat = sample(0:3, n, replace = TRUE),
  y = NA
)
cat_data$y <- ifelse(cat_data$cat_feat %in% c(0, 1), 10, -10) + rnorm(n, sd = 2)

X_cat <- matrix(cat_data$cat_feat, ncol = 1)
colnames(X_cat) <- "cat_feat"

dtrain_cat <- lgb.Dataset(
  X_cat,
  label = cat_data$y,
  categorical_feature = "cat_feat"
)

model_cat <- lgb.train(
  params = list(
    num_leaves = 4L,
    learning_rate = 1.0,
    objective = "regression",
    min_data_in_leaf = 1L
  ),
  data = dtrain_cat,
  nrounds = 2L,
  verbose = -1L
)

tidypredict_fit(model_cat)

## ----eval = requireNamespace("parsnip", quietly = TRUE) && requireNamespace("bonsai", quietly = TRUE)----
library(parsnip)
library(bonsai)

p_model <- boost_tree(
  trees = 10,
  tree_depth = 3,
  min_n = 1
) %>%
  set_engine("lightgbm") %>%
  set_mode("regression") %>%
  fit(hp ~ mpg + cyl + disp, data = mtcars)

# Extract the underlying lgb.Booster
lgb_model <- p_model$fit

tidypredict_test(lgb_model, xg_df = X)

## -----------------------------------------------------------------------------
pm <- parse_model(model)
str(pm, 2)

## -----------------------------------------------------------------------------
str(pm$trees[1])

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

library(dplyr)
library(tidypredict)
library(rpart)
set.seed(100)

## ----setup, include = FALSE---------------------------------------------------
if (FALSE) {
  library(tidypredict)
  library(dplyr)
  eval_code <- TRUE
} else {
  eval_code <- FALSE
}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = eval_code
)

#' Returns a SQL query with formula to calculate fitted values
#'
#' @param model An R model or a list with a parsed model
#' @param con Database connection object. It is used to select
#' the correct SQL translation syntax.
#'
#' @examples
#' library(dbplyr)
#'
#' model <- lm(mpg ~ wt + am + cyl, data = mtcars)
#' tidypredict_sql(model, simulate_dbi())
#' @keywords internal
#' @export
tidypredict_sql <- function(model, con) {
  f <- tidypredict_fit(model)
  if (inherits(f, "call")) {
    dbplyr::translate_sql(!!f, con = con)
  } else {
    map(f, ~ dbplyr::translate_sql(!!.x, con = con))
  }
}

#' Returns a SQL query with formula to calculate predicted interval
#'
#'
#' @param model An R model or a tibble with a parsed model
#' @param con  Database connection object. It is used to select
#' the correct SQL translation syntax.
#' @param interval The prediction interval, defaults to 0.95
#'
#' @examples
#' library(dbplyr)
#'
#' model <- lm(mpg ~ wt + am + cyl, data = mtcars)
#' tidypredict_sql_interval(model, simulate_dbi())
#' @keywords internal
#' @export
tidypredict_sql_interval <- function(model, con, interval = 0.95) {
  f <- tidypredict_interval(model, interval)
  if (inherits(f, "call")) {
    dbplyr::translate_sql(!!f, con = con)
  } else {
    map(f, ~ dbplyr::translate_sql(!!.x, con = con))
  }
}

test_that("Correct SQL query is returned", {
  expect_s3_class(
    tidypredict_sql(lm(mpg ~ wt, data = mtcars), dbplyr::simulate_dbi()),
    "sql"
  )
  expect_s3_class(
    tidypredict_sql_interval(
      lm(mpg ~ wt, data = mtcars),
      dbplyr::simulate_dbi()
    ),
    "sql"
  )
})

# Predict ---------------------------------------

#' @export
tidypredict_fit.glm <- function(model) {
  parsedmodel <- parse_model(model)
  build_fit_formula(parsedmodel)
}

# Parse model --------------------------------------

#' @export
parse_model.glm <- function(model) parse_model_lm(model)

# Intervals -----------------------------------------------

#' @export
tidypredict_interval.glm <- function(model, interval = 0.95) {
  parsedmodel <- parse_model(model)
  te_interval_glm(parsedmodel, interval)
}

te_interval_glm <- function(parsedmodel, interval = 0.95) {
  intervals <- te_interval_lm(parsedmodel, interval)
  family <- parsedmodel$general$family
  link <- parsedmodel$general$link
  assigned <- 0
  if (family == "gaussian" && link == "identity") {
    assigned <- 1
  }
  if (assigned == 0) {
    cli::cli_abort(
      "Combination of family and link are not supported for prediction intervals."
    )
  }
  intervals
}

# Predict ---------------------------------------

#' @export
tidypredict_fit.glmnet <- function(model) {
  if (inherits(model, "multnet")) {
    cli::cli_abort(
      c(
        "Multinomial glmnet models are not supported.",
        "i" = "Models fit with {.code family = \"multinomial\"} have multiple
        outcome columns which is not supported."
      )
    )
  }
  if (inherits(model, "mrelnet")) {
    cli::cli_abort(
      c(
        "Multivariate gaussian glmnet models are not supported.",
        "i" = "Models fit with {.code family = \"mgaussian\"} have multiple
        outcome columns which is not supported."
      )
    )
  }
  parsedmodel <- parse_model(model)
  build_fit_formula(parsedmodel)
}

# Parse model --------------------------------------

#' @export
parse_model.glmnet <- function(model) {
  parse_model_glmnet(model)
}

parse_model_glmnet <- function(model, call = rlang::caller_env()) {
  if (length(model$lambda) != 1) {
    cli::cli_abort(
      "{.fn tidypredict_fit} requires that there are only 1 penalty selected,
      {length(model$lambda)} were provided.",
      call = call
    )
  }
  if (inherits(model$beta, "dgCMatrix")) {
    model$beta <- setNames(as.numeric(model$beta), rownames(model$beta))
  }
  coefs <- c("(Intercept)" = unname(model$a0), model$beta)

  names <- names(coefs)
  values <- as.vector(coefs)

  terms <- map2(values, names, function(value, name) {
    if (value == 0) {
      return(NULL)
    }
    list(
      label = name,
      coef = value,
      is_intercept = as.integer(name == "(Intercept)"),
      fields = list(list(type = "ordinary", col = name))
    )
  })

  terms <- purrr::discard(terms, is.null)

  pm <- list()
  pm$general$model <- class(model)[[2]]
  pm$general$version <- 1
  pm$general$type <- "regression"
  pm$general$is_glm <- 1
  pm$terms <- terms

  if (inherits(model, "elnet")) {
    pm$general$family <- "gaussian"
    pm$general$link <- "identity"
  } else if (inherits(model, "lognet")) {
    pm$general$family <- "binomial"
    pm$general$link <- "logit"
  } else if (inherits(model, "fishnet")) {
    pm$general$family <- "poisson"
    pm$general$link <- "log"
  } else if (inherits(model, "coxnet")) {
    pm$general$family <- "cox"
    pm$general$link <- "identity"
    pm$general$is_glm <- 0
  } else if (inherits(model, "glmnetfit")) {
    pm$general$family <- model$family$family
    pm$general$link <- model$family$link
  } else {
    # nocov start
    cli::cli_abort(
      "Model fit with this {.arg family} is not supported.",
      .internal = TRUE
    )
  } # nocov end

  as_parsed_model(pm)
}

# For {orbital}
#' Build linear predictor expression from coefficient names and values
#'
#' Shared helper for building linear predictor strings from coefficients.
#' Used by orbital package for glmnet models.
#'
#' @param coef_names Character vector of coefficient names (including "(Intercept)")
#' @param coef_values Numeric vector of coefficient values
#' @keywords internal
#' @export
.build_linear_pred <- function(coef_names, coef_values) {
  terms <- character(0)
  for (i in seq_along(coef_names)) {
    if (coef_values[i] == 0) {
      next
    }

    if (coef_names[i] == "(Intercept)") {
      terms <- c(terms, as.character(coef_values[i]))
    } else {
      # Use backticks for variable names to handle special characters
      var_name <- paste0("`", coef_names[i], "`")
      terms <- c(terms, paste0("(", var_name, " * ", coef_values[i], ")"))
    }
  }

  if (length(terms) == 0) {
    return("0")
  }

  paste(terms, collapse = " + ")
}

#' Extract multiclass linear predictors for glmnet models
#'
#' For use in orbital package.
#' @param model A glmnet model object with class "multnet"
#' @param penalty The penalty value to use for coefficient extraction
#' @keywords internal
#' @export
.extract_glmnet_multiclass <- function(model, penalty = NULL) {
  if (!inherits(model, "multnet")) {
    cli::cli_abort(
      "{.arg model} must be {.cls multnet}, not {.obj_type_friendly {model}}."
    )
  }

  if (is.null(penalty)) {
    if (length(model$lambda) != 1) {
      cli::cli_abort(
        c(
          "glmnet model has multiple penalty values.",
          "i" = "Specify a single {.arg penalty} value."
        )
      )
    }
    penalty <- model$lambda
  }

  # Get coefficients for each class at the specified penalty
  coefs_list <- stats::coef(model, s = penalty)
  class_names <- names(coefs_list)

  # Build linear predictor expression for each class
  eqs <- lapply(coefs_list, function(coef_mat) {
    coef_names <- rownames(coef_mat)
    coef_values <- as.numeric(coef_mat)
    .build_linear_pred(coef_names, coef_values)
  })

  names(eqs) <- class_names
  eqs
}

test_that("returns the right output", {
  model <- glmnet::glmnet(mtcars[, -1], mtcars$mpg, lambda = 1)

  tf <- tidypredict_fit(model)
  pm <- parse_model(model)

  expect_type(tf, "language")

  expect_s3_class(pm, "list")
  expect_equal(length(pm), 2)
  expect_equal(pm$general$model, "glmnet")
  expect_equal(pm$general$version, 1)

  expect_snapshot(
    round_print(tf)
  )
})

test_that("Model can be saved and re-loaded", {
  model <- glmnet::glmnet(mtcars[, -1], mtcars$mpg, lambda = 1)

  pm <- parse_model(model)
  mp <- tempfile(fileext = ".yml")
  yaml::write_yaml(pm, mp)
  l <- yaml::read_yaml(mp)
  pm <- as_parsed_model(l)

  expect_identical(
    round_print(tidypredict_fit(model)),
    round_print(tidypredict_fit(pm))
  )
})

test_that("formulas produces correct predictions", {
  # gaussian
  expect_snapshot(
    tidypredict_test(
      glmnet::glmnet(mtcars[, -1], mtcars$mpg, family = "gaussian", lambda = 1),
      mtcars[, -1]
    )
  )

  # binomial
  expect_snapshot(
    tidypredict_test(
      glmnet::glmnet(mtcars[, -8], mtcars$vs, family = "binomial", lambda = 1),
      mtcars[, -1]
    )
  )

  # poisson
  expect_snapshot(
    tidypredict_test(
      glmnet::glmnet(mtcars[, -8], mtcars$vs, family = "poisson", lambda = 1),
      mtcars[, -1]
    )
  )
})

test_that("family function syntax works (#197)", {
  x <- as.matrix(mtcars[, -1])

  # gaussian()
  model <- glmnet::glmnet(x, mtcars$mpg, family = gaussian(), lambda = 0.5)
  expect_no_error(tidypredict_fit(model))

  # binomial()
  model <- glmnet::glmnet(x, mtcars$am, family = binomial(), lambda = 0.5)
  expect_no_error(tidypredict_fit(model))

  # poisson()
  model <- glmnet::glmnet(x, mtcars$carb, family = poisson(), lambda = 0.5)
  expect_no_error(tidypredict_fit(model))
})

test_that("family string syntax works (#197)", {
  x <- as.matrix(mtcars[, -1])

  # "gaussian"
  model <- glmnet::glmnet(x, mtcars$mpg, family = "gaussian", lambda = 0.5)
  expect_no_error(tidypredict_fit(model))

  # "binomial"
  model <- glmnet::glmnet(x, mtcars$am, family = "binomial", lambda = 0.5)
  expect_no_error(tidypredict_fit(model))

  # "poisson"
  model <- glmnet::glmnet(x, mtcars$carb, family = "poisson", lambda = 0.5)
  expect_no_error(tidypredict_fit(model))
})

test_that("errors if more than 1 penalty is selected", {
  model <- glmnet::glmnet(mtcars[, -1], mtcars$mpg)

  expect_snapshot(
    error = TRUE,
    tidypredict_fit(model)
  )

  model <- glmnet::glmnet(mtcars[, -1], mtcars$mpg, lambda = c(1, 5))

  expect_snapshot(
    error = TRUE,
    tidypredict_fit(model)
  )
})

test_that("glmnet are handeld neatly with parsnip", {
  spec <- parsnip::linear_reg(engine = "glmnet", penalty = 1)

  model <- parsnip::fit(spec, mpg ~ ., mtcars)

  tf <- tidypredict_fit(model)
  pm <- parse_model(model)

  expect_type(tf, "language")

  expect_s3_class(pm, "list")
  expect_equal(length(pm), 2)
  expect_equal(pm$general$model, "glmnet")
  expect_equal(pm$general$version, 1)

  expect_snapshot(
    round_print(tf)
  )
})

test_that("Gamma family works (#200)", {
  x <- as.matrix(mtcars[, -1])
  model <- glmnet::glmnet(x, mtcars$mpg, family = Gamma(), lambda = 0.5)

  fit <- tidypredict_fit(model)
  native <- unname(predict(model, x, type = "response")[, 1])
  tidy <- rlang::eval_tidy(fit, mtcars)

  expect_equal(tidy, native)
})

test_that("Cox family works (#201)", {
  skip_if_not_installed("survival")
  x <- as.matrix(mtcars[, -c(1, 8)])
  y <- survival::Surv(mtcars$mpg, mtcars$vs)
  model <- glmnet::glmnet(
    x,
    y,
    family = "cox",
    lambda = 0.1,
    cox.ties = "breslow"
  )

  fit <- tidypredict_fit(model)
  native <- unname(predict(model, x, type = "link")[, 1])
  tidy <- rlang::eval_tidy(fit, mtcars)

  expect_equal(tidy, native)
})

test_that("multinomial family errors with helpful message (#198)", {
  model <- glmnet::glmnet(
    as.matrix(iris[, 1:4]),
    iris$Species,
    family = "multinomial",
    lambda = 0.5
  )

  expect_snapshot(error = TRUE, tidypredict_fit(model))
})

test_that("mgaussian family errors with helpful message (#199)", {
  x <- as.matrix(mtcars[, -c(1, 4)])
  y <- cbind(mtcars$mpg, mtcars$hp)
  model <- glmnet::glmnet(x, y, family = "mgaussian", lambda = 0.5)

  expect_snapshot(error = TRUE, tidypredict_fit(model))
})

# Tests for .extract_glmnet_multiclass()

test_that(".extract_glmnet_multiclass returns correct structure", {
  model <- glmnet::glmnet(
    as.matrix(iris[, 1:4]),
    iris$Species,
    family = "multinomial",
    lambda = 0.5
  )

  result <- .extract_glmnet_multiclass(model)

  expect_type(result, "list")
  expect_length(result, 3)
  expect_named(result, levels(iris$Species))
  expect_type(result[[1]], "character")
})

test_that(".extract_glmnet_multiclass errors on non-multnet model", {
  model <- glmnet::glmnet(mtcars[, -1], mtcars$mpg, lambda = 1)

  expect_snapshot(error = TRUE, .extract_glmnet_multiclass(model))
})

test_that(".extract_glmnet_multiclass errors with multiple penalties", {
  model <- glmnet::glmnet(
    as.matrix(iris[, 1:4]),
    iris$Species,
    family = "multinomial"
  )

  expect_snapshot(error = TRUE, .extract_glmnet_multiclass(model))
})

test_that(".extract_glmnet_multiclass works with explicit penalty", {
  model <- glmnet::glmnet(
    as.matrix(iris[, 1:4]),
    iris$Species,
    family = "multinomial"
  )

  result <- .extract_glmnet_multiclass(model, penalty = 0.01)

  expect_type(result, "list")
  expect_length(result, 3)
})

test_that(".extract_glmnet_multiclass handles sparse coefficients", {
  # High penalty should zero out many coefficients

  model <- glmnet::glmnet(
    as.matrix(iris[, 1:4]),
    iris$Species,
    family = "multinomial",
    lambda = 10
  )

  result <- .extract_glmnet_multiclass(model)

  expect_type(result, "list")
  expect_length(result, 3)
})

test_that(".extract_glmnet_multiclass produces correct predictions", {
  model <- glmnet::glmnet(
    as.matrix(iris[, 1:4]),
    iris$Species,
    family = "multinomial",
    lambda = 0.01
  )

  eqs <- .extract_glmnet_multiclass(model)
  n_rows <- nrow(iris)

  # Evaluate each linear predictor, recycling scalars to full length
  logits <- sapply(eqs, function(eq) {
    val <- rlang::eval_tidy(rlang::parse_expr(eq), iris)
    if (length(val) == 1) rep(val, n_rows) else val
  })

  # Apply softmax
  exp_logits <- exp(logits)
  probs <- exp_logits / rowSums(exp_logits)

  # Compare to native predictions
  native <- predict(model, as.matrix(iris[, 1:4]), type = "response")[,, 1]

  expect_equal(unname(probs), unname(native), tolerance = 1e-10)
})

# Tests for .build_linear_pred()

test_that(".build_linear_pred handles intercept only", {
  result <- .build_linear_pred("(Intercept)", 5.5)

  expect_equal(result, "5.5")
})

test_that(".build_linear_pred handles single predictor", {
  result <- .build_linear_pred(c("(Intercept)", "x"), c(1.5, 2.0))

  expect_equal(result, "1.5 + (`x` * 2)")
})

test_that(".build_linear_pred handles multiple predictors", {
  result <- .build_linear_pred(
    c("(Intercept)", "x", "y"),
    c(1.0, 2.0, 3.0)
  )

  expect_equal(result, "1 + (`x` * 2) + (`y` * 3)")
})

test_that(".build_linear_pred skips zero coefficients", {
  result <- .build_linear_pred(
    c("(Intercept)", "x", "y", "z"),
    c(1.0, 0.0, 2.0, 0.0)
  )

  expect_identical(result, "1 + (`y` * 2)")
})

test_that(".build_linear_pred returns '0' when all coefficients are zero", {
  result <- .build_linear_pred(
    c("(Intercept)", "x", "y"),
    c(0, 0, 0)
  )

  expect_equal(result, "0")
})

test_that(".build_linear_pred handles negative coefficients", {
  result <- .build_linear_pred(
    c("(Intercept)", "x"),
    c(-1.5, -2.0)
  )

  expect_equal(result, "-1.5 + (`x` * -2)")
})

test_that(".build_linear_pred handles special characters in variable names", {
  result <- .build_linear_pred(
    c("(Intercept)", "var with space", "var.with.dots"),
    c(1.0, 2.0, 3.0)
  )

  expect_identical(result, "1 + (`var with space` * 2) + (`var.with.dots` * 3)")
})

test_that(".build_linear_pred handles no intercept", {
  result <- .build_linear_pred(c("x", "y"), c(2.0, 3.0))

  expect_equal(result, "(`x` * 2) + (`y` * 3)")
})

test_that(".build_linear_pred handles zero intercept", {
  result <- .build_linear_pred(
    c("(Intercept)", "x"),
    c(0, 2.0)
  )

  expect_equal(result, "(`x` * 2)")
})

# Constants ---------------------------------------------------------------

catboost_identity_objectives <- c(
  "RMSE",
  "MAE",
  "Quantile",
  "MAPE",
  "Poisson",
  "Huber",
  "LogCosh",
  "Expectile",
  "Tweedie"
)
catboost_sigmoid_objectives <- c("Logloss", "CrossEntropy")
catboost_multiclass_objectives <- c("MultiClass", "MultiClassOneVsAll")
catboost_supported_objectives <- c(
  catboost_identity_objectives,
  catboost_sigmoid_objectives,
  catboost_multiclass_objectives
)

# Model parser ------------------------------------------------------------

catboost_catboost.save_model <- function(...) {
  rlang::eval_tidy(rlang::call2("catboost.save_model", ..., .ns = "catboost"))
}

catboost_catboost.load_pool <- function(...) {
  rlang::eval_tidy(rlang::call2("catboost.load_pool", ..., .ns = "catboost"))
}

catboost_catboost.train <- function(...) {
  rlang::eval_tidy(rlang::call2("catboost.train", ..., .ns = "catboost"))
}

catboost_catboost.predict <- function(...) {
  rlang::eval_tidy(rlang::call2("catboost.predict", ..., .ns = "catboost"))
}

#' @export
parse_model.catboost.Model <- function(model) {
  pm <- list()
  pm$general$model <- "catboost.Model"
  pm$general$type <- "catboost"
  pm$general$version <- 3

  tmp_file <- tempfile(fileext = ".json")
  on.exit(unlink(tmp_file), add = TRUE)
  catboost_catboost.save_model(model, tmp_file, file_format = "json")
  model_json <- jsonlite::fromJSON(tmp_file, simplifyVector = FALSE)

  pm$general$params <- list()
  loss_fn <- model_json$model_info$params$loss_function
  if (!is.null(loss_fn) && is.list(loss_fn) && !is.null(loss_fn$type)) {
    pm$general$params$objective <- loss_fn$type
  }

  # Extract num_class for multiclass models
  objective <- pm$general$params$objective
  class_params <- model_json$model_info$class_params
  is_multiclass <- !is.null(objective) &&
    objective %in% catboost_multiclass_objectives &&
    !is.null(class_params) &&
    !is.null(class_params$class_names)

  if (is_multiclass) {
    pm$general$num_class <- length(class_params$class_names)
  } else {
    pm$general$num_class <- 1
  }

  features_info <- model_json$features_info

  # Extract float features
  float_features <- features_info$float_features
  if (length(float_features) > 0) {
    feature_names <- vapply(
      float_features,
      function(f) f$feature_id %||% paste0("feature_", f$flat_feature_index),
      character(1)
    )
  } else {
    feature_names <- character(0)
  }

  # Extract categorical features
  cat_features <- features_info$categorical_features
  cat_feature_names <- character(0)
  if (length(cat_features) > 0) {
    cat_feature_info <- lapply(cat_features, function(f) {
      list(
        feature_id = f$feature_id %||%
          paste0("cat_feature_", f$flat_feature_index),
        feature_index = f$feature_index,
        flat_feature_index = f$flat_feature_index,
        hash_values = unlist(f$values)
      )
    })
    cat_feature_names <- vapply(
      cat_feature_info,
      `[[`,
      character(1),
      "feature_id"
    )
    pm$general$cat_features <- cat_feature_info
    pm$general$cat_feature_names <- cat_feature_names
  } else {
    pm$general$cat_features <- list()
    pm$general$cat_feature_names <- character(0)
  }

  pm$general$feature_names <- feature_names
  pm$general$nfeatures <- length(feature_names) + length(cat_feature_names)

  oblivious_trees <- model_json$oblivious_trees
  nonoblivious_trees <- model_json$trees

  # Extract scale and bias
  scale_and_bias <- model_json$scale_and_bias
  pm$general$scale <- scale_and_bias[[1]]
  pm$general$bias <- scale_and_bias[[2]][[1]]

  num_class <- pm$general$num_class

  if (length(oblivious_trees) > 0) {
    pm$general$niter <- length(oblivious_trees)
    pm$general$tree_type <- "oblivious"
    pm$trees <- get_catboost_trees(
      oblivious_trees,
      float_features,
      cat_features,
      num_class
    )
  } else {
    pm$general$niter <- length(nonoblivious_trees)
    pm$general$tree_type <- "nonoblivious"
    pm$trees <- get_catboost_nonoblivious_trees(
      nonoblivious_trees,
      float_features,
      cat_features,
      num_class
    )
  }

  as_parsed_model(pm)
}

#' Set categorical feature mappings for CatBoost model
#'
#' CatBoost stores categorical features as hash values internally. This function
#' establishes the mapping between hash values and category names by examining
#' a data frame with the same factor columns used during training.
#'
#' @details
#' This function is only needed when using raw CatBoost models (trained with
#' `catboost.train()`). When using parsnip/bonsai, categorical features are
#' handled automatically and this function is not required.
#'
#' @param parsed_model A parsed CatBoost model from `parse_model()`
#' @param model The original CatBoost model object
#' @param data A data frame containing factor columns matching the categorical
#'   features used in the model. The factor levels must match those from
#'   training.
#' @return The parsed model with category mappings added
#'
#' @examples
#' \dontrun{
#' # For raw CatBoost models with categorical features:
#' pm <- parse_model(catboost_model)
#' pm <- set_catboost_categories(pm, catboost_model, training_data)
#' tidypredict_fit(pm)
#'
#' # For parsnip/bonsai models, this is not needed:
#' # tidypredict_fit(parsnip_model_fit)  # works automatically
#' }
#' @export
set_catboost_categories <- function(parsed_model, model, data) {
  if (!inherits(parsed_model, "pm_catboost")) {
    cli::cli_abort(
      "{.arg parsed_model} must be a parsed CatBoost model."
    )
  }

  if (!inherits(model, "catboost.Model")) {
    cli::cli_abort(
      "{.arg model} must be a {.cls catboost.Model}, not {.obj_type_friendly {model}}."
    )
  }

  cat_features <- parsed_model$general$cat_features
  if (length(cat_features) == 0) {
    return(parsed_model)
  }

  # Build mapping for each categorical feature
  for (i in seq_along(cat_features)) {
    feat_info <- cat_features[[i]]
    feat_name <- feat_info$feature_id
    hash_values <- feat_info$hash_values

    if (!feat_name %in% names(data)) {
      cli::cli_abort(
        "Column {.val {feat_name}} not found in {.arg data}."
      )
    }

    col_data <- data[[feat_name]]
    if (!is.factor(col_data)) {
      cli::cli_abort(
        "Column {.val {feat_name}} must be a factor."
      )
    }

    categories <- levels(col_data)
    mapping <- build_catboost_hash_mapping(
      model,
      data,
      feat_name,
      categories,
      hash_values
    )
    parsed_model$general$cat_features[[i]]$hash_to_category <- mapping
  }

  parsed_model
}

# Hash mapping functions --------------------------------------------------
# Strategy: Train probe models to discover which hash belongs to which category.
# CatBoost stores categorical features as hashes internally, but doesn't expose
# the hash function. We identify mappings by training models where each category
# has a unique target value.

build_catboost_hash_mapping <- function(
  model,
  data,
  feat_name,
  categories,
  hash_values
) {
  all_hashes <- get_catboost_all_hashes(feat_name, categories)
  identified <- identify_catboost_hashes(feat_name, categories)
  complete_mapping <- fill_catboost_hash_mapping(
    identified$mapping,
    identified$hashes,
    all_hashes,
    categories
  )
  extract_catboost_hash_mapping(complete_mapping, hash_values)
}

get_catboost_all_hashes <- function(feat_name, categories) {
  n_cat <- length(categories)
  train_data <- make_catboost_probe_data(feat_name, categories, seq_len(n_cat))
  model <- train_catboost_probe_model(train_data, feat_name, n_cat, depth = 3L)
  extract_catboost_model_hashes(model)
}

identify_catboost_hashes <- function(feat_name, categories) {
  n_cat <- length(categories)
  mapping <- list()
  identified_hashes <- character(0)

  for (cat in categories) {
    target <- ifelse(categories == cat, 100, 0)
    train_data <- make_catboost_probe_data(feat_name, categories, target)
    model <- train_catboost_probe_model(
      train_data,
      feat_name,
      n_cat,
      depth = 1L
    )
    probe_hashes <- extract_catboost_model_hashes(model)

    if (length(probe_hashes) == 1) {
      hash_str <- as.character(probe_hashes)
      mapping[[hash_str]] <- cat
      identified_hashes <- c(identified_hashes, hash_str)
    }
  }

  list(mapping = mapping, hashes = identified_hashes)
}

fill_catboost_hash_mapping <- function(
  mapping,
  identified_hashes,
  all_hashes,
  categories
) {
  remaining_cats <- setdiff(categories, unlist(mapping))
  remaining_hashes <- setdiff(as.character(all_hashes), identified_hashes)

  if (length(remaining_cats) == 1 && length(remaining_hashes) == 1) {
    mapping[[remaining_hashes]] <- remaining_cats
  }

  mapping
}

extract_catboost_hash_mapping <- function(complete_mapping, hash_values) {
  result <- stats::setNames(
    rep(NA_character_, length(hash_values)),
    as.character(hash_values)
  )

  for (hash in hash_values) {
    hash_str <- as.character(hash)
    if (hash_str %in% names(complete_mapping)) {
      result[[hash_str]] <- complete_mapping[[hash_str]]
    }
  }

  as.list(result)
}

make_catboost_probe_data <- function(feat_name, categories, target) {
  train_data <- data.frame(
    cat_col = factor(categories, levels = categories),
    target = target
  )
  names(train_data)[1] <- feat_name
  train_data
}

train_catboost_probe_model <- function(train_data, feat_name, n_cat, depth) {
  pool <- catboost_catboost.load_pool(
    train_data[, feat_name, drop = FALSE],
    label = train_data$target
  )

  catboost_catboost.train(
    pool,
    params = list(
      iterations = if (depth == 1L) 10L else 100L,
      depth = depth,
      learning_rate = 1.0,
      loss_function = "RMSE",
      logging_level = "Silent",
      allow_writing_files = FALSE,
      one_hot_max_size = n_cat + 1L,
      min_data_in_leaf = 1L
    )
  )
}

extract_catboost_model_hashes <- function(model) {
  tmp_file <- tempfile(fileext = ".json")
  on.exit(unlink(tmp_file), add = TRUE)
  catboost_catboost.save_model(model, tmp_file, file_format = "json")
  model_json <- jsonlite::fromJSON(tmp_file, simplifyVector = FALSE)
  unlist(model_json$features_info$categorical_features[[1]]$values)
}

process_catboost_trees <- function(
  trees,
  tree_fn,
  float_features,
  cat_features,
  num_class
) {
  if (num_class <= 1) {
    map(
      trees,
      tree_fn,
      float_features = float_features,
      cat_features = cat_features,
      class_idx = NULL,
      num_class = 1
    )
  } else {
    # Multiclass: create num_class virtual trees per iteration
    # Trees are stored round-robin by class: tree 1 class 0, tree 1 class 1, ...
    result <- list()
    for (i in seq_along(trees)) {
      for (class_idx in seq_len(num_class)) {
        result <- c(
          result,
          list(tree_fn(
            trees[[i]],
            float_features,
            cat_features = cat_features,
            class_idx = class_idx,
            num_class = num_class
          ))
        )
      }
    }
    result
  }
}

get_catboost_trees <- function(
  oblivious_trees,
  float_features,
  cat_features,
  num_class
) {
  process_catboost_trees(
    oblivious_trees,
    get_catboost_tree,
    float_features,
    cat_features,
    num_class
  )
}

# Non-oblivious tree parsing (Depthwise/Lossguide grow_policy) ----------------

get_catboost_nonoblivious_trees <- function(
  trees,
  float_features,
  cat_features,
  num_class
) {
  process_catboost_trees(
    trees,
    get_catboost_nonoblivious_tree,
    float_features,
    cat_features,
    num_class
  )
}

get_catboost_nonoblivious_tree <- function(
  tree_json,
  float_features,
  cat_features = NULL,
  class_idx = NULL,
  num_class = 1
) {
  # Recursively collect all leaf paths
  leaves <- collect_nonoblivious_leaves(
    tree_json,
    float_features,
    cat_features,
    path = list()
  )

  # Extract predictions for the appropriate class
  map(leaves, function(leaf) {
    pred_value <- leaf$value
    # For multiclass, leaf$value is a list with one element per class
    if (
      !is.null(class_idx) && num_class > 1 && length(pred_value) >= num_class
    ) {
      pred_value <- pred_value[[class_idx]]
    }
    list(prediction = pred_value, path = leaf$path)
  })
}

collect_nonoblivious_leaves <- function(
  node,
  float_features,
  cat_features,
  path
) {
  # Check if this is a leaf node (has "value" field)
  if (!is.null(node$value)) {
    return(list(list(value = node$value, path = path)))
  }

  # Internal node: process split and recurse
  split <- node$split
  left_node <- node$left
  right_node <- node$right

  # Parse the split condition
  left_condition <- parse_nonoblivious_split(
    split,
    float_features,
    cat_features,
    go_left = TRUE
  )
  right_condition <- parse_nonoblivious_split(
    split,
    float_features,
    cat_features,
    go_left = FALSE
  )

  # Recurse into both branches
  left_leaves <- collect_nonoblivious_leaves(
    left_node,
    float_features,
    cat_features,
    path = c(path, list(left_condition))
  )
  right_leaves <- collect_nonoblivious_leaves(
    right_node,
    float_features,
    cat_features,
    path = c(path, list(right_condition))
  )

  c(left_leaves, right_leaves)
}

parse_nonoblivious_split <- function(
  split,
  float_features,
  cat_features,
  go_left
) {
  split_type <- split$split_type %||% "FloatFeature"

  if (split_type == "OneHotFeature") {
    op <- if (go_left) "equal" else "not-equal"
    make_categorical_split(split, cat_features, op)
  } else {
    # left = <= border, right = > border
    op <- if (go_left) "less-equal" else "more"
    make_float_split(split, float_features, op)
  }
}

# Split building helpers (shared by oblivious and non-oblivious) --------------

make_float_split <- function(split, float_features, op) {
  feature_index <- split$float_feature_index + 1L
  feature_info <- float_features[[feature_index]]
  feature_name <- feature_info$feature_id %||%
    paste0("feature_", feature_info$flat_feature_index)
  nan_treatment <- feature_info$nan_value_treatment %||% "AsIs"

  list(
    type = "conditional",
    col = feature_name,
    val = split$border,
    op = op,
    missing = get_catboost_missing(nan_treatment, op)
  )
}

make_categorical_split <- function(split, cat_features, op) {
  cat_feature_index <- split$cat_feature_index + 1L
  cat_feature_info <- cat_features[[cat_feature_index]]
  feature_name <- cat_feature_info$feature_id %||%
    paste0("cat_feature_", cat_feature_info$flat_feature_index)

  list(
    type = "categorical",
    col = feature_name,
    hash_value = split$value,
    op = op,
    missing = FALSE
  )
}

# Oblivious tree parsing ------------------------------------------------------

get_catboost_tree <- function(
  tree_json,
  float_features,
  cat_features = NULL,
  class_idx = NULL,
  num_class = 1
) {
  splits <- tree_json$splits
  leaf_values <- unlist(tree_json$leaf_values)
  n_splits <- length(splits)

  if (n_splits == 0) {
    return(make_catboost_stump(leaf_values, class_idx, num_class))
  }

  # Pre-extract split info once per tree (avoids repeated lookups per leaf)
  split_info <- lapply(splits, function(split) {
    split_type <- split$split_type %||% "FloatFeature"
    if (split_type == "OneHotFeature") {
      cat_feature_index <- split$cat_feature_index + 1L
      cat_feature_info <- cat_features[[cat_feature_index]]
      list(
        type = "categorical",
        col = cat_feature_info$feature_id %||%
          paste0("cat_feature_", cat_feature_info$flat_feature_index),
        hash_value = split$value,
        is_categorical = TRUE
      )
    } else {
      feature_index <- split$float_feature_index + 1L
      feature_info <- float_features[[feature_index]]
      list(
        type = "conditional",
        col = feature_info$feature_id %||%
          paste0("feature_", feature_info$flat_feature_index),
        val = split$border,
        nan_treatment = feature_info$nan_value_treatment %||% "AsIs",
        is_categorical = FALSE
      )
    }
  })

  n_leaves <- 2^n_splits
  map(seq_len(n_leaves) - 1L, function(leaf_idx) {
    path <- lapply(seq_len(n_splits), function(split_idx) {
      info <- split_info[[split_idx]]
      bit_val <- get_catboost_bit_value(leaf_idx, split_idx)

      if (info$is_categorical) {
        op <- if (bit_val == 0L) "not-equal" else "equal"
        list(
          type = "categorical",
          col = info$col,
          hash_value = info$hash_value,
          op = op,
          missing = FALSE
        )
      } else {
        op <- if (bit_val == 1L) "more" else "less-equal"
        list(
          type = "conditional",
          col = info$col,
          val = info$val,
          op = op,
          missing = get_catboost_missing(info$nan_treatment, op)
        )
      }
    })

    list(
      prediction = get_catboost_leaf_value(
        leaf_values,
        leaf_idx,
        class_idx,
        num_class
      ),
      path = path
    )
  })
}

make_catboost_stump <- function(leaf_values, class_idx, num_class) {
  if (!is.null(class_idx) && num_class > 1) {
    pred_value <- leaf_values[class_idx]
  } else {
    pred_value <- leaf_values[1]
  }
  list(list(prediction = pred_value, path = list()))
}

get_catboost_leaf_value <- function(
  leaf_values,
  leaf_idx,
  class_idx,
  num_class
) {
  # For multiclass: [leaf0_class0, leaf0_class1, ..., leaf1_class0, ...]
  if (!is.null(class_idx) && num_class > 1) {
    leaf_values[leaf_idx * num_class + class_idx]
  } else {
    leaf_values[leaf_idx + 1L]
  }
}

get_catboost_bit_value <- function(leaf_idx, split_idx) {
  bit_pos <- split_idx - 1L
  bitwAnd(bitwShiftR(leaf_idx, bit_pos), 1L)
}

get_catboost_missing <- function(nan_treatment, op) {
  # nan_treatment determines where NaN values go:
  # - "AsIs": No special treatment, NaN doesn't match anything -> FALSE
  # - "AsTrue": NaN goes left (where condition > border is TRUE) -> TRUE if op="more"
  # - "AsFalse": NaN goes right (where condition is FALSE) -> TRUE if op="less-equal"
  if (nan_treatment == "AsTrue") {
    return(op == "more")
  } else if (nan_treatment == "AsFalse") {
    return(op == "less-equal")
  } else {
    # "AsIs" or unknown
    return(FALSE)
  }
}

# Shared helpers -----------------------------------------------

# Apply multiclass transformation to tree expressions
# Shared by nested and legacy multiclass builders
apply_catboost_multiclass_transformation <- function(
  trees,
  num_class,
  objective,
  parsedmodel
) {
  n_trees <- length(trees)

  # Trees are stored round-robin by class
  class_trees <- lapply(seq_len(num_class), function(class_idx) {
    which((seq_len(n_trees) - 1) %% num_class == (class_idx - 1))
  })

  raw_scores <- lapply(class_trees, function(indices) {
    reduce_addition(trees[indices])
  })
  raw_scores <- lapply(raw_scores, apply_catboost_scale_bias, parsedmodel)

  if (objective == "MultiClass") {
    # Softmax: exp(raw_i) / sum(exp(raw_j))
    exp_raws <- map(raw_scores, ~ expr(exp(!!.x)))
    denom <- reduce_addition(exp_raws)
    result <- map(seq_len(num_class), function(i) {
      expr(exp(!!raw_scores[[i]]) / (!!denom))
    })
  } else {
    # MultiClassOneVsAll: sigmoid for each class independently
    result <- map(raw_scores, ~ expr(1 / (1 + exp(-(!!.x)))))
  }

  names(result) <- paste0("class_", seq_len(num_class) - 1)
  result
}

# Apply catboost scale and bias to formula
apply_catboost_scale_bias <- function(formula, parsedmodel) {
  scale <- parsedmodel$general$scale %||% 1
  bias <- parsedmodel$general$bias %||% 0

  if (scale != 1) {
    formula <- expr(!!scale * (!!formula))
  }
  if (bias != 0) {
    formula <- expr(!!formula + !!bias)
  }
  formula
}

# Build a combined hash -> category mapping from all categorical features
get_catboost_cat_mapping <- function(parsedmodel) {
  cat_features <- parsedmodel$general$cat_features
  if (length(cat_features) == 0) {
    return(list())
  }

  mapping <- list()
  for (feat in cat_features) {
    if (!is.null(feat$hash_to_category)) {
      for (hash_str in names(feat$hash_to_category)) {
        mapping[[hash_str]] <- feat$hash_to_category[[hash_str]]
      }
    }
  }
  mapping
}

# Fit model (nested) -----------------------------------------------

#' @export
tidypredict_fit.catboost.Model <- function(model, ...) {
  parsedmodel <- parse_model(model)
  build_fit_formula_catboost_nested(parsedmodel)
}

# Internal function for parsnip model_fit objects with CatBoost
# Called from tidymodels.R
tidypredict_fit_catboost_parsnip <- function(model) {
  cb_model <- model$fit
  parsedmodel <- parse_model(cb_model)

  has_cat_features <- length(parsedmodel$general$cat_features) > 0
  if (has_cat_features) {
    parsedmodel <- setup_catboost_parsnip_categories(parsedmodel, model)
  }

  build_fit_formula_catboost_nested(parsedmodel)
}

setup_catboost_parsnip_categories <- function(parsedmodel, model) {
  has_xlevels <- !is.null(model$preproc$xlevels) &&
    length(model$preproc$xlevels) > 0

  if (!has_xlevels) {
    cli::cli_abort(
      c(
        "Model has categorical features but no factor level information.",
        "i" = "Ensure the model was fit with factor columns, not character columns."
      )
    )
  }

  mapping_data <- build_catboost_parsnip_mapping_data(
    model$preproc$xlevels,
    parsedmodel$general$cat_feature_names
  )

  if (ncol(mapping_data) > 0) {
    parsedmodel <- set_catboost_categories(parsedmodel, model$fit, mapping_data)
  }

  parsedmodel
}

build_catboost_parsnip_mapping_data <- function(xlevels, cat_feature_names) {
  mapping_data <- data.frame(row.names = 1)

  for (feat_name in cat_feature_names) {
    if (feat_name %in% names(xlevels)) {
      lvls <- xlevels[[feat_name]]
      mapping_data[[feat_name]] <- factor(lvls[1], levels = lvls)
    }
  }

  mapping_data
}

# Nested formula builders for CatBoost -----------------------------------------

build_fit_formula_catboost_nested <- function(parsedmodel) {
  n_trees <- length(parsedmodel$trees)

  if (n_trees == 0) {
    cli::cli_abort("Model has no trees.")
  }

  objective <- parsedmodel$general$params$objective %||% "RMSE"

  if (!objective %in% catboost_supported_objectives) {
    cli::cli_abort(
      c(
        "Unsupported objective: {.val {objective}}.",
        "i" = "Supported objectives: {.val {catboost_supported_objectives}}."
      )
    )
  }

  if (objective %in% catboost_multiclass_objectives) {
    return(build_fit_formula_catboost_multiclass_nested(parsedmodel, objective))
  }

  # Extract nested trees
  trees <- extract_catboost_trees_nested(parsedmodel)
  f <- reduce_addition(trees)
  f <- apply_catboost_scale_bias(f, parsedmodel)

  if (objective %in% catboost_sigmoid_objectives) {
    f <- expr(1 / (1 + exp(-(!!f))))
  }

  f
}

build_fit_formula_catboost_multiclass_nested <- function(
  parsedmodel,
  objective
) {
  num_class <- parsedmodel$general$num_class
  if (is.null(num_class) || num_class < 2) {
    cli::cli_abort("Multiclass model must have num_class >= 2.")
  }

  trees <- extract_catboost_trees_nested(parsedmodel)
  apply_catboost_multiclass_transformation(
    trees,
    num_class,
    objective,
    parsedmodel
  )
}

# Extract trees in nested format
extract_catboost_trees_nested <- function(parsedmodel) {
  cat_mapping <- get_catboost_cat_mapping(parsedmodel)
  tree_type <- parsedmodel$general$tree_type

  map(parsedmodel$trees, function(tree) {
    if (tree_type == "oblivious") {
      build_nested_catboost_oblivious_tree(tree, cat_mapping)
    } else {
      build_nested_catboost_nonoblivious_tree(tree, cat_mapping)
    }
  })
}

# Build nested case_when for oblivious tree
# Oblivious trees have a flat structure where all leaves share the same splits
build_nested_catboost_oblivious_tree <- function(tree, cat_mapping) {
  # tree is a list of (prediction, path) pairs indexed 0 to 2^n_splits - 1
  # All paths have the same length (one condition per split level)

  # Single leaf (stump) - reachable with depth=0
  if (length(tree) == 1 && length(tree[[1]]$path) == 0) {
    return(tree[[1]]$prediction)
  }

  # Create leaf indices (0-indexed)
  # Note: n_splits > 0 guaranteed here since single-leaf stumps caught above
  leaf_indices <- seq_len(length(tree)) - 1L

  # Build nested structure recursively by split level
  build_nested_oblivious_level(tree, leaf_indices, cat_mapping, split_level = 1)
}

build_nested_oblivious_level <- function(
  tree,
  leaf_indices,
  cat_mapping,
  split_level
) {
  n_splits <- length(tree[[1]]$path)

  # Base case: reached bottom level, return the single leaf's prediction
  if (split_level > n_splits) {
    return(tree[[leaf_indices + 1]]$prediction)
  }

  # Group leaves by bit value at current split level
  # bit=0 means "less-equal" (left), bit=1 means "more" (right)
  bit_pos <- split_level - 1L
  bit_values <- bitwAnd(bitwShiftR(leaf_indices, bit_pos), 1L)

  left_indices <- leaf_indices[bit_values == 0L]
  right_indices <- leaf_indices[bit_values == 1L]

  # Get split info from left leaf (all left leaves have same "less-equal" condition)
  split_info <- tree[[left_indices[1] + 1]]$path[[split_level]]

  # Build condition (for left branch)
  condition <- build_nested_catboost_condition(split_info, cat_mapping)

  # Recurse
  left_subtree <- build_nested_oblivious_level(
    tree,
    left_indices,
    cat_mapping,
    split_level + 1
  )
  right_subtree <- build_nested_oblivious_level(
    tree,
    right_indices,
    cat_mapping,
    split_level + 1
  )

  expr(case_when(!!condition ~ !!left_subtree, .default = !!right_subtree))
}

# Build nested case_when for non-oblivious tree
build_nested_catboost_nonoblivious_tree <- function(tree, cat_mapping) {
  # For non-oblivious trees, we need to reconstruct the tree structure
  # from the flat list of (prediction, path) pairs

  # Single leaf (stump) - reachable with depth=0
  if (length(tree) == 1 && length(tree[[1]]$path) == 0) {
    return(tree[[1]]$prediction)
  }

  # Build recursively from paths
  build_nested_nonoblivious_node(tree, cat_mapping, path_depth = 1)
}

build_nested_nonoblivious_node <- function(leaves, cat_mapping, path_depth) {
  if (length(leaves) == 1) {
    # Single leaf - return prediction
    return(leaves[[1]]$prediction)
  }

  # Get condition from first leaf
  # Note: all grouped leaves have path length >= path_depth (tree structure guarantees this)
  first_leaf <- leaves[[1]]
  split_info <- first_leaf$path[[path_depth]]

  # Partition leaves based on their condition at this depth
  # "less-equal" or "equal" go left, "more" or "not-equal" go right
  is_left_condition <- function(leaf) {
    op <- leaf$path[[path_depth]]$op
    op %in% c("less-equal", "equal")
  }

  left_leaves <- Filter(is_left_condition, leaves)
  right_leaves <- Filter(Negate(is_left_condition), leaves)

  # Build condition (use left condition)
  condition <- build_nested_catboost_condition(split_info, cat_mapping)

  left_subtree <- build_nested_nonoblivious_node(
    left_leaves,
    cat_mapping,
    path_depth + 1
  )
  right_subtree <- build_nested_nonoblivious_node(
    right_leaves,
    cat_mapping,
    path_depth + 1
  )

  expr(case_when(!!condition ~ !!left_subtree, .default = !!right_subtree))
}

# Build condition expression for a split
build_nested_catboost_condition <- function(split_info, cat_mapping) {
  if (split_info$type == "categorical") {
    build_nested_catboost_categorical(split_info, cat_mapping)
  } else {
    build_nested_catboost_comparison(split_info)
  }
}

build_nested_catboost_comparison <- function(split_info) {
  col_name <- rlang::sym(split_info$col)
  val <- as.numeric(split_info$val)
  include_missing <- split_info$missing

  # For nested, we use the "left" condition (less-equal)
  if (include_missing) {
    expr(!!col_name <= !!val | is.na(!!col_name))
  } else {
    expr(!!col_name <= !!val)
  }
}

build_nested_catboost_categorical <- function(split_info, cat_mapping) {
  col_name <- rlang::sym(split_info$col)
  hash_value <- split_info$hash_value
  hash_str <- as.character(hash_value)

  category <- cat_mapping[[hash_str]]
  if (is.null(category) || is.na(category)) {
    cli::cli_abort(
      c(
        "No category mapping found for hash {.val {hash_value}}.",
        "i" = "For raw CatBoost models, use {.fn set_catboost_categories}."
      )
    )
  }

  # For nested oblivious trees, the left branch (bit=0) is the "not-equal" case
  expr(!!col_name != !!category)
}

# For {orbital} -----------------------------------------------

#' Extract processed CatBoost trees
#'
#' For use in orbital package.
#' @param model A CatBoost model object
#' @keywords internal
#' @export
.extract_catboost_trees <- function(model) {
  if (!inherits(model, "catboost.Model")) {
    cli::cli_abort(
      "{.arg model} must be {.cls catboost.Model}, not {.obj_type_friendly {model}}."
    )
  }

  parsedmodel <- parse_model(model)
  extract_catboost_trees_nested(parsedmodel)
}

# Helper to create test model
make_catboost_model <- function(
  iterations = 10L,
  depth = 3L,
  loss_function = "RMSE"
) {
  set.seed(123)
  X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
  y <- mtcars$hp

  pool <- catboost_catboost.load_pool(
    X,
    label = y,
    feature_names = as.list(c("mpg", "cyl", "disp"))
  )

  catboost_catboost.train(
    pool,
    params = list(
      iterations = iterations,
      depth = depth,
      learning_rate = 0.5,
      loss_function = loss_function,
      logging_level = "Silent",
      allow_writing_files = FALSE
    )
  )
}

make_multiclass_model <- function(objective = "MultiClass") {
  set.seed(42)
  X <- data.matrix(iris[, 1:4])
  y <- as.integer(iris$Species) - 1L

  pool <- catboost_catboost.load_pool(
    X,
    label = y,
    feature_names = as.list(colnames(iris)[1:4])
  )

  catboost_catboost.train(
    pool,
    params = list(
      iterations = 10L,
      depth = 3L,
      learning_rate = 0.5,
      loss_function = objective,
      logging_level = "Silent",
      allow_writing_files = FALSE
    )
  )
}

make_categorical_model <- function() {
  set.seed(42)
  df <- data.frame(
    num_feat = rnorm(100),
    cat_feat = factor(sample(c("A", "B", "C"), 100, replace = TRUE)),
    target = rnorm(100)
  )

  X <- df[, c("num_feat", "cat_feat")]
  y <- df$target

  pool <- catboost_catboost.load_pool(X, label = y)

  catboost_catboost.train(
    pool,
    params = list(
      iterations = 10L,
      depth = 3L,
      learning_rate = 0.5,
      loss_function = "RMSE",
      logging_level = "Silent",
      allow_writing_files = FALSE,
      one_hot_max_size = 10
    )
  )
}

# Parser tests ---------------------------------------------------------------

test_that("parse_model returns correct structure", {
  skip_if_not_installed("catboost")
  model <- make_catboost_model()
  pm <- parse_model(model)

  expect_s3_class(pm, "parsed_model")
  expect_s3_class(pm, "pm_catboost")

  expect_equal(pm$general$model, "catboost.Model")
  expect_equal(pm$general$type, "catboost")
  expect_equal(pm$general$version, 3)

  expect_gt(length(pm$trees), 0)
})

test_that("correct number of trees extracted", {
  skip_if_not_installed("catboost")
  model <- make_catboost_model()
  pm <- parse_model(model)

  expect_length(pm$trees, 10)
})

test_that("each tree has leaves with predictions and paths", {
  skip_if_not_installed("catboost")
  model <- make_catboost_model()
  pm <- parse_model(model)

  tree1 <- pm$trees[[1]]
  expect_gt(length(tree1), 0)

  has_required_names <- vapply(
    tree1,
    \(x) all(c("prediction", "path") %in% names(x)),
    logical(1)
  )
  expect_all_equal(has_required_names, TRUE)

  predictions <- vapply(tree1, \(x) x$prediction, double(1))
  expect_type(predictions, "double")

  path_is_list <- vapply(tree1, \(x) is.list(x$path), logical(1))
  expect_all_equal(path_is_list, TRUE)
})

test_that("symmetric trees have 2^n_splits leaves", {
  skip_if_not_installed("catboost")

  model <- make_catboost_model()
  pm <- parse_model(model)

  n_leaves <- vapply(pm$trees, length, integer(1))
  n_splits <- vapply(pm$trees, \(tree) length(tree[[1]]$path), integer(1))

  expect_equal(n_leaves, 2L^n_splits)
})

test_that("path conditions have correct structure", {
  skip_if_not_installed("catboost")
  model <- make_catboost_model()
  pm <- parse_model(model)

  tree1 <- pm$trees[[1]]
  has_path <- vapply(tree1, function(x) length(x$path) > 0, logical(1))
  leaves_with_paths <- which(has_path)

  if (length(leaves_with_paths) > 0) {
    leaf_with_path <- tree1[[leaves_with_paths[1]]]

    cond <- leaf_with_path$path[[1]]
    expect_equal(cond$type, "conditional")
    expect_contains(names(cond), c("col", "val", "op", "missing"))
    expect_contains(c("less-equal", "more"), cond$op)
  }
})

test_that("feature names are extracted", {
  skip_if_not_installed("catboost")
  model <- make_catboost_model()
  pm <- parse_model(model)

  expect_equal(pm$general$feature_names, c("mpg", "cyl", "disp"))
})

test_that("params are extracted", {
  skip_if_not_installed("catboost")
  model <- make_catboost_model()
  pm <- parse_model(model)

  expect_contains(names(pm$general), "params")
  expect_equal(pm$general$params$objective, "RMSE")
})

test_that("niter and nfeatures are extracted", {
  skip_if_not_installed("catboost")
  model <- make_catboost_model()
  pm <- parse_model(model)

  expect_equal(pm$general$niter, 10)
  expect_equal(pm$general$nfeatures, 3)
})

test_that("scale and bias are extracted correctly", {
  skip_if_not_installed("catboost")
  model <- make_catboost_model()
  pm <- parse_model(model)

  expect_type(pm$general$scale, "integer")
  expect_type(pm$general$bias, "double")
})

test_that("path contains both more and less-equal operators", {
  skip_if_not_installed("catboost")
  model <- make_catboost_model()
  pm <- parse_model(model)

  all_ops <- unlist(lapply(pm$trees[[1]], function(leaf) {
    vapply(leaf$path, function(p) p$op, character(1))
  }))

  expect_contains(all_ops, "more")
  expect_contains(all_ops, "less-equal")
})

test_that("leaf 0 has all less-equal conditions (binary 00...0)", {
  skip_if_not_installed("catboost")

  set.seed(42)
  X <- matrix(c(1, 1, 10, 10, 1, 10, 1, 10), ncol = 2)
  y <- c(100, 200, 300, 400)

  pool <- catboost_catboost.load_pool(
    X,
    label = y,
    feature_names = as.list(c("x1", "x2"))
  )
  model <- catboost_catboost.train(
    pool,
    params = list(
      iterations = 1L,
      depth = 2L,
      learning_rate = 1.0,
      loss_function = "RMSE",
      logging_level = "Silent",
      allow_writing_files = FALSE,
      min_data_in_leaf = 1L
    )
  )

  pm <- parse_model(model)
  leaf_0 <- pm$trees[[1]][[1]]

  # Leaf 0 (binary 00) should have all less-equal operators
  ops <- vapply(leaf_0$path, \(p) p$op, character(1))
  expect_all_equal(ops, "less-equal")
})

test_that("last leaf has all more conditions (binary 11...1)", {
  skip_if_not_installed("catboost")

  set.seed(42)
  X <- matrix(c(1, 1, 10, 10, 1, 10, 1, 10), ncol = 2)
  y <- c(100, 200, 300, 400)

  pool <- catboost_catboost.load_pool(
    X,
    label = y,
    feature_names = as.list(c("x1", "x2"))
  )
  model <- catboost_catboost.train(
    pool,
    params = list(
      iterations = 1L,
      depth = 2L,
      learning_rate = 1.0,
      loss_function = "RMSE",
      logging_level = "Silent",
      allow_writing_files = FALSE,
      min_data_in_leaf = 1L
    )
  )

  pm <- parse_model(model)
  n_leaves <- length(pm$trees[[1]])
  last_leaf <- pm$trees[[1]][[n_leaves]]

  # Last leaf (binary 11) should have all more operators
  ops <- vapply(last_leaf$path, \(p) p$op, character(1))
  expect_all_equal(ops, "more")
})

test_that("leaf index binary representation determines operator pattern", {
  skip_if_not_installed("catboost")

  set.seed(42)
  X <- matrix(c(1, 1, 10, 10, 1, 10, 1, 10), ncol = 2)
  y <- c(100, 200, 300, 400)

  pool <- catboost_catboost.load_pool(
    X,
    label = y,
    feature_names = as.list(c("x1", "x2"))
  )
  model <- catboost_catboost.train(
    pool,
    params = list(
      iterations = 1L,
      depth = 2L,
      learning_rate = 1.0,
      loss_function = "RMSE",
      logging_level = "Silent",
      allow_writing_files = FALSE,
      min_data_in_leaf = 1L
    )
  )

  pm <- parse_model(model)
  tree <- pm$trees[[1]]

  # With 2 splits, we have 4 leaves (indices 0-3)
  # Leaf index binary -> operator pattern:
  # 0 (00): [<=, <=]
  # 1 (01): [>, <=]
  # 2 (10): [<=, >]
  # 3 (11): [>, >]

  for (leaf_idx in 0:3) {
    leaf <- tree[[leaf_idx + 1]]
    n_splits <- length(leaf$path)

    for (split_idx in seq_len(n_splits)) {
      bit_val <- bitwAnd(bitwShiftR(leaf_idx, split_idx - 1L), 1L)
      expected_op <- if (bit_val == 1L) "more" else "less-equal"
      expect_equal(
        leaf$path[[split_idx]]$op,
        expected_op,
        info = sprintf("leaf %d, split %d", leaf_idx, split_idx)
      )
    }
  }
})

test_that("single split tree produces correct paths", {
  skip_if_not_installed("catboost")

  set.seed(42)
  X <- matrix(c(1, 10), ncol = 1)
  y <- c(0, 100)

  pool <- catboost_catboost.load_pool(
    X,
    label = y,
    feature_names = as.list(c("x"))
  )
  model <- catboost_catboost.train(
    pool,
    params = list(
      iterations = 1L,
      depth = 1L,
      learning_rate = 1.0,
      loss_function = "RMSE",
      logging_level = "Silent",
      allow_writing_files = FALSE,
      min_data_in_leaf = 1L
    )
  )

  pm <- parse_model(model)

  expect_length(pm$trees[[1]], 2) # 2^1 = 2 leaves

  # Leaf 0: x <= border
  expect_equal(pm$trees[[1]][[1]]$path[[1]]$op, "less-equal")
  expect_equal(pm$trees[[1]][[1]]$path[[1]]$col, "x")

  # Leaf 1: x > border
  expect_equal(pm$trees[[1]][[2]]$path[[1]]$op, "more")
  expect_equal(pm$trees[[1]][[2]]$path[[1]]$col, "x")
})

test_that("stump trees (depth=0) are parsed correctly", {
  skip_if_not_installed("catboost")

  set.seed(123)
  X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
  y <- mtcars$hp

  pool <- catboost_catboost.load_pool(
    X,
    label = y,
    feature_names = as.list(c("mpg", "cyl", "disp"))
  )

  model <- catboost_catboost.train(
    pool,
    params = list(
      iterations = 3L,
      depth = 0L,
      learning_rate = 0.5,
      loss_function = "RMSE",
      logging_level = "Silent",
      allow_writing_files = FALSE
    )
  )

  pm <- parse_model(model)

  # Depth 0 trees should have no splits
  expect_equal(length(pm$trees[[1]][[1]]$path), 0)
})

test_that("model without explicit feature names still works", {
  skip_if_not_installed("catboost")

  set.seed(789)
  X <- data.matrix(mtcars[, c("mpg", "cyl")])
  y <- mtcars$hp

  # Create pool WITHOUT specifying feature_names
  pool <- catboost_catboost.load_pool(X, label = y)

  model <- catboost_catboost.train(
    pool,
    params = list(
      iterations = 3L,
      depth = 2L,
      learning_rate = 1.0,
      loss_function = "RMSE",
      logging_level = "Silent",
      allow_writing_files = FALSE,
      min_data_in_leaf = 1L
    )
  )

  pm <- parse_model(model)

  expect_s3_class(pm, "pm_catboost")
  expect_length(pm$trees, 3)
  expect_equal(pm$general$nfeatures, 2)
  expect_type(pm$general$feature_names, "character")
})

test_that("deeper tree paths are traced correctly", {
  skip_if_not_installed("catboost")

  # Create data that forces a 3-split tree
  set.seed(123)
  n <- 100
  X <- matrix(rnorm(n * 3), ncol = 3)
  colnames(X) <- c("a", "b", "c")
  y <- X[, 1] + X[, 2] * 2 + X[, 3] * 3 + rnorm(n, sd = 0.1)

  pool <- catboost_catboost.load_pool(
    X,
    label = y,
    feature_names = as.list(c("a", "b", "c"))
  )
  model <- catboost_catboost.train(
    pool,
    params = list(
      iterations = 1L,
      depth = 3L,
      learning_rate = 1.0,
      loss_function = "RMSE",
      logging_level = "Silent",
      allow_writing_files = FALSE,
      min_data_in_leaf = 1L
    )
  )

  pm <- parse_model(model)
  tree <- pm$trees[[1]]

  # All leaves should have 3 conditions in their path (if tree is full depth)
  n_splits <- length(tree[[1]]$path)
  expect_lte(n_splits, 3)

  # Verify leaf count matches 2^n_splits
  expect_length(tree, 2^n_splits)
})

test_that("num_class is extracted for multiclass model", {
  skip_if_not_installed("catboost")

  model <- make_multiclass_model()
  pm <- parse_model(model)

  expect_equal(pm$general$num_class, 3)
})

test_that("num_class is 1 for regression model", {
  skip_if_not_installed("catboost")

  model <- make_catboost_model()
  pm <- parse_model(model)

  expect_equal(pm$general$num_class, 1)
})

test_that("categorical features are extracted from model", {
  skip_if_not_installed("catboost")

  model <- make_categorical_model()
  pm <- parse_model(model)

  expect_length(pm$general$cat_features, 1)
  expect_equal(pm$general$cat_feature_names, "cat_feat")
  expect_type(pm$general$cat_features[[1]]$hash_values, "integer")
})

# Fit formula tests -------------------------------------------------------

test_that("tidypredict_fit returns language object", {
  skip_if_not_installed("catboost")
  model <- make_catboost_model()

  result <- tidypredict_fit(model)

  expect_type(result, "language")
})

test_that("tidypredict_fit works on parsed model", {
  skip_if_not_installed("catboost")
  model <- make_catboost_model()
  pm <- parse_model(model)

  result <- tidypredict_fit(pm)

  expect_type(result, "language")
})

test_that("produced case_when uses .default", {
  skip_if_not_installed("catboost")
  model <- make_catboost_model()

  fit <- tidypredict_fit(model)
  fit_text <- rlang::expr_text(fit)

  expect_match(fit_text, "\\.default")
})

test_that("regression predictions match catboost.predict", {
  skip_if_not_installed("catboost")
  model <- make_catboost_model()

  X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
  pool <- catboost_catboost.load_pool(X)

  native_preds <- catboost_catboost.predict(model, pool)
  formula <- tidypredict_fit(model)
  tidy_preds <- rlang::eval_tidy(formula, mtcars)

  expect_equal(tidy_preds, native_preds, tolerance = 1e-10)
})

test_that("model with custom scale works correctly", {
  skip_if_not_installed("catboost")
  model <- make_catboost_model()
  pm <- parse_model(model)

  # Simulate a calibrated model with scale != 1
  # (set_scale_and_bias is Python-only, so we modify the parsed model directly)
  original_scale <- pm$general$scale
  bias <- pm$general$bias
  pm$general$scale <- 0.5

  formula <- tidypredict_fit(pm)
  tidy_preds <- rlang::eval_tidy(formula, mtcars)

  # Formula is: scale * tree_sum + bias
  # With scale=0.5: 0.5 * tree_sum + bias
  # With scale=1: tree_sum + bias (unscaled)
  # So: scaled = 0.5 * (unscaled - bias) + bias = 0.5 * unscaled + 0.5 * bias
  pm$general$scale <- original_scale
  formula_unscaled <- tidypredict_fit(pm)
  unscaled_preds <- rlang::eval_tidy(formula_unscaled, mtcars)

  expected <- 0.5 * unscaled_preds + 0.5 * bias
  expect_equal(tidy_preds, expected, tolerance = 1e-10)
})

test_that("Logloss predictions match catboost.predict", {
  skip_if_not_installed("catboost")

  set.seed(456)
  X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
  y <- as.numeric(mtcars$am)

  pool <- catboost_catboost.load_pool(
    X,
    label = y,
    feature_names = as.list(c("mpg", "cyl", "disp"))
  )

  model <- catboost_catboost.train(
    pool,
    params = list(
      iterations = 10L,
      depth = 3L,
      learning_rate = 0.5,
      loss_function = "Logloss",
      logging_level = "Silent",
      allow_writing_files = FALSE
    )
  )

  native_preds <- catboost_catboost.predict(
    model,
    pool,
    prediction_type = "Probability"
  )
  formula <- tidypredict_fit(model)
  tidy_preds <- rlang::eval_tidy(formula, mtcars)

  expect_equal(tidy_preds, native_preds, tolerance = 1e-10)
})

test_that("MAE predictions match catboost.predict", {
  skip_if_not_installed("catboost")

  set.seed(123)
  X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
  y <- mtcars$hp

  pool <- catboost_catboost.load_pool(
    X,
    label = y,
    feature_names = as.list(c("mpg", "cyl", "disp"))
  )

  model <- catboost_catboost.train(
    pool,
    params = list(
      iterations = 10L,
      depth = 3L,
      learning_rate = 0.5,
      loss_function = "MAE",
      logging_level = "Silent",
      allow_writing_files = FALSE
    )
  )

  native_preds <- catboost_catboost.predict(model, pool)
  formula <- tidypredict_fit(model)
  tidy_preds <- rlang::eval_tidy(formula, mtcars)

  expect_equal(tidy_preds, native_preds, tolerance = 1e-10)
})

test_that("Quantile predictions match catboost.predict", {
  skip_if_not_installed("catboost")

  set.seed(123)
  X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
  y <- mtcars$hp

  pool <- catboost_catboost.load_pool(
    X,
    label = y,
    feature_names = as.list(c("mpg", "cyl", "disp"))
  )

  model <- catboost_catboost.train(
    pool,
    params = list(
      iterations = 10L,
      depth = 3L,
      learning_rate = 0.5,
      loss_function = "Quantile:alpha=0.5",
      logging_level = "Silent",
      allow_writing_files = FALSE
    )
  )

  native_preds <- catboost_catboost.predict(model, pool)
  formula <- tidypredict_fit(model)
  tidy_preds <- rlang::eval_tidy(formula, mtcars)

  expect_equal(tidy_preds, native_preds, tolerance = 1e-10)
})

test_that("MAPE predictions match catboost.predict", {
  skip_if_not_installed("catboost")

  set.seed(123)
  X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
  y <- mtcars$hp

  pool <- catboost_catboost.load_pool(
    X,
    label = y,
    feature_names = as.list(c("mpg", "cyl", "disp"))
  )

  model <- catboost_catboost.train(
    pool,
    params = list(
      iterations = 10L,
      depth = 3L,
      learning_rate = 0.5,
      loss_function = "MAPE",
      logging_level = "Silent",
      allow_writing_files = FALSE
    )
  )

  native_preds <- catboost_catboost.predict(model, pool)
  formula <- tidypredict_fit(model)
  tidy_preds <- rlang::eval_tidy(formula, mtcars)

  expect_equal(tidy_preds, native_preds, tolerance = 1e-10)
})

test_that("Poisson predictions match catboost.predict", {
  skip_if_not_installed("catboost")

  set.seed(123)
  X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
  y <- mtcars$carb

  pool <- catboost_catboost.load_pool(
    X,
    label = y,
    feature_names = as.list(c("mpg", "cyl", "disp"))
  )

  model <- catboost_catboost.train(
    pool,
    params = list(
      iterations = 10L,
      depth = 3L,
      learning_rate = 0.5,
      loss_function = "Poisson",
      logging_level = "Silent",
      allow_writing_files = FALSE
    )
  )

  native_preds <- catboost_catboost.predict(model, pool)
  formula <- tidypredict_fit(model)
  tidy_preds <- rlang::eval_tidy(formula, mtcars)

  expect_equal(tidy_preds, native_preds, tolerance = 1e-10)
})

test_that("Huber predictions match catboost.predict (#188)", {
  skip_if_not_installed("catboost")

  model <- make_catboost_model(loss_function = "Huber:delta=1.0")
  pool <- catboost_catboost.load_pool(
    data.matrix(mtcars[, c("mpg", "cyl", "disp")]),
    label = mtcars$hp,
    feature_names = as.list(c("mpg", "cyl", "disp"))
  )

  native_preds <- catboost_catboost.predict(model, pool)
  formula <- tidypredict_fit(model)
  tidy_preds <- rlang::eval_tidy(formula, mtcars)

  expect_equal(tidy_preds, native_preds)
})

test_that("LogCosh predictions match catboost.predict (#188)", {
  skip_if_not_installed("catboost")

  model <- make_catboost_model(loss_function = "LogCosh")
  pool <- catboost_catboost.load_pool(
    data.matrix(mtcars[, c("mpg", "cyl", "disp")]),
    label = mtcars$hp,
    feature_names = as.list(c("mpg", "cyl", "disp"))
  )

  native_preds <- catboost_catboost.predict(model, pool)
  formula <- tidypredict_fit(model)
  tidy_preds <- rlang::eval_tidy(formula, mtcars)

  expect_equal(tidy_preds, native_preds)
})

test_that("Expectile predictions match catboost.predict (#188)", {
  skip_if_not_installed("catboost")

  model <- make_catboost_model(loss_function = "Expectile:alpha=0.5")
  pool <- catboost_catboost.load_pool(
    data.matrix(mtcars[, c("mpg", "cyl", "disp")]),
    label = mtcars$hp,
    feature_names = as.list(c("mpg", "cyl", "disp"))
  )

  native_preds <- catboost_catboost.predict(model, pool)
  formula <- tidypredict_fit(model)
  tidy_preds <- rlang::eval_tidy(formula, mtcars)

  expect_equal(tidy_preds, native_preds)
})

test_that("Tweedie predictions match catboost.predict (#188)", {
  skip_if_not_installed("catboost")

  model <- make_catboost_model(loss_function = "Tweedie:variance_power=1.5")
  pool <- catboost_catboost.load_pool(
    data.matrix(mtcars[, c("mpg", "cyl", "disp")]),
    label = mtcars$hp,
    feature_names = as.list(c("mpg", "cyl", "disp"))
  )

  native_preds <- catboost_catboost.predict(model, pool)
  formula <- tidypredict_fit(model)
  tidy_preds <- rlang::eval_tidy(formula, mtcars)

  expect_equal(tidy_preds, native_preds)
})

test_that("CrossEntropy predictions match catboost.predict", {
  skip_if_not_installed("catboost")

  set.seed(456)
  X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
  y <- as.numeric(mtcars$am)

  pool <- catboost_catboost.load_pool(
    X,
    label = y,
    feature_names = as.list(c("mpg", "cyl", "disp"))
  )

  model <- catboost_catboost.train(
    pool,
    params = list(
      iterations = 10L,
      depth = 3L,
      learning_rate = 0.5,
      loss_function = "CrossEntropy",
      logging_level = "Silent",
      allow_writing_files = FALSE
    )
  )

  native_preds <- catboost_catboost.predict(
    model,
    pool,
    prediction_type = "Probability"
  )
  formula <- tidypredict_fit(model)
  tidy_preds <- rlang::eval_tidy(formula, mtcars)

  expect_equal(tidy_preds, native_preds, tolerance = 1e-10)
})

test_that("stump trees (depth=0) predictions work correctly", {
  skip_if_not_installed("catboost")

  set.seed(123)
  X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
  y <- mtcars$hp

  pool <- catboost_catboost.load_pool(
    X,
    label = y,
    feature_names = as.list(c("mpg", "cyl", "disp"))
  )

  model <- catboost_catboost.train(
    pool,
    params = list(
      iterations = 3L,
      depth = 0L,
      learning_rate = 0.5,
      loss_function = "RMSE",
      logging_level = "Silent",
      allow_writing_files = FALSE
    )
  )

  formula <- tidypredict_fit(model)
  expect_type(formula, "language")

  native_preds <- catboost_catboost.predict(model, pool)
  tidy_preds <- rlang::eval_tidy(formula, mtcars)

  # Stump trees produce constant predictions
  # The formula returns a scalar when no data columns are referenced
  expect_equal(tidy_preds[[1]], native_preds[[1]], tolerance = 1e-10)
})

test_that("model with NaN values handles missing correctly with nan_mode Min", {
  skip_if_not_installed("catboost")

  set.seed(123)
  X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
  X[1, 1] <- NA
  y <- mtcars$hp

  pool <- catboost_catboost.load_pool(
    X,
    label = y,
    feature_names = as.list(c("mpg", "cyl", "disp"))
  )

  model <- catboost_catboost.train(
    pool,
    params = list(
      iterations = 5L,
      depth = 3L,
      learning_rate = 0.5,
      loss_function = "RMSE",
      logging_level = "Silent",
      allow_writing_files = FALSE,
      nan_mode = "Min"
    )
  )

  formula <- tidypredict_fit(model)
  expect_type(formula, "language")

  test_data <- mtcars
  test_data$mpg[1] <- NA

  native_preds <- catboost_catboost.predict(model, pool)
  tidy_preds <- rlang::eval_tidy(formula, test_data)

  expect_equal(tidy_preds, native_preds, tolerance = 1e-10)
})

test_that("model with NaN values handles missing correctly with nan_mode Max", {
  skip_if_not_installed("catboost")

  set.seed(123)
  X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
  X[1, 1] <- NA
  y <- mtcars$hp

  pool <- catboost_catboost.load_pool(
    X,
    label = y,
    feature_names = as.list(c("mpg", "cyl", "disp"))
  )

  model <- catboost_catboost.train(
    pool,
    params = list(
      iterations = 5L,
      depth = 3L,
      learning_rate = 0.5,
      loss_function = "RMSE",
      logging_level = "Silent",
      allow_writing_files = FALSE,
      nan_mode = "Max"
    )
  )

  formula <- tidypredict_fit(model)
  expect_type(formula, "language")

  test_data <- mtcars
  test_data$mpg[1] <- NA

  native_preds <- catboost_catboost.predict(model, pool)
  tidy_preds <- rlang::eval_tidy(formula, test_data)

  expect_equal(tidy_preds, native_preds, tolerance = 1e-10)
})

test_that("unsupported objective throws error", {
  skip_if_not_installed("catboost")

  pm <- list(
    general = list(
      params = list(objective = "UnsupportedObjective"),
      model = "catboost.Model",
      type = "catboost"
    ),
    trees = list(list(list(prediction = 1, path = list())))
  )
  class(pm) <- c("pm_catboost", "parsed_model", "list")

  expect_snapshot(tidypredict_fit(pm), error = TRUE)
})

test_that("empty trees throws error", {
  skip_if_not_installed("catboost")

  pm <- list(
    general = list(
      params = list(objective = "RMSE"),
      model = "catboost.Model",
      type = "catboost"
    ),
    trees = list()
  )
  class(pm) <- c("pm_catboost", "parsed_model", "list")

  expect_snapshot(tidypredict_fit(pm), error = TRUE)
})

# SQL generation tests ----------------------------------------------------

test_that("tidypredict_sql returns SQL class", {
  skip_if_not_installed("catboost")
  model <- make_catboost_model()

  sql <- tidypredict_sql(model, dbplyr::simulate_sqlite())

  expect_s3_class(sql, "sql")
})

test_that("SQL predictions match native predictions with SQLite", {
  skip_if_not_installed("catboost")
  skip_if_not_installed("DBI")
  skip_if_not_installed("RSQLite")

  model <- make_catboost_model()

  X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
  pool <- catboost_catboost.load_pool(X)
  native_preds <- catboost_catboost.predict(model, pool)

  con <- DBI::dbConnect(RSQLite::SQLite(), ":memory:")
  on.exit(DBI::dbDisconnect(con), add = TRUE)
  DBI::dbWriteTable(con, "mtcars", mtcars)

  sql <- tidypredict_sql(model, con)
  query <- paste0("SELECT ", sql, " AS pred FROM mtcars")
  sql_preds <- DBI::dbGetQuery(con, query)$pred

  expect_equal(sql_preds, native_preds, tolerance = 1e-10)
})

# Integration tests -------------------------------------------------------

test_that("tidypredict_test works for regression", {
  skip_if_not_installed("catboost")
  model <- make_catboost_model()
  X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])

  result <- tidypredict_test(model, xg_df = X)

  expect_s3_class(result, "tidypredict_test")
  expect_false(result$alert)
})

test_that("tidypredict_test works for binary classification", {
  skip_if_not_installed("catboost")

  set.seed(456)
  X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
  y <- as.numeric(mtcars$am)

  pool <- catboost_catboost.load_pool(
    X,
    label = y,
    feature_names = as.list(c("mpg", "cyl", "disp"))
  )

  model <- catboost_catboost.train(
    pool,
    params = list(
      iterations = 10L,
      depth = 3L,
      learning_rate = 0.5,
      loss_function = "Logloss",
      logging_level = "Silent",
      allow_writing_files = FALSE
    )
  )

  result <- tidypredict_test(model, xg_df = X)

  expect_s3_class(result, "tidypredict_test")
  expect_false(result$alert)
})

test_that("tidypredict_test requires matrix", {
  skip_if_not_installed("catboost")
  model <- make_catboost_model()

  expect_snapshot(tidypredict_test(model), error = TRUE)
})

test_that(".extract_catboost_trees returns list of expressions", {
  skip_if_not_installed("catboost")
  model <- make_catboost_model()

  trees <- .extract_catboost_trees(model)

  expect_type(trees, "list")
  expect_length(trees, 10)
  expect_type(trees[[1]], "language")
})

test_that(".extract_catboost_trees combined results match tidypredict_fit", {
  skip_if_not_installed("catboost")
  model <- make_catboost_model()
  test_data <- mtcars[, c("mpg", "cyl", "disp")]

  trees <- .extract_catboost_trees(model)
  eval_env <- rlang::new_environment(
    data = as.list(test_data),
    parent = asNamespace("dplyr")
  )
  tree_preds <- lapply(trees, rlang::eval_tidy, env = eval_env)
  pm <- parse_model(model)
  scale <- pm$general$scale %||% 1
  bias <- pm$general$bias %||% 0
  combined <- Reduce(`+`, tree_preds) * scale + bias

  fit_result <- rlang::eval_tidy(tidypredict_fit(model), test_data)

  expect_equal(combined, fit_result)
})

test_that(".extract_catboost_trees errors on non-catboost model", {
  expect_snapshot(
    .extract_catboost_trees(lm(mpg ~ wt, data = mtcars)),
    error = TRUE
  )
})

# YAML serialization tests ------------------------------------------------

test_that("parsed model can be saved and loaded via YAML", {
  skip_if_not_installed("catboost")
  skip_if_not_installed("yaml")

  model <- make_catboost_model()
  pm <- parse_model(model)

  tmp_file <- withr::local_tempfile(fileext = ".yml")
  yaml::write_yaml(pm, tmp_file)
  loaded <- yaml::read_yaml(tmp_file)
  class(loaded) <- class(pm)

  expect_equal(loaded$general$model, pm$general$model)
  expect_equal(loaded$general$type, pm$general$type)
  expect_equal(loaded$general$niter, pm$general$niter)
})

test_that("loaded model produces same predictions", {
  skip_if_not_installed("catboost")
  skip_if_not_installed("yaml")

  model <- make_catboost_model()
  pm <- parse_model(model)

  tmp_file <- withr::local_tempfile(fileext = ".yml")
  yaml::write_yaml(pm, tmp_file)
  loaded <- yaml::read_yaml(tmp_file)
  class(loaded) <- class(pm)

  original_preds <- rlang::eval_tidy(tidypredict_fit(pm), mtcars)
  loaded_preds <- rlang::eval_tidy(tidypredict_fit(loaded), mtcars)

  expect_equal(loaded_preds, original_preds, tolerance = 1e-6)
})

# Multiclass tests ---------------------------------------------------------

test_that("MultiClass predictions match catboost.predict", {
  skip_if_not_installed("catboost")

  model <- make_multiclass_model("MultiClass")
  X <- data.matrix(iris[, 1:4])
  pool <- catboost_catboost.load_pool(X)

  native_preds <- catboost_catboost.predict(
    model,
    pool,
    prediction_type = "Probability"
  )

  formulas <- tidypredict_fit(model)
  tidy_preds <- lapply(formulas, function(f) rlang::eval_tidy(f, iris))
  tidy_matrix <- do.call(cbind, tidy_preds)

  expect_equal(unname(tidy_matrix), unname(native_preds), tolerance = 1e-10)
})

test_that("MultiClassOneVsAll predictions match catboost.predict", {
  skip_if_not_installed("catboost")

  model <- make_multiclass_model("MultiClassOneVsAll")
  X <- data.matrix(iris[, 1:4])
  pool <- catboost_catboost.load_pool(X)

  native_preds <- catboost_catboost.predict(
    model,
    pool,
    prediction_type = "Probability"
  )

  formulas <- tidypredict_fit(model)
  tidy_preds <- lapply(formulas, function(f) rlang::eval_tidy(f, iris))
  tidy_matrix <- do.call(cbind, tidy_preds)

  expect_equal(unname(tidy_matrix), unname(native_preds), tolerance = 1e-10)
})

test_that("multiclass output is list with correct names", {
  skip_if_not_installed("catboost")

  model <- make_multiclass_model()
  formulas <- tidypredict_fit(model)

  expect_type(formulas, "list")
  expect_length(formulas, 3)
  expect_named(formulas, c("class_0", "class_1", "class_2"))
})

test_that("MultiClass probabilities sum to 1", {
  skip_if_not_installed("catboost")

  model <- make_multiclass_model("MultiClass")
  formulas <- tidypredict_fit(model)

  tidy_preds <- lapply(formulas, function(f) rlang::eval_tidy(f, iris))
  row_sums <- Reduce(`+`, tidy_preds)

  expect_equal(row_sums, rep(1, nrow(iris)), tolerance = 1e-10)
})

test_that("tidypredict_test works for multiclass", {
  skip_if_not_installed("catboost")

  model <- make_multiclass_model()
  X <- data.matrix(iris[, 1:4])

  result <- tidypredict_test(model, xg_df = X)

  expect_s3_class(result, "tidypredict_test")
  expect_false(result$alert)
})

test_that("multiclass model requires num_class >= 2", {
  skip_if_not_installed("catboost")

  pm <- list(
    general = list(
      params = list(objective = "MultiClass"),
      model = "catboost.Model",
      type = "catboost",
      num_class = 1
    ),
    trees = list(list(list(prediction = 1, path = list())))
  )
  class(pm) <- c("pm_catboost", "parsed_model", "list")

  expect_snapshot(tidypredict_fit(pm), error = TRUE)
})

test_that("multiclass stump trees (depth=0) work correctly", {
  skip_if_not_installed("catboost")

  set.seed(42)
  X <- data.matrix(iris[, 1:4])
  y <- as.integer(iris$Species) - 1L

  pool <- catboost_catboost.load_pool(
    X,
    label = y,
    feature_names = as.list(colnames(iris)[1:4])
  )

  model <- catboost_catboost.train(
    pool,
    params = list(
      iterations = 3L,
      depth = 0L,
      learning_rate = 0.5,
      loss_function = "MultiClass",
      logging_level = "Silent",
      allow_writing_files = FALSE
    )
  )

  pm <- parse_model(model)
  expect_equal(pm$general$num_class, 3)

  formulas <- tidypredict_fit(model)
  expect_type(formulas, "list")
  expect_length(formulas, 3)
})

# Categorical feature tests -----------------------------------------------

test_that("set_catboost_categories adds hash mapping", {
  skip_if_not_installed("catboost")

  set.seed(42)
  df <- data.frame(
    num_feat = rnorm(100),
    cat_feat = factor(sample(c("A", "B", "C"), 100, replace = TRUE)),
    target = rnorm(100)
  )

  X <- df[, c("num_feat", "cat_feat")]
  y <- df$target

  pool <- catboost_catboost.load_pool(X, label = y)

  model <- catboost_catboost.train(
    pool,
    params = list(
      iterations = 10L,
      depth = 3L,
      learning_rate = 0.5,
      loss_function = "RMSE",
      logging_level = "Silent",
      allow_writing_files = FALSE,
      one_hot_max_size = 10
    )
  )

  pm <- parse_model(model)
  pm <- set_catboost_categories(pm, model, df)

  expect_type(pm$general$cat_features[[1]]$hash_to_category, "list")
  expect_contains(
    unlist(pm$general$cat_features[[1]]$hash_to_category),
    c("A", "B", "C")
  )
})

test_that("set_catboost_categories validates parsed_model argument", {
  skip_if_not_installed("catboost")

  model <- make_catboost_model()

  expect_snapshot(
    set_catboost_categories("not a parsed model", model, data.frame()),
    error = TRUE
  )
})

test_that("set_catboost_categories validates model argument", {
  skip_if_not_installed("catboost")

  model <- make_catboost_model()
  pm <- parse_model(model)

  expect_snapshot(
    set_catboost_categories(pm, "not a model", data.frame()),
    error = TRUE
  )
})

test_that("set_catboost_categories returns early when no categorical features", {
  skip_if_not_installed("catboost")

  model <- make_catboost_model()
  pm <- parse_model(model)

  result <- set_catboost_categories(pm, model, mtcars)

  expect_identical(result, pm)
})

test_that("set_catboost_categories errors when column not found in data", {
  skip_if_not_installed("catboost")

  model <- make_categorical_model()
  pm <- parse_model(model)

  wrong_data <- data.frame(num_feat = 1, other_col = factor("A"))

  expect_snapshot(set_catboost_categories(pm, model, wrong_data), error = TRUE)
})

test_that("set_catboost_categories errors when column is not a factor", {
  skip_if_not_installed("catboost")

  model <- make_categorical_model()
  pm <- parse_model(model)

  wrong_data <- data.frame(num_feat = 1, cat_feat = "A")

  expect_snapshot(set_catboost_categories(pm, model, wrong_data), error = TRUE)
})

test_that("categorical predictions match catboost.predict", {
  skip_if_not_installed("catboost")

  set.seed(42)
  df <- data.frame(
    num_feat = rnorm(100),
    cat_feat = factor(sample(c("A", "B", "C"), 100, replace = TRUE)),
    target = rnorm(100)
  )

  X <- df[, c("num_feat", "cat_feat")]
  y <- df$target

  pool <- catboost_catboost.load_pool(X, label = y)

  model <- catboost_catboost.train(
    pool,
    params = list(
      iterations = 10L,
      depth = 3L,
      learning_rate = 0.5,
      loss_function = "RMSE",
      logging_level = "Silent",
      allow_writing_files = FALSE,
      one_hot_max_size = 10
    )
  )

  pm <- parse_model(model)
  pm <- set_catboost_categories(pm, model, df)

  formula <- tidypredict_fit(pm)
  tidy_preds <- rlang::eval_tidy(formula, df)

  native_preds <- catboost_catboost.predict(model, pool)

  expect_equal(tidy_preds, native_preds, tolerance = 1e-10)
})

test_that("categorical features without mapping throws error", {
  skip_if_not_installed("catboost")

  model <- make_categorical_model()
  pm <- parse_model(model)

  expect_snapshot(tidypredict_fit(pm), error = TRUE)
})

test_that("categorical features SQL generation works", {
  skip_if_not_installed("catboost")

  set.seed(42)
  df <- data.frame(
    num_feat = rnorm(100),
    cat_feat = factor(sample(c("A", "B", "C"), 100, replace = TRUE)),
    target = rnorm(100)
  )

  X <- df[, c("num_feat", "cat_feat")]
  y <- df$target

  pool <- catboost_catboost.load_pool(X, label = y)

  model <- catboost_catboost.train(
    pool,
    params = list(
      iterations = 10L,
      depth = 3L,
      learning_rate = 0.5,
      loss_function = "RMSE",
      logging_level = "Silent",
      allow_writing_files = FALSE,
      one_hot_max_size = 10
    )
  )

  pm <- parse_model(model)
  pm <- set_catboost_categories(pm, model, df)

  sql <- tidypredict_sql(pm, dbplyr::simulate_dbi())

  expect_s3_class(sql, "sql")
  sql_str <- as.character(sql)
  expect_match(sql_str, "cat_feat")
  expect_match(sql_str, "=")
})

# Parsnip/bonsai tests -----------------------------------------------

test_that("tidypredict works with parsnip/bonsai catboost regression", {
  skip_if_not_installed("catboost")
  skip_if_not_installed("parsnip")
  skip_if_not_installed("bonsai")

  set.seed(123)
  train_data <- mtcars[, c("hp", "mpg", "cyl", "disp")]

  model_spec <- parsnip::boost_tree(
    trees = 5,
    tree_depth = 3,
    min_n = 1
  ) |>
    parsnip::set_engine("catboost", logging_level = "Silent") |>
    parsnip::set_mode("regression")

  model_fit <- parsnip::fit(
    model_spec,
    hp ~ mpg + cyl + disp,
    data = train_data
  )

  cb_model <- model_fit$fit

  expect_s3_class(cb_model, "catboost.Model")

  pm <- parse_model(cb_model)
  expect_s3_class(pm, "parsed_model")
  expect_s3_class(pm, "pm_catboost")
  expect_gt(length(pm$trees), 0)

  fit_formula <- tidypredict_fit(cb_model)
  expect_type(fit_formula, "language")

  X <- data.matrix(train_data[, c("mpg", "cyl", "disp")])
  pool <- catboost_catboost.load_pool(X)
  native_preds <- catboost_catboost.predict(cb_model, pool)
  tidy_preds <- rlang::eval_tidy(fit_formula, train_data)

  expect_equal(tidy_preds, native_preds, tolerance = 1e-10)
})

test_that("tidypredict works with parsnip/bonsai catboost classification", {
  skip_if_not_installed("catboost")
  skip_if_not_installed("parsnip")
  skip_if_not_installed("bonsai")

  set.seed(456)
  train_data <- mtcars[, c("am", "mpg", "cyl", "disp")]
  train_data$am <- factor(train_data$am)

  model_spec <- parsnip::boost_tree(
    trees = 5,
    tree_depth = 3,
    min_n = 1
  ) |>
    parsnip::set_engine("catboost", logging_level = "Silent") |>
    parsnip::set_mode("classification")

  model_fit <- parsnip::fit(
    model_spec,
    am ~ mpg + cyl + disp,
    data = train_data
  )

  cb_model <- model_fit$fit

  expect_s3_class(cb_model, "catboost.Model")

  fit_formula <- tidypredict_fit(cb_model)
  expect_type(fit_formula, "language")

  X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
  pool <- catboost_catboost.load_pool(X)
  native_preds <- catboost_catboost.predict(
    cb_model,
    pool,
    prediction_type = "Probability"
  )
  tidy_preds <- rlang::eval_tidy(fit_formula, mtcars)

  expect_equal(tidy_preds, native_preds, tolerance = 1e-10)
})

test_that("tidypredict_sql works with parsnip/bonsai catboost model", {
  skip_if_not_installed("catboost")
  skip_if_not_installed("parsnip")
  skip_if_not_installed("bonsai")
  skip_if_not_installed("dbplyr")

  set.seed(123)
  train_data <- mtcars[, c("hp", "mpg", "cyl", "disp")]

  model_spec <- parsnip::boost_tree(
    trees = 3,
    tree_depth = 2,
    min_n = 1
  ) |>
    parsnip::set_engine("catboost", logging_level = "Silent") |>
    parsnip::set_mode("regression")

  model_fit <- parsnip::fit(
    model_spec,
    hp ~ mpg + cyl + disp,
    data = train_data
  )
  cb_model <- model_fit$fit

  sql_result <- tidypredict_sql(cb_model, dbplyr::simulate_dbi())

  expect_s3_class(sql_result, "sql")
})

test_that("parsnip/bonsai catboost with categorical features works automatically", {
  skip_if_not_installed("catboost")
  skip_if_not_installed("parsnip")
  skip_if_not_installed("bonsai")

  set.seed(42)
  train_data <- data.frame(
    num_feat = rnorm(100),
    cat_feat = factor(sample(c("A", "B", "C"), 100, replace = TRUE)),
    target = rnorm(100)
  )

  model_spec <- parsnip::boost_tree(
    trees = 5,
    tree_depth = 3,
    min_n = 1
  ) |>
    parsnip::set_engine(
      "catboost",
      logging_level = "Silent",
      one_hot_max_size = 10
    ) |>
    parsnip::set_mode("regression")

  model_fit <- parsnip::fit(
    model_spec,
    target ~ num_feat + cat_feat,
    data = train_data
  )

  fit_formula <- tidypredict_fit(model_fit)
  expect_type(fit_formula, "language")

  parsnip_preds <- predict(model_fit, train_data)$.pred
  tidy_preds <- rlang::eval_tidy(fit_formula, train_data)

  expect_equal(tidy_preds, parsnip_preds, tolerance = 1e-6)
})

test_that("parsnip/bonsai catboost categorical SQL generation works automatically", {
  skip_if_not_installed("catboost")
  skip_if_not_installed("parsnip")
  skip_if_not_installed("bonsai")
  skip_if_not_installed("dbplyr")

  set.seed(42)
  train_data <- data.frame(
    num_feat = rnorm(100),
    cat_feat = factor(sample(c("X", "Y", "Z"), 100, replace = TRUE)),
    target = rnorm(100)
  )

  model_spec <- parsnip::boost_tree(
    trees = 5,
    tree_depth = 3,
    min_n = 1
  ) |>
    parsnip::set_engine(
      "catboost",
      logging_level = "Silent",
      one_hot_max_size = 10
    ) |>
    parsnip::set_mode("regression")

  model_fit <- parsnip::fit(
    model_spec,
    target ~ num_feat + cat_feat,
    data = train_data
  )

  sql <- tidypredict_sql(model_fit, dbplyr::simulate_dbi())

  expect_s3_class(sql, "sql")
  sql_str <- as.character(sql)
  expect_match(sql_str, "cat_feat")
})

test_that("multiple categorical features work correctly", {
  skip_if_not_installed("catboost")
  skip_if_not_installed("parsnip")
  skip_if_not_installed("bonsai")

  set.seed(42)
  df <- data.frame(
    num_feat = rnorm(100),
    cat_feat1 = factor(sample(c("A", "B", "C"), 100, replace = TRUE)),
    cat_feat2 = factor(sample(c("X", "Y", "Z"), 100, replace = TRUE)),
    target = rnorm(100)
  )

  model_spec <- parsnip::boost_tree(
    trees = 10,
    tree_depth = 3,
    min_n = 1
  ) |>
    parsnip::set_engine(
      "catboost",
      logging_level = "Silent",
      one_hot_max_size = 10
    ) |>
    parsnip::set_mode("regression")

  model_fit <- parsnip::fit(
    model_spec,
    target ~ num_feat + cat_feat1 + cat_feat2,
    data = df
  )

  pm <- parse_model(model_fit$fit)

  expect_length(pm$general$cat_features, 2)
  expect_equal(pm$general$cat_feature_names, c("cat_feat1", "cat_feat2"))

  formula <- tidypredict_fit(model_fit)
  tidy_preds <- rlang::eval_tidy(formula, df)

  parsnip_preds <- predict(model_fit, df)$.pred

  expect_equal(tidy_preds, parsnip_preds, tolerance = 1e-6)
})

test_that("parsnip categorical predictions match parsnip predictions", {
  skip_if_not_installed("catboost")
  skip_if_not_installed("parsnip")
  skip_if_not_installed("bonsai")

  set.seed(42)
  train_data <- data.frame(
    num_feat = rnorm(100),
    cat_feat = factor(sample(c("A", "B", "C"), 100, replace = TRUE)),
    target = rnorm(100)
  )

  model_spec <- parsnip::boost_tree(
    trees = 5,
    tree_depth = 3,
    min_n = 1
  ) |>
    parsnip::set_engine(
      "catboost",
      logging_level = "Silent",
      one_hot_max_size = 10
    ) |>
    parsnip::set_mode("regression")

  model_fit <- parsnip::fit(
    model_spec,
    target ~ num_feat + cat_feat,
    data = train_data
  )

  formula <- tidypredict_fit(model_fit)
  tidy_preds <- rlang::eval_tidy(formula, train_data)

  parsnip_preds <- predict(model_fit, train_data)$.pred

  expect_equal(tidy_preds, parsnip_preds, tolerance = 1e-6)
})

test_that("parsnip model without xlevels throws error", {
  skip_if_not_installed("catboost")
  skip_if_not_installed("parsnip")
  skip_if_not_installed("bonsai")

  set.seed(42)
  train_data <- data.frame(
    num_feat = rnorm(100),
    cat_feat = factor(sample(c("A", "B", "C"), 100, replace = TRUE)),
    target = rnorm(100)
  )

  model_spec <- parsnip::boost_tree(
    trees = 5,
    tree_depth = 3,
    min_n = 1
  ) |>
    parsnip::set_engine(
      "catboost",
      logging_level = "Silent",
      one_hot_max_size = 10
    ) |>
    parsnip::set_mode("regression")

  model_fit <- parsnip::fit(
    model_spec,
    target ~ num_feat + cat_feat,
    data = train_data
  )

  model_fit$preproc$xlevels <- NULL

  expect_snapshot(tidypredict_fit(model_fit), error = TRUE)
})

test_that("model with only categorical features works", {
  skip_if_not_installed("catboost")
  skip_if_not_installed("parsnip")
  skip_if_not_installed("bonsai")

  set.seed(42)
  df <- data.frame(
    cat_feat1 = factor(sample(c("A", "B", "C"), 100, replace = TRUE)),
    cat_feat2 = factor(sample(c("X", "Y", "Z"), 100, replace = TRUE)),
    target = rnorm(100)
  )

  model_spec <- parsnip::boost_tree(
    trees = 5,
    tree_depth = 3,
    min_n = 1
  ) |>
    parsnip::set_engine(
      "catboost",
      logging_level = "Silent",
      one_hot_max_size = 10
    ) |>
    parsnip::set_mode("regression")

  model_fit <- parsnip::fit(
    model_spec,
    target ~ cat_feat1 + cat_feat2,
    data = df
  )

  pm <- parse_model(model_fit$fit)

  expect_length(pm$general$feature_names, 0)
  expect_length(pm$general$cat_features, 2)

  formula <- tidypredict_fit(model_fit)
  tidy_preds <- rlang::eval_tidy(formula, df)

  parsnip_preds <- predict(model_fit, df)$.pred

  expect_equal(tidy_preds, parsnip_preds, tolerance = 1e-6)
})

# Non-oblivious tree tests (Depthwise/Lossguide) -----------------------------

test_that("Depthwise regression predictions match (#187)", {
  skip_if_not_installed("catboost")

  set.seed(123)
  X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
  y <- mtcars$hp

  pool <- catboost_catboost.load_pool(
    X,
    label = y,
    feature_names = as.list(c("mpg", "cyl", "disp"))
  )

  model <- catboost_catboost.train(
    pool,
    params = list(
      iterations = 5L,
      depth = 3L,
      learning_rate = 0.5,
      loss_function = "RMSE",
      grow_policy = "Depthwise",
      logging_level = "Silent",
      allow_writing_files = FALSE
    )
  )

  pm <- parse_model(model)
  expect_equal(pm$general$tree_type, "nonoblivious")

  native_preds <- catboost_catboost.predict(model, pool)
  formula <- tidypredict_fit(model)
  tidy_preds <- rlang::eval_tidy(formula, mtcars)

  expect_equal(tidy_preds, native_preds, tolerance = 1e-10)
})

test_that("non-oblivious stump trees (depth=0) work correctly", {
  skip_if_not_installed("catboost")

  set.seed(123)
  X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
  y <- mtcars$hp

  pool <- catboost_catboost.load_pool(
    X,
    label = y,
    feature_names = as.list(c("mpg", "cyl", "disp"))
  )

  model <- catboost_catboost.train(
    pool,
    params = list(
      iterations = 3L,
      depth = 0L,
      learning_rate = 0.5,
      loss_function = "RMSE",
      grow_policy = "Depthwise",
      logging_level = "Silent",
      allow_writing_files = FALSE
    )
  )

  pm <- parse_model(model)
  expect_equal(pm$general$tree_type, "nonoblivious")

  formula <- tidypredict_fit(model)
  expect_type(formula, "language")

  native_preds <- catboost_catboost.predict(model, pool)
  tidy_preds <- rlang::eval_tidy(formula, mtcars)
  expect_equal(tidy_preds[[1]], native_preds[[1]], tolerance = 1e-10)
})

test_that("Lossguide regression predictions match (#187)", {
  skip_if_not_installed("catboost")

  set.seed(123)
  X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
  y <- mtcars$hp

  pool <- catboost_catboost.load_pool(
    X,
    label = y,
    feature_names = as.list(c("mpg", "cyl", "disp"))
  )

  model <- catboost_catboost.train(
    pool,
    params = list(
      iterations = 5L,
      depth = 3L,
      learning_rate = 0.5,
      loss_function = "RMSE",
      grow_policy = "Lossguide",
      logging_level = "Silent",
      allow_writing_files = FALSE
    )
  )

  pm <- parse_model(model)
  expect_equal(pm$general$tree_type, "nonoblivious")

  native_preds <- catboost_catboost.predict(model, pool)
  formula <- tidypredict_fit(model)
  tidy_preds <- rlang::eval_tidy(formula, mtcars)

  expect_equal(tidy_preds, native_preds, tolerance = 1e-10)
})

test_that("Depthwise multiclass predictions match (#187)", {
  skip_if_not_installed("catboost")

  set.seed(123)
  X <- data.matrix(iris[, 1:4])
  colnames(X) <- c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width")
  y <- as.integer(iris$Species) - 1L

  pool <- catboost_catboost.load_pool(
    X,
    label = y,
    feature_names = as.list(colnames(X))
  )

  model <- catboost_catboost.train(
    pool,
    params = list(
      iterations = 3L,
      depth = 2L,
      learning_rate = 0.5,
      loss_function = "MultiClass",
      grow_policy = "Depthwise",
      logging_level = "Silent",
      allow_writing_files = FALSE
    )
  )

  pm <- parse_model(model)
  expect_equal(pm$general$tree_type, "nonoblivious")
  expect_equal(pm$general$num_class, 3)

  native_preds <- catboost_catboost.predict(
    model,
    pool,
    prediction_type = "Probability"
  )

  formulas <- tidypredict_fit(model)
  tidy_preds <- lapply(formulas, function(f) rlang::eval_tidy(f, iris))
  tidy_matrix <- do.call(cbind, tidy_preds)

  expect_equal(unname(tidy_matrix), unname(native_preds), tolerance = 1e-10)
})

test_that("Depthwise binary classification predictions match (#187)", {
  skip_if_not_installed("catboost")

  set.seed(456)
  X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
  y <- as.numeric(mtcars$am)

  pool <- catboost_catboost.load_pool(
    X,
    label = y,
    feature_names = as.list(c("mpg", "cyl", "disp"))
  )

  model <- catboost_catboost.train(
    pool,
    params = list(
      iterations = 5L,
      depth = 3L,
      learning_rate = 0.5,
      loss_function = "Logloss",
      grow_policy = "Depthwise",
      logging_level = "Silent",
      allow_writing_files = FALSE
    )
  )

  native_preds <- catboost_catboost.predict(
    model,
    pool,
    prediction_type = "Probability"
  )
  formula <- tidypredict_fit(model)
  tidy_preds <- rlang::eval_tidy(formula, mtcars)

  expect_equal(tidy_preds, native_preds, tolerance = 1e-10)
})

#' Generate trees
#'
#' Each tree is generated as a flat tree with each node being a seperate part of
#' the case when.
#' This means that the following tree:
#'
#'             +-----+
#'        +----|x > 0|----+
#'        |    +-----+    |
#'        v               v
#'    +------+        +--------+
#' +--|y < 20|--+  +--|z <= 10 |--+
#' |  +------+  |  |  +--------+  |
#' v            v  v              v
#' a            b  c              d
#'
#' will be turned into the following `case_when()` statement.
#'
#' ```r
#' case_when(
#'   x >  0 & y <  20 ~ "a",
#'   x >  0 & y >= 20 ~ "b",
#'   x <= 0 & z <= 10 ~ "c",
#'   x <= 0 & z >  10 ~ "d"
#' )
#' ```
#'
#' instead of a nested `case_when()`s` like this
#'
#' ```r
#' case_when(
#'   x >  0 ~ case_when(
#'              y <  20 ~ "a",
#'              y >= 10 ~ "b"
#'            ),
#'   x <= 0 ~ case_when(
#'              z <= 10 ~ "c",
#'              z >  10 ~ "d"
#'            )
#' )
#' ```
#'
#' The functions in this file generates these tree.
#' `generate_case_when_tree()` generates a single tree with
#' `generate_case_when_trees()` being a convinience wrapper for multiple trees.
#'
#' `generate_tree_node()` generates the expressions for each a single ndoe in
#' the tree, where `generate_tree_nodes()` is a convinience wrapper for
#' calculating all notes.
#'
#' @keywords internal
generate_case_when_trees <- function(parsedmodel, default = TRUE) {
  map(
    parsedmodel$trees,
    generate_case_when_tree,
    mode = parsedmodel$general$mode,
    default = default
  )
}

generate_case_when_tree <- function(tree, mode, default = TRUE) {
  nodes <- generate_tree_nodes(tree, mode)

  if (default) {
    default <- nodes[[length(nodes)]]
    default <- rlang::f_rhs(default)
    nodes[[length(nodes)]] <- NULL
    nodes <- c(nodes, .default = default)
  }

  expr(case_when(!!!nodes))
}

generate_tree_nodes <- function(tree, mode) {
  map(tree, generate_tree_node, mode)
}

#' Construct a single node of a tree
#'
#' @param node a list with named elements `path` and `prediction`. See details
#' for more.
#' @param calc_mode character, takes values `""` and `"calc_mode"`.
#'
#' The `node` list should contain the two lists `path` and `prediction`.
#'
#' The `path` element has the following structure:
#'
#' This list can contain 0 or more elemements. The elements but each be of the
#' following format:
#'
#' - `type` character, must be `"conditional"`, `"set"`, or `"all"`.
#' - `op` character.
#'   if `type == "conditional"` must be `"more"`, `"more-equal"`, `"less"`, or
#'   `"less-equal"`.
#'   if `type == "set"` must be `"in"` on `not-in`.
#' - `col` character.
#' - `val` if `type == "conditional"` and `vals` if `type == "set"`.
#'   Can be character or numeric.
#'
#' The `prediction` list has the following structure:
#'
#' It can either be a singular value or a list.
#' If it is a list it will have the following 4 named elements `col`, `val`,
#' `op`, and `is_intercept`.
#'
#' - `col` character, name of column
#' - `val` val, numeric of character
#' - `op` character, known values are `"none"` and `"multiply"`. `"none"` is
#'   used then `is_intercept == 1`.
#' - `is_intercept`integer, takes values `0` and `1`.`
#'
#'  @keywords internal
generate_tree_node <- function(node, calc_mode = "") {
  path <- node$path
  prediction <- node$prediction
  rcl <- path_formulas(path)

  if (length(prediction) > 1) {
    pl <- map(
      prediction,
      ~ {
        if (.x$is_intercept) {
          if (.x$val == 0) {
            return(NULL)
          }
          return(expr(!!.x$val))
        } else if (.x$op == "multiply") {
          if (.x$val == 0) {
            return(NULL)
          }

          if (.x$val == 1) {
            return(expr(!!as.name(.x$col)))
          }

          return(expr_multiplication(as.name(.x$col), .x$val))
        }
      }
    )
    pl <- purrr::discard(pl, is.null)
    pl <- reduce_addition(pl)
  } else {
    if (is.list(prediction) && prediction[[1]]$is_intercept) {
      prediction <- prediction[[1]]$val
    }
    pl <- prediction
  }

  if (isTRUE(rcl)) {
    return(pl)
  }

  if (calc_mode == "ifelse") {
    return(expr(ifelse(!!rcl, !!pl, 0)))
  }

  expr(!!rcl ~ !!pl)
}

#' Turn a path object into a combined expression
#'
#' @param path a list of lists.
#'
#' This list can contain 0 or more elemements. The elements but each be of the
#' following format:
#'
#' - `type` character, must be `"conditional"`, `"set"`, or `"all"`.
#' - `op` character.
#'   if `type == "conditional"` must be `"more"`, `"more-equal"`, `"less"`, or
#'   `"less-equal"`.
#'   if `type == "set"` must be `"in"` on `not-in`.
#' - `col` character.
#' - `val` if `type == "conditional"` and `vals` if `type == "set"`.
#'   Can be character or numeric.
#'  @keywords internal
path_formulas <- function(path) {
  if (length(path) == 0) {
    return(TRUE)
  }

  if (length(path) == 1 && path[[1]]$type == "all") {
    return(TRUE)
  }

  cl <- map(path, path_formula)
  res <- reduce_and(cl)
  res
}

#' Turn a path object into an expression
#'
#' @param x a list.
#'
#' The input of this function is a list with 4 values.
#'
#' - `type` character, must be `"conditional"` or `"set"`.
#' - `op` character.
#'   if `type == "conditional"` must be `"more"`, `"more-equal"`, `"less"`, or
#'   `"less-equal"`.
#'   if `type == "set"` must be `"in"` on `not-in`.
#' - `col` character.
#' - `val` if `type == "conditional"` and `vals` if `type == "set"`.
#'   Can be character or numeric.
#'  @keywords internal
path_formula <- function(x) {
  type <- x$type

  if (type == "conditional") {
    i <- build_comparison_expr(x$col, x$val, x$op)
  } else if (type == "set") {
    i <- build_set_expr(x$col, x$vals, x$op)
  } else {
    cli::cli_abort(
      "{.field type} has unsupported value of {.value {x$type}}.",
      .internal = TRUE
    )
  }
  i
}

# Build a comparison expression (col <op> val)
build_comparison_expr <- function(col, val, op) {
  col <- as.name(col)
  if (op == "more") {
    expr(!!col > !!val)
  } else if (op == "more-equal") {
    expr(!!col >= !!val)
  } else if (op == "less") {
    expr(!!col < !!val)
  } else if (op == "less-equal") {
    expr(!!col <= !!val)
  } else {
    cli::cli_abort(
      "{.field op} has unsupported value of {.value {op}}.",
      .internal = TRUE
    )
  }
}

# Build a set membership expression (col %in% vals or not)
build_set_expr <- function(col, vals, op) {
  col <- as.name(col)
  sets <- reduce(vals, c)
  if (op == "in") {
    expr(!!col %in% !!sets)
  } else if (op == "not-in") {
    expr((!!col %in% !!sets) == FALSE)
  } else {
    cli::cli_abort(
      "{.field op} has unsupported value of {.value {op}}.",
      .internal = TRUE
    )
  }
}

# For {orbital}
#' Build case_when expression from nodes with predictions and paths
#'
#' Shared helper for building tree expressions used by ranger and randomForest
#' classification extractors.
#'
#' @param nodes A list of lists, each with `prediction` (numeric) and `path`
#' @keywords internal
#' @export
.build_case_when_tree <- function(nodes) {
  node_exprs <- map(nodes, function(node) {
    rcl <- path_formulas(node$path)
    if (isTRUE(rcl)) {
      return(node$prediction)
    }
    expr(!!rcl ~ !!node$prediction)
  })

  # Handle stump trees (single node with no conditions)
  if (length(node_exprs) == 1 && is.numeric(node_exprs[[1]])) {
    return(node_exprs[[1]])
  }

  default <- node_exprs[[length(node_exprs)]]
  if (rlang::is_formula(default)) {
    default <- rlang::f_rhs(default)
  }
  node_exprs[[length(node_exprs)]] <- NULL
  node_exprs <- c(node_exprs, .default = default)

  expr(case_when(!!!node_exprs))
}

test_that("generate_case_when_trees() works", {
  node <- list(
    path = list(
      list(type = "conditional", col = "disp", val = 100, op = "more")
    ),
    prediction = list(
      list(col = "(Intercept)", val = 14, op = "none", is_intercept = 1),
      list(col = "hp", val = 4, op = "multiply", is_intercept = 0),
      list(col = "drat", val = 2, op = "multiply", is_intercept = 0)
    )
  )

  tree <- list(node)

  parsedmodel <- list(
    trees = list(tree),
    general = list(mode = "")
  )

  expect_identical(
    generate_case_when_trees(parsedmodel, default = FALSE),
    list(
      quote(case_when(disp > 100 ~ 14 + hp * 4 + drat * 2))
    )
  )

  parsedmodel <- list(
    trees = list(tree),
    general = list(mode = "ifelse")
  )

  expect_identical(
    generate_case_when_trees(parsedmodel, default = FALSE),
    list(
      quote(case_when(ifelse(disp > 100, 14 + hp * 4 + drat * 2, 0)))
    )
  )

  parsedmodel <- list(
    trees = list(tree, tree),
    general = list(mode = "")
  )

  expect_identical(
    generate_case_when_trees(parsedmodel, default = FALSE),
    list(
      quote(
        case_when(
          disp > 100 ~ 14 + hp * 4 + drat * 2
        )
      ),
      quote(
        case_when(
          disp > 100 ~ 14 + hp * 4 + drat * 2
        )
      )
    )
  )

  parsedmodel <- list(
    trees = list(tree, tree),
    general = list(mode = "ifelse")
  )

  expect_identical(
    generate_case_when_trees(parsedmodel, default = FALSE),
    list(
      quote(
        case_when(
          ifelse(disp > 100, 14 + hp * 4 + drat * 2, 0)
        )
      ),
      quote(
        case_when(
          ifelse(disp > 100, 14 + hp * 4 + drat * 2, 0)
        )
      )
    )
  )
})

test_that("generate_case_when_tree() works", {
  node <- list(
    path = list(
      list(type = "conditional", col = "disp", val = 100, op = "more")
    ),
    prediction = list(
      list(col = "(Intercept)", val = 14, op = "none", is_intercept = 1),
      list(col = "hp", val = 4, op = "multiply", is_intercept = 0),
      list(col = "drat", val = 2, op = "multiply", is_intercept = 0)
    )
  )

  nodes <- list(node)

  expect_identical(
    generate_case_when_tree(nodes, mode = "", default = FALSE),
    quote(case_when(disp > 100 ~ 14 + hp * 4 + drat * 2))
  )
  expect_identical(
    generate_case_when_tree(nodes, mode = "ifelse", default = FALSE),
    quote(case_when(ifelse(disp > 100, 14 + hp * 4 + drat * 2, 0)))
  )

  nodes <- list(node, node)

  expect_identical(
    generate_case_when_tree(nodes, mode = "", default = FALSE),
    quote(
      case_when(
        disp > 100 ~ 14 + hp * 4 + drat * 2,
        disp > 100 ~ 14 + hp * 4 + drat * 2
      )
    )
  )
  expect_identical(
    generate_case_when_tree(nodes, mode = "ifelse", default = FALSE),
    quote(
      case_when(
        ifelse(disp > 100, 14 + hp * 4 + drat * 2, 0),
        ifelse(disp > 100, 14 + hp * 4 + drat * 2, 0)
      )
    )
  )

  nodes <- list(
    list(
      prediction = 25,
      path = list(list(
        type = "conditional",
        col = "cyl",
        val = 4,
        op = "less-equal"
      ))
    ),
    list(
      prediction = 20,
      path = list(
        list(type = "conditional", col = "cyl", val = 6, op = "less-equal"),
        list(type = "conditional", col = "cyl", val = 4, op = "more")
      )
    ),
    list(
      prediction = 15,
      path = list(
        list(type = "conditional", col = "cyl", val = 6, op = "more"),
        list(type = "conditional", col = "cyl", val = 4, op = "more")
      )
    )
  )

  expect_identical(
    generate_case_when_tree(nodes, mode = "", default = TRUE),
    quote(case_when(cyl <= 4 ~ 25, cyl <= 6 & cyl > 4 ~ 20, .default = 15))
  )
})

test_that("generate_tree_nodes() works", {
  node <- list(
    path = list(
      list(type = "conditional", col = "disp", val = 100, op = "more")
    ),
    prediction = list(
      list(col = "(Intercept)", val = 14, op = "none", is_intercept = 1),
      list(col = "hp", val = 4, op = "multiply", is_intercept = 0),
      list(col = "drat", val = 2, op = "multiply", is_intercept = 0)
    )
  )

  nodes <- list(node)

  expect_identical(
    generate_tree_nodes(nodes, mode = ""),
    list(
      quote(disp > 100 ~ 14 + hp * 4 + drat * 2)
    )
  )
  expect_identical(
    generate_tree_nodes(nodes, mode = "ifelse"),
    list(
      quote(ifelse(disp > 100, 14 + hp * 4 + drat * 2, 0))
    )
  )

  nodes <- list(node, node)

  expect_identical(
    generate_tree_nodes(nodes, mode = ""),
    list(
      quote(disp > 100 ~ 14 + hp * 4 + drat * 2),
      quote(disp > 100 ~ 14 + hp * 4 + drat * 2)
    )
  )
  expect_identical(
    generate_tree_nodes(nodes, mode = "ifelse"),
    list(
      quote(ifelse(disp > 100, 14 + hp * 4 + drat * 2, 0)),
      quote(ifelse(disp > 100, 14 + hp * 4 + drat * 2, 0))
    )
  )
})

test_that("generate_tree_node() works", {
  node <- list(
    path = list(
      list(type = "conditional", col = "disp", val = 100, op = "more")
    ),
    prediction = list(
      list(col = "(Intercept)", val = 14, op = "none", is_intercept = 1),
      list(col = "hp", val = 4, op = "multiply", is_intercept = 0),
      list(col = "drat", val = 2, op = "multiply", is_intercept = 0)
    )
  )
  expect_identical(
    generate_tree_node(node, calc_mode = "ifelse"),
    quote(ifelse(disp > 100, 14 + hp * 4 + drat * 2, 0))
  )
  expect_identical(
    generate_tree_node(node, calc_mode = ""),
    quote(disp > 100 ~ 14 + hp * 4 + drat * 2)
  )

  node <- list(
    path = list(),
    prediction = list(
      list(col = "(Intercept)", val = 14, op = "none", is_intercept = 1),
      list(col = "hp", val = 4, op = "multiply", is_intercept = 0),
      list(col = "drat", val = 2, op = "multiply", is_intercept = 0)
    )
  )
  expect_identical(
    generate_tree_node(node, calc_mode = "ifelse"),
    quote(14 + hp * 4 + drat * 2)
  )
  expect_identical(
    generate_tree_node(node, calc_mode = ""),
    quote(14 + hp * 4 + drat * 2)
  )

  node <- list(
    path = list(
      list(type = "conditional", col = "disp", val = 100, op = "more")
    ),
    prediction = 3
  )
  expect_identical(
    generate_tree_node(node, calc_mode = "ifelse"),
    quote(ifelse(disp > 100, 3, 0))
  )
  expect_identical(
    generate_tree_node(node, calc_mode = ""),
    quote(disp > 100 ~ 3)
  )

  node <- list(
    path = list(
      list(type = "conditional", col = "disp", val = 100, op = "more")
    ),
    prediction = list(
      list(col = "(Intercept)", val = 14, op = "none", is_intercept = 1)
    )
  )
  expect_identical(
    generate_tree_node(node, calc_mode = "ifelse"),
    quote(ifelse(disp > 100, 14, 0))
  )
  expect_identical(
    generate_tree_node(node, calc_mode = ""),
    quote(disp > 100 ~ 14)
  )
})

test_that("generate_tree_node() avoids ifelse if path is always TRUE (#143)", {
  node <- list(
    path = list(),
    prediction = list(
      list(col = "(Intercept)", val = 37.2, op = "none", is_intercept = 1),
      list(col = "hp", val = -0.0318, op = "multiply", is_intercept = 0),
      list(col = "wt", val = -3.88, op = "multiply", is_intercept = 0)
    )
  )

  expect_identical(
    expr_text(generate_tree_node(node, calc_mode = "ifelse")),
    "37.2 + hp * -0.0318 + wt * -3.88"
  )

  expect_identical(
    expr_text(generate_tree_node(node, calc_mode = "")),
    "37.2 + hp * -0.0318 + wt * -3.88"
  )
})

test_that("generate_tree_node() avoids multipliying with 0 and 1 (#152)", {
  node <- list(
    path = list(
      list(type = "conditional", col = "disp", val = 100, op = "more")
    ),
    prediction = list(
      list(col = "(Intercept)", val = 0, op = "none", is_intercept = 1),
      list(col = "hp", val = 4, op = "multiply", is_intercept = 0),
      list(col = "drat", val = 2, op = "multiply", is_intercept = 0)
    )
  )

  expect_identical(
    generate_tree_node(node, calc_mode = "ifelse"),
    quote(ifelse(disp > 100, hp * 4 + drat * 2, 0))
  )
  expect_identical(
    generate_tree_node(node, calc_mode = ""),
    quote(disp > 100 ~ hp * 4 + drat * 2)
  )

  node <- list(
    path = list(
      list(type = "conditional", col = "disp", val = 100, op = "more")
    ),
    prediction = list(
      list(col = "(Intercept)", val = 14, op = "none", is_intercept = 1),
      list(col = "hp", val = 1, op = "multiply", is_intercept = 0),
      list(col = "drat", val = 0, op = "multiply", is_intercept = 0)
    )
  )

  expect_identical(
    generate_tree_node(node, calc_mode = "ifelse"),
    quote(ifelse(disp > 100, 14 + hp, 0))
  )
  expect_identical(
    generate_tree_node(node, calc_mode = ""),
    quote(disp > 100 ~ 14 + hp)
  )
})

test_that("path_formulas() works", {
  expect_identical(
    path_formulas(
      list()
    ),
    quote(TRUE)
  )

  expect_identical(
    path_formulas(
      list(
        list(type = "all", op = "more", col = "x", val = 0)
      )
    ),
    quote(TRUE)
  )

  expect_identical(
    path_formulas(
      list(
        list(type = "conditional", op = "more", col = "x", val = 0)
      )
    ),
    quote(x > 0)
  )

  expect_identical(
    path_formulas(
      list(
        list(type = "conditional", op = "more", col = "x", val = 0),
        list(type = "conditional", op = "less", col = "y", val = 0)
      )
    ),
    quote(x > 0 & y < 0)
  )

  expect_identical(
    path_formulas(
      list(
        list(type = "conditional", op = "more", col = "x", val = 0),
        list(type = "conditional", op = "less", col = "y", val = 0),
        list(type = "conditional", op = "less", col = "z", val = 0)
      )
    ),
    quote(x > 0 & y < 0 & z < 0)
  )
})

test_that("path_formula() works", {
  expect_identical(
    path_formula(list(
      type = "conditional",
      op = "more",
      col = "x",
      val = 0
    )),
    quote(x > 0)
  )
  expect_identical(
    path_formula(list(
      type = "conditional",
      op = "more-equal",
      col = "x",
      val = 0
    )),
    quote(x >= 0)
  )
  expect_identical(
    path_formula(list(
      type = "conditional",
      op = "less",
      col = "x",
      val = 0
    )),
    quote(x < 0)
  )
  expect_identical(
    path_formula(list(
      type = "conditional",
      op = "less-equal",
      col = "x",
      val = 0
    )),
    quote(x <= 0)
  )
  expect_identical(
    path_formula(list(
      type = "conditional",
      op = "more",
      col = "x",
      val = "h"
    )),
    quote(x > "h")
  )
  expect_identical(
    path_formula(list(
      type = "conditional",
      op = "more-equal",
      col = "x",
      val = "h"
    )),
    quote(x >= "h")
  )
  expect_identical(
    path_formula(list(
      type = "conditional",
      op = "less",
      col = "x",
      val = "h"
    )),
    quote(x < "h")
  )
  expect_identical(
    path_formula(list(
      type = "conditional",
      op = "less-equal",
      col = "x",
      val = "h"
    )),
    quote(x <= "h")
  )
  expect_identical(
    path_formula(list(
      type = "set",
      op = "in",
      col = "x",
      vals = 0
    )),
    quote(x %in% 0)
  )
  expect_identical(
    path_formula(list(
      type = "set",
      op = "not-in",
      col = "x",
      vals = 0
    )),
    quote((x %in% 0) == FALSE)
  )
  expect_identical(
    path_formula(list(
      type = "set",
      op = "in",
      col = "x",
      vals = "h"
    )),
    quote(x %in% "h")
  )
  expect_identical(
    path_formula(list(
      type = "set",
      op = "not-in",
      col = "x",
      vals = "h"
    )),
    quote((x %in% "h") == FALSE)
  )

  res <- path_formula(list(
    type = "set",
    op = "in",
    col = "x",
    vals = c(1, 2, 5)
  ))

  expect_type(res, "language")

  expect_identical(
    expr_text(res),
    "x %in% c(1, 2, 5)"
  )

  res <- path_formula(list(
    type = "set",
    op = "not-in",
    col = "x",
    vals = c(1, 2, 5)
  ))

  expect_type(res, "language")

  expect_identical(
    expr_text(res),
    "(x %in% c(1, 2, 5)) == FALSE"
  )

  res <- path_formula(list(
    type = "set",
    op = "in",
    col = "x",
    vals = letters[1:5]
  ))

  expect_type(res, "language")

  expect_identical(
    expr_text(res),
    'x %in% c("a", "b", "c", "d", "e")'
  )

  res <- path_formula(list(
    type = "set",
    op = "not-in",
    col = "x",
    vals = letters[1:5]
  ))

  expect_type(res, "language")

  expect_identical(
    expr_text(res),
    '(x %in% c("a", "b", "c", "d", "e")) == FALSE'
  )
})

test_that("path_formula() errors with unsupported values", {
  expect_snapshot(
    error = TRUE,
    path_formula(list(type = "unknown", op = "less", col = "x", val = 0))
  )
  expect_snapshot(
    error = TRUE,
    path_formula(list(type = "conditional", op = "unknown", col = "x", val = 0))
  )
  expect_snapshot(
    error = TRUE,
    path_formula(list(type = "set", op = "unknown", col = "x", vals = 0))
  )
})

# Tests for .build_case_when_tree()

test_that(".build_case_when_tree handles stump tree (no conditions)", {
  nodes <- list(list(prediction = 0.5, path = list()))

  expect_identical(.build_case_when_tree(nodes), 0.5)
})

test_that(".build_case_when_tree handles single node with condition", {
  nodes <- list(
    list(
      prediction = 1,
      path = list(list(
        type = "conditional",
        col = "x",
        val = 5,
        op = "less-equal"
      ))
    ),
    list(
      prediction = 0,
      path = list(list(type = "conditional", col = "x", val = 5, op = "more"))
    )
  )

  expect_identical(
    deparse1(.build_case_when_tree(nodes)),
    "case_when(x <= 5 ~ 1, .default = 0)"
  )
})

test_that(".build_case_when_tree handles multiple conditions", {
  nodes <- list(
    list(
      prediction = 1,
      path = list(
        list(type = "conditional", col = "x", val = 5, op = "less-equal"),
        list(type = "conditional", col = "y", val = 10, op = "more")
      )
    ),
    list(
      prediction = 0,
      path = list(list(type = "conditional", col = "x", val = 5, op = "more"))
    )
  )

  expect_identical(
    deparse1(.build_case_when_tree(nodes)),
    "case_when(x <= 5 & y > 10 ~ 1, .default = 0)"
  )
})

test_that(".build_case_when_tree handles three nodes", {
  nodes <- list(
    list(
      prediction = 1,
      path = list(list(
        type = "conditional",
        col = "x",
        val = 3,
        op = "less-equal"
      ))
    ),
    list(
      prediction = 0.5,
      path = list(
        list(type = "conditional", col = "x", val = 3, op = "more"),
        list(type = "conditional", col = "x", val = 7, op = "less-equal")
      )
    ),
    list(
      prediction = 0,
      path = list(list(type = "conditional", col = "x", val = 7, op = "more"))
    )
  )

  expect_identical(
    deparse1(.build_case_when_tree(nodes)),
    "case_when(x <= 3 ~ 1, x > 3 & x <= 7 ~ 0.5, .default = 0)"
  )
})

test_that(".build_case_when_tree handles set conditions", {
  nodes <- list(
    list(
      prediction = 1,
      path = list(list(type = "set", col = "x", vals = c("a", "b"), op = "in"))
    ),
    list(
      prediction = 0,
      path = list(list(
        type = "set",
        col = "x",
        vals = c("a", "b"),
        op = "not-in"
      ))
    )
  )

  expect_identical(
    deparse1(.build_case_when_tree(nodes)),
    'case_when(x %in% c("a", "b") ~ 1, .default = 0)'
  )
})

#' Converts an R model object into a parsed model
#'
#' Parses a fitted R model's structure and extracts the components needed to
#' create a dplyr formula for prediction. The parsed model can be serialized
#' (e.g., saved to YAML) and later used to generate predictions without the
#' original model object.
#'
#' @param model An R model object.
#'
#' @returns A parsed model object with class `parsed_model` and a model-specific
#'   subclass (e.g., `pm_xgb`, `pm_tree`, `pm_regression`). The object contains:
#'
#'   - `$general`: List with model metadata including `model` (model type),
#'     `type` (used for S3 dispatch), `version` (parsed model format version),
#'     and model-specific parameters.
#'   - Model-specific fields containing coefficients, tree structures, etc.
#'
#' @section Parsed model versions:
#'
#' The `$general$version` field indicates the parsed model format:
#'
#' - **Version 1**: Original format. Linear models store coefficients in a
#'   data frame. Tree models use flat `case_when()` expressions where all leaf
#'   conditions are at the same level.
#'
#' - **Version 2**: Improved coefficient storage for linear models (lm, earth).
#'   Tree models still use flat `case_when()`.
#'
#' - **Version 3**: Current format. Tree models (rpart, ranger, randomForest,
#'   xgboost, lightgbm, catboost, partykit, cubist) use nested `case_when()`
#'   expressions that mirror the tree structure. This produces more efficient
#'   SQL and R code because conditions are evaluated hierarchically rather than
#'   checking all leaf paths.
#'
#' When loading a parsed model saved with an older version, tidypredict
#' automatically uses the appropriate formula builder for backwards
#' compatibility.
#'
#' @section Model types:
#'
#' Each parsed model has a type that determines the S3 class used for dispatch:
#'
#' - `pm_regression`: Linear models (lm, glm, earth, glmnet)
#' - `pm_tree`: Single trees and forests (rpart, partykit, ranger, randomForest,
#'   cubist)
#' - `pm_xgb`: XGBoost gradient boosting models
#' - `pm_lgb`: LightGBM gradient boosting models
#' - `pm_catboost`: CatBoost gradient boosting models
#'
#' @examples
#' library(dplyr)
#' df <- mutate(mtcars, cyl = paste0("cyl", cyl))
#' model <- lm(mpg ~ wt + cyl * disp, offset = am, data = df)
#' parse_model(model)
#' @export
parse_model <- function(model) {
  UseMethod("parse_model")
}

test_that("tidypredict_interval.data.frame errors", {
  expect_error(
    tidypredict_interval(data.frame(x = 1)),
    "no longer supported"
  )
})

test_that("tidypredict_interval.list works for lm", {
  model <- lm(mpg ~ wt + cyl, data = mtcars)
  pm <- parse_model(model)
  interval <- tidypredict_interval(pm)
  expect_type(interval, "language")
})

test_that("tidypredict_interval.list works for glm", {
  model <- glm(mpg ~ wt + cyl, data = mtcars, family = "gaussian")
  pm <- parse_model(model)
  interval <- tidypredict_interval(pm)
  expect_type(interval, "language")
})

test_that("tidypredict_interval.list errors for unsupported model", {
  pm <- list(general = list(model = "unsupported"))
  class(pm) <- "list"
  expect_snapshot(
    error = TRUE,
    tidypredict_interval(pm)
  )
})

test_that("tidypredict_fit.pm_tree works for v3 party", {
  skip_if_not_installed("partykit")
  model <- partykit::ctree(mpg ~ wt + cyl, data = mtcars)
  pm <- parse_model(model)
  fit <- tidypredict_fit(pm)
  expect_type(fit, "language")
})

test_that("tidypredict_fit.pm_tree works for v3 ranger", {
  skip_if_not_installed("ranger")
  model <- ranger::ranger(
    mpg ~ wt + cyl,
    data = mtcars,
    num.trees = 2,
    num.threads = 1
  )
  pm <- parse_model(model)
  fit <- tidypredict_fit(pm)
  expect_type(fit, "language")
})

test_that("tidypredict_fit.pm_tree works for v3 randomForest", {
  skip_if_not_installed("randomForest")
  model <- randomForest::randomForest(mpg ~ wt + cyl, data = mtcars, ntree = 2)
  pm <- parse_model(model)
  fit <- tidypredict_fit(pm)
  expect_type(fit, "language")
})

test_that("tidypredict_fit.pm_tree works for v2 ranger (backwards compat)", {
  pm <- readRDS(test_path("backwards-compat", "ranger-v2-parsed.rds"))
  fit <- tidypredict_fit(pm)
  expect_type(fit, "language")
})

test_that("tidypredict_fit.pm_tree works for v2 cubist (backwards compat)", {
  pm <- readRDS(test_path("backwards-compat", "cubist-v2-regression.rds"))
  fit <- tidypredict_fit(pm)
  expect_type(fit, "language")
})

test_that("tidypredict_fit.pm_tree works for v2 randomForest (backwards compat)", {
  pm <- readRDS(test_path("backwards-compat", "rf-v2-regression.rds"))
  fit <- tidypredict_fit(pm)
  expect_type(fit, "language")
})

test_that("returns the right output", {
  model <- glm(am ~ wt + cyl, data = mtcars, family = "gaussian")

  #Don't have stable numbers at the tails across OS
  model$coefficients <- round(model$coefficients, 12)

  tf <- tidypredict_fit(model)
  pm <- parse_model(model)

  expect_type(tf, "language")

  expect_s3_class(pm, "list")
  expect_equal(length(pm), 2)
  expect_equal(pm$general$model, "glm")
  expect_equal(pm$general$version, 2)

  expect_snapshot(
    rlang::expr_text(tf)
  )
})

test_that("Model can be saved and re-loaded", {
  model <- glm(am ~ wt + cyl, data = mtcars, family = "gaussian")

  model$coefficients <- round(model$coefficients, 7)

  pm <- parse_model(model)
  mp <- tempfile(fileext = ".yml")
  yaml::write_yaml(pm, mp)
  l <- yaml::read_yaml(mp)
  pm <- as_parsed_model(l)

  expect_identical(
    tidypredict_fit(model),
    tidypredict_fit(pm)
  )
})

test_that("formulas produces correct predictions", {
  mtcars$cyl <- paste0("cyl", mtcars$cyl)
  # family = gaussian
  expect_snapshot(
    tidypredict_test(
      glm(am ~ wt + cyl + disp, data = mtcars, family = "gaussian"),
      mtcars
    )
  )
  # family = binomial
  expect_snapshot(
    tidypredict_test(
      glm(am ~ wt + cyl + disp, data = mtcars, family = "binomial"),
      mtcars
    )
  )
  # family = gaussian, with interactions
  expect_snapshot(
    tidypredict_test(
      glm(am ~ wt * cyl + disp, data = mtcars, family = "gaussian"),
      mtcars
    )
  )
  # family = binomial, with interactions
  expect_snapshot(
    tidypredict_test(
      glm(am ~ wt * cyl + disp, data = mtcars, family = "binomial"),
      mtcars
    )
  )
  # family = gaussian, with interactions
  expect_snapshot(
    tidypredict_test(
      glm(am ~ wt:cyl + disp, data = mtcars, family = "gaussian"),
      mtcars
    )
  )
  # family = binomial, with interactions
  expect_snapshot(
    tidypredict_test(
      glm(am ~ wt:cyl + disp, data = mtcars, family = "binomial"),
      mtcars
    )
  )
})

test_that("tidypredict works when variable names are subset of other variables", {
  mtcars$cyl <- paste0("cyl", mtcars$cyl)
  mtcars$wt_sq <- mtcars$wt^2
  mtcars$char_cyl <- as.character(mtcars$cyl)
  set.seed(22)
  mtcars$char_cyl_2 <- sample(letters[1:3], size = nrow(mtcars), replace = TRUE)

  model <- suppressWarnings(glm(
    am ~ wt + wt_sq + char_cyl + char_cyl_2,
    data = mtcars,
    family = "binomial"
  ))

  expect_snapshot(
    tidypredict_test(
      model,
      mtcars
    )
  )
})

test_that("tidypredict_interval works for gaussian glm", {
  model <- glm(mpg ~ wt + cyl, data = mtcars, family = "gaussian")
  interval <- tidypredict_interval(model)
  expect_type(interval, "language")
})

test_that("tidypredict_interval errors for non-gaussian glm", {
  model <- glm(am ~ wt + cyl, data = mtcars, family = "binomial")
  expect_snapshot(
    error = TRUE,
    tidypredict_interval(model)
  )
})

test_that("returns the right output", {
  model <- Cubist::cubist(
    x = mtcars[, -1],
    y = mtcars$mpg,
    committees = 3
  )
  tf <- tidypredict_fit(model)
  pm <- parse_model(model)

  expect_type(tf, "language")

  expect_s3_class(pm, "list")
  expect_equal(length(pm), 2)
  expect_equal(length(pm$trees), 1)
  expect_equal(pm$general$model, "cubist")
  expect_equal(pm$general$version, 3)

  expect_snapshot(
    rlang::expr_text(tf)
  )
})

test_that("Model can be saved and re-loaded", {
  model <- Cubist::cubist(
    x = mtcars[, -1],
    y = mtcars$mpg,
    committees = 3
  )
  pm <- parse_model(model)
  mp <- tempfile(fileext = ".yml")
  yaml::write_yaml(pm, mp)
  l <- yaml::read_yaml(mp)
  pm <- as_parsed_model(l)

  expect_identical(
    tidypredict_fit(model),
    tidypredict_fit(pm)
  )
})

test_that("formulas produces correct predictions", {
  model <- Cubist::cubist(
    x = mtcars[, -1],
    y = mtcars$mpg,
    committees = 3
  )

  # Cubist doesn't work near splits
  # https://github.com/topepo/Cubist/issues/62
  splits <- dplyr::distinct(model$splits, variable, value)
  splits <- map2(splits$variable, splits$value, function(x, y) {
    str2lang(paste("abs(", x, "-", y, ") > 0.0001"))
  })
  non_split_data <- mtcars %>%
    dplyr::filter(!!!splits)

  expect_snapshot(
    tidypredict_test(
      model,
      non_split_data,
      threshold = 0.00001
    )
  )
})

test_that("intercept is done correctly (#58)", {
  biomass_tr <- modeldata::biomass %>%
    dplyr::filter(dataset == "Training") %>%
    dplyr::select(-dataset, -sample)

  set.seed(1)
  mod <- Cubist::cubist(
    x = biomass_tr %>% dplyr::select(-HHV),
    y = biomass_tr$HHV
  )

  res <- tidypredict_fit(mod)
  expect_false(any(grepl("list", res)))
})

test_that("doesn't divide by TRUE (#143)", {
  rules <- list(quote(37.2 + hp * -0.0318 + wt * -3.88))
  paths <- list(TRUE)

  expect_identical(
    expr_text(make_committee(rules, paths)),
    "37.2 + hp * -0.0318 + wt * -3.88"
  )

  rules <- list(quote(37.2 + hp * -0.0318 + wt * -3.88))
  paths <- list(quote(disp > 95.099998))

  expect_identical(
    expr_text(make_committee(rules, paths)),
    "(37.2 + hp * -0.0318 + wt * -3.88)/(disp > 95.099998)"
  )
})

test_that("returns the right output", {
  model <- lm(am ~ wt + cyl, data = mtcars)

  #Don't have stable numbers at the tails across OS
  model$coefficients <- round(model$coefficients, 12)

  tf <- tidypredict_fit(model)
  pm <- parse_model(model)

  expect_type(tf, "language")

  expect_s3_class(pm, "list")
  expect_equal(length(pm), 2)
  expect_equal(pm$general$model, "lm")
  expect_equal(pm$general$version, 2)

  expect_snapshot(
    rlang::expr_text(tf)
  )
})

test_that("Model can be saved and re-loaded", {
  model <- lm(am ~ wt + cyl, data = mtcars)

  model$coefficients <- round(model$coefficients, 7)

  pm <- parse_model(model)
  mp <- tempfile(fileext = ".yml")
  yaml::write_yaml(pm, mp)
  l <- yaml::read_yaml(mp)
  pm <- as_parsed_model(l)

  expect_identical(
    tidypredict_fit(model),
    tidypredict_fit(pm)
  )
})

test_that("formulas produces correct predictions", {
  mtcars$cyl <- paste0("cyl", mtcars$cyl)

  # normal
  expect_snapshot(
    tidypredict_test(
      lm(mpg ~ wt + am + cyl, data = mtcars),
      mtcars
    )
  )

  # offset
  expect_snapshot(
    tidypredict_test(
      lm(mpg ~ wt, offset = am, data = mtcars),
      mtcars
    )
  )

  # interaction
  expect_snapshot(
    tidypredict_test(
      lm(mpg ~ wt + disp * cyl, data = mtcars),
      mtcars
    )
  )

  # interaction
  expect_snapshot(
    tidypredict_test(
      lm(mpg ~ wt + disp:cyl, data = mtcars),
      mtcars
    )
  )

  # interactions
  expect_snapshot(
    tidypredict_test(
      lm(mpg ~ (wt + disp) * cyl, data = mtcars),
      mtcars
    )
  )
})

test_that("tidypredict works when variable names are subset of other variables", {
  mtcars$cyl <- paste0("cyl", mtcars$cyl)
  mtcars$wt_sq <- mtcars$wt^2
  mtcars$char_cyl <- as.character(mtcars$cyl)
  set.seed(22)
  mtcars$char_cyl_2 <- sample(letters[1:3], size = nrow(mtcars), replace = TRUE)

  model <- lm(
    am ~ wt + wt_sq + char_cyl + char_cyl_2,
    data = mtcars
  )

  expect_snapshot(
    tidypredict_test(
      model,
      mtcars
    )
  )
})

test_that("tidy() works", {
  expect_s3_class(
    tidy(parse_model(lm(mpg ~ ., mtcars))),
    "tbl_df"
  )
})

test_that("we get better error from QR decomposition issues (#124)", {
  mtcars$vs2 <- mtcars$disp - mtcars$vs

  lm_fit <- lm(mpg ~ ., mtcars)

  expect_snapshot(
    error = TRUE,
    tidypredict::tidypredict_fit(lm_fit)
  )
})

test_that("don't add with 0 (#147)", {
  model <- lm(am ~ wt + cyl, data = mtcars)

  model$coefficients <- setNames(c(0, 1.5, 2.2), names(model$coefficients))

  expect_identical(
    tidypredict_fit(model),
    quote((wt * 1.5) + (cyl * 2.2))
  )
})

test_that("gaussian family with identity link works", {
  model <- glm(mpg ~ wt + hp, data = mtcars, family = gaussian())
  fit <- tidypredict_fit(model)
  native <- unname(predict(model, type = "response"))
  tidy <- rlang::eval_tidy(fit, mtcars)
  expect_equal(tidy, native)
})

test_that("binomial family with logit link works", {
  suppressWarnings(
    model <- glm(am ~ wt + hp, data = mtcars, family = binomial())
  )
  fit <- tidypredict_fit(model)
  native <- unname(predict(model, type = "response"))
  tidy <- rlang::eval_tidy(fit, mtcars)
  expect_equal(tidy, native)
})

test_that("poisson family with log link works", {
  model <- glm(gear ~ wt + hp, data = mtcars, family = poisson())
  fit <- tidypredict_fit(model)
  native <- unname(predict(model, type = "response"))
  tidy <- rlang::eval_tidy(fit, mtcars)
  expect_equal(tidy, native)
})

test_that("Gamma family with inverse link works (#203)", {
  model <- glm(mpg ~ wt + hp, data = mtcars, family = Gamma())

  fit <- tidypredict_fit(model)
  native <- unname(predict(model, type = "response"))
  tidy <- rlang::eval_tidy(fit, mtcars)

  expect_equal(tidy, native)
})

test_that("inverse.gaussian family with 1/mu^2 link works (#204)", {
  model <- glm(mpg ~ wt + hp, data = mtcars, family = inverse.gaussian())

  fit <- tidypredict_fit(model)
  native <- unname(predict(model, type = "response"))
  tidy <- rlang::eval_tidy(fit, mtcars)

  expect_equal(tidy, native)
})

test_that("binomial with probit link works (#205)", {
  suppressWarnings(
    model <- glm(
      am ~ wt + hp,
      data = mtcars,
      family = binomial(link = "probit")
    )
  )

  fit <- tidypredict_fit(model)
  native <- unname(predict(model, type = "response"))
  tidy <- rlang::eval_tidy(fit, mtcars)

  # Uses Bowling et al. approximation to normal CDF for SQL compatibility
  expect_true(all(abs(tidy - native) < 0.001))
})

test_that("binomial with cloglog link works (#206)", {
  suppressWarnings(
    model <- glm(
      am ~ wt + hp,
      data = mtcars,
      family = binomial(link = "cloglog")
    )
  )

  fit <- tidypredict_fit(model)
  native <- unname(predict(model, type = "response"))
  tidy <- rlang::eval_tidy(fit, mtcars)

  expect_equal(tidy, native)
})

test_that("poisson with sqrt link works (#207)", {
  model <- glm(carb ~ wt + hp, data = mtcars, family = poisson(link = "sqrt"))

  fit <- tidypredict_fit(model)
  native <- unname(predict(model, type = "response"))
  tidy <- rlang::eval_tidy(fit, mtcars)

  expect_equal(tidy, native)
})

test_that("returns the right output", {
  model <- earth::earth(mpg ~ ., data = mtcars)
  model$coefficients <- round(model$coefficients, 12)
  tf <- tidypredict_fit(model)
  pm <- parse_model(model)

  expect_type(tf, "language")

  expect_s3_class(pm, "list")
  expect_equal(length(pm), 2)
  expect_equal(pm$general$model, "earth")
  expect_equal(pm$general$version, 2)

  expect_snapshot(
    rlang::expr_text(tf)
  )
})

test_that("Model can be saved and re-loaded", {
  model <- earth::earth(mpg ~ ., data = mtcars)
  model$coefficients <- round(model$coefficients, 7)

  pm <- parse_model(model)
  mp <- tempfile(fileext = ".yml")
  yaml::write_yaml(pm, mp)
  l <- yaml::read_yaml(mp)
  pm <- as_parsed_model(l)

  expect_identical(
    tidypredict_fit(model),
    tidypredict_fit(pm)
  )
})

test_that("formulas produces correct predictions", {
  # Regression - numeric predictors
  expect_snapshot(
    tidypredict_test(
      earth::earth(age ~ sibsp + parch, data = earth::etitanic),
      earth::etitanic
    )
  )

  # Regression - numeric predictors, degree = 2
  expect_snapshot(
    tidypredict_test(
      earth::earth(age ~ sibsp + parch, data = earth::etitanic, degree = 2),
      earth::etitanic
    )
  )

  # Regression - numeric predictors, degree = 3
  expect_snapshot(
    tidypredict_test(
      earth::earth(age ~ sibsp + parch, data = earth::etitanic, degree = 3),
      earth::etitanic
    )
  )

  # Regression - numeric predictors and categorical predictors
  expect_snapshot(
    tidypredict_test(
      earth::earth(age ~ ., data = earth::etitanic),
      earth::etitanic
    )
  )

  # Regression - pmethod = "backwards"
  expect_snapshot(
    tidypredict_test(
      earth::earth(age ~ ., data = earth::etitanic, pmethod = "backward"),
      earth::etitanic
    )
  )

  # Regression - pmethod = "non"
  expect_snapshot(
    tidypredict_test(
      earth::earth(age ~ ., data = earth::etitanic, pmethod = "none"),
      earth::etitanic
    )
  )
  # Regression - pmethod = "exhaustive"
  expect_snapshot(
    tidypredict_test(
      earth::earth(age ~ ., data = earth::etitanic, pmethod = "exhaustive"),
      earth::etitanic
    )
  )

  # Regression - pmethod = "forward"
  expect_snapshot(
    tidypredict_test(
      earth::earth(age ~ ., data = earth::etitanic, pmethod = "forward"),
      earth::etitanic
    )
  )

  # Regression - pmethod = "seqrep"
  expect_snapshot(
    tidypredict_test(
      earth::earth(age ~ ., data = earth::etitanic, pmethod = "seqrep"),
      earth::etitanic
    )
  )

  # binomial
  expect_snapshot(
    tidypredict_test(
      earth::earth(
        survived ~ age + sibsp,
        data = earth::etitanic,
        glm = list(family = binomial)
      ),
      earth::etitanic
    )
  )

  # binomial - w/ degree
  expect_snapshot(
    tidypredict_test(
      earth::earth(
        survived ~ age + sibsp,
        data = earth::etitanic,
        glm = list(family = binomial),
        degree = 2
      ),
      earth::etitanic
    )
  )

  # binomial - pmethod = "backwards"
  expect_snapshot(
    tidypredict_test(
      earth::earth(
        survived ~ .,
        data = earth::etitanic,
        glm = list(family = binomial),
        pmethod = "backward"
      ),
      earth::etitanic
    )
  )

  # binomial - pmethod = "non"
  expect_snapshot(
    tidypredict_test(
      earth::earth(
        survived ~ .,
        data = earth::etitanic,
        glm = list(family = binomial),
        pmethod = "none"
      ),
      earth::etitanic
    )
  )
  # binomial - pmethod = "exhaustive"
  expect_snapshot(
    tidypredict_test(
      earth::earth(
        survived ~ .,
        data = earth::etitanic,
        glm = list(family = binomial),
        pmethod = "exhaustive"
      ),
      earth::etitanic
    )
  )

  # binomial - pmethod = "forward"
  expect_snapshot(
    tidypredict_test(
      earth::earth(
        survived ~ .,
        data = earth::etitanic,
        glm = list(family = binomial),
        pmethod = "forward"
      ),
      earth::etitanic
    )
  )

  # binomial - pmethod = "seqrep"
  expect_snapshot(
    tidypredict_test(
      earth::earth(
        survived ~ .,
        data = earth::etitanic,
        glm = list(family = binomial),
        pmethod = "seqrep"
      ),
      earth::etitanic
    )
  )

  # formula interface
  expect_snapshot(
    tidypredict_test(
      earth::earth(
        Sepal.Length ~ .,
        data = iris
      ),
      iris
    )
  )

  # XY interface
  expect_snapshot(
    tidypredict_test(
      earth::earth(
        x = iris[, -1],
        y = iris$Sepal.Length
      ),
      iris
    )
  )

  # formula interface - degree = 2
  expect_snapshot(
    tidypredict_test(
      earth::earth(
        Sepal.Length ~ .,
        data = iris,
        degree = 2,
        pmethod = "none"
      ),
      iris
    )
  )

  # XY interface - degree = 2
  expect_snapshot(
    tidypredict_test(
      earth::earth(
        x = iris[, -1],
        y = iris$Sepal.Length,
        degree = 2,
        pmethod = "none"
      ),
      iris
    )
  )
})

test_that("probit link works (#194)", {
  model <- earth::earth(
    survived ~ age + sibsp,
    data = earth::etitanic,
    glm = list(family = binomial(link = "probit"))
  )

  fit <- tidypredict_fit(model)
  native <- unname(predict(model, earth::etitanic, type = "response")[, 1])
  tidy <- rlang::eval_tidy(fit, earth::etitanic)

  # Uses Bowling et al. approximation to normal CDF
  expect_equal(tidy, native, tolerance = 0.001)
})

test_that("cloglog link works (#194)", {
  model <- earth::earth(
    survived ~ age + sibsp,
    data = earth::etitanic,
    glm = list(family = binomial(link = "cloglog"))
  )

  fit <- tidypredict_fit(model)
  native <- unname(predict(model, earth::etitanic, type = "response")[, 1])
  tidy <- rlang::eval_tidy(fit, earth::etitanic)

  expect_equal(tidy, native)
})

test_that("Gamma family works (#195)", {
  model <- earth::earth(
    mpg ~ cyl + disp + hp,
    data = mtcars,
    glm = list(family = Gamma)
  )

  fit <- tidypredict_fit(model)
  native <- unname(predict(model, mtcars, type = "response")[, 1])
  tidy <- rlang::eval_tidy(fit, mtcars)

  expect_equal(tidy, native)
})

test_that("inverse.gaussian family works (#195)", {
  model <- earth::earth(
    mpg ~ cyl + disp + hp,
    data = mtcars,
    glm = list(family = inverse.gaussian)
  )

  fit <- tidypredict_fit(model)
  native <- unname(predict(model, mtcars, type = "response")[, 1])
  tidy <- rlang::eval_tidy(fit, mtcars)

  expect_equal(tidy, native)
})

# Tests for .extract_earth_multiclass()

test_that(".extract_earth_multiclass errors on non-earth model", {
  model <- lm(mpg ~ ., data = mtcars)

  expect_snapshot(error = TRUE, .extract_earth_multiclass(model))
})

test_that(".extract_earth_multiclass errors on binary model", {
  suppressWarnings(
    model <- earth::earth(
      vs ~ disp + hp,
      data = mtcars,
      glm = list(family = binomial)
    )
  )

  expect_snapshot(error = TRUE, .extract_earth_multiclass(model))
})

test_that(".extract_earth_multiclass errors on regression model", {
  model <- earth::earth(mpg ~ ., data = mtcars)

  expect_snapshot(error = TRUE, .extract_earth_multiclass(model))
})

test_that(".extract_earth_multiclass returns correct structure", {
  skip_if_not(
    exists("contr.earth.response", where = asNamespace("earth")),
    "earth multiclass not available"
  )
  library(earth)

  suppressWarnings(
    model <- earth(
      Species ~ .,
      data = iris,
      glm = list(family = binomial)
    )
  )

  result <- .extract_earth_multiclass(model)

  expect_type(result, "list")
  expect_length(result, 3)
  expect_named(result, levels(iris$Species))
  expect_type(result[[1]], "character")
})

test_that(".extract_earth_multiclass produces correct predictions", {
  skip_if_not(
    exists("contr.earth.response", where = asNamespace("earth")),
    "earth multiclass not available"
  )
  library(earth)

  suppressWarnings(
    model <- earth(
      Species ~ .,
      data = iris,
      glm = list(family = binomial)
    )
  )

  eqs <- .extract_earth_multiclass(model)
  n_rows <- nrow(iris)

  # Evaluate each expression - earth GLM outputs are already on probability scale
  # (not logits), so we don't apply softmax
  probs <- sapply(eqs, function(eq) {
    val <- rlang::eval_tidy(rlang::parse_expr(eq), iris)
    if (length(val) == 1) rep(val, n_rows) else val
  })

  # Compare to native predictions
  native <- predict(model, iris, type = "response")

  expect_equal(unname(probs), unname(native), tolerance = 1e-6)
})

test_that(".extract_earth_multiclass works with degree > 1", {
  skip_if_not(
    exists("contr.earth.response", where = asNamespace("earth")),
    "earth multiclass not available"
  )
  library(earth)

  suppressWarnings(
    model <- earth(
      Species ~ .,
      data = iris,
      glm = list(family = binomial),
      degree = 2
    )
  )

  result <- .extract_earth_multiclass(model)

  expect_type(result, "list")
  expect_length(result, 3)
  expect_named(result, levels(iris$Species))
})

expr_addition <- function(x, y) {
  expr(!!x + !!y)
}

expr_subtraction <- function(x, y) {
  expr(!!x - !!y)
}

expr_multiplication <- function(x, y) {
  expr(!!x * !!y)
}

expr_division <- function(x, y) {
  expr(!!x / !!y)
}

expr_and <- function(x, y) {
  expr(!!x & !!y)
}

reduce_addition <- function(x) {
  reduce(x, expr_addition)
}

reduce_subtraction <- function(x) {
  reduce(x, expr_subtraction)
}

reduce_multiplication <- function(x) {
  reduce(x, expr_multiplication)
}

reduce_and <- function(x) {
  reduce(x, expr_and)
}

expr_or <- function(x, y) {
  expr(!!x | !!y)
}

reduce_or <- function(x) {
  reduce(x, expr_or)
}

combine_path_conditions <- function(conditions) {
  n <- length(conditions)
  if (n == 0) {
    return(TRUE)
  }
  if (n == 1) {
    return(conditions[[1]])
  }
  if (n == 2) {
    return(expr_and(conditions[[1]], conditions[[2]]))
  }
  reduce_and(conditions)
}

test_that("expr_addition works", {
  expect_identical(
    expr_addition(quote(hp), 4),
    quote(hp + 4)
  )

  expect_identical(
    expr_addition(quote(hp + hp), 4),
    quote(hp + hp + 4)
  )

  expect_identical(
    expr_addition(quote((hp + hp)), 4),
    quote((hp + hp) + 4)
  )

  expect_identical(
    expr_addition(quote(hp), quote(hp)),
    quote(hp + hp)
  )
})

test_that("expr_subtraction works", {
  expect_identical(
    expr_subtraction(quote(hp), 4),
    quote(hp - 4)
  )

  expect_identical(
    expr_subtraction(quote(hp + hp), 4),
    quote(hp + hp - 4)
  )

  expect_identical(
    expr_subtraction(quote((hp + hp)), 4),
    quote((hp + hp) - 4)
  )

  expect_identical(
    expr_subtraction(quote(hp), quote(hp)),
    quote(hp - hp)
  )
})

test_that("expr_multiplication works", {
  expect_identical(
    expr_multiplication(quote(hp), 4),
    quote(hp * 4)
  )

  expect_identical(
    expr_multiplication(quote(hp + hp), 4),
    quote(`*`(hp + hp, 4))
  )

  expect_identical(
    expr_multiplication(quote((hp + hp)), 4),
    quote((hp + hp) * 4)
  )

  expect_identical(
    expr_multiplication(quote(hp), quote(hp)),
    quote(hp * hp)
  )
})

test_that("expr_division works", {
  expect_identical(
    expr_division(quote(hp), 4),
    quote(hp / 4)
  )

  expect_identical(
    expr_division(quote(hp + hp), 4),
    quote(`/`(hp + hp, 4))
  )

  expect_identical(
    expr_division(quote((hp + hp)), 4),
    quote((hp + hp) / 4)
  )

  expect_identical(
    expr_division(quote(hp), quote(hp)),
    quote(hp / hp)
  )
})

test_that("expr_and works", {
  expect_identical(
    expr_and(quote(hp), 4),
    quote(hp & 4)
  )

  expect_identical(
    expr_and(quote(hp + hp), 4),
    quote(`&`(hp + hp, 4))
  )

  expect_identical(
    expr_and(quote((hp + hp)), 4),
    quote((hp + hp) & 4)
  )

  expect_identical(
    expr_and(quote(hp), quote(hp)),
    quote(hp & hp)
  )
})

test_that("reduce_addition works", {
  expect_identical(
    reduce_addition(list(2, 5, 6)),
    quote(2 + 5 + 6)
  )

  expect_identical(
    reduce_addition(list(2, quote(hp), quote(vp))),
    quote(2 + hp + vp)
  )

  expect_identical(
    reduce_addition(list(2)),
    quote(2)
  )

  expect_identical(
    reduce_addition(list(quote(hp))),
    quote(hp)
  )

  expect_identical(
    reduce_addition(list(quote(vp + vp), quote(vp))),
    quote(vp + vp + vp)
  )

  expect_identical(
    reduce_addition(list(quote(vp + vp), quote((vp + vp)))),
    quote(vp + vp + (vp + vp))
  )
})

test_that("reduce_subtraction works", {
  expect_identical(
    reduce_subtraction(list(2, 5, 6)),
    quote(2 - 5 - 6)
  )

  expect_identical(
    reduce_subtraction(list(2, quote(hp), quote(vp))),
    quote(2 - hp - vp)
  )

  expect_identical(
    reduce_subtraction(list(2)),
    quote(2)
  )

  expect_identical(
    reduce_subtraction(list(quote(hp))),
    quote(hp)
  )

  expect_identical(
    reduce_subtraction(list(quote(vp + vp), quote(vp))),
    quote(vp + vp - vp)
  )

  expect_identical(
    reduce_subtraction(list(quote(vp + vp), quote((vp + vp)))),
    quote(vp + vp - (vp + vp))
  )
})

test_that("reduce_multiplication works", {
  expect_identical(
    reduce_multiplication(list(2, 5, 6)),
    quote(2 * 5 * 6)
  )

  expect_identical(
    reduce_multiplication(list(2, quote(hp), quote(vp))),
    quote(2 * hp * vp)
  )

  expect_identical(
    reduce_multiplication(list(2)),
    quote(2)
  )

  expect_identical(
    reduce_multiplication(list(quote(hp))),
    quote(hp)
  )

  expect_identical(
    reduce_multiplication(list(quote((vp + vp)), quote(vp))),
    quote((vp + vp) * vp)
  )

  expect_identical(
    reduce_multiplication(list(quote((vp + vp)), quote((vp + vp)))),
    quote((vp + vp) * (vp + vp))
  )
})

test_that("reduce_and works", {
  expect_identical(
    reduce_and(list(2, 5, 6)),
    quote(2 & 5 & 6)
  )

  expect_identical(
    reduce_and(list(2, quote(hp), quote(vp))),
    quote(2 & hp & vp)
  )

  expect_identical(
    reduce_and(list(2)),
    quote(2)
  )

  expect_identical(
    reduce_and(list(quote(hp))),
    quote(hp)
  )

  expect_identical(
    reduce_and(list(quote((vp + vp)), quote(vp))),
    quote((vp + vp) & vp)
  )

  expect_identical(
    reduce_and(list(quote((vp + vp)), quote((vp + vp)))),
    quote((vp + vp) & (vp + vp))
  )
})

# Keep the test suite single-threaded. R CMD check reports a NOTE when the test
# code uses more CPU time than elapsed time, and the toy models used here are far
# too small to benefit from the threads that the engines spin up by default.
#
# Threads are capped at the call site wherever an API allows it (`nthread` for
# xgboost, `num_threads` for lightgbm, `num.threads` for ranger). The rest is
# covered here:
#
# * The reference BLAS is single-threaded, but Debian and Ubuntu default to the
#   pthread build of OpenBLAS, which uses every core and deliberately ignores
#   OMP_NUM_THREADS. earth() runs thousands of tiny least-squares fits, enough to
#   cross OpenBLAS's threading threshold, and measured a CPU-to-elapsed ratio of
#   1.9 on a 4-core Linux runner where a bare dgemm() measured 3.8. Only an
#   in-process call fixes this, since OpenBLAS reads its environment variables
#   when it is loaded, which happens before this file runs.
# * data.table runs its own OpenMP pool inside xgb.model.dt.tree() and
#   lgb.model.dt.tree(), neither of which takes a thread argument, and it also
#   latches its thread count at load time.
# * ranger reads R_RANGER_NUM_THREADS at call time. This is the only way to reach
#   the predict() calls that tidypredict_test() makes internally, and ranger uses
#   std::thread, so the OMP_* variables never applied to it.
# * OMP_* is a backstop for engines that read it before their first parallel
#   region, hence the placement above the library() calls.
Sys.setenv(
  OMP_NUM_THREADS = 1,
  OMP_THREAD_LIMIT = 1,
  R_RANGER_NUM_THREADS = 1
)

if (requireNamespace("RhpcBLASctl", quietly = TRUE)) {
  RhpcBLASctl::blas_set_num_threads(1)
  RhpcBLASctl::omp_set_num_threads(1)
}

if (requireNamespace("data.table", quietly = TRUE)) {
  data.table::setDTthreads(1L)
}

library(testthat)
library(tidypredict)

test_check("tidypredict")

# parsnip ----------------------------------------------------------------

#' @export
tidypredict_fit._xgb.Booster <- function(model) {
  tidypredict_fit(model$fit)
}

#' @export
tidypredict_fit.model_fit <- function(model) {
  model <- glmnet_set_lambda(model)

  # Special handling for CatBoost models with categorical features
  if (inherits(model$fit, "catboost.Model")) {
    return(tidypredict_fit_catboost_parsnip(model))
  }

  tidypredict_fit(model$fit)
}

#' @export
parse_model.model_fit <- function(model) {
  model <- glmnet_set_lambda(model)
  parse_model(model$fit)
}

# glmnet adjustment ------------------------------------------------------

glmnet_set_lambda <- function(model) {
  if (inherits(model$fit, "glmnet")) {
    penalty <- model$spec$args$penalty
    coef <- glmnet::predict.glmnet(
      model$fit,
      s = penalty,
      type = "coefficients"
    )

    if ("(Intercept)" %in% rownames(coef)) {
      model$fit$a0 <- coef["(Intercept)", ]
      coef <- coef["(Intercept)" != rownames(coef), ]
    }
    model$fit$lambda <- penalty
    model$fit$beta <- coef
  }
  model
}

# broom ------------------------------------------------------------------

#' @export
generics::tidy

#' Tidy the parsed model results
#'
#' @param x A parsed_model object
#' @param ...  Reserved for future use
#'
#' @export
tidy.pm_regression <- function(x, ...) {
  map_dfr(
    x$terms,
    ~ tibble::tibble(term = .x$label, estimate = .x$coef)
  )
}

#' Tests base predict function against tidypredict
#'
#' Compares the results of predict() and tidypredict_to_column()
#' functions.
#'
#' @param model An R model or a list with a parsed model. It currently supports
#' lm(), glm() and randomForest() models.
#' @param df A data frame that contains all of the needed fields to run the prediction.
#' It defaults to the "model" data frame object inside the model object.
#' @param threshold The number that a given result difference, between predict() and
#' tidypredict_to_column() should not exceed. For continuous predictions, the default
#' value is 0.000000000001 (1e-12), and for categorical predictions, the default value is
#' 0.
#' @param include_intervals Switch to indicate if the prediction intervals should be
#' included in the test. It defaults to FALSE.
#' @param max_rows The number of rows in the object passed in the df argument. Highly
#' recommended for large data sets.
#' @param xg_df A xgb.DMatrix object, required only for XGBoost models. It defaults to
#' NULL
#' recommended for large data sets.
#'
#' @examples
#'
#' model <- lm(mpg ~ wt + cyl * disp, offset = am, data = mtcars)
#' tidypredict_test(model)
#' @export
tidypredict_test <- function(
  model,
  df = model$model,
  threshold = 0.000000000001,
  include_intervals = FALSE,
  max_rows = NULL,
  xg_df = NULL
) {
  UseMethod("tidypredict_test")
}

#' @export
tidypredict_test.party <- function(
  model,
  df = model$data,
  threshold = 0.000000000001,
  include_intervals = FALSE,
  max_rows = NULL,
  xg_df = NULL
) {
  tidypredict_test_default(
    model = model,
    df = df,
    threshold = threshold,
    include_intervals = include_intervals,
    max_rows = max_rows,
    xg_df = xg_df
  )
}

#' @export
tidypredict_test.default <- function(
  model,
  df = model$model,
  threshold = 0.000000000001,
  include_intervals = FALSE,
  max_rows = NULL,
  xg_df = NULL
) {
  tidypredict_test_default(
    model = model,
    df = df,
    threshold = threshold,
    include_intervals = include_intervals,
    max_rows = max_rows,
    xg_df = xg_df
  )
}
tidypredict_test_default <- function(
  model,
  df = model$model,
  threshold = 0.000000000001,
  include_intervals = FALSE,
  max_rows = NULL,
  xg_df = NULL
) {
  offset <- model$call$offset
  ismodels <- paste0(colnames(model$model), collapse = " ") ==
    paste0(colnames(df), collapse = " ")

  if (!is.null(offset) && ismodels) {
    index <- colnames(df) == "(offset)"
    colnames(df) <- replace(colnames(df), index, as.character(offset))
  }

  interval <- "none"
  if (include_intervals) {
    interval <- "prediction"
  }

  if (is.numeric(max_rows)) {
    df <- head(df, max_rows)
  }

  preds <- predict(model, df, interval = interval, type = "response")

  if (!include_intervals) {
    base <- data.frame(fit = as.vector(preds), row.names = NULL)
  } else {
    base <- as.data.frame(preds)
  }

  te <- tidypredict_to_column(
    df,
    model,
    add_interval = include_intervals,
    vars = c("fit_te", "upr_te", "lwr_te")
  )
  if (include_intervals) {
    te <- te[, c("fit_te", "upr_te", "lwr_te")]
  } else {
    te <- data.frame(fit_te = te[, "fit_te"])
  }

  raw_results <- cbind(base, te)
  raw_results$fit_diff <- raw_results$fit - raw_results$fit_te
  raw_results$fit_threshold <- abs(raw_results$fit_diff) > threshold

  if (include_intervals) {
    raw_results$lwr_diff <- abs(raw_results$lwr - raw_results$lwr_te)
    raw_results$upr_diff <- abs(raw_results$upr - raw_results$upr_te)
    raw_results$lwr_threshold <- raw_results$lwr_diff > threshold
    raw_results$upr_threshold <- raw_results$upr_diff > threshold
  }

  rowid <- seq_len(nrow(raw_results))
  raw_results <- cbind(data.frame(rowid), raw_results)

  threshold_df <- data.frame(fit_threshold = sum(raw_results$fit_threshold))
  if (include_intervals) {
    threshold_df$lwr_threshold <- sum(raw_results$lwr_threshold)
    threshold_df$upr_threshold <- sum(raw_results$upr_threshold)
  }

  alert <- any(threshold_df > 0)

  message <- paste0(
    "tidypredict test results\n",
    "Difference threshold: ",
    threshold,
    "\n"
  )

  if (alert) {
    difference <- data.frame(fit_diff = max(raw_results$fit_diff))
    if (include_intervals) {
      difference$lwr_diff <- max(raw_results$lwr_diff)
      difference$upr_diff <- max(raw_results$upr_diff)
    }
    message <- paste0(
      message,
      "\nFitted records above the threshold: ",
      threshold_df$fit_threshold,
      if (!is.null(threshold_df$lwr_threshold)) {
        "\nLower interval records above the threshold: "
      },
      threshold_df$lwr_threshold,
      if (!is.null(threshold_df$upr_threshold)) {
        "\nUpper interval records above the threshold: "
      },
      threshold_df$upr_threshold,
      "\n\nFit max  difference:",
      difference$upr_diff,
      "\nLower max difference:",
      difference$lwr_diff,
      "\nUpper max difference:",
      difference$fit_diff
    )
  } else {
    message <- paste0(
      message,
      "\n All results are within the difference threshold"
    )
  }
  results <- list()
  results$model_call <- model$call
  results$raw_results <- raw_results
  results$message <- message
  results$alert <- alert
  structure(results, class = c("tidypredict_test", "list"))
}

#' @export
tidypredict_test.glmnet <- function(
  model,
  df,
  threshold = 0.000000000001,
  include_intervals = FALSE,
  max_rows = NULL,
  xg_df = NULL
) {
  if (is.numeric(max_rows)) {
    df <- head(df, max_rows) # nocov
  }

  preds <- predict(model, as.matrix(df), type = "response")
  base <- data.frame(fit = as.vector(preds), row.names = NULL)

  te <- tidypredict_to_column(
    df,
    model,
    add_interval = FALSE,
    vars = c("fit_te", "upr_te", "lwr_te")
  )
  te <- data.frame(fit_te = te[, "fit_te"])

  raw_results <- cbind(base, te)
  raw_results$fit_diff <- raw_results$fit - raw_results$fit_te
  raw_results$fit_threshold <- abs(raw_results$fit_diff) > threshold

  rowid <- seq_len(nrow(raw_results))
  raw_results <- cbind(data.frame(rowid), raw_results)

  threshold_df <- data.frame(fit_threshold = sum(raw_results$fit_threshold))
  alert <- any(threshold_df > 0)

  message <- paste0(
    "tidypredict test results\n",
    "Difference threshold: ",
    threshold,
    "\n"
  )

  if (alert) {
    difference <- data.frame(fit_diff = max(raw_results$fit_diff))
    message <- paste0(
      message,
      "\nFitted records above the threshold: ",
      threshold_df$fit_threshold,
      "\n\nMax difference: ",
      difference$fit_diff
    )
  } else {
    message <- paste0(
      message,
      "\n All results are within the difference threshold"
    )
  }

  results <- list()
  results$model_call <- model$call
  results$raw_results <- raw_results
  results$message <- message
  results$alert <- alert
  structure(results, class = c("tidypredict_test", "list"))
}

#' @export
tidypredict_test.xgb.Booster <- function(
  model,
  df = model$model,
  threshold = 0.000000000001,
  include_intervals = FALSE,
  max_rows = NULL,
  xg_df = NULL
) {
  xgb_booster(
    model = model,
    df = df,
    threshold = threshold,
    include_intervals = include_intervals,
    max_rows = max_rows,
    xg_df = xg_df
  )
}

# Legacy method for old xgboost models with underscore prefix class
#' @export
tidypredict_test._xgb.Booster <- function(
  model,
  df = model$model,
  threshold = 0.000000000001,
  include_intervals = FALSE,
  max_rows = NULL,
  xg_df = NULL
) {
  # If this is also a model_fit (parsnip), delegate to that method
  if (inherits(model, "model_fit")) {
    return(tidypredict_test.model_fit(
      model = model,
      df = df,
      threshold = threshold,
      include_intervals = include_intervals,
      max_rows = max_rows,
      xg_df = xg_df
    ))
  }
  xgb_booster(
    model = model,
    df = df,
    threshold = threshold,
    include_intervals = include_intervals,
    max_rows = max_rows,
    xg_df = xg_df %||% df
  )
}

xgb_booster <- function(
  model,
  df = model$model,
  threshold = 0.000000000001,
  include_intervals = FALSE,
  max_rows = NULL,
  xg_df = NULL
) {
  if (is.numeric(max_rows)) {
    df <- head(df, max_rows)
  }
  base <- predict(model, xg_df)
  te <- tidypredict_to_column(
    df,
    model,
    add_interval = FALSE,
    vars = c("fit_te", "upr_te", "lwr_te")
  )
  raw_results <- cbind(base, te)
  raw_results$fit_diff <- raw_results$base - raw_results$fit_te
  raw_results$fit_threshold <- raw_results$fit_diff > threshold

  rowid <- seq_len(nrow(raw_results))
  raw_results <- cbind(data.frame(rowid), raw_results)

  threshold_df <- data.frame(fit_threshold = sum(raw_results$fit_threshold))
  alert <- any(threshold_df > 0)
  message <- paste0(
    "tidypredict test results\n",
    "Difference threshold: ",
    threshold,
    "\n"
  )

  if (alert) {
    difference <- data.frame(fit_diff = max(raw_results$fit_diff))
    message <- paste0(
      message,
      "\nFitted records above the threshold: ",
      threshold_df$fit_threshold,
      "\n\nMax difference: ",
      difference$fit_diff
    )
  } else {
    message <- paste0(
      message,
      "\n All results are within the difference threshold"
    )
  }
  results <- list()
  results$model_call <- model$call
  results$raw_results <- raw_results
  results$message <- message
  results$alert <- alert
  structure(results, class = c("tidypredict_test", "list"))
}

setOldClass(c("tidypredict_test", "list"))

#' @export
tidypredict_test.lgb.Booster <- function(
  model,
  df = NULL,
  threshold = 0.000000000001,
  include_intervals = FALSE,
  max_rows = NULL,
  xg_df = NULL
) {
  lgb_booster(
    model = model,
    df = df,
    threshold = threshold,
    include_intervals = include_intervals,
    max_rows = max_rows,
    lgb_df = xg_df
  )
}

lgb_booster <- function(
  model,
  df = NULL,
  threshold = 0.000000000001,
  include_intervals = FALSE,
  max_rows = NULL,
  lgb_df = NULL
) {
  if (is.null(lgb_df)) {
    cli::cli_abort(
      c(
        "LightGBM models require a matrix for predictions.",
        "i" = "Pass the prediction matrix via the {.arg xg_df} argument."
      )
    )
  }

  if (is.null(df)) {
    df <- as.data.frame(lgb_df)
  }

  if (is.numeric(max_rows)) {
    df <- head(df, max_rows)
    lgb_df <- lgb_df[seq_len(max_rows), , drop = FALSE]
  }

  base <- predict(model, lgb_df)

  # Check for multiclass (returns matrix)
  if (is.matrix(base)) {
    cli::cli_abort(
      c(
        "tidypredict_test does not support multiclass LightGBM models.",
        "i" = "Use tidypredict_fit() directly for multiclass predictions."
      )
    )
  }

  te <- tidypredict_to_column(
    df,
    model,
    add_interval = FALSE,
    vars = c("fit_te", "upr_te", "lwr_te")
  )

  raw_results <- cbind(data.frame(base = base), te)
  raw_results$fit_diff <- abs(raw_results$base - raw_results$fit_te)
  raw_results$fit_threshold <- raw_results$fit_diff > threshold

  rowid <- seq_len(nrow(raw_results))
  raw_results <- cbind(data.frame(rowid), raw_results)

  threshold_df <- data.frame(fit_threshold = sum(raw_results$fit_threshold))
  alert <- any(threshold_df > 0)

  message <- paste0(
    "tidypredict test results\n",
    "Difference threshold: ",
    threshold,
    "\n"
  )

  if (alert) {
    difference <- data.frame(fit_diff = max(raw_results$fit_diff))
    message <- paste0(
      message,
      "\nFitted records above the threshold: ",
      threshold_df$fit_threshold,
      "\n\nMax difference: ",
      difference$fit_diff
    )
  } else {
    message <- paste0(
      message,
      "\n All results are within the difference threshold"
    )
  }

  results <- list()
  results$raw_results <- raw_results
  results$message <- message
  results$alert <- alert
  structure(results, class = c("tidypredict_test", "list"))
}

#' @export
tidypredict_test.model_fit <- function(
  model,
  df = model$model,
  threshold = 0.000000000001,
  include_intervals = FALSE,
  max_rows = NULL,
  xg_df = NULL
) {
  tidypredict_test(
    model = model$fit,
    df = df,
    threshold = threshold,
    include_intervals = include_intervals,
    max_rows = max_rows,
    xg_df = xg_df
  )
}

#' @export
tidypredict_test.catboost.Model <- function(
  model,
  df = NULL,
  threshold = 0.000000000001,
  include_intervals = FALSE,
  max_rows = NULL,
  xg_df = NULL
) {
  catboost_model(
    model = model,
    df = df,
    threshold = threshold,
    include_intervals = include_intervals,
    max_rows = max_rows,
    cb_df = xg_df
  )
}

catboost_model <- function(
  model,
  df = NULL,
  threshold = 0.000000000001,
  include_intervals = FALSE,
  max_rows = NULL,
  cb_df = NULL
) {
  if (is.null(cb_df)) {
    cli::cli_abort(
      c(
        "CatBoost models require a matrix for predictions.",
        "i" = "Pass the prediction matrix via the {.arg xg_df} argument."
      )
    )
  }

  if (is.null(df)) {
    df <- as.data.frame(cb_df)
  }

  if (is.numeric(max_rows)) {
    df <- head(df, max_rows)
    cb_df <- cb_df[seq_len(max_rows), , drop = FALSE]
  }

  # Create pool for prediction
  pool <- catboost_catboost.load_pool(cb_df)

  # Detect objective type
  pm <- parse_model(model)
  objective <- pm$general$params$objective
  is_binary <- !is.null(objective) &&
    objective %in% c("Logloss", "CrossEntropy")
  is_multiclass <- !is.null(objective) &&
    objective %in% c("MultiClass", "MultiClassOneVsAll")

  if (is_multiclass) {
    return(catboost_model_multiclass(
      model,
      df,
      threshold,
      pool,
      pm,
      objective
    ))
  }

  if (is_binary) {
    base <- catboost_catboost.predict(
      model,
      pool,
      prediction_type = "Probability"
    )
  } else {
    base <- catboost_catboost.predict(model, pool)
  }

  te <- tidypredict_to_column(
    df,
    model,
    add_interval = FALSE,
    vars = c("fit_te", "upr_te", "lwr_te")
  )

  raw_results <- cbind(data.frame(base = base), te)
  raw_results$fit_diff <- abs(raw_results$base - raw_results$fit_te)
  raw_results$fit_threshold <- raw_results$fit_diff > threshold

  rowid <- seq_len(nrow(raw_results))
  raw_results <- cbind(data.frame(rowid), raw_results)

  threshold_df <- data.frame(fit_threshold = sum(raw_results$fit_threshold))
  alert <- any(threshold_df > 0)

  message <- paste0(
    "tidypredict test results\n",
    "Difference threshold: ",
    threshold,
    "\n"
  )

  if (alert) {
    difference <- data.frame(fit_diff = max(raw_results$fit_diff))
    message <- paste0(
      message,
      "\nFitted records above the threshold: ",
      threshold_df$fit_threshold,
      "\n\nMax difference: ",
      difference$fit_diff
    )
  } else {
    message <- paste0(
      message,
      "\n All results are within the difference threshold"
    )
  }

  results <- list()
  results$raw_results <- raw_results
  results$message <- message
  results$alert <- alert
  structure(results, class = c("tidypredict_test", "list"))
}

catboost_model_multiclass <- function(
  model,
  df,
  threshold,
  pool,
  pm,
  objective
) {
  num_class <- pm$general$num_class

  # Get native predictions as matrix
  base <- catboost_catboost.predict(
    model,
    pool,
    prediction_type = "Probability"
  )

  # Get tidypredict formulas (returns a list)
  formulas <- tidypredict_fit(model)

  # Evaluate each class formula
  te_preds <- lapply(formulas, function(f) rlang::eval_tidy(f, df))
  te_matrix <- do.call(cbind, te_preds)

  # Compare predictions
  diffs <- abs(base - te_matrix)
  alert <- any(diffs > threshold)

  message <- paste0(
    "tidypredict test results (multiclass: ",
    num_class,
    " classes)\n",
    "Difference threshold: ",
    threshold,
    "\n"
  )

  message <- paste0(
    message,
    "\n All results are within the difference threshold"
  )

  # Build raw_results for consistency
  raw_results <- data.frame(rowid = seq_len(nrow(df)))
  for (i in seq_len(num_class)) {
    raw_results[[paste0("base_class_", i - 1)]] <- base[, i]
    raw_results[[paste0("te_class_", i - 1)]] <- te_matrix[, i]
    raw_results[[paste0("diff_class_", i - 1)]] <- diffs[, i]
  }
  raw_results$max_diff <- apply(diffs, 1, max)
  raw_results$fit_threshold <- raw_results$max_diff > threshold

  results <- list()
  results$raw_results <- raw_results
  results$message <- message
  results$alert <- alert
  structure(results, class = c("tidypredict_test", "list"))
}

#' print method for test predictions results
#' @keywords internal
#' @export
print.tidypredict_test <- function(x, ...) {
  cat(x$message)
}

#' Knit print method for test predictions results
#' @keywords internal
#' @export
knit_print.tidypredict_test <- function(x, ...) {
  x$message
}

skip_if_not_installed("randomForest")

# Helper to normalize floating-point numbers in snapshot output
# Replaces scientific notation numbers with "<number>" to avoid platform differences
scrub_floats <- function(x) {
  gsub("[0-9]+\\.?[0-9]*e[+-]?[0-9]+|[0-9]+\\.[0-9]{10,}", "<number>", x)
}

test_that("Tester returns warning", {
  t <- tidypredict_test(
    lm(mpg ~ wt, data = mtcars),
    threshold = 1
  )
  expect_false(t$alert)
})

test_that("Intervals returns list", {
  expect_s3_class(
    tidypredict_test(lm(mpg ~ ., data = mtcars), include_intervals = TRUE),
    "tidypredict_test"
  )
})

test_that("alert triggered with threshold = 0 (mocked)", {
  local_mocked_bindings(
    tidypredict_to_column = function(df, model, ...) {
      df$fit_te <- rep(999, nrow(df))
      df
    }
  )

  t <- tidypredict_test(
    lm(mpg ~ wt + cyl, data = mtcars),
    threshold = 0
  )
  expect_true(t$alert)
  expect_snapshot(cat(t$message), transform = scrub_floats)
})

test_that("alert with intervals (mocked)", {
  local_mocked_bindings(
    tidypredict_to_column = function(df, model, ...) {
      df$fit_te <- rep(999, nrow(df))
      df$upr_te <- rep(999, nrow(df))
      df$lwr_te <- rep(999, nrow(df))
      df
    }
  )

  t <- tidypredict_test(
    lm(mpg ~ wt + cyl + disp + hp + drat, data = mtcars),
    threshold = 0,
    include_intervals = TRUE
  )
  expect_true(t$alert)
  expect_snapshot(cat(t$message), transform = scrub_floats)
})

test_that("max_rows limits data", {
  t <- tidypredict_test(
    lm(mpg ~ wt, data = mtcars),
    max_rows = 5
  )
  expect_equal(nrow(t$raw_results), 5)
})

test_that("print method works", {
  t <- tidypredict_test(lm(mpg ~ wt, data = mtcars))
  expect_output(print(t), "tidypredict test results")
})

test_that("knit_print method works", {
  t <- tidypredict_test(lm(mpg ~ wt, data = mtcars))
  expect_match(knitr::knit_print(t), "tidypredict test results")
})

test_that("offset handling in tidypredict_test", {
  model <- lm(mpg ~ wt + cyl, offset = am, data = mtcars)
  t <- tidypredict_test(model)
  expect_false(t$alert)
})

test_that("xgboost alert branch", {
  skip_if_not_installed("xgboost")
  df <- mtcars[, c("wt", "cyl", "disp")]
  xg_mat <- xgboost::xgb.DMatrix(
    nthread = 1L,
    as.matrix(df),
    label = mtcars$mpg
  )
  model <- xgboost::xgb.train(
    params = list(
      max_depth = 2,
      eta = 0.5,
      objective = "reg:squarederror",
      nthread = 1
    ),
    data = xg_mat,
    nrounds = 3,
    verbose = 0
  )
  t <- tidypredict_test(model, df = df, threshold = 0, xg_df = xg_mat)
  expect_true(t$alert)
  expect_snapshot(cat(t$message), transform = scrub_floats)
})

test_that("glmnet alert branch (mocked)", {
  skip_if_not_installed("glmnet")
  df <- mtcars[, c("wt", "cyl", "disp")]
  model <- glmnet::glmnet(as.matrix(df), mtcars$mpg)
  model$lambda <- model$lambda[5]

  local_mocked_bindings(
    tidypredict_to_column = function(df, model, ...) {
      df$fit_te <- rep(999, nrow(df))
      df
    }
  )

  t <- tidypredict_test(model, df = df, threshold = 0)
  expect_true(t$alert)
  expect_snapshot(cat(t$message), transform = scrub_floats)
})

test_that("lightgbm alert branch (mocked)", {
  skip_if_not_installed("lightgbm")
  df <- mtcars[, c("wt", "cyl", "disp")]
  lgb_mat <- as.matrix(df)
  dtrain <- lightgbm::lgb.Dataset(
    params = list(num_threads = 1L),
    lgb_mat,
    label = mtcars$mpg,
    colnames = colnames(df)
  )
  model <- lightgbm::lgb.train(
    params = list(
      objective = "regression",
      num_leaves = 4L,
      min_data_in_leaf = 1L,
      num_threads = 1
    ),
    data = dtrain,
    nrounds = 10L,
    verbose = -1L
  )

  local_mocked_bindings(
    tidypredict_to_column = function(df, model, ...) {
      df$fit_te <- rep(999, nrow(df))
      df
    }
  )

  t <- tidypredict_test(model, df = df, threshold = 0, xg_df = lgb_mat)
  expect_true(t$alert)
  expect_snapshot(cat(t$message), transform = scrub_floats)
})

test_that("catboost max_rows", {
  skip_if_not_installed("catboost")
  df <- mtcars[, c("wt", "cyl", "disp")]
  cb_mat <- as.matrix(df)
  pool <- catboost::catboost.load_pool(
    cb_mat,
    label = mtcars$mpg,
    feature_names = as.list(colnames(df))
  )
  model <- catboost::catboost.train(
    pool,
    params = list(
      iterations = 3,
      depth = 2,
      loss_function = "RMSE",
      logging_level = "Silent",
      allow_writing_files = FALSE
    )
  )
  t <- tidypredict_test(model, df = df, xg_df = cb_mat, max_rows = 5)
  expect_equal(nrow(t$raw_results), 5)
})


test_that("catboost alert branch (mocked)", {
  skip_if_not_installed("catboost")
  df <- mtcars[, c("wt", "cyl", "disp")]
  cb_mat <- as.matrix(df)
  pool <- catboost::catboost.load_pool(
    cb_mat,
    label = mtcars$mpg,
    feature_names = as.list(colnames(df))
  )
  model <- catboost::catboost.train(
    pool,
    params = list(
      iterations = 3,
      depth = 2,
      loss_function = "RMSE",
      logging_level = "Silent",
      allow_writing_files = FALSE
    )
  )

  local_mocked_bindings(
    tidypredict_to_column = function(df, model, ...) {
      df$fit_te <- rep(999, nrow(df))
      df
    }
  )

  t <- tidypredict_test(model, df = df, threshold = 0, xg_df = cb_mat)
  expect_true(t$alert)
  expect_snapshot(cat(t$message), transform = scrub_floats)
})

test_that("works with parsnip model specification", {
  etitanic_fac <- dplyr::mutate(earth::etitanic, survived = as.factor(survived))

  # Classification
  model <- parsnip::fit(
    parsnip::set_engine(parsnip::mars(mode = "classification"), "earth"),
    survived ~ age + sibsp,
    data = etitanic_fac
  )

  expect_snapshot(
    tidypredict_test(
      model,
      df = etitanic_fac
    )
  )

  # Regression
  model <- parsnip::fit(
    parsnip::set_engine(parsnip::mars(mode = "regression"), "earth"),
    survived ~ age + sibsp,
    data = etitanic
  )

  expect_snapshot(
    tidypredict_test(
      model,
      df = etitanic_fac
    )
  )
})

# Script to regenerate xgb-v2-regression.rds
# Uses CRAN tidypredict 1.0.1 which produced v1 format (legacy)
# Note: Despite the filename, this is actually version 1 from CRAN

tmp_lib <- tempfile("lib")
dir.create(tmp_lib)
install.packages(
  "tidypredict",
  lib = tmp_lib,
  repos = "https://cloud.r-project.org"
)
library(tidypredict, lib.loc = tmp_lib)

library(xgboost)
set.seed(123)
dtrain <- xgb.DMatrix(
  as.matrix(mtcars[, c("mpg", "cyl")]),
  label = mtcars$hp,
  feature_names = c("mpg", "cyl")
)
model <- xgb.train(
  params = list(
    max_depth = 2,
    objective = "reg:squarederror",
    base_score = 0.5
  ),
  data = dtrain,
  nrounds = 3,
  verbose = 0
)

pm <- parse_model(model)
saveRDS(pm, "tests/testthat/backwards-compat/xgb-v2-regression.rds")

# Keep vignette building single-threaded, for the same reasons as
# tests/testthat.R: R CMD check reports a NOTE when the code uses more CPU time
# than elapsed time, and the toy models here are too small to gain anything from
# extra threads. Sourced from the setup chunk of every vignette that fits a
# model, and it has to run before the engines are loaded, since libgomp and
# OpenBLAS both latch their thread counts when they are loaded.
Sys.setenv(
  OMP_NUM_THREADS = 1,
  OMP_THREAD_LIMIT = 1,
  R_RANGER_NUM_THREADS = 1
)

# Debian and Ubuntu default to the pthread build of OpenBLAS, which uses every
# core and ignores OMP_NUM_THREADS, so it needs an in-process call.
if (requireNamespace("RhpcBLASctl", quietly = TRUE)) {
  RhpcBLASctl::blas_set_num_threads(1)
  RhpcBLASctl::omp_set_num_threads(1)
}

if (requireNamespace("data.table", quietly = TRUE)) {
  data.table::setDTthreads(1L)
}

# For {orbital}
#' Extract comprehensive tree info for rpart models
#'
#' Returns tree structure in format needed by nested case_when generator.
#' For use in orbital package.
#' @param model An rpart model object
#' @keywords internal
#' @export
.rpart_tree_info_full <- function(model) {
  rpart_tree_info_full(model)
}

# Extract comprehensive tree info including surrogate splits
rpart_tree_info_full <- function(model) {
  frame <- model$frame
  splits <- model$splits
  orig_node_ids <- as.integer(rownames(frame))
  is_terminal <- frame$var == "<leaf>"

  # Check if surrogates are used during prediction
  # usesurrogate: 0 = don't use (NAs stop at internal nodes)
  #               1 = use surrogates, if all NA stop at internal node
  #               2 = use surrogates, if all NA go to majority (default)
  usesurr <- model$control$usesurrogate
  use_surrogates <- !identical(usesurr, 0L) && !identical(usesurr, 0)

  # Create mapping from original rpart node IDs to sequential 0-indexed IDs
  id_map <- setNames(seq_along(orig_node_ids) - 1L, orig_node_ids)

  # Build child relationships
  left_candidates <- 2L * orig_node_ids
  right_candidates <- 2L * orig_node_ids + 1L

  rpart_left <- ifelse(
    left_candidates %in% orig_node_ids,
    id_map[as.character(left_candidates)],
    NA_integer_
  )
  rpart_right <- ifelse(
    right_candidates %in% orig_node_ids,
    id_map[as.character(right_candidates)],
    NA_integer_
  )

  # Get predictions for ALL nodes (needed for usesurrogate=0)
  if (model$method == "class") {
    ylevels <- attr(model, "ylevels")
    prediction <- ylevels[frame$yval]
  } else {
    prediction <- frame$yval
  }

  # Extract split info per node, including surrogates
  n_nodes <- nrow(frame)
  node_splits <- vector("list", n_nodes)
  needs_swap <- rep(FALSE, n_nodes)
  majority_left <- rep(NA, n_nodes)

  if (!is.null(splits) && nrow(splits) > 0) {
    split_idx <- 1
    for (i in seq_len(n_nodes)) {
      if (!is_terminal[i]) {
        n_compete <- frame$ncompete[i]
        n_surr <- frame$nsurrogate[i]

        # Primary split
        primary <- extract_one_split(
          splits,
          split_idx,
          model,
          as.character(frame$var[i])
        )
        needs_swap[i] <- primary$needs_swap

        # Surrogate splits (skip competing splits)
        # Only extract if usesurrogate != 0
        surrogates <- list()
        if (use_surrogates) {
          surr_start <- split_idx + 1 + n_compete
          for (j in seq_len(n_surr)) {
            surr_idx <- surr_start + j - 1
            surr_var <- rownames(splits)[surr_idx]
            surr_info <- extract_one_split(splits, surr_idx, model, surr_var)
            # Adjust surrogate direction based on primary direction
            if (primary$needs_swap) {
              surr_info$needs_swap <- !surr_info$needs_swap
            }
            surrogates[[j]] <- surr_info
          }
        }

        # Majority direction: left child has more observations
        left_id <- 2L * orig_node_ids[i]
        right_id <- 2L * orig_node_ids[i] + 1L
        left_n <- frame$n[orig_node_ids == left_id]
        right_n <- frame$n[orig_node_ids == right_id]
        majority_left[i] <- left_n >= right_n

        node_splits[[i]] <- list(
          primary = primary,
          surrogates = surrogates
        )

        n_splits <- 1 + n_compete + n_surr
        split_idx <- split_idx + n_splits
      }
    }
  }

  # Swap left/right based on ncat direction
  # This aligns with tidypredict convention: "left" means "< split value"
  left_child <- ifelse(needs_swap, rpart_right, rpart_left)
  right_child <- ifelse(needs_swap, rpart_left, rpart_right)

  # Also swap majority_left to stay consistent with the new left/right
  majority_left_adjusted <- ifelse(needs_swap, !majority_left, majority_left)

  list(
    nodeID = seq_along(orig_node_ids) - 1L,
    leftChild = left_child,
    rightChild = right_child,
    splitvarName = as.character(frame$var),
    terminal = is_terminal,
    prediction = prediction,
    node_splits = node_splits,
    majority_left = majority_left_adjusted,
    use_surrogates = use_surrogates
  )
}

# Extract info for a single split from the splits matrix
extract_one_split <- function(splits, idx, model, var_name) {
  ncat <- splits[idx, "ncat"]
  index <- splits[idx, "index"]

  if (abs(ncat) == 1) {
    # Continuous split
    list(
      col = var_name,
      val = index,
      is_categorical = FALSE,
      needs_swap = ncat == 1
    )
  } else {
    # Categorical split
    csplit_row <- model$csplit[index, , drop = TRUE]
    xlevels <- attr(model, "xlevels")[[var_name]]
    left_levels <- xlevels[csplit_row == 1]
    list(
      col = var_name,
      vals = as.list(left_levels),
      is_categorical = TRUE,
      needs_swap = FALSE
    )
  }
}

# Simplified tree info for tests (returns data frame like other models)
rpart_tree_info <- function(model) {
  info <- rpart_tree_info_full(model)

  split_val <- rep(NA_real_, length(info$nodeID))
  split_class <- rep(NA_character_, length(info$nodeID))

  for (i in seq_along(info$node_splits)) {
    if (!is.null(info$node_splits[[i]])) {
      primary <- info$node_splits[[i]]$primary
      if (primary$is_categorical) {
        split_class[i] <- paste0(primary$vals, collapse = ", ")
      } else {
        split_val[i] <- primary$val
      }
    }
  }

  data.frame(
    nodeID = info$nodeID,
    leftChild = info$leftChild,
    rightChild = info$rightChild,
    splitvarName = info$splitvarName,
    splitval = split_val,
    splitclass = split_class,
    terminal = info$terminal,
    prediction = info$prediction
  )
}

#' @export
parse_model.rpart <- function(model) {
  pm <- list()
  pm$general$model <- "rpart"
  pm$general$type <- "tree"
  pm$general$version <- 3
  pm$tree_info <- rpart_tree_info_full(model)
  as_parsed_model(pm)
}

#' @export
tidypredict_fit.rpart <- function(model, ...) {
  tree_info <- rpart_tree_info_full(model)
  generate_nested_case_when_tree(tree_info)
}

#' @export
tidypredict_test.rpart <- function(
  model,
  df = model$model,
  threshold = 0.000000000001,
  include_intervals = FALSE,

  max_rows = NULL,
  xg_df = NULL
) {
  if (is.numeric(max_rows)) {
    df <- head(df, max_rows)
  }

  # rpart uses type = "vector" for regression, type = "class" for classification
  pred_type <- if (model$method == "class") "class" else "vector"
  base <- predict(model, df, type = pred_type)

  # For classification, threshold should be 0 (exact match)
  if (model$method == "class") {
    threshold <- 0
    base <- as.character(base)
  }

  te <- tidypredict_to_column(
    df,
    model,
    add_interval = FALSE,
    vars = c("fit_te", "upr_te", "lwr_te")
  )

  raw_results <- data.frame(fit = base, fit_te = te$fit_te)
  raw_results$fit_diff <- if (model$method == "class") {
    as.numeric(raw_results$fit != raw_results$fit_te)
  } else {
    raw_results$fit - raw_results$fit_te
  }
  raw_results$fit_threshold <- abs(raw_results$fit_diff) > threshold

  rowid <- seq_len(nrow(raw_results))
  raw_results <- cbind(data.frame(rowid), raw_results)

  threshold_df <- data.frame(fit_threshold = sum(raw_results$fit_threshold))
  alert <- any(threshold_df > 0)

  message <- paste0(
    "tidypredict test results\n",
    "Difference threshold: ",
    threshold,
    "\n"
  )

  if (alert) {
    difference <- max(abs(raw_results$fit_diff))
    message <- paste0(
      message,
      "\nFitted records above the threshold: ",
      threshold_df$fit_threshold,
      "\n\nMax difference: ",
      difference
    )
  } else {
    message <- paste0(
      message,
      "\n All results are within the difference threshold"
    )
  }

  results <- list()
  results$model_call <- model$call
  results$raw_results <- raw_results
  results$message <- message
  results$alert <- alert
  structure(results, class = c("tidypredict_test", "list"))
}

# For {orbital}
#' Extract classprob trees for rpart models
#'
#' For use in orbital package.
#' @param model An rpart model object
#' @keywords internal
#' @export
.extract_rpart_classprob <- function(model) {
  if (!inherits(model, "rpart")) {
    cli::cli_abort(
      "{.arg model} must be {.cls rpart}, not {.obj_type_friendly {model}}."
    )
  }

  if (model$method != "class") {
    cli::cli_abort(
      "{.arg model} must be a classification model (method = 'class')."
    )
  }

  # Extract class probabilities from yval2
  # yval2 structure: [yval, count_class1, ..., count_classN, prob_class1, ..., prob_classN, nodeprob]
  yval2 <- model$frame$yval2
  ylevels <- attr(model, "ylevels")
  n_classes <- length(ylevels)

  # Probability columns are at positions (n_classes + 2) to (2 * n_classes + 1)
  prob_cols <- seq(n_classes + 2, 2 * n_classes + 1)
  probs <- yval2[, prob_cols, drop = FALSE]
  colnames(probs) <- ylevels

  # Get tree structure
  tree_info <- rpart_tree_info_full(model)

  res <- list()
  for (i in seq_len(ncol(probs))) {
    tree_info_copy <- tree_info
    tree_info_copy$prediction <- probs[, i]
    res[[i]] <- generate_nested_case_when_tree(tree_info_copy)
  }
  res
}

test_that("rpart_tree_info returns correct structure", {
  model <- rpart::rpart(mpg ~ cyl + wt, data = mtcars)
  tree_info <- rpart_tree_info(model)

  expect_s3_class(tree_info, "data.frame")
  expect_named(
    tree_info,
    c(
      "nodeID",
      "leftChild",
      "rightChild",
      "splitvarName",
      "splitval",
      "splitclass",
      "terminal",
      "prediction"
    )
  )
})

test_that("rpart_tree_info handles categorical predictors", {
  mtcars2 <- mtcars
  mtcars2$cyl <- factor(mtcars2$cyl)
  model <- rpart::rpart(mpg ~ cyl + wt, data = mtcars2)
  tree_info <- rpart_tree_info(model)

  # Should have splitclass values for categorical splits
  expect_true(any(!is.na(tree_info$splitclass)))
})

test_that("returns the right output", {
  model <- rpart::rpart(mpg ~ am + cyl, data = mtcars)
  tf <- tidypredict_fit(model)
  pm <- parse_model(model)

  expect_type(tf, "language")
  expect_s3_class(pm, "list")
  expect_equal(pm$general$model, "rpart")
  expect_equal(pm$general$version, 3)

  expect_snapshot(rlang::expr_text(tf))
})

test_that("tidypredict_fit produces correct predictions", {
  model <- rpart::rpart(mpg ~ am + cyl, data = mtcars)

  fit_expr <- tidypredict_fit(model)
  fit_pred <- dplyr::mutate(mtcars, pred = !!fit_expr)$pred
  original_pred <- predict(model, mtcars)

  expect_equal(fit_pred, unname(original_pred))
})

test_that("formulas produce correct predictions - regression", {
  expect_snapshot(
    tidypredict_test(
      rpart::rpart(mpg ~ am + cyl + wt, data = mtcars),
      mtcars
    )
  )
})

test_that("tidypredict_test.rpart max_rows parameter works", {
  model <- rpart::rpart(mpg ~ am + cyl + wt, data = mtcars)
  result <- tidypredict_test(model, mtcars, max_rows = 10)

  expect_equal(nrow(result$raw_results), 10)
})

test_that("tidypredict_test.rpart alert message works", {
  model <- rpart::rpart(mpg ~ am + cyl + wt, data = mtcars)

  # Use negative threshold to trigger alert
  result <- tidypredict_test(model, mtcars, threshold = -1)

  expect_true(result$alert)
  expect_match(result$message, "Fitted records above the threshold")
  expect_match(result$message, "Max difference")
})

test_that("formulas produce correct predictions - classification", {
  expect_snapshot(
    tidypredict_test(
      rpart::rpart(Species ~ ., data = iris),
      iris
    )
  )
})

test_that("categorical predictors work correctly", {
  mtcars2 <- mtcars
  mtcars2$cyl <- factor(mtcars2$cyl)

  expect_snapshot(
    tidypredict_test(
      rpart::rpart(mpg ~ cyl + wt, data = mtcars2),
      mtcars2
    )
  )
})

test_that("stump trees work correctly", {
  ctrl <- rpart::rpart.control(minsplit = 100, cp = 1)
  model <- rpart::rpart(mpg ~ cyl + disp, data = mtcars, control = ctrl)

  fit <- tidypredict_fit(model)

  expect_type(fit, "double")
  expect_equal(fit, mean(mtcars$mpg))
})

test_that("produced case_when uses .default", {
  model <- rpart::rpart(mpg ~ am + cyl, data = mtcars)

  fit <- tidypredict_fit(model)
  fit_text <- rlang::expr_text(fit)

  expect_match(fit_text, "\\.default")
})

# .extract_rpart_classprob tests ------------------------------------------

test_that(".extract_rpart_classprob returns list of expressions", {
  model <- rpart::rpart(Species ~ Sepal.Length + Sepal.Width, data = iris)

  exprs <- .extract_rpart_classprob(model)

  expect_type(exprs, "list")
  expect_length(exprs, 3)
  expect_true(all(vapply(exprs, typeof, character(1)) == "language"))
})

test_that(".extract_rpart_classprob results match predict probabilities", {
  model <- rpart::rpart(Species ~ Sepal.Length + Sepal.Width, data = iris)

  exprs <- .extract_rpart_classprob(model)
  eval_env <- rlang::new_environment(
    data = as.list(iris),
    parent = asNamespace("dplyr")
  )
  probs <- lapply(exprs, rlang::eval_tidy, env = eval_env)
  combined <- do.call(cbind, probs)

  native <- predict(model, type = "prob")

  expect_equal(unname(combined), unname(native))
})

test_that(".extract_rpart_classprob errors on non-rpart model", {
  expect_snapshot(.extract_rpart_classprob(list()), error = TRUE)
})

test_that(".extract_rpart_classprob errors on regression model", {
  model <- rpart::rpart(mpg ~ cyl + wt, data = mtcars)
  expect_snapshot(.extract_rpart_classprob(model), error = TRUE)
})

# Nested case_when tests --------------------------------------------------

test_that("tidypredict_fit matches original model predictions", {
  model <- rpart::rpart(mpg ~ cyl + wt, data = mtcars)

  fit_expr <- tidypredict_fit(model)
  fit_pred <- dplyr::mutate(mtcars, pred = !!fit_expr)$pred
  original_pred <- predict(model, mtcars)

  expect_equal(fit_pred, unname(original_pred))
})

test_that("tidypredict_fit works for classification", {
  model <- rpart::rpart(Species ~ ., data = iris)

  fit_expr <- tidypredict_fit(model)
  fit_pred <- dplyr::mutate(iris, pred = !!fit_expr)$pred
  original_pred <- as.character(predict(model, iris, type = "class"))

  expect_equal(fit_pred, original_pred)
})

test_that(".extract_rpart_classprob matches original model probabilities", {
  model <- rpart::rpart(Species ~ Sepal.Length + Sepal.Width, data = iris)

  exprs <- .extract_rpart_classprob(model)

  eval_env <- rlang::new_environment(
    data = as.list(iris),
    parent = asNamespace("dplyr")
  )

  probs <- lapply(exprs, rlang::eval_tidy, env = eval_env)
  combined <- do.call(cbind, probs)
  native <- predict(model, type = "prob")

  expect_equal(unname(combined), unname(native))
})

test_that(".rpart_tree_info_full is exported and works", {
  model <- rpart::rpart(mpg ~ cyl + wt, data = mtcars)

  tree_info <- .rpart_tree_info_full(model)

  expect_type(tree_info, "list")
  expect_named(
    tree_info,
    c(
      "nodeID",
      "leftChild",
      "rightChild",
      "splitvarName",
      "terminal",
      "prediction",
      "node_splits",
      "majority_left",
      "use_surrogates"
    )
  )
})

#' @export
parse_model.cubist <- function(model) {
  coefs <- model$coefficients
  splits <- model$splits
  if (!is.null(splits)) {
    splits$variable <- as.character(splits$variable)
    splits$dir <- as.character(splits$dir)
  }

  # Pre-split data by committee and rule to avoid O(n) scans in nested loops
  coefs_by_comm_rule <- split(coefs, list(coefs$committee, coefs$rule))
  if (!is.null(splits)) {
    splits_by_comm_rule <- split(splits, list(splits$committee, splits$rule))
  }

  committees2 <- map(
    unique(coefs$committee),
    ~ {
      comm <- .x
      rules <- map(
        coefs$rule[coefs$committee == comm],
        ~ {
          key <- paste(comm, .x, sep = ".")
          cc <- coefs_by_comm_rule[[key]]
          if (!is.null(model$splits)) {
            cs <- splits_by_comm_rule[[key]]
            if (!is.null(cs) && nrow(cs) > 0) {
              tcs <- transpose(cs)
              mcs <- map(
                tcs,
                ~ list(
                  type = "conditional",
                  col = .x$variable,
                  val = .x$value,
                  op = ifelse(.x$dir == ">", "more", "less-equal")
                )
              )
            } else {
              mcs <- list(list(type = "all"))
            }
          } else {
            mcs <- list(list(type = "all"))
          }
          cc_names <- names(cc)
          f_coefs <- map(
            seq_along(cc_names),
            ~ {
              if (cc_names[.x] == "(Intercept)") {
                op <- "none"
                is_intercept <- 1
              } else {
                op <- "multiply"
                is_intercept <- 0
              }
              list(
                col = cc_names[.x],
                val = cc[, .x],
                op = op,
                is_intercept = is_intercept
              )
            }
          )

          f_na <- map_lgl(
            seq_along(cc_names),
            ~ !is.na(cc[, .x])
          )
          f_coefs <- f_coefs[f_na]
          f_coefs <- f_coefs[1:(length(f_coefs) - 2)]

          list(
            prediction = f_coefs,
            path = mcs
          )
        }
      )
    }
  )
  comm <- purrr::list_flatten(committees2)

  if (model$committees == 1) {
    ommittee_id <- rep(1, length(comm))
  } else {
    model_print <- utils::capture.output(print(model))
    model_print <- model_print[grep(
      "Number of rules per committee",
      model_print
    )]
    model_print <- regmatches(
      model_print,
      m = gregexpr("[0-9]+", model_print)
    )[[
      1
    ]]
    ommittee_id <- as.integer(model_print)
    ommittee_id <- rep(seq_along(ommittee_id), times = ommittee_id)
  }

  pm <- list(
    general = list(
      model = "cubist",
      type = "tree",
      version = 3,
      mode = "ifelse",
      n_committees = model$committees,
      ommittee_id = ommittee_id
    ),
    trees = list(comm)
  )
  as_parsed_model(pm)
}

#' @export
tidypredict_fit.cubist <- function(model) {
  parsedmodel <- parse_model(model)
  tidypredict_fit_cubist(parsedmodel)
}

tidypredict_fit_cubist <- function(parsedmodel) {
  rules <- generate_tree_nodes(parsedmodel$trees[[1]], parsedmodel$general$mode)
  paths <- lapply(parsedmodel$trees[[1]], function(x) path_formulas(x$path))

  n_committees <- parsedmodel$general$n_committees
  ommittee_id <- parsedmodel$general$ommittee_id

  committees <- purrr::map2(
    split(rules, ommittee_id),
    split(paths, ommittee_id),
    make_committee
  )

  out <- reduce_addition(committees)
  if (n_committees > 1) {
    # Average the committes
    out <- expr_division(out, n_committees)
  }

  out
}

make_committee <- function(rules, paths) {
  # cubist averages out rules if multiple apply
  paths <- lapply(paths, function(x) x %||% TRUE)
  paths <- reduce_addition(paths)
  rules <- reduce_addition(rules)
  if (identical(paths, TRUE)) {
    res <- rules
  } else {
    res <- expr_division(rules, paths)
  }
  res
}

#' Returns a Tidy Eval formula to calculate fitted values
#'
#' It parses a model or uses an already parsed model to return a
#' Tidy Eval formula that can then be used inside a dplyr command.
#'
#' @param model An R model or a list with a parsed model.
#'
#' @examples
#'
#' model <- lm(mpg ~ wt + cyl * disp, offset = am, data = mtcars)
#' tidypredict_fit(model)
#' @export
tidypredict_fit <- function(model) {
  UseMethod("tidypredict_fit")
}

#' @export
tidypredict_fit.pm_regression <- function(model) {
  build_fit_formula(model)
}

#' @export
tidypredict_fit.pm_tree <- function(model) {
  version <- model$general$version %||% 1

  # Version 3: nested case_when format
  if (version >= 3) {
    model_type <- model$general$model
    if (model_type == "cubist") {
      return(tidypredict_fit_cubist(model))
    }
    if (model_type %in% c("rpart", "party")) {
      return(generate_nested_case_when_tree(model$tree_info))
    }
    if (model_type %in% c("ranger", "randomForest")) {
      # For forests, average all trees
      tree_exprs <- map(model$tree_info_list, generate_nested_case_when_tree)
      res <- reduce_addition(tree_exprs)
      return(expr_division(res, length(tree_exprs)))
    }
  }

  # Version 1/2: flat case_when format (backwards compatibility with saved models)
  if (model$general$model == "cubist") {
    return(tidypredict_fit_cubist(model))
  }
  if (model$general$model == "randomForest") {
    return(tidypredict_fit_randomForest(model))
  }
  if (model$general$model == "ranger") {
    return(tidypredict_fit_ranger(model))
  }
}

#' @export
tidypredict_fit.pm_xgb <- function(model) {
  version <- model$general$version %||% 1

  if (version >= 3) {
    return(build_fit_formula_xgb_from_parsed(model))
  }

  # Version 1/2: flat case_when (backwards compatibility)
  build_fit_formula_xgb(model)
}

#' @export
tidypredict_fit.pm_lgb <- function(model) {
  build_fit_formula_lgb_from_parsed(model)
}

#' @export
tidypredict_fit.pm_catboost <- function(model) {
  build_fit_formula_catboost_nested(model)
}

# For {orbital}
#' Extract comprehensive tree info for partykit models
#'
#' Returns tree structure in format needed by nested case_when generator.
#' For use in orbital package.
#' @param model A partykit model object
#' @keywords internal
#' @export
.partykit_tree_info_full <- function(model) {
  partykit_tree_info_full(model)
}

# Convert partykit tree info to the format needed by nested generator
partykit_tree_info_full <- function(model) {
  tree_df <- partykit_tree_info(model)

  # Build node_splits list in the format expected by nested generator
  node_splits <- vector("list", nrow(tree_df))

  for (i in seq_len(nrow(tree_df))) {
    if (!tree_df$terminal[i]) {
      var_name <- tree_df$splitvarName[i]

      if (!is.na(tree_df$splitclass[i])) {
        # Categorical split
        vals <- strsplit(tree_df$splitclass[i], ", ")[[1]]
        node_splits[[i]] <- list(
          primary = list(
            col = var_name,
            vals = as.list(vals),
            is_categorical = TRUE,
            needs_swap = FALSE
          ),
          surrogates = list()
        )
      } else {
        # Continuous split
        node_splits[[i]] <- list(
          primary = list(
            col = var_name,
            val = tree_df$splitval[i],
            is_categorical = FALSE,
            needs_swap = FALSE
          ),
          surrogates = list()
        )
      }
    }
  }

  list(
    nodeID = tree_df$nodeID,
    leftChild = tree_df$leftChild,
    rightChild = tree_df$rightChild,
    splitvarName = tree_df$splitvarName,
    terminal = tree_df$terminal,
    prediction = tree_df$prediction,
    node_splits = node_splits,
    majority_left = rep(NA, nrow(tree_df)),
    use_surrogates = FALSE
  )
}

partykit_tree_info <- function(model) {
  # Get all node IDs at once (avoids repeated tree traversals)
  all_node_ids <- partykit::nodeids(model)
  n_nodes <- length(all_node_ids)

  # Extract all nodes at once using nodeapply with all IDs
  all_nodes <- partykit::nodeapply(model, ids = all_node_ids, FUN = identity)

  # Pre-extract node properties to avoid repeated list access
  is_split <- logical(n_nodes)
  splitvarID <- integer(n_nodes)
  splitval <- numeric(n_nodes)
  split_index <- vector("list", n_nodes)
  left_child <- integer(n_nodes)
  right_child <- integer(n_nodes)

  for (i in seq_len(n_nodes)) {
    node <- all_nodes[[i]]
    is_split[i] <- !partykit::is.terminal(node)
    if (is_split[i]) {
      splitvarID[i] <- node$split$varid
      splitval[i] <- node$split$breaks %||% NA_real_
      split_index[[i]] <- node$split$index
      kids <- partykit::kids_node(node)
      left_child[i] <- partykit::id_node(kids[[1]])
      right_child[i] <- partykit::id_node(kids[[2]])
    } else {
      splitvarID[i] <- NA_integer_
      splitval[i] <- NA_real_
      left_child[i] <- NA_integer_
      right_child[i] <- NA_integer_
    }
  }

  # Extract predictions from fitted data (only need to access once)
  fitted_data <- model$fitted
  response_col <- fitted_data[["(response)"]]
  node_col <- fitted_data[["(fitted)"]]

  if (is.numeric(response_col)) {
    # Regression: compute mean per node
    node_means <- tapply(response_col, node_col, mean)
    prediction <- ifelse(!is_split, node_means[as.character(all_node_ids)], NA)
  } else {
    # Classification: compute mode per node
    stat_mode <- function(x) {
      counts <- rev(sort(table(x)))
      if (length(counts) > 1 && counts[[1]] == counts[[2]]) {
        ties <- counts[counts[1] == counts]
        return(names(rev(ties))[1])
      }
      names(counts)[1]
    }
    node_modes <- tapply(response_col, node_col, stat_mode)
    prediction <- ifelse(!is_split, node_modes[as.character(all_node_ids)], NA)
  }

  # Get variable info
  vars <- as.character(attr(model$terms, "variables"))
  vars <- vars[2:length(vars)]

  var_details <- map_chr(model$data, class)
  var_class <- as.character(var_details)
  var_name <- names(var_details)

  # Build categorical split strings
  class_splits <- character(n_nodes)
  for (i in seq_len(n_nodes)) {
    if (is.na(splitvarID[i])) {
      class_splits[i] <- NA_character_
    } else {
      v <- vars[splitvarID[i]]
      if (var_class[var_name == v] == "factor") {
        lvls <- levels(model$data[, colnames(model$data) == v])
        pn <- split_index[[i]]
        pn <- ifelse(is.na(pn), 0, pn)
        class_splits[i] <- paste0(lvls[pn == 1], collapse = ", ")
      } else {
        class_splits[i] <- NA_character_
      }
    }
  }

  data.frame(
    nodeID = all_node_ids - 1L,
    leftChild = left_child - 1L,
    rightChild = right_child - 1L,
    splitvarID = splitvarID,
    splitvarName = vars[splitvarID],
    splitval = splitval,
    splitclass = class_splits,
    terminal = !is_split,
    prediction = prediction
  )
}

#' @export
parse_model.party <- function(model) {
  pm <- list()
  pm$general$model <- "party"
  pm$general$type <- "tree"
  pm$general$version <- 3
  pm$tree_info <- partykit_tree_info_full(model)
  as_parsed_model(pm)
}

# Fit formula -----------------------------------

#' @export
tidypredict_fit.party <- function(model, ...) {
  tree_info <- partykit_tree_info_full(model)
  generate_nested_case_when_tree(tree_info)
}

# For {orbital}
#' Extract classprob trees for partykit models
#'
#' For use in orbital package.
#' @param model A partykit model object
#' @keywords internal
#' @export
.extract_partykit_classprob <- function(model) {
  if (!inherits(model, "party")) {
    cli::cli_abort(
      "{.arg model} must be {.cls party}, not {.obj_type_friendly {model}}."
    )
  }

  extract_classprob <- function(model) {
    mod <- model$fitted
    response <- mod[["(response)"]]
    weights <- mod[["(weights)"]]

    lvls <- levels(response)
    weights_sum <- tapply(weights, response, sum)
    weights_sum[is.na(weights_sum)] <- 0
    res <- weights_sum / sum(weights)
    names(res) <- lvls
    res
  }

  preds <- map(seq_along(model), ~ extract_classprob(model[[.x]]))
  preds <- matrix(
    unlist(preds),
    nrow = length(preds),
    byrow = TRUE,
    dimnames = list(NULL, names(preds[[1]]))
  )

  tree_info_full <- partykit_tree_info_full(model)

  res <- list()
  for (i in seq_len(ncol(preds))) {
    tree_info_copy <- tree_info_full
    tree_info_copy$prediction <- preds[, i]
    res[[i]] <- generate_nested_case_when_tree(tree_info_copy)
  }
  res
}

test_that("returns the right output", {
  model <- partykit::ctree(mpg ~ am + cyl, data = mtcars)
  tf <- tidypredict_fit(model)
  pm <- parse_model(model)

  expect_type(tf, "language")

  expect_s3_class(pm, "list")
  expect_equal(length(pm), 2)
  expect_equal(pm$general$model, "party")
  expect_equal(pm$general$version, 3)

  expect_snapshot(
    rlang::expr_text(tf)
  )
})

test_that("tidypredict_fit produces correct predictions", {
  model <- partykit::ctree(mpg ~ am + cyl, data = mtcars)

  fit_expr <- tidypredict_fit(model)
  fit_pred <- dplyr::mutate(mtcars, pred = !!fit_expr)$pred
  original_pred <- as.vector(predict(model, mtcars))

  expect_equal(fit_pred, original_pred)
})

test_that("formulas produces correct predictions", {
  mtcars <- mtcars
  mtcars$am1 <- mtcars$am
  mtcars$am <- ifelse(mtcars$am == 1, "auto", "man")
  mtcars$am <- as.factor(mtcars$am)

  mtcars$cyl <- ifelse(mtcars$cyl == 4, "four", mtcars$cyl)
  mtcars$cyl <- ifelse(mtcars$cyl == 6, "six", mtcars$cyl)
  mtcars$cyl <- ifelse(mtcars$cyl == 8, "eight", mtcars$cyl)
  mtcars$cyl <- as.factor(mtcars$cyl)

  # normal
  expect_snapshot(
    tidypredict_test(
      partykit::ctree(mpg ~ am + cyl, data = mtcars),
      mtcars
    )
  )

  # offset
  expect_snapshot(
    tidypredict_test(
      partykit::ctree(mpg ~ wt, offset = am1, data = mtcars),
      mtcars
    )
  )

  # interaction
  expect_snapshot(
    tidypredict_test(
      partykit::ctree(mpg ~ wt + disp * cyl, data = mtcars),
      mtcars
    )
  )

  # interactions
  expect_snapshot(
    tidypredict_test(
      partykit::ctree(mpg ~ (wt + disp) * cyl, data = mtcars),
      mtcars
    )
  )
})

# .extract_partykit_classprob tests ------------------------------------------

test_that(".extract_partykit_classprob returns list of expressions", {
  model <- partykit::ctree(Species ~ Sepal.Length + Sepal.Width, data = iris)

  exprs <- .extract_partykit_classprob(model)

  expect_type(exprs, "list")
  expect_length(exprs, 3)
  expect_true(all(vapply(exprs, typeof, character(1)) == "language"))
})

test_that(".extract_partykit_classprob results match predict probabilities", {
  model <- partykit::ctree(Species ~ Sepal.Length + Sepal.Width, data = iris)

  exprs <- .extract_partykit_classprob(model)
  eval_env <- rlang::new_environment(
    data = as.list(iris),
    parent = asNamespace("dplyr")
  )
  probs <- lapply(exprs, rlang::eval_tidy, env = eval_env)
  combined <- do.call(cbind, probs)

  native <- predict(model, type = "prob")

  expect_equal(unname(combined), unname(native))
})

test_that(".extract_partykit_classprob errors on non-party model", {
  expect_snapshot(.extract_partykit_classprob(list()), error = TRUE)
})

test_that("stump trees (no splits) work correctly (#196)", {
  ctrl <- partykit::ctree_control(mincriterion = 0.9999999)
  model <- partykit::ctree(mpg ~ cyl + disp + hp, data = mtcars, control = ctrl)

  # Verify it's a stump (only root node, no splits)
  expect_equal(length(partykit::nodeids(model, terminal = TRUE)), 1)

  fit <- tidypredict_fit(model)

  expect_type(fit, "double")
  expect_equal(fit, mean(mtcars$mpg))
})

test_that("produced case_when uses .default", {
  model <- partykit::ctree(mpg ~ am + cyl, data = mtcars)

  fit <- tidypredict_fit(model)
  fit_text <- rlang::expr_text(fit)

  expect_match(fit_text, "\\.default")
})

# Nested case_when tests --------------------------------------------------

test_that("tidypredict_fit matches original model predictions", {
  model <- partykit::ctree(mpg ~ cyl + wt, data = mtcars)

  fit_expr <- tidypredict_fit(model)
  fit_pred <- dplyr::mutate(mtcars, pred = !!fit_expr)$pred
  original_pred <- predict(model, mtcars)

  expect_equal(fit_pred, as.vector(original_pred))
})

test_that("tidypredict_fit works for classification", {
  model <- partykit::ctree(Species ~ ., data = iris)

  fit_expr <- tidypredict_fit(model)
  fit_pred <- dplyr::mutate(iris, pred = !!fit_expr)$pred
  original_pred <- as.character(predict(model, iris, type = "response"))

  expect_equal(fit_pred, original_pred)
})

test_that(".extract_partykit_classprob matches original model probabilities", {
  model <- partykit::ctree(Species ~ Sepal.Length + Sepal.Width, data = iris)

  exprs <- .extract_partykit_classprob(model)

  eval_env <- rlang::new_environment(
    data = as.list(iris),
    parent = asNamespace("dplyr")
  )

  probs <- lapply(exprs, rlang::eval_tidy, env = eval_env)
  combined <- do.call(cbind, probs)
  native <- predict(model, type = "prob")

  expect_equal(unname(combined), unname(native))
})

test_that(".partykit_tree_info_full is exported and works", {
  model <- partykit::ctree(mpg ~ cyl + wt, data = mtcars)

  tree_info <- .partykit_tree_info_full(model)

  expect_type(tree_info, "list")
  expect_named(
    tree_info,
    c(
      "nodeID",
      "leftChild",
      "rightChild",
      "splitvarName",
      "terminal",
      "prediction",
      "node_splits",
      "majority_left",
      "use_surrogates"
    )
  )
})

#' Returns a Tidy Eval formula to calculate prediction interval.
#'
#' It parses a model or uses an already parsed model to return a
#' Tidy Eval formula that can then be used inside a dplyr command.
#'
#' The result still has to be added to and subtracted from the fit to obtain the upper and
#' lower bound respectively.
#'
#' @param model An R model or a list with a parsed model
#' @param interval The prediction interval, defaults to 0.95
#'
#' @examples
#'
#' model <- lm(mpg ~ wt + cyl * disp, offset = am, data = mtcars)
#' tidypredict_interval(model)
#' @export
tidypredict_interval <- function(model, interval = 0.95) {
  UseMethod("tidypredict_interval")
}

#' @export
`tidypredict_interval.data.frame` <- function(model, interval = 0.95) {
  cli::cli_abort("data.frame based parsed models are no longer supported.")
}

#' @export
tidypredict_interval.list <- function(model, interval = 0.95) {
  mt <- model$general$model
  fit <- NULL
  if (mt == "lm") {
    fit <- te_interval_lm(model)
  }
  if (mt == "glm") {
    fit <- te_interval_glm(model)
  }
  if (is.null(fit)) {
    cli::cli_abort("Model type not supported.")
  }
  fit
}

#' Checks that the formula can be parsed
#'
#' Uses an S3 method to check that a given formula can be parsed based on its class.
#' It currently scans for contrasts that are not supported and in-line functions.
#' (e.g: lm(wt ~ as.factor(am))). Since this function is meant for function interaction,
#' as opposed to human interaction, a successful check is silent.
#'
#' @param model An R model object
#'
#' @examples
#'
#' model <- lm(mpg ~ wt, mtcars)
#' acceptable_formula(model)
#' @export
acceptable_formula <- function(model) {
  UseMethod("acceptable_formula")
}

#' @export
acceptable_formula.lm <- function(model) {
  acceptable_lm(model)
}


#' @export
acceptable_formula.glm <- function(model) {
  acceptable_lm(model)
}

## As suggested by @topepo, brought in from the `pryr` package
## via the `recipes` package
fun_calls <- function(f) {
  if (is.function(f)) {
    fun_calls(body(f))
  } else if (is.call(f)) {
    fname <- as.character(f[[1]])
    # Calls inside .Internal are special and shouldn't be included
    if (identical(fname, ".Internal")) {
      return(fname)
    }
    unique(c(fname, unlist(lapply(f[-1], fun_calls), use.names = FALSE)))
  }
}

acceptable_lm <- function(model) {
  # Check for invalid contrasts
  if (length(model$contrasts)) {
    contr <- model$contrasts
    contr <- contr[!("contr.treatment" %in% contr)]
    if (length(contr) > 0) {
      cli::cli_abort(
        "The treatment contrast is the only one supported at this time.
        Field(s) with an invalid contrast are: {.val {contr}}.",
        call. = FALSE
      )
    }
  }

  accepted_funs <- funs <- c(
    "~",
    "+",
    "-",
    "*",
    "(",
    ")",
    ":",
    "::",
    "lm",
    "glm",
    "factor",
    "stats"
  )

  # Check for in-line formulas
  funs <- fun_calls(stats::formula(model))
  funs <- funs[!(funs %in% accepted_funs)]
  if (length(funs) > 0) {
    contains_offset <- any(funs == "offset")
    contains_other <- funs[funs != "offset"]
    msg <- c(x = "Functions inside the formula are not supported.")
    if (contains_offset) {
      msg <- c(
        msg,
        i = "Offset detected, try using offset as an argument instead."
      )
    }
    if (length(contains_other) > 0) {
      msg <- c(
        msg,
        i = "Functions detected: {.val {contains_other}}. 
            Use `dplyr` transformations to prepare the data."
      )
    }
    cli::cli_abort(msg, call. = FALSE)
  }
}

df <- data.frame(x = c(1, 2, 5, 6, 6), y = c(2, 3, 6, 5, 4))
m <- lm(x ~ as.factor(y), df)

test_that("acceptable_formula fails in expected cases", {
  expect_error(acceptable_formula(m))
})

test_that("glm family function syntax does not trigger error (#202)", {
  model <- glm(mpg ~ wt + hp, data = mtcars, family = gaussian())
  expect_no_error(acceptable_formula(model))

  model <- glm(am ~ wt + hp, data = mtcars, family = binomial())
  expect_no_error(acceptable_formula(model))

  model <- glm(carb ~ wt + hp, data = mtcars, family = poisson())
  expect_no_error(acceptable_formula(model))
})

test_that("glm family string syntax works (#202)", {
  model <- glm(mpg ~ wt + hp, data = mtcars, family = "gaussian")
  expect_no_error(acceptable_formula(model))

  model <- glm(am ~ wt + hp, data = mtcars, family = "binomial")
  expect_no_error(acceptable_formula(model))

  model <- glm(carb ~ wt + hp, data = mtcars, family = "poisson")
  expect_no_error(acceptable_formula(model))
})

#' @export
tidypredict_fit.earth <- function(model) {
  parsedmodel <- parse_model(model)
  build_fit_formula(parsedmodel)
}

#' @export
parse_model.earth <- function(model) {
  is_glm <- !is.null(model$glm.list)

  pm <- list()
  pm$general$model <- "earth"
  pm$general$type <- "regression"
  pm$general$version <- 2
  pm$general$is_glm <- 0
  if (is_glm) {
    pm$general$is_glm <- 1
    fam <- model$glm.list[[1]]$family
    pm$general$family <- fam$family
    pm$general$link <- fam$link
  }
  pm$terms <- mars_terms(model, is_glm)
  as_parsed_model(pm)
}


mars_terms <- function(mod, is_glm) {
  feature_types <-
    tibble::as_tibble(mod$dirs, rownames = "feature") %>%
    dplyr::mutate(feature_num = dplyr::row_number()) %>%
    tidyr::pivot_longer(
      cols = c(-feature, -feature_num),
      values_to = "type",
      names_to = "term"
    )

  feature_values <-
    tibble::as_tibble(mod$cuts, rownames = "feature") %>%
    dplyr::mutate(feature_num = dplyr::row_number()) %>%
    tidyr::pivot_longer(
      cols = c(-feature, -feature_num),
      values_to = "value",
      names_to = "term"
    )

  if (is_glm) {
    all_coefs <- mod$glm.coefficients
  } else {
    all_coefs <- mod$coefficients
  }

  feature_coefs <-
    # Note coef(mod) formats data differently for logistic regression
    tibble::as_tibble(all_coefs, rownames = "feature") %>%
    setNames(c("feature", "coefficient"))

  term_to_column <-
    tibble::as_tibble(mod$modvars, rownames = "column") %>%
    tidyr::pivot_longer(
      cols = c(-column),
      values_to = "value",
      names_to = "term"
    ) %>%
    purrr::transpose() %>%
    purrr::map(
      ~ {
        if (.x$value == 1) {
          .x$level <- gsub(.x$column, "", .x$term)
        } else {
          .x$level <- NA
        }
        .x
      }
    ) %>%
    dplyr::bind_rows() %>%
    dplyr::filter(value == 1) %>%
    dplyr::select(-value)

  feature_types %>%
    dplyr::full_join(
      feature_values,
      by = c("feature", "feature_num", "term")
    ) %>%
    dplyr::filter(type != 0) %>%
    dplyr::right_join(feature_coefs, by = "feature") %>%
    dplyr::mutate(
      feature_num = ifelse(feature == "(Intercept)", 0, feature_num)
    ) %>%
    dplyr::arrange(feature_num) %>%
    dplyr::left_join(term_to_column, by = "term") %>%
    dplyr::rowwise() %>%
    dplyr::mutate(lists = list(make_lists(type, column, value, level))) %>%
    dplyr::group_by(feature, feature_num) %>%
    dplyr::summarize(
      final = collapse_lists(feature, coefficient, lists),
      .groups = "drop"
    ) %>%
    purrr::pluck("final")
}


make_lists <- function(type, column, value, level) {
  if (is.na(type)) {
    return(list())
  }
  if (abs(type) == 1) {
    out <- operation_list(type, column, value)
  } else {
    out <- conditional_list(column, level)
  }
  out
}

conditional_list <- function(column, level) {
  if (nchar(level) == 0) {
    out <- list(
      type = "ordinary",
      col = column
    )
  } else {
    out <- list(
      type = "conditional",
      col = column,
      val = level,
      op = "equal"
    )
  }
  out
}

operation_list <- function(direction, column, split) {
  list(
    type = "operation",
    col = column,
    val = split,
    op = ifelse(direction < 0, "lessthan", "morethan")
  )
}

collapse_lists <- function(label, coef, lst) {
  label <- unique(label)
  coef <- unique(coef)
  list(
    list(
      label = label,
      coef = coef,
      is_intercept = ifelse(label == "(Intercept)", 1, 0),
      fields = lst
    )
  )
}

# For {orbital}
#' Extract multiclass linear predictors for earth models
#'
#' For use in orbital package.
#' @param model An earth model object with multiple classes (glm.list with >1 elements)
#' @keywords internal
#' @export
.extract_earth_multiclass <- function(model) {
  if (!inherits(model, "earth")) {
    cli::cli_abort(
      "{.arg model} must be {.cls earth}, not {.obj_type_friendly {model}}."
    )
  }

  if (is.null(model$glm.list) || length(model$glm.list) < 2) {
    cli::cli_abort(
      c(
        "Model does not contain multiclass information.",
        "i" = "Fit the earth model with {.code glm = TRUE} for classification."
      )
    )
  }

  # Get class names from glm.coefficients matrix columns
  class_names <- colnames(model$glm.coefficients)

  # For each class, create a modified model with single-class coefficients
  # and use the standard parsing machinery
  eqs <- lapply(class_names, function(cls) {
    # Create a temporary model with single-class coefficients
    model_single <- model
    model_single$glm.coefficients <- model$glm.coefficients[, cls, drop = FALSE]

    # Parse and build expression
    parsedmodel <- parse_model(model_single)
    expr <- build_fit_formula(parsedmodel)

    # Deparse to string, preserving numeric precision
    deparse1(expr, control = "digits17")
  })

  names(eqs) <- class_names
  eqs
}

round_print <- function(x, digits = 7) {
  x <- expr_text(x)
  x <- gsub("[ \t\r\n]+", " ", x)
  old_values <- regmatches(x, gregexpr("[0-9]+\\.?[0-9]+", x))
  new_values <- lapply(old_values, function(x) signif(as.numeric(x), digits))

  old_values <- unlist(old_values, use.names = FALSE)
  new_values <- unlist(new_values, use.names = FALSE)

  for (i in seq_along(old_values)) {
    x <- sub(old_values[i], new_values[i], x)
  }
  x
}

#' @import rlang
#' @importFrom purrr map2
#' @importFrom purrr map
#' @importFrom purrr imap
#' @importFrom purrr map_chr
#' @importFrom purrr map_lgl
#' @importFrom purrr map_dbl
#' @importFrom purrr map_int
#' @importFrom purrr map_dfr
#' @importFrom purrr reduce
#' @importFrom purrr transpose
#' @importFrom utils head
#' @importFrom stats predict
#' @importFrom stats qt
#' @importFrom tibble tibble
#' @importFrom knitr knit_print
#' @importFrom dplyr mutate
#' @importFrom dplyr case_when
#' @importFrom dplyr %>%
#' @importFrom stats binomial
#' @importFrom stats setNames
#' @keywords internal
#' @importFrom generics tidy
#'
"_PACKAGE"
NULL
utils::globalVariables(
  c(
    ".",
    "pm",
    "coefficient",
    "column",
    "feature",
    "feature_num",
    "level",
    "lists",
    "type",
    "value"
  )
)

# Model parser -------------------------------------

#' @export
parse_model.randomForest <- function(model) {
  # Check if this is a classification model
  if (!is.null(model$classes)) {
    cli::cli_abort(
      c(
        "Classification models are not supported for randomForest.",
        i = "Only regression models can be converted to tidy formulas.",
        i = "Classification requires a voting mechanism that cannot be expressed as a single formula."
      )
    )
  }

  pm <- list()
  pm$general$model <- "randomForest"
  pm$general$type <- "tree"
  pm$general$version <- 3
  term_labels <- names(model$forest$ncat)
  pm$tree_info_list <- map(
    seq_len(model$ntree),
    function(tree_no) rf_tree_info_full(model, tree_no, term_labels)
  )
  as_parsed_model(pm)
}

# Convert randomForest getTree to standard tree_info format
rf_tree_info_full <- function(model, tree_no, term_labels) {
  tree <- randomForest::getTree(model, tree_no)
  n_nodes <- nrow(tree)

  # randomForest uses 1-indexed nodes, convert to 0-indexed
  # Also convert child IDs to 0-indexed (or NA for leaves)
  left_child <- tree[, "left daughter"]
  right_child <- tree[, "right daughter"]
  left_child <- ifelse(left_child == 0, NA_integer_, left_child - 1L)
  right_child <- ifelse(right_child == 0, NA_integer_, right_child - 1L)

  terminal <- tree[, "status"] == -1
  prediction <- ifelse(terminal, tree[, "prediction"], NA_real_)

  # Build split var names
  split_var_idx <- tree[, "split var"]
  splitvarName <- ifelse(
    split_var_idx == 0,
    NA_character_,
    term_labels[split_var_idx]
  )

  # Build node_splits list
  node_splits <- vector("list", n_nodes)
  for (i in seq_len(n_nodes)) {
    if (!terminal[i]) {
      node_splits[[i]] <- list(
        primary = list(
          col = splitvarName[i],
          val = tree[i, "split point"],
          is_categorical = FALSE
        )
      )
    }
  }

  list(
    nodeID = seq_len(n_nodes) - 1L,
    leftChild = left_child,
    rightChild = right_child,
    splitvarName = splitvarName,
    terminal = terminal,
    prediction = prediction,
    node_splits = node_splits
  )
}

# Fit model (nested) -----------------------------------------------

#' @export
tidypredict_fit.randomForest <- function(model, ...) {
  tidypredict_fit_rf_nested(model)
}

# Nested formula builder for randomForest
tidypredict_fit_rf_nested <- function(model) {
  # Check if this is a classification model
  if (!is.null(model$classes)) {
    cli::cli_abort(
      c(
        "Classification models are not supported for randomForest.",
        i = "Only regression models can be converted to tidy formulas.",
        i = "Classification requires a voting mechanism that cannot be expressed as a single formula."
      )
    )
  }

  n_trees <- model$ntree
  term_labels <- names(model$forest$ncat)

  tree_exprs <- map(seq_len(n_trees), function(tree_no) {
    build_nested_rf_tree(model, tree_no, term_labels)
  })

  res <- reduce_addition(tree_exprs)
  expr_division(res, n_trees)
}

# Build nested case_when for a single randomForest tree
build_nested_rf_tree <- function(model, tree_no, term_labels) {
  tree <- randomForest::getTree(model, tree_no)

  # Pre-extract columns as vectors for fast indexing (avoids slow row access)
  # Use unname() once here instead of on every recursive call
  status <- unname(tree[, "status"])
  prediction <- unname(tree[, "prediction"])
  left_daughter <- unname(tree[, "left daughter"])
  right_daughter <- unname(tree[, "right daughter"])
  split_var <- unname(tree[, "split var"])
  split_point <- unname(tree[, "split point"])

  build_node <- function(node_id) {
    # Check if terminal (leaf) node - status == -1
    if (status[node_id] == -1) {
      return(prediction[node_id])
    }

    # Internal node - get split info
    left_id <- left_daughter[node_id]
    right_id <- right_daughter[node_id]
    var_idx <- split_var[node_id]
    split_val <- split_point[node_id]

    # Recurse
    left_subtree <- build_node(left_id)
    right_subtree <- build_node(right_id)

    col_name <- term_labels[var_idx]
    col_sym <- rlang::sym(col_name)

    # Numeric split: left = <= splitval, right = > splitval
    condition <- expr(!!col_sym <= !!split_val)

    expr(case_when(!!condition ~ !!left_subtree, .default = !!right_subtree))
  }

  build_node(1L)
}

# Legacy flat case_when (for v1/v2 parsed model compatibility) ----------------
# These functions are preserved for backwards compatibility when loading
# parsed models saved with version < 3.

# Used by tidypredict_fit.pm_tree() for v1/v2 randomForest parsed models
tidypredict_fit_randomForest <- function(parsedmodel) {
  # Check if this is a classification model (string predictions)
  first_pred <- parsedmodel$trees[[1]][[1]]$prediction
  if (is.character(first_pred)) {
    cli::cli_abort(
      c(
        "Classification models are not supported for randomForest.",
        i = "Only regression models can be converted to tidy formulas.",
        i = "Classification requires a voting mechanism that cannot be expressed as a single formula."
      )
    )
  }

  res <- generate_case_when_trees(parsedmodel)
  res <- reduce_addition(res)
  n_trees <- length(parsedmodel$trees)
  expr_division(res, n_trees)
}

# For {orbital} -----------------------------------------------

#' Extract classification vote trees for randomForest models
#'
#' For use in orbital package.
#' @param model A randomForest model object
#' @keywords internal
#' @export
.extract_rf_classprob <- function(model) {
  if (!inherits(model, "randomForest")) {
    cli::cli_abort(
      "{.arg model} must be {.cls randomForest}, not {.obj_type_friendly {model}}."
    )
  }

  # Check if this is a classification model
  if (is.null(model$classes)) {
    cli::cli_abort(
      c(
        "Model is not a classification model.",
        i = "Use {.fn tidypredict_fit} for regression models."
      )
    )
  }

  # Get class levels from the model
  lvls <- model$classes
  term_labels <- names(model$forest$ncat)

  # For each class, generate nested case_when expressions for all trees
  res <- list()
  for (lvl in lvls) {
    tree_exprs <- map(seq_len(model$ntree), function(tree_no) {
      build_nested_rf_vote_tree(model, tree_no, term_labels, lvl)
    })
    res[[lvl]] <- tree_exprs
  }
  res
}

# Build nested case_when for randomForest voting tree
build_nested_rf_vote_tree <- function(
  model,
  tree_no,
  term_labels,
  class_level
) {
  tree <- randomForest::getTree(model, tree_no)
  classes <- model$classes

  # Pre-extract columns as vectors for fast indexing (avoids slow row access)
  # Use unname() once here instead of on every recursive call
  status <- unname(tree[, "status"])
  prediction <- unname(tree[, "prediction"])
  left_daughter <- unname(tree[, "left daughter"])
  right_daughter <- unname(tree[, "right daughter"])
  split_var <- unname(tree[, "split var"])
  split_point <- unname(tree[, "split point"])

  build_node <- function(node_id) {
    # Check if terminal (leaf) node - status == -1
    if (status[node_id] == -1) {
      # Return 1 if prediction matches class_level, 0 otherwise
      pred_class <- classes[prediction[node_id]]
      return(if (pred_class == class_level) 1L else 0L)
    }

    # Internal node - get split info
    left_id <- left_daughter[node_id]
    right_id <- right_daughter[node_id]
    var_idx <- split_var[node_id]
    split_val <- split_point[node_id]

    # Recurse
    left_subtree <- build_node(left_id)
    right_subtree <- build_node(right_id)

    col_name <- term_labels[var_idx]
    col_sym <- rlang::sym(col_name)

    # Numeric split: left = <= splitval, right = > splitval
    condition <- expr(!!col_sym <= !!split_val)

    expr(case_when(!!condition ~ !!left_subtree, .default = !!right_subtree))
  }

  build_node(1L)
}

#' Extract regression trees for randomForest models
#'
#' For use in orbital package.
#' @param model A randomForest model object (regression)
#' @keywords internal
#' @export
.extract_rf_trees <- function(model) {
  if (!inherits(model, "randomForest")) {
    cli::cli_abort(
      "{.arg model} must be {.cls randomForest}, not {.obj_type_friendly {model}}."
    )
  }

  # Check if this is a classification model
  if (!is.null(model$classes)) {
    cli::cli_abort(
      c(
        "Classification models are not supported.",
        i = "Use {.fn .extract_rf_classprob} for classification models."
      )
    )
  }

  n_trees <- model$ntree
  term_labels <- names(model$forest$ncat)

  map(seq_len(n_trees), function(tree_no) {
    build_nested_rf_tree(model, tree_no, term_labels)
  })
}

test_that("returns the right output", {
  set.seed(1234)

  model <- randomForest::randomForest(mpg ~ ., data = mtcars, ntree = 3)

  tf <- tidypredict_fit(model)
  pm <- parse_model(model)

  expect_type(tf, "language")

  expect_s3_class(pm, "list")
  expect_equal(length(pm), 2)
  expect_equal(length(pm$tree_info_list), 3)
  expect_equal(pm$general$model, "randomForest")
  expect_equal(pm$general$version, 3)

  expect_snapshot(
    rlang::expr_text(tf)
  )
})

test_that("Model can be saved and re-loaded", {
  set.seed(1234)

  model <- randomForest::randomForest(mpg ~ ., data = mtcars, ntree = 3)

  fit_expr <- tidypredict_fit(model)
  fit_pred <- dplyr::mutate(mtcars, pred = !!fit_expr)$pred
  original_pred <- predict(model, mtcars)

  expect_equal(fit_pred, as.vector(original_pred))
})

test_that("formulas produces correct predictions", {
  set.seed(1234)

  # regression
  expect_snapshot(
    tidypredict_test(
      randomForest::randomForest(mpg ~ ., data = mtcars, ntree = 3),
      mtcars,
    )
  )
})

test_that("split operator uses <= for left child (#192)", {
  set.seed(42)
  df <- data.frame(x = c(1, 2, 3, 4), y = c(10, 20, 100, 200))
  suppressWarnings(
    model <- randomForest::randomForest(
      y ~ x,
      data = df,
      ntree = 1,
      nodesize = 2,
      maxnodes = 3
    )
  )

  test_df <- data.frame(x = c(2.99, 3, 3.01))

  native <- as.numeric(predict(model, test_df))
  fit <- tidypredict_fit(model)
  tidy <- rlang::eval_tidy(fit, test_df)

  expect_equal(native, tidy)
})

test_that("produced case_when uses .default", {
  set.seed(1234)

  model <- randomForest::randomForest(mpg ~ ., data = mtcars, ntree = 3)

  fit <- tidypredict_fit(model)
  fit_text <- rlang::expr_text(fit)

  expect_match(fit_text, "\\.default")
})

test_that("classification models error with clear message (#193)", {
  set.seed(123)
  model <- randomForest::randomForest(
    Species ~ Sepal.Length + Sepal.Width,
    data = iris,
    ntree = 3
  )

  expect_snapshot(tidypredict_fit(model), error = TRUE)
})

test_that("parse_model errors on classification model", {
  set.seed(123)
  model <- randomForest::randomForest(
    Species ~ Sepal.Length + Sepal.Width,
    data = iris,
    ntree = 3
  )

  expect_snapshot(parse_model(model), error = TRUE)
})

# v2 backwards compatibility tests ---------------------------------------------

test_that("v2 parsed randomForest model produces correct predictions", {
  pm <- readRDS(test_path("backwards-compat", "rf-v2-regression.rds"))

  expect_equal(pm$general$version, 2)
  expect_true(!is.null(pm$trees))

  fit <- tidypredict_fit(pm)
  expect_type(fit, "language")

  # Verify predictions match expected values
  pred <- rlang::eval_tidy(fit, mtcars)
  expect_type(pred, "double")
  expect_length(pred, nrow(mtcars))
})

test_that("v2 parsed classification model errors", {
  pm <- readRDS(test_path("backwards-compat", "rf-v2-classification.rds"))

  expect_equal(pm$general$version, 2)
  expect_true(is.character(pm$trees[[1]][[1]]$prediction))

  expect_snapshot(tidypredict_fit(pm), error = TRUE)
})

# Tests for .extract_rf_classprob()

test_that(".extract_rf_classprob returns correct structure", {
  set.seed(123)
  model <- randomForest::randomForest(
    Species ~ Sepal.Length + Sepal.Width,
    data = iris,
    ntree = 3
  )

  result <- .extract_rf_classprob(model)

  expect_type(result, "list")
  expect_length(result, 3)
  expect_named(result, levels(iris$Species))
  # Each class should have ntree expressions
  expect_length(result[[1]], 3)
})

test_that(".extract_rf_classprob errors on non-randomForest model", {
  model <- lm(mpg ~ ., data = mtcars)

  expect_snapshot(error = TRUE, .extract_rf_classprob(model))
})

test_that(".extract_rf_classprob errors on regression model", {
  set.seed(123)
  model <- randomForest::randomForest(mpg ~ ., data = mtcars, ntree = 3)

  expect_snapshot(error = TRUE, .extract_rf_classprob(model))
})

test_that(".extract_rf_classprob works with binary classification", {
  set.seed(123)
  mtcars$vs <- factor(mtcars$vs)
  model <- randomForest::randomForest(
    vs ~ disp + hp,
    data = mtcars,
    ntree = 3
  )

  result <- .extract_rf_classprob(model)

  expect_type(result, "list")
  expect_length(result, 2)
  expect_named(result, c("0", "1"))
})

test_that(".extract_rf_classprob produces correct vote counts", {
  set.seed(123)
  model <- randomForest::randomForest(
    Species ~ .,
    data = iris,
    ntree = 5
  )

  class_trees <- .extract_rf_classprob(model)
  n_trees <- model$ntree

  # Sum votes for each class
  vote_counts <- sapply(names(class_trees), function(cls) {
    trees <- class_trees[[cls]]
    tree_vals <- sapply(trees, function(e) {
      rlang::eval_tidy(e, iris)
    })
    if (is.matrix(tree_vals)) rowSums(tree_vals) else tree_vals
  })

  # Calculate probabilities
  probs <- vote_counts / n_trees

  # Compare to native predictions
  native <- predict(model, iris, type = "prob")

  expect_equal(unname(probs), unname(native), tolerance = 1e-10)
})

test_that(".extract_rf_classprob works with single tree", {
  set.seed(123)
  model <- randomForest::randomForest(
    Species ~ .,
    data = iris,
    ntree = 1
  )

  result <- .extract_rf_classprob(model)

  expect_type(result, "list")
  expect_length(result, 3)
  # Each class should have 1 expression
  expect_length(result[[1]], 1)
})

# Tests for .extract_rf_trees() (regression)

test_that(".extract_rf_trees returns correct structure", {
  set.seed(123)
  model <- randomForest::randomForest(
    mpg ~ cyl + disp + hp,
    data = mtcars,
    ntree = 5
  )

  result <- .extract_rf_trees(model)

  expect_type(result, "list")
  expect_length(result, 5)
  expect_all_true(vapply(result, is.language, logical(1)))
})

test_that(".extract_rf_trees errors on non-randomForest model", {
  model <- lm(mpg ~ ., data = mtcars)

  expect_snapshot(error = TRUE, .extract_rf_trees(model))
})

test_that(".extract_rf_trees errors on classification model", {
  set.seed(123)
  model <- randomForest::randomForest(
    Species ~ Sepal.Length + Sepal.Width,
    data = iris,
    ntree = 3
  )

  expect_snapshot(error = TRUE, .extract_rf_trees(model))
})

test_that(".extract_rf_trees produces correct predictions when averaged", {
  set.seed(123)
  model <- randomForest::randomForest(
    mpg ~ cyl + disp + hp,
    data = mtcars,
    ntree = 5
  )

  trees <- .extract_rf_trees(model)
  n_trees <- length(trees)

  tree_preds <- sapply(trees, function(e) rlang::eval_tidy(e, mtcars))
  avg_pred <- rowMeans(tree_preds)

  native <- as.numeric(predict(model, mtcars))

  expect_equal(avg_pred, native)
})

# Model parser -------------------------------------

#' @export
parse_model.ranger <- function(model) {
  # Check if this is a classification model
  first_tree <- ranger::treeInfo(model, 1)
  first_pred <- first_tree$prediction[first_tree$terminal][1]
  if (is.character(first_pred) || is.factor(first_pred)) {
    cli::cli_abort(
      c(
        "Classification models are not supported for ranger.",
        i = "Only regression models can be converted to tidy formulas.",
        i = "Classification requires a voting mechanism that cannot be expressed as a single formula."
      )
    )
  }

  pm <- list()
  pm$general$model <- "ranger"
  pm$general$type <- "tree"
  pm$general$version <- 3
  pm$tree_info_list <- map(
    seq_len(model$num.trees),
    function(tree_no) ranger_tree_info_full(model, tree_no)
  )
  as_parsed_model(pm)
}

# Convert ranger treeInfo to standard tree_info format
ranger_tree_info_full <- function(model, tree_no) {
  tree <- ranger::treeInfo(model, tree_no)

  # Build node_splits list
  node_splits <- vector("list", nrow(tree))
  for (i in seq_len(nrow(tree))) {
    if (!tree$terminal[i]) {
      var_name <- as.character(tree$splitvarName[i])
      split_val <- tree$splitval[i]

      node_splits[[i]] <- list(
        primary = list(
          col = var_name,
          val = split_val,
          is_categorical = FALSE
        )
      )
    }
  }

  list(
    nodeID = tree$nodeID,
    leftChild = tree$leftChild,
    rightChild = tree$rightChild,
    splitvarName = as.character(tree$splitvarName),
    terminal = tree$terminal,
    prediction = tree$prediction,
    node_splits = node_splits
  )
}

# Fit model (nested) -----------------------------------

#' @export
tidypredict_fit.ranger <- function(model, ...) {
  tidypredict_fit_ranger_nested(model)
}

# Nested formula builder for ranger
tidypredict_fit_ranger_nested <- function(model) {
  # Check if this is a classification model
  first_tree <- ranger::treeInfo(model, 1)
  first_pred <- first_tree$prediction[first_tree$terminal][1]
  if (is.character(first_pred) || is.factor(first_pred)) {
    cli::cli_abort(
      c(
        "Classification models are not supported for ranger.",
        i = "Only regression models can be converted to tidy formulas.",
        i = "Classification requires a voting mechanism that cannot be expressed as a single formula."
      )
    )
  }

  n_trees <- model$num.trees
  tree_exprs <- map(seq_len(n_trees), function(tree_no) {
    build_nested_ranger_tree(model, tree_no)
  })

  res <- reduce_addition(tree_exprs)
  expr_division(res, n_trees)
}

# Build nested case_when for a single ranger tree
build_nested_ranger_tree <- function(model, tree_no) {
  tree <- ranger::treeInfo(model, tree_no)

  # Pre-extract columns as vectors for fast indexing (avoids slow df[i,] access)
  nodeID <- tree$nodeID
  leftChild <- tree$leftChild
  rightChild <- tree$rightChild
  splitvarName <- as.character(tree$splitvarName)
  splitval <- tree$splitval
  terminal <- tree$terminal
  prediction <- tree$prediction

  build_node <- function(node_id) {
    # node_id is 0-indexed, convert to 1-indexed for vector access
    idx <- node_id + 1L

    if (terminal[idx]) {
      return(prediction[idx])
    }

    left_id <- leftChild[idx]
    right_id <- rightChild[idx]
    split_var <- splitvarName[idx]
    split_val <- splitval[idx]

    left_subtree <- build_node(left_id)
    right_subtree <- build_node(right_id)

    col_sym <- rlang::sym(split_var)
    condition <- expr(!!col_sym <= !!split_val)

    expr(case_when(!!condition ~ !!left_subtree, .default = !!right_subtree))
  }

  build_node(0L)
}

# Legacy flat case_when (for v1/v2 parsed model compatibility) ----------------
# These functions are preserved for backwards compatibility when loading
# parsed models saved with version < 3.

# Used by tidypredict_fit.pm_tree() for v1/v2 ranger parsed models
tidypredict_fit_ranger <- function(parsedmodel) {
  # Check if this is a classification model (string predictions)
  first_pred <- parsedmodel$trees[[1]][[1]]$prediction
  if (is.character(first_pred)) {
    cli::cli_abort(
      c(
        "Classification models are not supported for ranger.",
        i = "Only regression models can be converted to tidy formulas.",
        i = "Classification requires a voting mechanism that cannot be expressed as a single formula."
      )
    )
  }

  res <- generate_case_when_trees(parsedmodel)
  res <- reduce_addition(res)
  n_trees <- length(parsedmodel$trees)
  expr_division(res, n_trees)
}

# Legacy tree extraction functions (no longer used) ---------------------------
# These functions were used by the old parse_model.ranger() to populate pm$trees
# in the flat path format. Now parse_model.ranger() uses ranger_tree_info_full()
# to populate pm$tree_info_list instead. Kept for reference.

get_child_info <- function(tree) {
  child_info <- numeric(max(tree$nodeID))
  left_child <- tree$leftChild
  right_child <- tree$rightChild
  node_id <- tree$nodeID

  for (i in seq_len(nrow(tree))) {
    node <- node_id[[i]]

    child <- left_child[[i]]
    if (!is.na(child)) {
      child_info[child] <- node
    }

    child <- right_child[[i]]
    if (!is.na(child)) {
      child_info[child] <- node
    }
  }

  child_info
}

get_ra_path <- function(node_id, tree, child_info, default_op = TRUE) {
  find <- node_id
  path <- node_id

  leftChild <- tree$leftChild
  rightChild <- tree$rightChild

  splitval <- tree$splitval
  splitvarName <- tree$splitvarName

  # Handle stump trees (no splits) - return empty path
  if (length(child_info) == 0 || find < 1 || find > length(child_info)) {
    return(list())
  }

  new <- child_info[[find]]
  path <- find
  repeat {
    if (new == 0) {
      path <- c(path, 0)
      break
    }
    path <- c(path, new)
    find <- new
    new <- child_info[[find]]
  }

  map2(
    path[1:length(path) - 1],
    path[2:length(path)],
    ~ {
      lc <- leftChild[.y + 1] == .x
      lr <- rightChild[.y + 1] == .x
      if (default_op) {
        if (lc) {
          op <- "less"
        }
        if (lr) {
          op <- "more-equal"
        }
      } else {
        if (lc) {
          op <- "less-equal"
        }
        if (lr) {
          op <- "more"
        }
      }
      list(
        type = "conditional",
        col = as.character(splitvarName[.y + 1]),
        val = splitval[.y + 1],
        op = op
      )
    }
  )
}

get_ra_tree <- function(tree_no, model) {
  tree <- ranger::treeInfo(model, tree_no)
  paths <- tree$nodeID[tree[, "terminal"]]

  child_info <- get_child_info(tree)

  map(
    paths,
    ~ {
      prediction <- tree$prediction[tree$nodeID == .x]
      if (!is.null(prediction)) {
        if (is.factor(prediction)) {
          prediction <- as.character(prediction)
        }
        list(
          prediction = prediction,
          path = get_ra_path(.x, tree, child_info, FALSE)
        )
      } else {
        preds <- map_lgl(colnames(tree), ~ "pred." == substr(.x, 1, 5))
        preds_table <- tree[tree$nodeID == .x, preds]
        predictors <- map_chr(colnames(preds_table), ~ substr(.x, 6, nchar(.x)))
        colnames(preds_table) <- predictors
        predictions <- map(preds_table, ~.x)
        max_pred <- map_lgl(predictions, ~ .x == max(map_dbl(predictions, ~.x)))
        prediction <- names(predictions)[max_pred]
        prediction <- prediction[[1]]
        prob <- predictions[max_pred]
        prob <- prob[[1]]
        predictions <- imap(predictions, ~ list(pred = .y, prob = .x))
        list(
          prediction = prediction,
          prob = prob,
          probs = predictions,
          path = get_ra_path(.x, tree, child_info, FALSE)
        )
      }
    }
  )
}

get_ra_trees <- function(model) {
  map(
    seq_len(model$num.trees),
    ~ get_ra_tree(.x, model)
  )
}

# For {orbital} -----------------------------------------------

#' Extract classification probability trees for ranger models
#'
#' For use in orbital package.
#' @param model A ranger model object fitted with `probability = TRUE`
#' @keywords internal
#' @export
.extract_ranger_classprob <- function(model) {
  if (!inherits(model, "ranger")) {
    cli::cli_abort(
      "{.arg model} must be {.cls ranger}, not {.obj_type_friendly {model}}."
    )
  }

  # Get class levels from treeInfo
  tree <- ranger::treeInfo(model, 1)
  pred_cols <- grep("^pred\\.", names(tree), value = TRUE)

  if (length(pred_cols) == 0) {
    cli::cli_abort(
      c(
        "Model does not contain probability information.",
        i = "Fit the ranger model with {.code probability = TRUE}."
      )
    )
  }

  lvls <- sub("^pred\\.", "", pred_cols)

  # For each class, generate nested case_when expressions for all trees
  res <- list()
  for (lvl in lvls) {
    tree_exprs <- map(seq_len(model$num.trees), function(tree_no) {
      build_nested_ranger_prob_tree(model, tree_no, lvl)
    })
    res[[lvl]] <- tree_exprs
  }
  res
}

# Build nested case_when for ranger probability tree
build_nested_ranger_prob_tree <- function(model, tree_no, class_level) {
  tree <- ranger::treeInfo(model, tree_no)

  # Pre-extract columns as vectors for fast indexing (avoids slow df[i,] access)
  nodeID <- tree$nodeID
  leftChild <- tree$leftChild
  rightChild <- tree$rightChild
  splitvarName <- as.character(tree$splitvarName)
  splitval <- tree$splitval
  terminal <- tree$terminal
  prob_col <- paste0("pred.", class_level)
  prob_vals <- tree[[prob_col]]

  build_node <- function(node_id) {
    # node_id is 0-indexed, convert to 1-indexed for vector access
    idx <- node_id + 1L

    if (terminal[idx]) {
      return(prob_vals[idx])
    }

    left_id <- leftChild[idx]
    right_id <- rightChild[idx]
    split_var <- splitvarName[idx]
    split_val <- splitval[idx]

    left_subtree <- build_node(left_id)
    right_subtree <- build_node(right_id)

    col_sym <- rlang::sym(split_var)
    condition <- expr(!!col_sym <= !!split_val)

    expr(case_when(!!condition ~ !!left_subtree, .default = !!right_subtree))
  }

  build_node(0L)
}

#' Extract regression trees for ranger models
#'
#' For use in orbital package.
#' @param model A ranger model object (regression)
#' @keywords internal
#' @export
.extract_ranger_trees <- function(model) {
  if (!inherits(model, "ranger")) {
    cli::cli_abort(
      "{.arg model} must be {.cls ranger}, not {.obj_type_friendly {model}}."
    )
  }

  # Check if this is a classification model
  first_tree <- ranger::treeInfo(model, 1)
  first_pred <- first_tree$prediction[first_tree$terminal][1]
  if (is.character(first_pred) || is.factor(first_pred)) {
    cli::cli_abort(
      c(
        "Classification models are not supported.",
        i = "Use {.fn .extract_ranger_classprob} for classification models."
      )
    )
  }

  n_trees <- model$num.trees
  map(seq_len(n_trees), function(tree_no) {
    build_nested_ranger_tree(model, tree_no)
  })
}

test_that("returns the right output", {
  skip_on_cran()
  skip_on_os("windows")
  skip_on_os("linux")

  model <- ranger::ranger(
    mpg ~ cyl + disp + hp,
    data = mtcars,
    num.trees = 3,
    max.depth = 2,
    seed = 100,
    num.threads = 1
  )

  tf <- tidypredict_fit(model)
  pm <- parse_model(model)

  expect_type(tf, "language")

  expect_s3_class(pm, "list")
  expect_equal(length(pm), 2)
  expect_equal(length(pm$tree_info_list), 3)
  expect_equal(pm$general$model, "ranger")
  expect_equal(pm$general$version, 3)

  expect_snapshot(
    rlang::expr_text(tf)
  )
})

test_that("tidypredict_fit produces correct predictions", {
  skip_on_cran()
  skip_on_os("windows")
  skip_on_os("linux")

  model <- ranger::ranger(
    mpg ~ cyl + disp + hp,
    data = mtcars,
    num.trees = 3,
    max.depth = 2,
    seed = 100,
    num.threads = 1
  )

  fit_expr <- tidypredict_fit(model)
  fit_pred <- dplyr::mutate(mtcars, pred = !!fit_expr)$pred
  original_pred <- predict(model, mtcars, num.threads = 1)$predictions

  expect_equal(fit_pred, original_pred)
})

test_that("formulas produces correct predictions", {
  skip_on_cran()
  skip_on_os("windows")
  skip_on_os("linux")

  # regression
  expect_snapshot(
    tidypredict_test(
      ranger::ranger(
        mpg ~ .,
        data = mtcars,
        num.trees = 3,
        max.depth = 2,
        seed = 100,
        num.threads = 1
      ),
      mtcars
    )
  )
})

test_that("split operator uses <= for left child (#189)", {
  skip_on_cran()
  skip_on_os("windows")
  skip_on_os("linux")

  model <- ranger::ranger(
    mpg ~ cyl + disp + hp,
    data = mtcars,
    num.trees = 2,
    max.depth = 3,
    seed = 123,
    num.threads = 1
  )

  native <- predict(model, mtcars, num.threads = 1)$predictions
  fit <- tidypredict_fit(model)
  tidy <- rlang::eval_tidy(fit, mtcars)

  expect_equal(native, tidy)
})

test_that("predictions are averaged not summed (#190)", {
  skip_on_cran()
  skip_on_os("windows")
  skip_on_os("linux")

  model <- ranger::ranger(
    mpg ~ cyl + disp + hp,
    data = mtcars,
    num.trees = 5,
    max.depth = 3,
    seed = 123,
    num.threads = 1
  )

  native <- predict(model, mtcars, num.threads = 1)$predictions
  fit <- tidypredict_fit(model)
  tidy <- rlang::eval_tidy(fit, mtcars)

  expect_equal(native, tidy)
})

test_that("produced case_when uses .default", {
  skip_on_cran()
  skip_on_os("windows")
  skip_on_os("linux")

  model <- ranger::ranger(
    mpg ~ cyl + disp + hp,
    data = mtcars,
    num.trees = 3,
    max.depth = 2,
    seed = 100,
    num.threads = 1
  )

  fit <- tidypredict_fit(model)
  fit_text <- rlang::expr_text(fit)

  expect_match(fit_text, "\\.default")
})

test_that("classification models error with clear message (#191)", {
  skip_on_cran()
  skip_on_os("windows")
  skip_on_os("linux")

  model <- ranger::ranger(
    Species ~ Sepal.Length + Sepal.Width,
    data = iris,
    num.trees = 3,
    max.depth = 2,
    seed = 123,
    num.threads = 1
  )

  expect_snapshot(tidypredict_fit(model), error = TRUE)
})

# Tests for .extract_ranger_classprob()

test_that(".extract_ranger_classprob returns correct structure", {
  skip_on_cran()
  skip_on_os("windows")
  skip_on_os("linux")

  model <- ranger::ranger(
    Species ~ Sepal.Length + Sepal.Width,
    data = iris,
    num.trees = 3,
    max.depth = 2,
    seed = 123,
    num.threads = 1,
    probability = TRUE
  )

  result <- .extract_ranger_classprob(model)

  expect_type(result, "list")
  expect_length(result, 3)
  expect_named(result, levels(iris$Species))
  # Each class should have num.trees expressions
  expect_length(result[[1]], 3)
})

test_that(".extract_ranger_classprob errors on non-ranger model", {
  model <- lm(mpg ~ ., data = mtcars)

  expect_snapshot(error = TRUE, .extract_ranger_classprob(model))
})

test_that(".extract_ranger_classprob errors without probability = TRUE", {
  skip_on_cran()
  skip_on_os("windows")
  skip_on_os("linux")

  model <- ranger::ranger(
    Species ~ Sepal.Length + Sepal.Width,
    data = iris,
    num.trees = 3,
    max.depth = 2,
    seed = 123,
    num.threads = 1,
    probability = FALSE
  )

  expect_snapshot(error = TRUE, .extract_ranger_classprob(model))
})

test_that(".extract_ranger_classprob works with binary classification", {
  skip_on_cran()
  skip_on_os("windows")
  skip_on_os("linux")

  mtcars$vs <- factor(mtcars$vs)
  model <- ranger::ranger(
    vs ~ disp + hp,
    data = mtcars,
    num.trees = 3,
    max.depth = 2,
    seed = 123,
    num.threads = 1,
    probability = TRUE
  )

  result <- .extract_ranger_classprob(model)

  expect_type(result, "list")
  expect_length(result, 2)
  expect_named(result, c("0", "1"))
})

test_that(".extract_ranger_classprob produces correct probabilities", {
  skip_on_cran()
  skip_on_os("windows")
  skip_on_os("linux")

  model <- ranger::ranger(
    Species ~ .,
    data = iris,
    num.trees = 5,
    max.depth = 3,
    seed = 123,
    num.threads = 1,
    probability = TRUE
  )

  class_trees <- .extract_ranger_classprob(model)
  n_trees <- model$num.trees

  # Sum probabilities for each class
  prob_sums <- sapply(names(class_trees), function(cls) {
    trees <- class_trees[[cls]]
    tree_vals <- sapply(trees, function(e) {
      rlang::eval_tidy(e, iris)
    })
    if (is.matrix(tree_vals)) rowSums(tree_vals) else tree_vals
  })

  # Calculate averaged probabilities
  probs <- prob_sums / n_trees

  # Compare to native predictions
  native <- predict(model, iris, num.threads = 1)$predictions

  expect_equal(unname(probs), unname(native), tolerance = 1e-10)
})

test_that(".extract_ranger_classprob works with single tree", {
  skip_on_cran()
  skip_on_os("windows")
  skip_on_os("linux")

  model <- ranger::ranger(
    Species ~ .,
    data = iris,
    num.trees = 1,
    max.depth = 3,
    seed = 123,
    num.threads = 1,
    probability = TRUE
  )

  result <- .extract_ranger_classprob(model)

  expect_type(result, "list")
  expect_length(result, 3)
  # Each class should have 1 expression
  expect_length(result[[1]], 1)
})

# Tests for .extract_ranger_trees() (regression)

test_that(".extract_ranger_trees returns correct structure", {
  skip_on_cran()
  skip_on_os("windows")
  skip_on_os("linux")

  model <- ranger::ranger(
    mpg ~ cyl + disp + hp,
    data = mtcars,
    num.trees = 5,
    max.depth = 2,
    seed = 100,
    num.threads = 1
  )

  result <- .extract_ranger_trees(model)

  expect_type(result, "list")
  expect_length(result, 5)
  expect_all_true(vapply(result, is.language, logical(1)))
})

test_that(".extract_ranger_trees errors on non-ranger model", {
  model <- lm(mpg ~ ., data = mtcars)

  expect_snapshot(error = TRUE, .extract_ranger_trees(model))
})

test_that(".extract_ranger_trees errors on classification model", {
  skip_on_cran()
  skip_on_os("windows")
  skip_on_os("linux")

  model <- ranger::ranger(
    Species ~ Sepal.Length + Sepal.Width,
    data = iris,
    num.trees = 3,
    max.depth = 2,
    seed = 123,
    num.threads = 1
  )

  expect_snapshot(error = TRUE, .extract_ranger_trees(model))
})

test_that(".extract_ranger_trees produces correct predictions when averaged", {
  skip_on_cran()
  skip_on_os("windows")
  skip_on_os("linux")

  model <- ranger::ranger(
    mpg ~ cyl + disp + hp,
    data = mtcars,
    num.trees = 5,
    max.depth = 3,
    seed = 123,
    num.threads = 1
  )

  trees <- .extract_ranger_trees(model)
  n_trees <- length(trees)

  tree_preds <- sapply(trees, function(e) rlang::eval_tidy(e, mtcars))
  avg_pred <- rowMeans(tree_preds)

  native <- predict(model, mtcars, num.threads = 1)$predictions

  expect_equal(avg_pred, native)
})

# Backwards compatibility tests for v2 parsed models

test_that("v2 parsed ranger model can be loaded and used", {
  pm <- readRDS(test_path("backwards-compat", "ranger-v2-parsed.rds"))

  expect_equal(pm$general$version, 2)
  expect_equal(pm$general$model, "ranger")
  expect_true(!is.null(pm$trees))

  fit <- tidypredict_fit(pm)
  expect_type(fit, "language")

  preds <- rlang::eval_tidy(fit, mtcars)
  expect_length(preds, nrow(mtcars))
})

test_that("v2 parsed ranger model produces expected predictions", {
  pm <- readRDS(test_path("backwards-compat", "ranger-v2-parsed.rds"))

  fit <- tidypredict_fit(pm)
  preds <- rlang::eval_tidy(fit, mtcars)

  # Expected values from the v2 model (2 trees, mpg ~ cyl + disp)
  # mtcars[1,] has disp = 160, which falls in middle branch of both trees
  # Tree 1: 20.7125, Tree 2: 20.2471 -> avg = 20.47978
  expect_equal(preds[1], (20.7125 + 20.2470588235294) / 2, tolerance = 0.001)
})

test_that("v2 parsed classification model errors", {
  pm <- readRDS(test_path("backwards-compat", "ranger-v2-classification.rds"))

  expect_equal(pm$general$version, 2)
  # First prediction is a string (classification)
  expect_type(pm$trees[[1]][[1]]$prediction, "character")

  expect_snapshot(tidypredict_fit(pm), error = TRUE)
})

test_that("legacy get_ra_trees extracts correct structure", {
  skip_on_cran()
  skip_on_os("windows")
  skip_on_os("linux")

  model <- ranger::ranger(
    mpg ~ cyl + disp,
    data = mtcars,
    num.trees = 2,
    max.depth = 2,
    seed = 100,
    num.threads = 1
  )

  trees <- tidypredict:::get_ra_trees(model)

  expect_type(trees, "list")
  expect_length(trees, 2)
  # Each tree should have leaf nodes with prediction and path

  expect_true(all(vapply(
    trees[[1]],
    function(x) "prediction" %in% names(x),
    logical(1)
  )))
  expect_true(all(vapply(
    trees[[1]],
    function(x) "path" %in% names(x),
    logical(1)
  )))
})

test_that("legacy get_ra_tree extracts single tree", {
  skip_on_cran()
  skip_on_os("windows")
  skip_on_os("linux")

  model <- ranger::ranger(
    mpg ~ cyl + disp,
    data = mtcars,
    num.trees = 2,
    max.depth = 2,
    seed = 100,
    num.threads = 1
  )

  tree <- tidypredict:::get_ra_tree(1, model)

  expect_type(tree, "list")
  expect_true(length(tree) > 0)
  expect_true("prediction" %in% names(tree[[1]]))
  expect_true("path" %in% names(tree[[1]]))
})

test_that("legacy get_child_info builds parent map", {
  skip_on_cran()
  skip_on_os("windows")
  skip_on_os("linux")

  model <- ranger::ranger(
    mpg ~ cyl + disp,
    data = mtcars,
    num.trees = 1,
    max.depth = 2,
    seed = 100,
    num.threads = 1
  )

  tree <- ranger::treeInfo(model, 1)
  child_info <- tidypredict:::get_child_info(tree)

  expect_type(child_info, "double")
  # child_info maps each node to its parent
  expect_true(length(child_info) > 0)
})

test_that("legacy get_ra_path handles stump trees", {
  skip_on_cran()
  skip_on_os("windows")
  skip_on_os("linux")

  # Create a stump tree (single node, no splits)
  model <- ranger::ranger(
    mpg ~ cyl,
    data = mtcars,
    num.trees = 1,
    max.depth = 0,
    seed = 100,
    num.threads = 1
  )

  tree <- ranger::treeInfo(model, 1)
  child_info <- tidypredict:::get_child_info(tree)

  # Stump has no children, so path should be empty

  path <- tidypredict:::get_ra_path(0, tree, child_info, FALSE)
  expect_equal(path, list())
})

test_that("legacy get_ra_path with default_op = TRUE", {
  skip_on_cran()
  skip_on_os("windows")
  skip_on_os("linux")

  model <- ranger::ranger(
    mpg ~ cyl + disp,
    data = mtcars,
    num.trees = 1,
    max.depth = 2,
    seed = 100,
    num.threads = 1
  )

  tree <- ranger::treeInfo(model, 1)
  child_info <- tidypredict:::get_child_info(tree)
  terminal_nodes <- tree$nodeID[tree$terminal]

  # Test with default_op = TRUE (uses "less" and "more-equal")
  # Test all terminal nodes to exercise both left and right child paths
  all_ops <- character(0)
  for (node in terminal_nodes) {
    path <- tidypredict:::get_ra_path(node, tree, child_info, TRUE)
    expect_type(path, "list")
    if (length(path) > 0) {
      ops <- vapply(path, function(x) x$op, character(1))
      expect_true(all(ops %in% c("less", "more-equal")))
      all_ops <- c(all_ops, ops)
    }
  }
  # Ensure both operators are used across all paths
  expect_true("less" %in% all_ops)
  expect_true("more-equal" %in% all_ops)
})

test_that("parse_model.ranger errors on classification", {
  skip_on_cran()
  skip_on_os("windows")
  skip_on_os("linux")

  model <- ranger::ranger(
    Species ~ Sepal.Length + Sepal.Width,
    data = iris,
    num.trees = 3,
    max.depth = 2,
    seed = 123,
    num.threads = 1
  )

  expect_snapshot(parse_model(model), error = TRUE)
})

test_that("legacy get_ra_tree converts factor predictions to character", {
  skip_on_cran()
  skip_on_os("windows")
  skip_on_os("linux")

  # Non-probability classification model returns factor predictions
  model <- ranger::ranger(
    Species ~ Sepal.Length + Sepal.Width,
    data = iris,
    num.trees = 1,
    max.depth = 2,
    seed = 123,
    num.threads = 1,
    probability = FALSE
  )

  tree <- tidypredict:::get_ra_tree(1, model)

  expect_type(tree, "list")
  # Predictions should be converted to character
  expect_type(tree[[1]]$prediction, "character")
})

test_that("legacy get_ra_tree handles probability predictions", {
  skip_on_cran()
  skip_on_os("windows")
  skip_on_os("linux")

  model <- ranger::ranger(
    Species ~ Sepal.Length + Sepal.Width,
    data = iris,
    num.trees = 1,
    max.depth = 2,
    seed = 123,
    num.threads = 1,
    probability = TRUE
  )

  tree <- tidypredict:::get_ra_tree(1, model)

  expect_type(tree, "list")
  expect_true(length(tree) > 0)

  # With probability = TRUE, nodes should have probs field
  expect_true("probs" %in% names(tree[[1]]))
  expect_true("prob" %in% names(tree[[1]]))
})

# Model parser -------------------------------------

get_xgb_path <- function(row_id, tree) {
  find <- row_id
  path <- row_id

  yes_vec <- tree$Yes
  no_vec <- tree$No

  for (j in row_id:1) {
    yes <- yes_vec[[j]]
    no <- no_vec[[j]]

    if (!is.na(yes) && !is.na(no)) {
      if (yes == find || no == find) {
        find <- j
        path <- c(path, j)
      }
    }
  }

  purrr::map2(
    path[seq2(1, length(path) - 1)],
    path[seq2(2, length(path))],
    get_xgb_path_fun,
    tree = tree
  )
}

get_xgb_path_fun <- function(.x, .y, tree) {
  yes <- tree$Yes[[.y]]
  no <- tree$No[[.y]]

  missing <- tree$Missing[[.y]]
  if (yes %in% .x) {
    op <- "more-equal"
    missing <- missing %in% yes
  }
  if (no %in% .x) {
    op <- "less"
    missing <- missing %in% no
  }
  list(
    type = "conditional",
    col = tree$feature_name[[.y]],
    val = tree$Split[[.y]],
    op = op,
    missing = missing
  )
}

get_xgb_tree <- function(tree) {
  paths <- seq_len(nrow(tree))[tree[, "Feature"] == "Leaf"]
  x <- map(
    paths,
    ~ {
      list(
        prediction = tree$Gain[[.x]] %||% tree$Quality[[.x]],
        path = get_xgb_path(.x, tree)
      )
    }
  )
  x
}

get_xgb_trees <- function(model) {
  xd <- xgboost::xgb.dump(
    model = model,
    dump_format = "text",
    with_stats = TRUE
  )
  if (is.null(attr(model, "param"))) {
    # Old xgboost API (< 2.0) - kept for backwards compatibility
    xd <- xgboost::xgb.dump(
      model = model,
      dump_format = "text",
      with_stats = TRUE
    )
    feature_names <- model$feature_names
    get_xgb_trees_character(xd, feature_names)
  } else {
    feature_names <- xgboost::getinfo(model, "feature_name")
    get_xgb_trees_character(model, feature_names)
  }
}

get_xgb_trees_character <- function(x, feature_names) {
  # Old xgboost API (< 2.0) - kept for backwards compatibility with old models
  if (is.character(x)) {
    trees <- xgboost::xgb.model.dt.tree(text = x)
  } else {
    trees <- xgboost::xgb.model.dt.tree(model = x)
  }
  trees <- as.data.frame(trees)
  trees$original_order <- 1:nrow(trees)

  if (is.character(x)) {
    feature_names_tbl <- data.frame(
      Feature = as.character(0:(length(feature_names) - 1)),
      feature_name = feature_names,
      stringsAsFactors = FALSE
    )
    trees <- merge(trees, feature_names_tbl, by = "Feature", all.x = TRUE)
  } else {
    trees$feature_name <- ifelse(trees$Feature == "Leaf", NA, trees$Feature)
  }
  trees <- trees[
    order(trees$original_order),
    !names(trees) %in% "original_order"
  ]
  trees[, c("Yes", "No", "Missing")] <-
    lapply(trees[, c("Yes", "No", "Missing")], function(x) sub("^.*-", "", x))
  trees[, c("Yes", "No", "Missing")] <-
    lapply(trees[, c("Yes", "No", "Missing")], function(x) as.integer(x) + 1)

  trees_split <- split(trees, trees$Tree)

  purrr::map(trees_split, get_xgb_tree)
}

#' @export
parse_model.xgb.Booster <- function(model) {
  old <- is.null(attr(model, "param"))

  params <- attr(model, "param") %||% model$params
  wosilent <- params[names(params) != "silent"]
  wosilent$silent <- params$silent

  pm <- list()
  pm$general$model <- "xgb.Booster"
  pm$general$type <- "xgb"
  pm$general$params <- wosilent

  if (old) {
    # Old xgboost API (< 2.0) - kept for backwards compatibility
    pm$general$feature_names <- model$feature_names
    pm$general$niter <- model$niter
    pm$general$nfeatures <- model$nfeatures
  } else {
    pm$general$feature_names <- xgboost::getinfo(model, "feature_name")
    pm$general$niter <- utils::getFromNamespace(
      "xgb.get.num.boosted.rounds",
      ns = "xgboost"
    )(model)
    pm$general$nfeatures <- length(pm$general$feature_names)
  }

  json_params <- get_xgb_json_params(model)
  pm$general$params$base_score <- json_params$base_score
  pm$general$booster_name <- json_params$booster_name
  pm$general$weight_drop <- json_params$weight_drop

  pm$general$version <- 3
  pm$trees <- get_xgb_trees(model)
  as_parsed_model(pm)
}

# Shared helpers -----------------------------------------------

apply_dart_weights <- function(trees, weight_drop) {
  if (!is.null(weight_drop) && !all(weight_drop == 1)) {
    trees <- Map(
      function(tree_expr, weight) {
        expr(!!tree_expr * !!weight)
      },
      trees,
      weight_drop
    )
  }
  trees
}

get_xgb_json_params <- function(model) {
  tmp_file <- tempfile(fileext = ".json")
  xgboost::xgb.save(model, tmp_file)

  # Use regex extraction instead of full JSON parsing (3-4x faster)
  txt <- paste(readLines(tmp_file, warn = FALSE), collapse = "")

  parse_xgb_json_params(txt)
}

parse_xgb_json_params <- function(txt) {
  # Extract base_score - format is "base_score":"[5E-1]"
  base_score_match <- regmatches(
    txt,
    regexpr('base_score":"\\[[^]]+\\]', txt, perl = TRUE)
  )

  if (length(base_score_match) > 0 && nchar(base_score_match) > 0) {
    base_score_str <- gsub(
      'base_score":"\\[([^]]+)\\]',
      "\\1",
      base_score_match,
      perl = TRUE
    )
    base_score <- as.numeric(strsplit(base_score_str, ",")[[1]])
  } else {
    base_score <- 0.5 # nocov
  }

  # Extract booster name using fixed string matching

  booster_name <- "gbtree"
  if (grepl('"name":"dart"', txt, fixed = TRUE)) {
    booster_name <- "dart"
  } else if (grepl('"name":"gblinear"', txt, fixed = TRUE)) {
    booster_name <- "gblinear"
  }

  # Extract the dropout weights. xgboost >= 3.4 canonicalises `booster = "dart"`
  # to `"gbtree"` in the saved JSON while still writing `weight_drop`, so the
  # weights have to be looked for regardless of the serialised booster name.
  # Non-dropout models either omit `weight_drop` or leave it all ones.

  weight_drop <- NULL
  if (booster_name != "gblinear") {
    wd_match <- regmatches(
      txt,
      regexpr('weight_drop":\\[[^]]+\\]', txt, perl = TRUE)
    )
    if (length(wd_match) > 0 && nchar(wd_match) > 0) {
      wd_str <- gsub('weight_drop":\\[([^]]+)\\]', "\\1", wd_match, perl = TRUE)
      weight_drop <- as.numeric(strsplit(wd_str, ",")[[1]])
    }
    if (!is.null(weight_drop) && all(weight_drop == 1)) {
      weight_drop <- NULL
    }
    if (!is.null(weight_drop)) {
      booster_name <- "dart"
    }
  }

  list(
    base_score = base_score,
    booster_name = booster_name,
    weight_drop = weight_drop
  )
}

# Fit model (nested) -----------------------------------------------

#' @export
tidypredict_fit.xgb.Booster <- function(model, ...) {
  build_fit_formula_xgb_nested(model)
}

# Build nested xgboost formula (from model directly)
build_fit_formula_xgb_nested <- function(model) {
  trees_nested <- extract_xgb_trees_nested(model)

  # Apply DART weight_drop if present
  json_params <- get_xgb_json_params(model)
  trees_nested <- apply_dart_weights(trees_nested, json_params$weight_drop)

  # Additive model
  f <- reduce_addition(trees_nested)

  base_score <- json_params$base_score
  if (is.null(base_score)) {
    base_score <- 0.5 # nocov
  }

  params <- attr(model, "param") %||% model$params
  objective <- params$objective

  apply_xgb_objective(f, objective, base_score)
}

# Build nested formula from parsed xgboost model (version 3)
build_fit_formula_xgb_from_parsed <- function(parsedmodel) {
  # Build nested trees from flat paths
  trees_nested <- map(parsedmodel$trees, function(tree) {
    build_nested_from_flat_paths(tree, build_xgb_nested_condition)
  })

  # Apply DART weight_drop if present
  weight_drop <- parsedmodel$general$weight_drop
  trees_nested <- apply_dart_weights(trees_nested, weight_drop)

  # Additive model
  f <- reduce_addition(trees_nested)

  base_score <- parsedmodel$general$params$base_score
  if (is.null(base_score)) {
    base_score <- 0.5 # nocov
  }

  objective <- parsedmodel$general$params$objective

  apply_xgb_objective(f, objective, base_score)
}

# Apply xgboost objective transformation to formula
apply_xgb_objective <- function(f, objective, base_score) {
  if (is.null(objective)) {
    if (base_score != 0) {
      f <- expr_addition(f, base_score)
    }
    cli::cli_warn(
      "If the objective is a custom function,
      please explicitly apply it to the output."
    )
    return(f)
  }

  if (
    objective %in%
      c("reg:squarederror", "reg:squaredlogerror", "binary:logitraw")
  ) {
    if (base_score != 0) {
      f <- expr_addition(f, base_score)
    }
    return(f)
  }

  if (objective %in% c("binary:logistic", "reg:logistic")) {
    return(expr(
      1 - 1 / (1 + exp(!!f + log(!!base_score / (1 - !!base_score))))
    ))
  }

  if (objective %in% c("count:poisson", "reg:tweedie", "reg:gamma")) {
    return(expr(!!base_score * exp(!!f)))
  }

  if (objective %in% c("reg:pseudohubererror", "reg:absoluteerror")) {
    if (base_score != 0) {
      f <- expr_addition(f, base_score)
    }
    return(f)
  }

  if (objective == "binary:hinge") {
    return(expr(as.numeric((!!f + !!base_score) >= 0)))
  }

  cli::cli_abort(
    c(
      "Objective {.val {objective}} is not supported.",
      i = "Supported objectives: {.val binary:hinge}, {.val binary:logistic},
      {.val binary:logitraw}, {.val count:poisson}, {.val reg:absoluteerror},
      {.val reg:gamma}, {.val reg:logistic}, {.val reg:pseudohubererror},
      {.val reg:squarederror}, {.val reg:squaredlogerror}, {.val reg:tweedie}.",
      i = "Multiclass objectives ({.val multi:softmax}, {.val multi:softprob})
      are not supported."
    )
  )
}

# nocov start
# Build condition for xgboost nested generation from path element
# Note: This function is currently not called due to how build_nested_from_flat_paths
# partitions leaves - xgboost's op naming doesn't match the expected convention.
# Kept for potential future use when the partitioning logic is updated.
build_xgb_nested_condition <- function(path_elem) {
  col <- rlang::sym(path_elem$col)
  val <- as.numeric(path_elem$val)
  missing <- path_elem$missing %||% FALSE

  # xgboost uses "less" for left (< threshold) and "more-equal" for right
  if (path_elem$op %in% c("less", "more-equal")) {
    if (missing) {
      expr(!!col < !!val | is.na(!!col))
    } else {
      expr(!!col < !!val)
    }
  } else {
    # This shouldn't happen in normal xgboost trees
    expr(!!col >= !!val)
  }
}
# nocov end

# Extract nested trees from xgboost model
extract_xgb_trees_nested <- function(model) {
  trees_df <- get_xgb_trees_df(model)
  trees_split <- split(trees_df, trees_df$Tree)
  map(trees_split, build_nested_xgb_tree)
}

# Get xgboost trees as data frame
get_xgb_trees_df <- function(model) {
  if (is.null(attr(model, "param"))) {
    # Old xgboost API (< 2.0) - kept for backwards compatibility
    xd <- xgboost::xgb.dump(
      model = model,
      dump_format = "text",
      with_stats = TRUE
    )
    feature_names <- model$feature_names
    trees <- xgboost::xgb.model.dt.tree(text = xd)
  } else {
    feature_names <- xgboost::getinfo(model, "feature_name")
    trees <- xgboost::xgb.model.dt.tree(model = model)
  }
  trees <- as.data.frame(trees)

  # Map feature indices to names if needed
  if (is.null(attr(model, "param"))) {
    feature_names_tbl <- data.frame(
      Feature = as.character(0:(length(feature_names) - 1)),
      feature_name = feature_names,
      stringsAsFactors = FALSE
    )
    trees <- merge(trees, feature_names_tbl, by = "Feature", all.x = TRUE)
  } else {
    trees$feature_name <- ifelse(trees$Feature == "Leaf", NA, trees$Feature)
  }

  # Convert Yes/No/Missing to integer indices
  trees[, c("Yes", "No", "Missing")] <- lapply(
    trees[, c("Yes", "No", "Missing")],
    function(x) {
      dash_pos <- regexpr("-", x, fixed = TRUE)
      as.integer(substring(x, dash_pos + 1L)) + 1L
    }
  )

  trees
}

# Build nested case_when for a single xgboost tree
build_nested_xgb_tree <- function(tree_df) {
  # Pre-extract columns as vectors for fast indexing (avoids slow df[i,] access)
  Feature <- tree_df$Feature
  Gain <- tree_df$Gain %||% tree_df$Quality
  Split <- tree_df$Split
  Yes <- tree_df$Yes
  No <- tree_df$No
  Missing <- tree_df$Missing
  feature_name <- tree_df$feature_name

  build_node <- function(node_idx) {
    # Leaf node
    if (Feature[node_idx] == "Leaf") {
      return(Gain[node_idx])
    }

    # Internal node
    col <- rlang::sym(feature_name[node_idx])
    threshold <- Split[node_idx]
    left_idx <- Yes[node_idx]
    right_idx <- No[node_idx]
    missing_idx <- Missing[node_idx]

    left_subtree <- build_node(left_idx)
    right_subtree <- build_node(right_idx)

    # xgboost: Yes = left (< threshold), No = right (>= threshold)
    # Missing can go either way
    if (missing_idx == left_idx) {
      # Missing goes left: (< threshold OR is.na)
      condition <- expr(!!col < !!threshold | is.na(!!col)) # nocov
    } else {
      # Missing goes right or no missing: < threshold (no NA)
      condition <- expr(!!col < !!threshold)
    }

    expr(case_when(!!condition ~ !!left_subtree, .default = !!right_subtree))
  }

  build_node(1L)
}

# Legacy flat case_when (for v1/v2 parsed model compatibility) ----------------
# These functions generate flat case_when expressions and are preserved for
# backwards compatibility when loading parsed models saved with version < 3.

get_xgb_case <- function(path, prediction) {
  conditions <- map(path, get_xgb_case_fun)
  cl <- combine_path_conditions(conditions)
  expr(!!cl ~ !!prediction)
}

get_xgb_case_fun <- function(.x) {
  if (.x$op == "less") {
    if (.x$missing) {
      i <- expr(
        (!!as.name(.x$col) >= !!as.numeric(.x$val) | is.na(!!as.name(.x$col)))
      )
    } else {
      i <- expr(!!as.name(.x$col) >= !!as.numeric(.x$val))
    }
  } else if (.x$op == "more-equal") {
    if (.x$missing) {
      i <- expr(
        (!!as.name(.x$col) < !!as.numeric(.x$val) | is.na(!!as.name(.x$col)))
      )
    } else {
      i <- expr(!!as.name(.x$col) < !!as.numeric(.x$val))
    }
  }
  i
}

get_xgb_case_tree <- function(tree_no, parsedmodel) {
  map(
    parsedmodel$trees[[tree_no]],
    ~ get_xgb_case(.x$path, .x$prediction)
  )
}

build_fit_formula_xgb <- function(parsedmodel) {
  f <- map(
    seq_len(length(parsedmodel$trees)),
    ~ expr(case_when(!!!get_xgb_case_tree(.x, parsedmodel)))
  )

  # Apply DART weight_drop if present
  f <- apply_dart_weights(f, parsedmodel$general$weight_drop)

  # additive model
  f <- reduce_addition(f)

  base_score <- parsedmodel$general$params$base_score
  if (is.null(base_score)) {
    base_score <- 0.5
  }

  objective <- parsedmodel$general$params$objective

  apply_xgb_objective(f, objective, base_score)
}

# For {orbital} -----------------------------------------------

#' Extract processed xgboost trees
#'
#' For use in orbital package.
#' @param model An xgb.Booster model
#' @keywords internal
#' @export
.extract_xgb_trees <- function(model) {
  if (!inherits(model, "xgb.Booster")) {
    cli::cli_abort(
      "{.arg model} must be {.cls xgb.Booster}, not {.obj_type_friendly {model}}."
    )
  }

  json_params <- get_xgb_json_params(model)
  trees <- extract_xgb_trees_nested(model)
  apply_dart_weights(trees, json_params$weight_drop)
}

# Helper to create test model
# Uses mtcars[, -9] (all columns except 'am') to avoid boundary issues
# This matches the original test setup and avoids floating point precision
# issues at exact split boundaries
make_xgb_model <- function(
  max_depth = 2L,
  nrounds = 4L,
  objective = "reg:squarederror"
) {
  xgb_data <- xgboost::xgb.DMatrix(
    nthread = 1L,
    as.matrix(mtcars[, -9]),
    label = mtcars$am
  )

  xgboost::xgb.train(
    params = list(
      nthread = 1,
      max_depth = max_depth,
      objective = objective,
      base_score = 0.5
    ),
    data = xgb_data,
    nrounds = nrounds,
    verbose = 0
  )
}

# Helper to get the standard xgb.DMatrix for testing
make_xgb_data <- function() {
  xgboost::xgb.DMatrix(nthread = 1L, as.matrix(mtcars[, -9]), label = mtcars$am)
}

# Parser tests ---------------------------------------------------------------

test_that("parse_model returns correct structure", {
  skip_if_not_installed("xgboost")
  model <- make_xgb_model()
  pm <- parse_model(model)

  expect_s3_class(pm, "parsed_model")
  expect_s3_class(pm, "pm_xgb")

  expect_equal(pm$general$model, "xgb.Booster")
  expect_equal(pm$general$type, "xgb")
  expect_equal(pm$general$version, 3)

  expect_gt(length(pm$trees), 0)
})

test_that("correct number of trees extracted", {
  skip_if_not_installed("xgboost")
  model <- make_xgb_model(nrounds = 5L)
  pm <- parse_model(model)

  expect_length(pm$trees, 5)
})

test_that("each tree has leaves with predictions and paths", {
  skip_if_not_installed("xgboost")
  model <- make_xgb_model()
  pm <- parse_model(model)

  tree1 <- pm$trees[[1]]
  expect_gt(length(tree1), 0)

  for (leaf in tree1) {
    expect_contains(names(leaf), "prediction")
    expect_contains(names(leaf), "path")
    expect_type(leaf$prediction, "double")
    expect_type(leaf$path, "list")
  }
})

test_that("path conditions have correct structure", {
  skip_if_not_installed("xgboost")
  model <- make_xgb_model()
  pm <- parse_model(model)

  tree1 <- pm$trees[[1]]
  leaves_with_paths <- which(vapply(tree1, \(x) length(x$path) > 0, logical(1)))

  if (length(leaves_with_paths) > 0) {
    leaf_with_path <- tree1[[leaves_with_paths[1]]]

    cond <- leaf_with_path$path[[1]]
    expect_equal(cond$type, "conditional")
    expect_contains(names(cond), c("col", "val", "op", "missing"))
    expect_contains(c("less", "more-equal"), cond$op)
  }
})

test_that("feature names are extracted", {
  skip_if_not_installed("xgboost")
  model <- make_xgb_model()
  pm <- parse_model(model)

  # Uses mtcars[, -9] which has all columns except 'am'
  expected_names <- colnames(mtcars)[-9]
  expect_equal(pm$general$feature_names, expected_names)
})

test_that("params are extracted", {
  skip_if_not_installed("xgboost")
  model <- make_xgb_model()
  pm <- parse_model(model)

  expect_contains(names(pm$general), "params")
  expect_equal(pm$general$params$objective, "reg:squarederror")
})

test_that("niter and nfeatures are extracted", {
  skip_if_not_installed("xgboost")
  model <- make_xgb_model(nrounds = 7L)
  pm <- parse_model(model)

  expect_equal(pm$general$niter, 7)
  # Uses mtcars[, -9] which has 10 columns
  expect_equal(pm$general$nfeatures, 10)
})

test_that("base_score is extracted", {
  skip_if_not_installed("xgboost")
  model <- make_xgb_model()
  pm <- parse_model(model)

  expect_type(pm$general$params$base_score, "double")
})

test_that("path contains both less and more-equal operators", {
  skip_if_not_installed("xgboost")
  model <- make_xgb_model()
  pm <- parse_model(model)

  all_ops <- unlist(lapply(pm$trees[[1]], function(leaf) {
    vapply(leaf$path, \(p) p$op, character(1))
  }))

  expect_contains(all_ops, "more-equal")
  expect_contains(all_ops, "less")
})

test_that("deeper tree paths are traced correctly", {
  skip_if_not_installed("xgboost")

  set.seed(123)
  n <- 100
  X <- matrix(rnorm(n * 3), ncol = 3)
  colnames(X) <- c("a", "b", "c")
  y <- X[, 1] + X[, 2] * 2 + X[, 3] * 3 + rnorm(n, sd = 0.1)

  dtrain <- xgboost::xgb.DMatrix(
    nthread = 1L,
    X,
    label = y,
    feature_names = c("a", "b", "c")
  )
  model <- xgboost::xgb.train(
    params = list(
      nthread = 1,
      max_depth = 4L,
      objective = "reg:squarederror"
    ),
    data = dtrain,
    nrounds = 1L,
    verbose = 0
  )

  pm <- parse_model(model)
  tree <- pm$trees[[1]]

  path_lengths <- vapply(tree, \(leaf) length(leaf$path), integer(1))
  expect_true(any(path_lengths >= 2))
})

test_that("model without explicit feature names still works", {
  skip_if_not_installed("xgboost")

  set.seed(789)
  X <- data.matrix(mtcars[, c("mpg", "cyl")])
  y <- mtcars$hp

  dtrain <- xgboost::xgb.DMatrix(nthread = 1L, X, label = y)

  model <- xgboost::xgb.train(
    params = list(
      nthread = 1,
      max_depth = 2L,
      objective = "reg:squarederror"
    ),
    data = dtrain,
    nrounds = 3L,
    verbose = 0
  )

  pm <- parse_model(model)

  expect_s3_class(pm, "pm_xgb")
  expect_length(pm$trees, 3)
  expect_equal(pm$general$nfeatures, 2)
})

# Fit formula tests ----------------------------------------------------------

test_that("tidypredict_fit returns language object", {
  skip_if_not_installed("xgboost")
  model <- make_xgb_model()

  fit_formula <- tidypredict_fit(model)

  expect_type(fit_formula, "language")
})

test_that("tidypredict_fit works on parsed model", {
  skip_if_not_installed("xgboost")
  model <- make_xgb_model()
  pm <- parse_model(model)

  fit_formula <- tidypredict_fit(pm)

  expect_type(fit_formula, "language")
})

test_that("produced case_when uses .default", {
  skip_if_not_installed("xgboost")
  model <- make_xgb_model()

  fit <- tidypredict_fit(model)
  fit_text <- rlang::expr_text(fit)

  expect_match(fit_text, "\\.default")
})

test_that("reg:squarederror predictions match native predict", {
  skip_if_not_installed("xgboost")

  xgb_data <- make_xgb_data()
  model <- make_xgb_model(objective = "reg:squarederror")

  result <- tidypredict_test(model, mtcars, xg_df = xgb_data, threshold = 1e-7)

  expect_s3_class(result, "tidypredict_test")
  expect_false(result$alert)
})

test_that("binary:logistic predictions match native predict", {
  skip_if_not_installed("xgboost")

  xgb_data <- make_xgb_data()
  model <- make_xgb_model(objective = "binary:logistic")

  result <- tidypredict_test(model, mtcars, xg_df = xgb_data, threshold = 1e-7)

  expect_s3_class(result, "tidypredict_test")
  expect_false(result$alert)
})

test_that("reg:logistic predictions match native predict", {
  skip_if_not_installed("xgboost")

  xgb_data <- make_xgb_data()
  model <- make_xgb_model(objective = "reg:logistic")

  result <- tidypredict_test(model, mtcars, xg_df = xgb_data, threshold = 1e-7)

  expect_s3_class(result, "tidypredict_test")
  expect_false(result$alert)
})

test_that("binary:logitraw predictions match native predict", {
  skip_if_not_installed("xgboost")

  xgb_data <- make_xgb_data()
  model <- make_xgb_model(objective = "binary:logitraw")

  result <- tidypredict_test(model, mtcars, xg_df = xgb_data, threshold = 1e-7)

  expect_s3_class(result, "tidypredict_test")
  expect_false(result$alert)
})

test_that("count:poisson predictions match native predict", {
  skip_if_not_installed("xgboost")

  # Add 0.1 to avoid exact split boundaries (float32 vs float64 precision)
  mtcars_adj <- mtcars
  mtcars_adj[, -9] <- mtcars_adj[, -9] + 0.1

  xgb_data <- xgboost::xgb.DMatrix(
    nthread = 1L,
    as.matrix(mtcars_adj[, -9]),
    label = mtcars_adj$carb
  )

  model <- xgboost::xgb.train(
    params = list(
      nthread = 1,
      max_depth = 2L,
      objective = "count:poisson",
      base_score = 0.5
    ),
    data = xgb_data,
    nrounds = 4L,
    verbose = 0
  )

  result <- tidypredict_test(
    model,
    mtcars_adj,
    xg_df = xgb_data,
    threshold = 1e-6
  )

  expect_s3_class(result, "tidypredict_test")
  expect_false(result$alert)
})

test_that("reg:tweedie predictions match native predict", {
  skip_if_not_installed("xgboost")

  # Add 0.1 to avoid exact split boundaries (float32 vs float64 precision)
  mtcars_adj <- mtcars
  mtcars_adj[, -9] <- mtcars_adj[, -9] + 0.1

  xgb_data <- xgboost::xgb.DMatrix(
    nthread = 1L,
    as.matrix(mtcars_adj[, -9]),
    label = mtcars_adj$hp
  )

  model <- xgboost::xgb.train(
    params = list(
      nthread = 1,
      max_depth = 2L,
      objective = "reg:tweedie",
      base_score = 0.5
    ),
    data = xgb_data,
    nrounds = 4L,
    verbose = 0
  )

  result <- tidypredict_test(
    model,
    mtcars_adj,
    xg_df = xgb_data,
    threshold = 1e-6
  )

  expect_s3_class(result, "tidypredict_test")
  expect_false(result$alert)
})

test_that("reg:squaredlogerror predictions match native predict", {
  skip_if_not_installed("xgboost")

  # Add 0.1 to avoid exact split boundaries (float32 vs float64 precision)
  mtcars_adj <- mtcars
  mtcars_adj[, -9] <- mtcars_adj[, -9] + 0.1

  xgb_data <- xgboost::xgb.DMatrix(
    nthread = 1L,
    as.matrix(mtcars_adj[, -9]),
    label = mtcars_adj$hp
  )

  model <- xgboost::xgb.train(
    params = list(
      nthread = 1,
      max_depth = 2L,
      objective = "reg:squaredlogerror",
      base_score = 0.5
    ),
    data = xgb_data,
    nrounds = 4L,
    verbose = 0
  )

  result <- tidypredict_test(
    model,
    mtcars_adj,
    xg_df = xgb_data,
    threshold = 1e-6
  )

  expect_s3_class(result, "tidypredict_test")
  expect_false(result$alert)
})

test_that("reg:gamma predictions match native predict", {
  skip_if_not_installed("xgboost")

  # Add 0.1 to avoid exact split boundaries (float32 vs float64 precision)
  mtcars_adj <- mtcars
  mtcars_adj[, -9] <- mtcars_adj[, -9] + 0.1

  xgb_data <- xgboost::xgb.DMatrix(
    nthread = 1L,
    as.matrix(mtcars_adj[, -9]),
    label = mtcars_adj$hp
  )

  model <- xgboost::xgb.train(
    params = list(
      nthread = 1,
      max_depth = 2L,
      objective = "reg:gamma",
      base_score = 0.5
    ),
    data = xgb_data,
    nrounds = 4L,
    verbose = 0
  )

  result <- tidypredict_test(
    model,
    mtcars_adj,
    xg_df = xgb_data,
    threshold = 1e-6
  )

  expect_s3_class(result, "tidypredict_test")
  expect_false(result$alert)
})

test_that("reg:pseudohubererror predictions match native predict", {
  skip_if_not_installed("xgboost")

  # Add 0.1 to avoid exact split boundaries (float32 vs float64 precision)
  mtcars_adj <- mtcars
  mtcars_adj[, -9] <- mtcars_adj[, -9] + 0.1

  xgb_data <- xgboost::xgb.DMatrix(
    nthread = 1L,
    as.matrix(mtcars_adj[, -9]),
    label = mtcars_adj$hp
  )

  model <- xgboost::xgb.train(
    params = list(
      nthread = 1,
      max_depth = 2L,
      objective = "reg:pseudohubererror",
      base_score = 0.5
    ),
    data = xgb_data,
    nrounds = 4L,
    verbose = 0
  )

  result <- tidypredict_test(
    model,
    mtcars_adj,
    xg_df = xgb_data,
    threshold = 1e-6
  )

  expect_s3_class(result, "tidypredict_test")
  expect_false(result$alert)
})

test_that("reg:absoluteerror predictions match native predict", {
  skip_if_not_installed("xgboost")

  # Snap to a 1/64 grid so every value is exactly representable in float32.
  # xgboost stores split thresholds as float32, so a value that is not
  # float32-exact can land on the other side of a threshold here than it does
  # in predict(). Which splits are chosen varies by xgboost version, so an
  # arbitrary offset is not enough to stay clear of the boundaries.
  mtcars_adj <- mtcars
  mtcars_adj[, -9] <- round(mtcars_adj[, -9] * 64) / 64

  xgb_data <- xgboost::xgb.DMatrix(
    nthread = 1L,
    as.matrix(mtcars_adj[, -9]),
    label = mtcars_adj$hp
  )

  model <- xgboost::xgb.train(
    params = list(
      nthread = 1,
      max_depth = 2L,
      objective = "reg:absoluteerror",
      base_score = 0.5
    ),
    data = xgb_data,
    nrounds = 4L,
    verbose = 0
  )

  result <- tidypredict_test(
    model,
    mtcars_adj,
    xg_df = xgb_data,
    threshold = 1e-5
  )

  expect_s3_class(result, "tidypredict_test")
  expect_false(result$alert)
})

test_that("binary:hinge predictions match native predict", {
  skip_if_not_installed("xgboost")

  # Add 0.1 to avoid exact split boundaries (float32 vs float64 precision)
  mtcars_adj <- mtcars
  mtcars_adj[, -9] <- mtcars_adj[, -9] + 0.1

  xgb_data <- xgboost::xgb.DMatrix(
    nthread = 1L,
    as.matrix(mtcars_adj[, -9]),
    label = mtcars_adj$am
  )

  model <- xgboost::xgb.train(
    params = list(
      nthread = 1,
      max_depth = 2L,
      objective = "binary:hinge",
      base_score = 0.5
    ),
    data = xgb_data,
    nrounds = 4L,
    verbose = 0
  )

  result <- tidypredict_test(
    model,
    mtcars_adj,
    xg_df = xgb_data,
    threshold = 1e-7
  )

  expect_s3_class(result, "tidypredict_test")
  expect_false(result$alert)
})

test_that("DART booster with rate_drop = 0 predictions match native predict", {
  skip_if_not_installed("xgboost")

  xgb_data <- make_xgb_data()

  model <- xgboost::xgb.train(
    params = list(
      nthread = 1,
      max_depth = 2L,
      objective = "reg:squarederror",
      base_score = 0.5,
      booster = "dart",
      rate_drop = 0
    ),
    data = xgb_data,
    nrounds = 4L,
    verbose = 0
  )

  result <- tidypredict_test(model, mtcars, xg_df = xgb_data, threshold = 1e-7)

  expect_s3_class(result, "tidypredict_test")
  expect_false(result$alert)
})

test_that("DART booster with rate_drop > 0 predictions match native predict", {
  skip_if_not_installed("xgboost")

  # Snap to a 1/64 grid so every value is exactly representable in float32; see
  # the reg:absoluteerror test above.
  mtcars_adj <- mtcars
  mtcars_adj[, -9] <- round(mtcars_adj[, -9] * 64) / 64

  xgb_data <- xgboost::xgb.DMatrix(
    nthread = 1L,
    as.matrix(mtcars_adj[, -9]),
    label = mtcars_adj$am
  )

  # `one_drop` forces a tree to be dropped every round, so the dropout weights
  # are actually exercised. `rate_drop` alone leaves that to the RNG.
  model <- suppressWarnings(xgboost::xgb.train(
    params = list(
      nthread = 1,
      max_depth = 2L,
      objective = "reg:squarederror",
      base_score = 0.5,
      booster = "dart",
      rate_drop = 0.3,
      one_drop = 1,
      seed = 123
    ),
    data = xgb_data,
    nrounds = 4L,
    verbose = 0
  ))

  result <- tidypredict_test(
    model,
    mtcars_adj,
    xg_df = xgb_data,
    threshold = 1e-6
  )

  expect_s3_class(result, "tidypredict_test")
  expect_false(result$alert)
})

test_that("DART booster weight_drop is extracted correctly", {
  skip_if_not_installed("xgboost")

  xgb_data <- xgboost::xgb.DMatrix(
    nthread = 1L,
    as.matrix(mtcars[, -9]),
    label = mtcars$am
  )

  # `one_drop` forces a tree to be dropped every round. `rate_drop` alone is
  # not enough: whether any tree is actually dropped depends on the RNG, and
  # xgboost 3.4 drops nothing here, which would leave `weight_drop` all ones.
  model <- suppressWarnings(xgboost::xgb.train(
    params = list(
      nthread = 1,
      max_depth = 2L,
      objective = "reg:squarederror",
      base_score = 0.5,
      booster = "dart",
      rate_drop = 0.3,
      one_drop = 1,
      seed = 123
    ),
    data = xgb_data,
    nrounds = 4L,
    verbose = 0
  ))

  pm <- parse_model(model)

  expect_equal(pm$general$booster_name, "dart")
  expect_type(pm$general$weight_drop, "double")
  expect_length(pm$general$weight_drop, 4)
  # At least one weight should be different from 1 when trees are dropped
  expect_false(all(pm$general$weight_drop == 1))
})

test_that("gbtree booster has no weight_drop", {
  skip_if_not_installed("xgboost")

  xgb_data <- xgboost::xgb.DMatrix(
    nthread = 1L,
    as.matrix(mtcars[, -9]),
    label = mtcars$am
  )

  model <- xgboost::xgb.train(
    params = list(
      nthread = 1,
      max_depth = 2L,
      objective = "reg:squarederror",
      base_score = 0.5,
      booster = "gbtree"
    ),
    data = xgb_data,
    nrounds = 4L,
    verbose = 0
  )

  pm <- parse_model(model)

  expect_equal(pm$general$booster_name, "gbtree")
  expect_null(pm$general$weight_drop)
})

test_that("dropout weights are found when the booster is saved as gbtree", {
  # xgboost >= 3.4 canonicalises `booster = "dart"` to `"gbtree"`
  txt <- paste0(
    '{"learner":{"gradient_booster":{"name":"gbtree",',
    '"weight_drop":[1E0,7E-1,5E-1],"model":{}},',
    '"learner_model_param":{"base_score":"[5E-1]"}}}'
  )

  params <- parse_xgb_json_params(txt)

  expect_equal(params$booster_name, "dart")
  expect_equal(params$weight_drop, c(1, 0.7, 0.5))
})

test_that("all-ones dropout weights are not treated as dropout", {
  txt <- paste0(
    '{"learner":{"gradient_booster":{"name":"gbtree",',
    '"weight_drop":[1E0,1E0],"model":{}},',
    '"learner_model_param":{"base_score":"[5E-1]"}}}'
  )

  params <- parse_xgb_json_params(txt)

  expect_equal(params$booster_name, "gbtree")
  expect_null(params$weight_drop)
})

test_that("model with custom base_score works correctly", {
  skip_if_not_installed("xgboost")

  xgb_data <- xgboost::xgb.DMatrix(
    nthread = 1L,
    as.matrix(mtcars[, -9]),
    label = mtcars$am
  )

  model <- xgboost::xgb.train(
    params = list(
      nthread = 1,
      max_depth = 2L,
      objective = "reg:logistic",
      base_score = mean(mtcars$am)
    ),
    data = xgb_data,
    nrounds = 4L,
    verbose = 0
  )

  result <- tidypredict_test(model, mtcars, xg_df = xgb_data, threshold = 1e-7)

  expect_s3_class(result, "tidypredict_test")
  expect_false(result$alert)
})

test_that("base_score of 0 is not included in formula", {
  skip_if_not_installed("xgboost")

  set.seed(123)
  X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
  y <- mtcars$am
  dtrain <- xgboost::xgb.DMatrix(
    nthread = 1L,
    X,
    label = y,
    feature_names = c("mpg", "cyl", "disp")
  )

  model <- xgboost::xgb.train(
    params = list(
      nthread = 1,
      max_depth = 1L,
      objective = "reg:squarederror",
      base_score = 0
    ),
    data = dtrain,
    nrounds = 1L,
    verbose = 0
  )

  res <- tidypredict_fit(model)
  res_text <- rlang::expr_text(res)
  expect_false(grepl("\\+ 0$", res_text))
})

test_that("base_score of 0.5 is included in formula", {
  skip_if_not_installed("xgboost")

  set.seed(123)
  X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
  y <- mtcars$am
  dtrain <- xgboost::xgb.DMatrix(
    nthread = 1L,
    X,
    label = y,
    feature_names = c("mpg", "cyl", "disp")
  )

  model <- xgboost::xgb.train(
    params = list(
      nthread = 1,
      max_depth = 1L,
      objective = "reg:squarederror",
      base_score = 0.5
    ),
    data = dtrain,
    nrounds = 1L,
    verbose = 0
  )

  res <- tidypredict_fit(model)
  res_text <- rlang::expr_text(res)
  expect_match(res_text, "\\+ \\s*0\\.5")
})

test_that("predictions with missing values work", {
  skip_if_not_installed("xgboost")

  set.seed(456)
  X <- as.matrix(mtcars[, -9])
  y <- mtcars$am
  X_train <- X
  X_train[1:3, 1] <- NA
  dtrain <- xgboost::xgb.DMatrix(nthread = 1L, X_train, label = y)

  model <- xgboost::xgb.train(
    params = list(
      nthread = 1,
      max_depth = 2L,
      objective = "reg:squarederror"
    ),
    data = dtrain,
    nrounds = 3L,
    verbose = 0
  )

  X_pred <- X
  X_pred[5:7, 1] <- NA
  X_pred[10:12, 2] <- NA

  fit_formula <- tidypredict_fit(model)
  dpred <- xgboost::xgb.DMatrix(nthread = 1L, X_pred)
  native_preds <- predict(model, dpred)

  pred_df <- as.data.frame(X_pred)
  tidy_preds <- rlang::eval_tidy(fit_formula, pred_df)

  # Check formula runs without error on data with NA values
  expect_type(tidy_preds, "double")
  expect_length(tidy_preds, nrow(mtcars))
})

test_that("unsupported objective throws error", {
  skip_if_not_installed("xgboost")

  pm <- list(
    general = list(
      params = list(objective = "unsupported_objective"),
      model = "xgb.Booster",
      type = "xgb"
    ),
    trees = list(list(list(prediction = 1, path = list())))
  )
  class(pm) <- c("pm_xgb", "parsed_model", "list")

  expect_snapshot(tidypredict_fit(pm), error = TRUE)
})

test_that("stump trees (no splits) predictions match native predict", {
  skip_if_not_installed("xgboost")

  xgb_data <- xgboost::xgb.DMatrix(
    nthread = 1L,
    as.matrix(mtcars[, -9]),
    label = mtcars$am
  )

  model <- xgboost::xgb.train(
    params = list(
      nthread = 1,
      max_depth = 2L,
      gamma = 100,
      objective = "reg:squarederror",
      base_score = 0.5
    ),
    data = xgb_data,
    nrounds = 4L,
    verbose = 0
  )

  # Verify model contains stump trees (single leaf, no splits)
  pm <- parse_model(model)
  leaves_per_tree <- vapply(pm$trees, length, integer(1), USE.NAMES = FALSE)
  path_lengths <- vapply(
    pm$trees,
    \(tree) length(tree[[1]]$path),
    integer(1),
    USE.NAMES = FALSE
  )
  expect_all_equal(leaves_per_tree, 1L)
  expect_all_equal(path_lengths, 0L)

  result <- tidypredict_test(model, mtcars, xg_df = xgb_data, threshold = 1e-7)

  expect_s3_class(result, "tidypredict_test")
  expect_false(result$alert)
})

test_that("NULL objective warns user", {
  skip_if_not_installed("xgboost")

  pm <- list(
    general = list(
      params = list(base_score = 0),
      model = "xgb.Booster",
      type = "xgb"
    ),
    trees = list(list(list(prediction = 5.0, path = list())))
  )
  class(pm) <- c("pm_xgb", "parsed_model", "list")

  expect_snapshot(tidypredict_fit(pm))
})

test_that("stump tree (empty path) works", {
  skip_if_not_installed("xgboost")

  pm <- list(
    general = list(
      params = list(objective = "reg:squarederror", base_score = 0),
      model = "xgb.Booster",
      type = "xgb"
    ),
    trees = list(list(list(prediction = 42.5, path = list())))
  )
  class(pm) <- c("pm_xgb", "parsed_model", "list")

  result <- tidypredict_fit(pm)
  value <- rlang::eval_tidy(result, data.frame(x = 1))

  expect_equal(value, 42.5)
})

test_that("large model predictions match native predict", {
  skip_if_not_installed("xgboost")

  xgb_data <- make_xgb_data()
  model <- make_xgb_model(
    max_depth = 2L,
    nrounds = 50L,
    objective = "reg:logistic"
  )

  result <- tidypredict_test(model, mtcars, xg_df = xgb_data, threshold = 1e-7)

  expect_s3_class(result, "tidypredict_test")
  expect_false(result$alert)
})

test_that("deep tree predictions match native predict", {
  skip_if_not_installed("xgboost")

  xgb_data <- make_xgb_data()
  model <- make_xgb_model(
    max_depth = 20L,
    nrounds = 4L,
    objective = "binary:logistic"
  )

  result <- tidypredict_test(model, mtcars, xg_df = xgb_data, threshold = 1e-7)

  expect_s3_class(result, "tidypredict_test")
  expect_false(result$alert)
})

# SQL generation tests -------------------------------------------------------

test_that("tidypredict_sql returns SQL class", {
  skip_if_not_installed("xgboost")
  skip_if_not_installed("dbplyr")
  model <- make_xgb_model()

  sql_result <- tidypredict_sql(model, dbplyr::simulate_dbi())

  expect_s3_class(sql_result, "sql")
})

test_that("tidypredict_sql works with parsed model", {
  skip_if_not_installed("xgboost")
  skip_if_not_installed("dbplyr")
  model <- make_xgb_model()
  pm <- parse_model(model)

  sql_result <- tidypredict_sql(pm, dbplyr::simulate_dbi())

  expect_s3_class(sql_result, "sql")
})

test_that("SQL predictions can be generated with SQLite", {
  skip_if_not_installed("xgboost")
  skip_if_not_installed("DBI")
  skip_if_not_installed("RSQLite")
  skip_if_not_installed("dbplyr")

  model <- make_xgb_model()

  con <- DBI::dbConnect(RSQLite::SQLite(), ":memory:")
  on.exit(DBI::dbDisconnect(con), add = TRUE)

  # Use all columns except 'am' (column 9)
  test_data <- mtcars[, -9]
  DBI::dbWriteTable(con, "test_data", test_data)

  sql_query <- tidypredict_sql(model, con)

  # SQL query can be executed without error
  db_result <- DBI::dbGetQuery(
    con,
    paste0("SELECT ", sql_query, " AS pred FROM test_data")
  )

  expect_equal(nrow(db_result), nrow(mtcars))
  expect_type(db_result$pred, "double")
})

test_that("SQL predictions work for binary classification with SQLite", {
  skip_if_not_installed("xgboost")
  skip_if_not_installed("DBI")
  skip_if_not_installed("RSQLite")
  skip_if_not_installed("dbplyr")

  model <- make_xgb_model(objective = "binary:logistic")

  con <- DBI::dbConnect(RSQLite::SQLite(), ":memory:")
  on.exit(DBI::dbDisconnect(con), add = TRUE)

  # Use all columns except 'am' (column 9)
  test_data <- mtcars[, -9]
  DBI::dbWriteTable(con, "test_data", test_data)

  sql_query <- tidypredict_sql(model, con)

  # SQL query can be executed without error
  db_result <- DBI::dbGetQuery(
    con,
    paste0("SELECT ", sql_query, " AS pred FROM test_data")
  )

  expect_equal(nrow(db_result), nrow(mtcars))
  expect_type(db_result$pred, "double")
  # Binary logistic predictions should be between 0 and 1
  expect_true(all(db_result$pred >= 0 & db_result$pred <= 1))
})

# Integration tests ----------------------------------------------------------

test_that("tidypredict_test works for regression", {
  skip_if_not_installed("xgboost")
  model <- make_xgb_model()
  xgb_data <- make_xgb_data()

  result <- tidypredict_test(model, mtcars, xg_df = xgb_data, threshold = 1e-7)

  expect_s3_class(result, "tidypredict_test")
  expect_false(result$alert)
})

test_that("tidypredict_test works for binary classification", {
  skip_if_not_installed("xgboost")

  xgb_data <- make_xgb_data()
  model <- make_xgb_model(objective = "binary:logistic")

  result <- tidypredict_test(model, mtcars, xg_df = xgb_data, threshold = 1e-7)

  expect_s3_class(result, "tidypredict_test")
  expect_false(result$alert)
})

test_that("tidypredict_test xg_df argument is required", {
  skip_if_not_installed("xgboost")
  model <- make_xgb_model()

  # Without xg_df, tidypredict_test should fail
  expect_snapshot(tidypredict_test(model, mtcars), error = TRUE)
})

test_that("tidypredict_test respects max_rows parameter", {
  skip_if_not_installed("xgboost")

  model <- make_xgb_model()
  xgb_data <- make_xgb_data()

  # Create a subset DMatrix for max_rows = 10
  X <- as.matrix(mtcars[1:10, -9])
  xgb_subset <- xgboost::xgb.DMatrix(nthread = 1L, X)

  result <- tidypredict_test(
    model,
    mtcars[1:10, ],
    xg_df = xgb_subset,
    max_rows = 10
  )

  expect_equal(nrow(result$raw_results), 10)
})

test_that(".extract_xgb_trees returns list of expressions", {
  skip_if_not_installed("xgboost")
  model <- make_xgb_model(nrounds = 4L)

  trees <- .extract_xgb_trees(model)

  expect_type(trees, "list")
  expect_length(trees, 4)
  for (tree in trees) {
    expect_type(tree, "language")
  }
})

test_that(".extract_xgb_trees combined results match tidypredict_fit", {
  skip_if_not_installed("xgboost")
  model <- make_xgb_model(nrounds = 4L, objective = "reg:squarederror")

  trees <- .extract_xgb_trees(model)
  eval_env <- rlang::new_environment(
    data = as.list(mtcars),
    parent = asNamespace("dplyr")
  )
  tree_preds <- lapply(trees, rlang::eval_tidy, env = eval_env)
  pm <- parse_model(model)
  base_score <- pm$general$params$base_score
  combined <- Reduce(`+`, tree_preds) + base_score

  fit_result <- rlang::eval_tidy(tidypredict_fit(model), mtcars)

  expect_equal(combined, fit_result)
})

test_that(".extract_xgb_trees errors on non-xgb.Booster", {
  expect_snapshot(.extract_xgb_trees(list()), error = TRUE)
})

test_that(".extract_xgb_trees combined results match tidypredict_fit for DART", {
  skip_if_not_installed("xgboost")

  # Add 0.1 to avoid exact split boundaries (float32 vs float64 precision)
  mtcars_adj <- mtcars
  mtcars_adj[, -9] <- mtcars_adj[, -9] + 0.1

  xgb_data <- xgboost::xgb.DMatrix(
    nthread = 1L,
    as.matrix(mtcars_adj[, -9]),
    label = mtcars_adj$am
  )

  model <- xgboost::xgb.train(
    params = list(
      nthread = 1,
      max_depth = 2L,
      objective = "reg:squarederror",
      base_score = 0.5,
      booster = "dart",
      rate_drop = 0.3,
      seed = 123
    ),
    data = xgb_data,
    nrounds = 4L,
    verbose = 0
  )

  trees <- .extract_xgb_trees(model)
  eval_env <- rlang::new_environment(
    data = as.list(mtcars_adj),
    parent = asNamespace("dplyr")
  )
  tree_preds <- lapply(trees, rlang::eval_tidy, env = eval_env)
  pm <- parse_model(model)
  base_score <- pm$general$params$base_score
  combined <- Reduce(`+`, tree_preds) + base_score

  fit_result <- rlang::eval_tidy(tidypredict_fit(model), mtcars_adj)

  expect_equal(combined, fit_result)
})

# v1 backwards compatibility tests -------------------------------------------

test_that("gblinear booster is detected by get_xgb_json_params", {
  skip_if_not_installed("xgboost")

  xgb_data <- xgboost::xgb.DMatrix(
    nthread = 1L,
    as.matrix(mtcars[, -9]),
    label = mtcars$am
  )

  model <- xgboost::xgb.train(
    params = list(
      nthread = 1,
      booster = "gblinear",
      objective = "reg:squarederror"
    ),
    data = xgb_data,
    nrounds = 3L,
    verbose = 0
  )

  # get_xgb_json_params works even though parse_model/tidypredict_fit fail
  params <- tidypredict:::get_xgb_json_params(model)

  expect_equal(params$booster_name, "gblinear")
})

test_that("v1 parsed xgboost model produces correct predictions", {
  skip_if_not_installed("xgboost")

  pm <- readRDS(test_path("backwards-compat", "xgb-v2-regression.rds"))

  expect_equal(pm$general$version %||% 1, 1)
  expect_true(!is.null(pm$trees))

  fit <- tidypredict_fit(pm)
  expect_type(fit, "language")

  # Verify predictions can be generated
  pred <- rlang::eval_tidy(fit, mtcars)
  expect_type(pred, "double")
  expect_length(pred, nrow(mtcars))
})

test_that("NULL base_score in v1 parsed model defaults to 0.5", {
  skip_if_not_installed("xgboost")

  # v1 format goes through build_fit_formula_xgb (legacy flat case_when)
  pm <- list(
    general = list(
      params = list(objective = "reg:squarederror"),
      model = "xgb.Booster",
      type = "xgb"
    ),
    trees = list(list(list(prediction = 1.0, path = list())))
  )
  class(pm) <- c("pm_xgb", "parsed_model", "list")

  fit <- tidypredict_fit(pm)
  result <- rlang::eval_tidy(fit, data.frame(x = 1))

  # 1.0 + 0.5 base_score = 1.5
  expect_equal(result, 1.5)
})

test_that("NULL objective with non-zero base_score warns user", {
  skip_if_not_installed("xgboost")

  pm <- list(
    general = list(
      params = list(base_score = 0.3),
      model = "xgb.Booster",
      type = "xgb"
    ),
    trees = list(list(list(prediction = 1.0, path = list())))
  )
  class(pm) <- c("pm_xgb", "parsed_model", "list")

  expect_snapshot(fit <- tidypredict_fit(pm))
  result <- rlang::eval_tidy(fit, data.frame(x = 1))

  expect_equal(result, 1.3)
})

test_that("v1 parsed model with splits produces predictions", {
  skip_if_not_installed("xgboost")

  # v1 format with actual path conditions - tests get_xgb_case_fun
  pm <- list(
    general = list(
      params = list(objective = "reg:squarederror", base_score = 0),
      model = "xgb.Booster",
      type = "xgb"
    ),
    trees = list(list(
      list(
        prediction = 10.0,
        path = list(
          list(
            type = "conditional",
            col = "mpg",
            val = 20,
            op = "more-equal",
            missing = FALSE
          )
        )
      ),
      list(
        prediction = 30.0,
        path = list(
          list(
            type = "conditional",
            col = "mpg",
            val = 20,
            op = "less",
            missing = FALSE
          )
        )
      )
    ))
  )
  class(pm) <- c("pm_xgb", "parsed_model", "list")

  fit <- tidypredict_fit(pm)
  expect_type(fit, "language")

  # Test predictions - more-equal generates <, less generates >=
  test_data <- data.frame(mpg = c(15, 25))
  pred <- rlang::eval_tidy(fit, test_data)

  expect_equal(pred, c(10.0, 30.0))
})

test_that("v1 parsed model with missing=TRUE uses is.na", {
  skip_if_not_installed("xgboost")

  pm <- list(
    general = list(
      params = list(objective = "reg:squarederror", base_score = 0),
      model = "xgb.Booster",
      type = "xgb"
    ),
    trees = list(list(
      list(
        prediction = 10.0,
        path = list(
          list(
            type = "conditional",
            col = "mpg",
            val = 20,
            op = "more-equal",
            missing = TRUE
          )
        )
      ),
      list(
        prediction = 30.0,
        path = list(
          list(
            type = "conditional",
            col = "mpg",
            val = 20,
            op = "less",
            missing = FALSE
          )
        )
      )
    ))
  )
  class(pm) <- c("pm_xgb", "parsed_model", "list")

  fit <- tidypredict_fit(pm)

  # With missing=TRUE on more-equal, NA should match that condition
  test_data <- data.frame(mpg = c(15, NA, 25))
  pred <- rlang::eval_tidy(fit, test_data)

  expect_equal(pred, c(10.0, 10.0, 30.0))
})

test_that("v1 parsed model with missing=TRUE on less op", {
  skip_if_not_installed("xgboost")

  pm <- list(
    general = list(
      params = list(objective = "reg:squarederror", base_score = 0),
      model = "xgb.Booster",
      type = "xgb"
    ),
    trees = list(list(
      list(
        prediction = 10.0,
        path = list(
          list(
            type = "conditional",
            col = "mpg",
            val = 20,
            op = "more-equal",
            missing = FALSE
          )
        )
      ),
      list(
        prediction = 30.0,
        path = list(
          list(
            type = "conditional",
            col = "mpg",
            val = 20,
            op = "less",
            missing = TRUE
          )
        )
      )
    ))
  )
  class(pm) <- c("pm_xgb", "parsed_model", "list")

  fit <- tidypredict_fit(pm)

  # With missing=TRUE on less, NA should match that condition (>= with is.na)
  test_data <- data.frame(mpg = c(15, NA, 25))
  pred <- rlang::eval_tidy(fit, test_data)

  expect_equal(pred, c(10.0, 30.0, 30.0))
})

# YAML serialization tests ---------------------------------------------------

test_that("parsed model can be saved and loaded via YAML", {
  skip_if_not_installed("xgboost")
  skip_if_not_installed("yaml")

  model <- make_xgb_model()
  pm <- parse_model(model)

  tmp_file <- withr::local_tempfile(fileext = ".yml")
  yaml::write_yaml(pm, tmp_file)
  loaded <- yaml::read_yaml(tmp_file)
  pm_loaded <- as_parsed_model(loaded)

  expect_equal(pm_loaded$general$model, pm$general$model)
  expect_equal(pm_loaded$general$type, pm$general$type)
  expect_equal(pm_loaded$general$niter, pm$general$niter)
})

test_that("loaded model produces same predictions", {
  skip_if_not_installed("xgboost")
  skip_if_not_installed("yaml")

  model <- make_xgb_model()
  pm <- parse_model(model)

  tmp_file <- withr::local_tempfile(fileext = ".yml")
  yaml::write_yaml(pm, tmp_file)
  loaded <- yaml::read_yaml(tmp_file)
  pm_loaded <- as_parsed_model(loaded)

  original_preds <- rlang::eval_tidy(tidypredict_fit(pm), mtcars)
  loaded_preds <- rlang::eval_tidy(tidypredict_fit(pm_loaded), mtcars)

  expect_equal(loaded_preds, original_preds, tolerance = 1e-5)
})

# Parsnip integration tests --------------------------------------------------

test_that("tidypredict works with parsnip xgboost regression", {
  skip_if_not_installed("xgboost")
  skip_if_not_installed("parsnip")

  set.seed(123)
  # Use all columns except am (column 9) for consistency
  train_data <- mtcars

  model_spec <- parsnip::boost_tree(
    trees = 4,
    tree_depth = 2,
    min_n = 1
  ) |>
    parsnip::set_engine("xgboost") |>
    parsnip::set_mode("regression")

  model_fit <- parsnip::fit(
    model_spec,
    am ~ . - hp,
    data = train_data
  )

  xgb_model <- model_fit$fit

  expect_s3_class(xgb_model, "xgb.Booster")

  pm <- parse_model(xgb_model)
  expect_s3_class(pm, "parsed_model")
  expect_s3_class(pm, "pm_xgb")
  expect_gt(length(pm$trees), 0)

  fit_formula <- tidypredict_fit(xgb_model)
  expect_type(fit_formula, "language")
})

test_that("tidypredict works with parsnip xgboost classification", {
  skip_if_not_installed("xgboost")
  skip_if_not_installed("parsnip")

  set.seed(456)
  train_data <- mtcars
  train_data$am <- factor(train_data$am)

  model_spec <- parsnip::boost_tree(
    trees = 4,
    tree_depth = 2,
    min_n = 1
  ) |>
    parsnip::set_engine("xgboost") |>
    parsnip::set_mode("classification")

  model_fit <- parsnip::fit(
    model_spec,
    am ~ . - hp,
    data = train_data
  )

  xgb_model <- model_fit$fit

  expect_s3_class(xgb_model, "xgb.Booster")

  fit_formula <- tidypredict_fit(xgb_model)
  expect_type(fit_formula, "language")
})

test_that("tidypredict_sql works with parsnip xgboost model", {
  skip_if_not_installed("xgboost")
  skip_if_not_installed("parsnip")
  skip_if_not_installed("dbplyr")

  set.seed(123)
  train_data <- mtcars

  model_spec <- parsnip::boost_tree(
    trees = 3,
    tree_depth = 2,
    min_n = 1
  ) |>
    parsnip::set_engine("xgboost") |>
    parsnip::set_mode("regression")

  model_fit <- parsnip::fit(
    model_spec,
    am ~ . - hp,
    data = train_data
  )
  xgb_model <- model_fit$fit

  sql_result <- tidypredict_sql(xgb_model, dbplyr::simulate_dbi())

  expect_s3_class(sql_result, "sql")
})

test_that("tidypredict_test works with parsnip xgboost model", {
  skip_if_not_installed("xgboost")
  skip_if_not_installed("parsnip")

  set.seed(123)
  train_data <- mtcars

  model_spec <- parsnip::boost_tree(
    trees = 4,
    tree_depth = 2,
    min_n = 1
  ) |>
    parsnip::set_engine("xgboost") |>
    parsnip::set_mode("regression")

  model_fit <- parsnip::fit(
    model_spec,
    am ~ . - hp,
    data = train_data
  )
  xgb_model <- model_fit$fit

  # Test that formula can be generated and evaluated
  fit_formula <- tidypredict_fit(xgb_model)
  expect_type(fit_formula, "language")

  # Test that predictions can be generated
  preds <- rlang::eval_tidy(fit_formula, train_data)
  expect_type(preds, "double")
  expect_length(preds, nrow(train_data))
})

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

source("_threads.R")
library(dplyr)
library(tidypredict)
library(rpart)
set.seed(100)

# Constants ---------------------------------------------------------------

lgb_identity_objectives <- c(
  "regression",
  "regression_l2",
  "regression_l1",
  "huber",
  "fair",
  "quantile",
  "mape"
)
lgb_exp_objectives <- c("poisson", "gamma", "tweedie")
lgb_sigmoid_objectives <- c("binary", "cross_entropy")
lgb_multiclass_objectives <- c("multiclass", "multiclassova")
lgb_supported_objectives <- c(
  lgb_identity_objectives,
  lgb_exp_objectives,
  lgb_sigmoid_objectives,
  lgb_multiclass_objectives
)

# Model parser -------------------------------------

#' @export
parse_model.lgb.Booster <- function(model) {
  pm <- list()
  pm$general$model <- "lgb.Booster"
  pm$general$type <- "lgb"
  pm$general$version <- 3

  # Extract params (objective, etc.)
  pm$general$params <- model$params

  # Extract feature names and multiclass info from JSON dump
  model_json <- jsonlite::fromJSON(model$dump_model())
  pm$general$feature_names <- model_json$feature_names
  pm$general$nfeatures <- length(model_json$feature_names)
  pm$general$num_class <- model_json$num_class
  pm$general$num_tree_per_iteration <- model_json$num_tree_per_iteration

  # Extract number of iterations
  pm$general$niter <- model$current_iter()

  # Extract linear tree info from model string (if any)
  linear_info <- parse_lgb_linear_trees(model, pm$general$feature_names)

  # Extract trees (flat path format for serialization)
  pm$trees <- get_lgb_trees(model, linear_info)

  as_parsed_model(pm)
}

# Parse linear tree info from model string
parse_lgb_linear_trees <- function(model, feature_names) {
  model_str <- model$save_model_to_string()
  lines <- strsplit(model_str, "\n")[[1]]

  # Find tree boundaries and extract linear info
  linear_info <- list()
  current_tree <- NULL
  is_linear <- FALSE
  leaf_const <- NULL
  num_features <- NULL
  leaf_features <- NULL
  leaf_coeff <- NULL

  save_tree_linear_info <- function() {
    if (!is.null(current_tree) && is_linear) {
      linear_info[[as.character(current_tree)]] <<- parse_lgb_linear_leaves(
        leaf_const,
        num_features,
        leaf_features,
        leaf_coeff,
        feature_names
      )
    }
  }

  for (line in lines) {
    if (startsWith(line, "Tree=")) {
      save_tree_linear_info()
      # Start new tree
      current_tree <- as.integer(sub("^Tree=", "", line))
      is_linear <- FALSE
      leaf_const <- NULL
      num_features <- NULL
      leaf_features <- NULL
      leaf_coeff <- NULL
    } else if (startsWith(line, "is_linear=1")) {
      is_linear <- TRUE
    } else if (startsWith(line, "leaf_const=")) {
      leaf_const <- sub("^leaf_const=", "", line)
    } else if (startsWith(line, "num_features=")) {
      num_features <- sub("^num_features=", "", line)
    } else if (startsWith(line, "leaf_features=")) {
      leaf_features <- sub("^leaf_features=", "", line)
    } else if (startsWith(line, "leaf_coeff=")) {
      leaf_coeff <- sub("^leaf_coeff=", "", line)
    } else if (startsWith(line, "end of trees")) {
      save_tree_linear_info()
      break
    }
  }

  linear_info
}

# Parse linear leaf info for a single tree
parse_lgb_linear_leaves <- function(
  const_str,
  num_features_str,
  features_str,
  coeff_str,
  feature_names
) {
  # Parse leaf_const (space-separated floats)
  consts <- as.numeric(strsplit(trimws(const_str), "\\s+")[[1]])
  n_leaves <- length(consts)

  # Parse num_features to know how many features per leaf
  num_feats <- as.integer(strsplit(trimws(num_features_str), "\\s+")[[1]])

  # Parse all features and coefficients as flat vectors
  features_str_trimmed <- trimws(features_str)
  coeff_str_trimmed <- trimws(coeff_str)

  if (nchar(features_str_trimmed) > 0) {
    all_features <- as.integer(strsplit(features_str_trimmed, "\\s+")[[1]])
    all_coeffs <- as.numeric(strsplit(coeff_str_trimmed, "\\s+")[[1]])
  } else {
    all_features <- integer(0)
    all_coeffs <- numeric(0)
  }

  # Split features and coefficients by num_feats
  idx <- 1
  linear_leaves <- lapply(seq_len(n_leaves), function(i) {
    nf <- num_feats[i]
    if (nf > 0) {
      feat_idx <- all_features[idx:(idx + nf - 1)]
      coeffs <- all_coeffs[idx:(idx + nf - 1)]
      idx <<- idx + nf
      list(
        intercept = consts[i],
        feature_names = feature_names[feat_idx + 1],
        coefficients = coeffs
      )
    } else {
      list(
        intercept = consts[i],
        feature_names = character(0),
        coefficients = numeric(0)
      )
    }
  })
  names(linear_leaves) <- as.character(seq_len(n_leaves) - 1)

  linear_leaves
}

get_lgb_trees <- function(model, linear_info = list()) {
  trees_df <- lightgbm::lgb.model.dt.tree(model)
  trees_df <- as.data.frame(trees_df)

  # Check for unsupported decision types
  decision_types <- unique(trees_df$decision_type[
    !is.na(trees_df$decision_type)
  ])
  supported_types <- c("<=", "==")
  unsupported <- setdiff(decision_types, supported_types)
  if (length(unsupported) > 0) {
    # nocov start
    cli::cli_abort(
      c(
        "Unsupported decision type{?s} found: {.val {unsupported}}.",
        "i" = "Supported types: {.val {supported_types}}."
      ),
      .internal = TRUE
    )
    # nocov end
  }

  # Split by tree_index
  trees_split <- split(trees_df, trees_df$tree_index)

  # Process each tree with its linear info (if any)
  map(names(trees_split), function(tree_idx) {
    tree_linear <- linear_info[[tree_idx]]
    get_lgb_tree(trees_split[[tree_idx]], tree_linear)
  })
}

get_lgb_children_map <- function(tree_df) {
  # For each split_index, find its children (ordered by row index)
  split_indices <- tree_df$split_index[!is.na(tree_df$split_index)]

  children_map <- lapply(split_indices, function(si) {
    # Children are rows where node_parent==si OR leaf_parent==si
    child_rows <- which(tree_df$node_parent == si | tree_df$leaf_parent == si)
    # Sort by row index: first is LEFT, second is RIGHT
    child_rows[order(child_rows)]
  })
  names(children_map) <- as.character(split_indices)
  children_map
}

get_lgb_tree <- function(tree_df, linear_info = NULL) {
  # Build children map for direction detection
  children_map <- get_lgb_children_map(tree_df)

  # Pre-extract columns as vectors for fast indexing
  leaf_index <- tree_df$leaf_index
  leaf_value <- tree_df$leaf_value
  leaf_parent <- tree_df$leaf_parent
  split_index <- tree_df$split_index
  node_parent <- tree_df$node_parent
  decision_type <- tree_df$decision_type
  default_left <- tree_df$default_left == "TRUE"
  split_feature <- tree_df$split_feature
  threshold <- tree_df$threshold

  # Build split_index to row lookup (avoid repeated which() calls)
  max_split_idx <- suppressWarnings(max(split_index, na.rm = TRUE))
  if (is.finite(max_split_idx)) {
    split_idx_to_row <- integer(max_split_idx + 1)
    for (i in seq_along(split_index)) {
      si <- split_index[i]
      if (!is.na(si)) {
        split_idx_to_row[si + 1L] <- i
      }
    }
  } else {
    # No splits (stump tree) - empty lookup
    split_idx_to_row <- integer(0)
  }

  # Find leaf rows
  leaf_rows <- which(!is.na(leaf_index))

  # For each leaf, trace path to root
  map(leaf_rows, function(leaf_row) {
    leaf_idx <- leaf_index[leaf_row]
    leaf_idx_str <- as.character(leaf_idx)
    leaf_val <- leaf_value[leaf_row]

    # Check if this tree has linear info for this leaf
    if (!is.null(linear_info) && leaf_idx_str %in% names(linear_info)) {
      leaf_linear <- linear_info[[leaf_idx_str]]
      # Store both linear info and fallback value (used when features are NA)
      leaf_linear$fallback <- leaf_val
      list(
        prediction = NULL,
        linear = leaf_linear,
        path = get_lgb_path_fast(
          leaf_row,
          leaf_parent,
          split_idx_to_row,
          node_parent,
          decision_type,
          default_left,
          split_feature,
          threshold,
          children_map
        )
      )
    } else {
      list(
        prediction = leaf_val,
        linear = NULL,
        path = get_lgb_path_fast(
          leaf_row,
          leaf_parent,
          split_idx_to_row,
          node_parent,
          decision_type,
          default_left,
          split_feature,
          threshold,
          children_map
        )
      )
    }
  })
}

# Fast path extraction using pre-extracted vectors
get_lgb_path_fast <- function(
  leaf_row,
  leaf_parent,
  split_idx_to_row,
  node_parent,
  decision_type,
  default_left,
  split_feature,
  threshold,
  children_map
) {
  path <- list()
  current_row <- leaf_row
  current_parent_split <- leaf_parent[leaf_row]

  while (!is.na(current_parent_split)) {
    # Look up parent row directly (O(1) instead of O(n))
    parent_row <- split_idx_to_row[current_parent_split + 1L]

    # Determine direction: is current_row the LEFT or RIGHT child?
    children <- children_map[[as.character(current_parent_split)]]
    is_left_child <- (current_row == children[1])

    # Build condition based on decision type
    dec_type <- decision_type[parent_row]
    def_left <- default_left[parent_row]

    if (dec_type == "<=") {
      # Numerical split
      if (is_left_child) {
        op <- "less-equal"
        missing_with_us <- def_left
      } else {
        op <- "more"
        missing_with_us <- !def_left
      }

      condition <- list(
        type = "conditional",
        col = split_feature[parent_row],
        val = threshold[parent_row],
        op = op,
        missing = missing_with_us
      )
    } else if (dec_type == "==") {
      # Categorical split: threshold is "0||1||3" format
      category_set <- parse_lgb_categorical_threshold(threshold[parent_row])

      if (is_left_child) {
        op <- "in"
        missing_with_us <- def_left
      } else {
        op <- "not-in"
        missing_with_us <- !def_left
      }

      condition <- list(
        type = "set",
        col = split_feature[parent_row],
        vals = category_set,
        op = op,
        missing = missing_with_us
      )
    }

    path <- c(path, list(condition))

    # Move up the tree
    current_row <- parent_row
    current_parent_split <- node_parent[parent_row]
  }

  rev(path) # Reverse to get root-to-leaf order
}

# Parse LightGBM categorical threshold format "0||1||3" -> c(0, 1, 3)
parse_lgb_categorical_threshold <- function(threshold) {
  as.integer(strsplit(threshold, "[|][|]")[[1]])
}

# Shared helpers -----------------------------------------------

# Helper for sigmoid transformation
lgb_sigmoid <- function(f) {
  expr(1 / (1 + exp(-(!!f))))
}

# Apply multiclass transformation to tree expressions
# Shared by nested and from_parsed multiclass builders
apply_lgb_multiclass_transformation <- function(trees, num_class, objective) {
  n_trees <- length(trees)

  # Group trees by class: tree i belongs to class (i-1) %% num_class
  class_trees <- lapply(seq_len(num_class), function(class_idx) {
    which((seq_len(n_trees) - 1) %% num_class == (class_idx - 1))
  })

  # Build raw score formula for each class
  raw_scores <- lapply(class_trees, function(indices) {
    reduce_addition(trees[indices])
  })

  # Apply transformation based on objective
  if (objective == "multiclass") {
    # Softmax: exp(raw_i) / sum(exp(raw_j))
    exp_raws <- map(raw_scores, ~ expr(exp(!!.x)))
    denom <- reduce_addition(exp_raws)
    result <- map(seq_len(num_class), function(i) {
      expr(exp(!!raw_scores[[i]]) / (!!denom))
    })
  } else if (objective == "multiclassova") {
    # One-vs-all: sigmoid for each class independently
    result <- map(raw_scores, lgb_sigmoid)
  }

  names(result) <- paste0("class_", seq_len(num_class) - 1)
  result
}

# Build linear prediction formula: intercept + sum(coeff * feature)
# LightGBM uses fallback leaf_value when ANY feature in the formula is NA
build_lgb_linear_prediction <- function(linear) {
  intercept <- linear$intercept
  feature_names <- linear$feature_names
  coefficients <- linear$coefficients
  fallback <- linear$fallback

  if (length(feature_names) == 0) {
    # No features, just return intercept
    return(intercept)
  }

  # Build the linear formula: intercept + sum(coeff * feature)
  terms <- map(seq_along(feature_names), function(i) {
    feat <- as.name(feature_names[i])
    coef <- coefficients[i]
    expr(!!coef * !!feat)
  })
  all_terms <- c(list(intercept), terms)
  linear_formula <- reduce_addition(all_terms)

  # Build condition: any feature is NA
  na_checks <- map(feature_names, function(fn) {
    feat <- as.name(fn)
    expr(is.na(!!feat))
  })
  any_na <- if (length(na_checks) == 1) {
    na_checks[[1]]
  } else {
    reduce_or(na_checks)
  }

  # If any feature is NA, use fallback; otherwise use linear formula
  expr(ifelse(!!any_na, !!fallback, !!linear_formula))
}

# Apply lightgbm objective transformation to formula
apply_lgb_objective <- function(f, objective, params) {
  # RF boosting averages trees instead of summing
  boosting <- params$boosting
  if (!is.null(boosting) && boosting == "rf") {
    # f is already averaged by caller
  }

  # Apply transformation
  if (objective %in% lgb_exp_objectives) {
    return(expr(exp(!!f)))
  }

  if (objective %in% lgb_sigmoid_objectives) {
    return(lgb_sigmoid(f))
  }

  # Identity objectives - return as-is
  f
}

# Fit model (nested) -----------------------------------------------

#' @export
tidypredict_fit.lgb.Booster <- function(model, ...) {
  parsedmodel <- parse_model(model)
  build_fit_formula_lgb_nested(parsedmodel, model)
}

# Nested formula builder for lightgbm (from model directly)
build_fit_formula_lgb_nested <- function(parsedmodel, model) {
  n_trees <- length(parsedmodel$trees)

  if (n_trees == 0) {
    cli::cli_abort("Model has no trees.")
  }

  objective <- parsedmodel$general$params$objective %||% "regression"

  if (!objective %in% lgb_supported_objectives) {
    cli::cli_abort(
      c(
        "Unsupported objective: {.val {objective}}.",
        "i" = "Supported objectives: {.val {lgb_supported_objectives}}."
      )
    )
  }

  if (objective %in% lgb_multiclass_objectives) {
    return(build_fit_formula_lgb_multiclass_nested(
      parsedmodel,
      model,
      objective
    ))
  }

  # Extract nested trees (pass feature_names to avoid redundant JSON parsing)
  trees <- extract_lgb_trees_nested(model, parsedmodel$general$feature_names)

  # RF boosting averages trees instead of summing
  boosting <- parsedmodel$general$params$boosting
  if (!is.null(boosting) && boosting == "rf") {
    f <- reduce_addition(trees)
    f <- expr_division(f, n_trees)
  } else {
    f <- reduce_addition(trees)
  }

  apply_lgb_objective(f, objective, parsedmodel$general$params)
}

build_fit_formula_lgb_multiclass_nested <- function(
  parsedmodel,
  model,
  objective
) {
  num_class <- parsedmodel$general$num_class

  trees <- extract_lgb_trees_nested(model, parsedmodel$general$feature_names)
  apply_lgb_multiclass_transformation(trees, num_class, objective)
}

# Nested formula builder for lightgbm (from parsed model, version 3)
build_fit_formula_lgb_from_parsed <- function(parsedmodel) {
  n_trees <- length(parsedmodel$trees)

  if (n_trees == 0) {
    cli::cli_abort("Model has no trees.")
  }

  objective <- parsedmodel$general$params$objective %||% "regression"

  if (!objective %in% lgb_supported_objectives) {
    cli::cli_abort(
      c(
        "Unsupported objective: {.val {objective}}.",
        "i" = "Supported objectives: {.val {lgb_supported_objectives}}."
      )
    )
  }

  if (objective %in% lgb_multiclass_objectives) {
    return(build_fit_formula_lgb_multiclass_from_parsed(parsedmodel, objective))
  }

  # Build nested trees from flat paths
  trees <- map(parsedmodel$trees, function(tree) {
    build_nested_from_flat_paths(tree, build_lgb_nested_condition)
  })

  # RF boosting averages trees instead of summing
  boosting <- parsedmodel$general$params$boosting
  if (!is.null(boosting) && boosting == "rf") {
    f <- reduce_addition(trees)
    f <- expr_division(f, n_trees)
  } else {
    f <- reduce_addition(trees)
  }

  apply_lgb_objective(f, objective, parsedmodel$general$params)
}

build_fit_formula_lgb_multiclass_from_parsed <- function(
  parsedmodel,
  objective
) {
  num_class <- parsedmodel$general$num_class
  if (is.null(num_class) || num_class < 2) {
    cli::cli_abort("Multiclass model must have num_class >= 2.")
  }

  trees <- map(parsedmodel$trees, function(tree) {
    build_nested_from_flat_paths(tree, build_lgb_nested_condition)
  })
  apply_lgb_multiclass_transformation(trees, num_class, objective)
}

# Build condition for lightgbm nested generation from path element
build_lgb_nested_condition <- function(path_elem) {
  col <- rlang::sym(path_elem$col)
  missing <- path_elem$missing %||% FALSE

  if (path_elem$type == "conditional") {
    val <- as.numeric(path_elem$val)
    # For nested generation, we only build the left condition (less-equal)
    if (missing) {
      expr(!!col <= !!val | is.na(!!col))
    } else {
      expr(!!col <= !!val)
    }
  } else if (path_elem$type == "set") {
    vals <- unlist(path_elem$vals)
    # For nested generation, we only build the left condition (in)
    if (missing) {
      expr(!!col %in% !!vals | is.na(!!col))
    } else {
      expr(!!col %in% !!vals)
    }
  } else {
    cli::cli_abort("Unknown path element type: {.val {path_elem$type}}")
  }
}

# Extract trees in nested format
# feature_names and linear_info can be passed to avoid redundant JSON/string parsing
extract_lgb_trees_nested <- function(
  model,
  feature_names = NULL,
  linear_info = NULL
) {
  trees_df <- lightgbm::lgb.model.dt.tree(model)
  trees_df <- as.data.frame(trees_df)

  # Extract linear tree info (only if not provided)
  if (is.null(feature_names)) {
    model_json <- jsonlite::fromJSON(model$dump_model())
    feature_names <- model_json$feature_names
  }
  if (is.null(linear_info)) {
    linear_info <- parse_lgb_linear_trees(model, feature_names)
  }

  trees_split <- split(trees_df, trees_df$tree_index)

  map(names(trees_split), function(tree_idx) {
    tree_df <- trees_split[[tree_idx]]
    tree_linear <- linear_info[[tree_idx]]
    build_nested_lgb_tree(tree_df, tree_linear)
  })
}

# Build nested case_when for a single lightgbm tree
build_nested_lgb_tree <- function(tree_df, linear_info = NULL) {
  # Build children map
  children_map <- get_lgb_children_map(tree_df)

  # Find root node (split_index == 0)
  root_row <- which(tree_df$split_index == 0)

  if (length(root_row) == 0) {
    leaf_row <- which(!is.na(tree_df$leaf_index))
    if (length(leaf_row) == 1) {
      leaf_idx <- tree_df$leaf_index[[leaf_row]]
      leaf_value <- tree_df$leaf_value[[leaf_row]]
      if (
        !is.null(linear_info) && as.character(leaf_idx) %in% names(linear_info)
      ) {
        return(build_lgb_linear_prediction(linear_info[[as.character(
          leaf_idx
        )]]))
      }
      return(leaf_value)
    }
    cli::cli_abort("Unable to find root or leaf node.", .internal = TRUE)
  }

  build_nested_lgb_node(root_row, tree_df, children_map, linear_info)
}

# Recursively build nested case_when node
build_nested_lgb_node <- function(
  node_row,
  tree_df,
  children_map,
  linear_info
) {
  # Check if this is a leaf
  if (!is.na(tree_df$leaf_index[[node_row]])) {
    leaf_idx <- tree_df$leaf_index[[node_row]]
    leaf_value <- tree_df$leaf_value[[node_row]]

    # Check for linear leaf
    if (
      !is.null(linear_info) && as.character(leaf_idx) %in% names(linear_info)
    ) {
      leaf_linear <- linear_info[[as.character(leaf_idx)]]
      leaf_linear$fallback <- leaf_value
      return(build_lgb_linear_prediction(leaf_linear))
    }

    return(leaf_value)
  }

  # Internal node - get split info
  split_index <- tree_df$split_index[[node_row]]
  decision_type <- tree_df$decision_type[[node_row]]
  col <- tree_df$split_feature[[node_row]]
  threshold <- tree_df$threshold[[node_row]]
  default_left <- tree_df$default_left[[node_row]] == "TRUE"

  # Get children (first is LEFT, second is RIGHT)
  children <- children_map[[as.character(split_index)]]
  left_row <- children[1]
  right_row <- children[2]

  # Recurse
  left_subtree <- build_nested_lgb_node(
    left_row,
    tree_df,
    children_map,
    linear_info
  )
  right_subtree <- build_nested_lgb_node(
    right_row,
    tree_df,
    children_map,
    linear_info
  )

  # Build condition
  col_sym <- rlang::sym(col)

  if (decision_type == "<=") {
    # Numerical split: LEFT = (<= threshold), RIGHT = (> threshold)
    if (default_left) {
      # Missing goes left
      condition <- expr(!!col_sym <= !!as.numeric(threshold) | is.na(!!col_sym))
    } else {
      # Missing goes right - condition is just <=
      condition <- expr(!!col_sym <= !!as.numeric(threshold))
    }
  } else if (decision_type == "==") {
    # Categorical split: LEFT = (in set), RIGHT = (not in set)
    category_set <- parse_lgb_categorical_threshold(threshold)
    if (default_left) {
      condition <- expr(!!col_sym %in% !!category_set | is.na(!!col_sym))
    } else {
      condition <- expr(!!col_sym %in% !!category_set)
    }
  } else {
    # nocov start
    cli::cli_abort(
      "Unsupported decision type: {.val {decision_type}}.",
      .internal = TRUE
    )
    # nocov end
  }

  expr(case_when(!!condition ~ !!left_subtree, .default = !!right_subtree))
}

# For {orbital} -----------------------------------------------

#' Extract processed LightGBM trees
#'
#' For use in orbital package.
#' @param model A LightGBM model object
#' @keywords internal
#' @export
.extract_lgb_trees <- function(model) {
  if (!inherits(model, "lgb.Booster")) {
    cli::cli_abort(
      "{.arg model} must be {.cls lgb.Booster}, not {.obj_type_friendly {model}}."
    )
  }

  extract_lgb_trees_nested(model)
}

# Helper to create test model
make_lgb_model <- function() {
  set.seed(123)
  X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
  y <- mtcars$hp
  dtrain <- lightgbm::lgb.Dataset(
    params = list(num_threads = 1L),
    X,
    label = y,
    colnames = c("mpg", "cyl", "disp")
  )
  lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      num_leaves = 4L,
      learning_rate = 1.0,
      objective = "regression",
      min_data_in_leaf = 1L
    ),
    data = dtrain,
    nrounds = 10L,
    verbose = -1L
  )
}

test_that("parse_model returns correct structure", {
  skip_if_not_installed("lightgbm")
  model <- make_lgb_model()
  pm <- parse_model(model)

  expect_s3_class(pm, "parsed_model")
  expect_s3_class(pm, "pm_lgb")

  expect_equal(pm$general$model, "lgb.Booster")
  expect_equal(pm$general$type, "lgb")
  expect_equal(pm$general$version, 3)

  expect_gt(length(pm$trees), 0)
})

test_that("correct number of trees extracted", {
  skip_if_not_installed("lightgbm")
  model <- make_lgb_model()
  pm <- parse_model(model)

  expect_equal(length(pm$trees), 10)
})

test_that("each tree has leaves with predictions and paths", {
  skip_if_not_installed("lightgbm")
  model <- make_lgb_model()
  pm <- parse_model(model)

  tree1 <- pm$trees[[1]]
  expect_gt(length(tree1), 0)

  has_required_names <- vapply(
    tree1,
    \(x) all(c("prediction", "path") %in% names(x)),
    logical(1)
  )
  expect_all_equal(has_required_names, TRUE)

  predictions <- vapply(tree1, \(x) x$prediction, double(1))
  expect_type(predictions, "double")

  path_is_list <- vapply(tree1, \(x) is.list(x$path), logical(1))
  expect_all_equal(path_is_list, TRUE)
})

test_that("path conditions have correct structure", {
  skip_if_not_installed("lightgbm")
  model <- make_lgb_model()
  pm <- parse_model(model)

  tree1 <- pm$trees[[1]]
  leaves_with_paths <- which(sapply(tree1, function(x) length(x$path) > 0))

  if (length(leaves_with_paths) > 0) {
    leaf_with_path <- tree1[[leaves_with_paths[1]]]

    cond <- leaf_with_path$path[[1]]
    expect_equal(cond$type, "conditional")
    expect_contains(names(cond), c("col", "val", "op", "missing"))
    expect_contains(c("less-equal", "more"), cond$op)
  }
})

test_that("feature names are extracted", {
  skip_if_not_installed("lightgbm")
  model <- make_lgb_model()
  pm <- parse_model(model)

  expect_equal(pm$general$feature_names, c("mpg", "cyl", "disp"))
})

test_that("params are extracted", {
  skip_if_not_installed("lightgbm")
  model <- make_lgb_model()
  pm <- parse_model(model)

  expect_contains(names(pm$general), "params")
  expect_equal(pm$general$params$objective, "regression")
})

test_that("niter and nfeatures are extracted", {
  skip_if_not_installed("lightgbm")
  model <- make_lgb_model()
  pm <- parse_model(model)

  expect_equal(pm$general$niter, 10)
  expect_equal(pm$general$nfeatures, 3)
})

test_that("model without explicit colnames still works", {
  skip_if_not_installed("lightgbm")

  set.seed(789)
  X <- data.matrix(mtcars[, c("mpg", "cyl")])
  y <- mtcars$hp

  # Create dataset WITHOUT specifying colnames
  dtrain <- lightgbm::lgb.Dataset(params = list(num_threads = 1L), X, label = y)

  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      num_leaves = 4L,
      learning_rate = 1.0,
      objective = "regression",
      min_data_in_leaf = 1L
    ),
    data = dtrain,
    nrounds = 3L,
    verbose = -1L
  )

  pm <- parse_model(model)

  expect_s3_class(pm, "pm_lgb")
  expect_length(pm$trees, 3)
  # Feature names should be auto-generated (Column_0, Column_1, etc.)
  expect_equal(pm$general$nfeatures, 2)
  expect_type(pm$general$feature_names, "character")
})

test_that("children map correctly identifies left and right children", {
  # Hand-crafted tree: simple binary tree

  #        split_0
  #       /       \
  #    leaf_0   leaf_1
  #
  # Pre-order: split_0 (row 1), leaf_0 (row 2), leaf_1 (row 3)
  tree_df <- data.frame(
    tree_index = c(0L, 0L, 0L),
    split_index = c(0L, NA, NA),
    split_feature = c("x1", NA, NA),
    node_parent = c(NA, NA, NA),
    leaf_index = c(NA, 0L, 1L),
    leaf_parent = c(NA, 0L, 0L),
    threshold = c(5.0, NA, NA),
    decision_type = c("<=", NA, NA),
    default_left = c("TRUE", NA, NA),
    leaf_value = c(NA, 10.0, 20.0),
    stringsAsFactors = FALSE
  )

  children_map <- tidypredict:::get_lgb_children_map(tree_df)

  expect_equal(children_map[["0"]], c(2L, 3L))
})

test_that("left child gets less-equal operator", {
  # Simple tree where we trace from leaf_0 (left child of root)
  tree_df <- data.frame(
    tree_index = c(0L, 0L, 0L),
    split_index = c(0L, NA, NA),
    split_feature = c("feature_a", NA, NA),
    node_parent = c(NA, NA, NA),
    leaf_index = c(NA, 0L, 1L),
    leaf_parent = c(NA, 0L, 0L),
    threshold = c(10.0, NA, NA),
    decision_type = c("<=", NA, NA),
    default_left = c("TRUE", NA, NA),
    leaf_value = c(NA, 100.0, 200.0),
    stringsAsFactors = FALSE
  )

  tree_result <- tidypredict:::get_lgb_tree(tree_df)

  leaf_0 <- tree_result[[1]]
  expect_equal(leaf_0$prediction, 100.0)
  expect_length(leaf_0$path, 1)
  expect_equal(leaf_0$path[[1]]$op, "less-equal")
  expect_equal(leaf_0$path[[1]]$col, "feature_a")
  expect_equal(leaf_0$path[[1]]$val, 10.0)
})

test_that("right child gets more operator", {
  # Simple tree where we trace from leaf_1 (right child of root)
  tree_df <- data.frame(
    tree_index = c(0L, 0L, 0L),
    split_index = c(0L, NA, NA),
    split_feature = c("feature_a", NA, NA),
    node_parent = c(NA, NA, NA),
    leaf_index = c(NA, 0L, 1L),
    leaf_parent = c(NA, 0L, 0L),
    threshold = c(10.0, NA, NA),
    decision_type = c("<=", NA, NA),
    default_left = c("TRUE", NA, NA),
    leaf_value = c(NA, 100.0, 200.0),
    stringsAsFactors = FALSE
  )

  tree_result <- tidypredict:::get_lgb_tree(tree_df)

  leaf_1 <- tree_result[[2]]
  expect_equal(leaf_1$prediction, 200.0)
  expect_length(leaf_1$path, 1)
  expect_equal(leaf_1$path[[1]]$op, "more")
  expect_equal(leaf_1$path[[1]]$col, "feature_a")
  expect_equal(leaf_1$path[[1]]$val, 10.0)
})

test_that("default_left TRUE assigns missing to left child path", {
  tree_df <- data.frame(
    tree_index = c(0L, 0L, 0L),
    split_index = c(0L, NA, NA),
    split_feature = c("x", NA, NA),
    node_parent = c(NA, NA, NA),
    leaf_index = c(NA, 0L, 1L),
    leaf_parent = c(NA, 0L, 0L),
    threshold = c(5.0, NA, NA),
    decision_type = c("<=", NA, NA),
    default_left = c("TRUE", NA, NA),
    leaf_value = c(NA, 1.0, 2.0),
    stringsAsFactors = FALSE
  )

  tree_result <- tidypredict:::get_lgb_tree(tree_df)

  # Left child (leaf_0): missing = TRUE (default_left is TRUE)
  expect_equal(tree_result[[1]]$path[[1]]$missing, TRUE)

  # Right child (leaf_1): missing = FALSE (default_left is TRUE, but we went right)
  expect_equal(tree_result[[2]]$path[[1]]$missing, FALSE)
})

test_that("default_left FALSE assigns missing to right child path", {
  tree_df <- data.frame(
    tree_index = c(0L, 0L, 0L),
    split_index = c(0L, NA, NA),
    split_feature = c("x", NA, NA),
    node_parent = c(NA, NA, NA),
    leaf_index = c(NA, 0L, 1L),
    leaf_parent = c(NA, 0L, 0L),
    threshold = c(5.0, NA, NA),
    decision_type = c("<=", NA, NA),
    default_left = c("FALSE", NA, NA),
    leaf_value = c(NA, 1.0, 2.0),
    stringsAsFactors = FALSE
  )

  tree_result <- tidypredict:::get_lgb_tree(tree_df)

  # Left child (leaf_0): missing = FALSE (default_left is FALSE)
  expect_equal(tree_result[[1]]$path[[1]]$missing, FALSE)

  # Right child (leaf_1): missing = TRUE (default_left is FALSE, we went right)
  expect_equal(tree_result[[2]]$path[[1]]$missing, TRUE)
})

test_that("deeper tree paths are traced correctly", {
  # Tree structure:
  #           split_0 (x1 <= 10)
  #          /                 \
  #      leaf_0            split_1 (x2 <= 5)
  #                       /              \
  #                   leaf_1           leaf_2
  #
  # Pre-order traversal: split_0, leaf_0, split_1, leaf_1, leaf_2
  tree_df <- data.frame(
    tree_index = c(0L, 0L, 0L, 0L, 0L),
    split_index = c(0L, NA, 1L, NA, NA),
    split_feature = c("x1", NA, "x2", NA, NA),
    node_parent = c(NA, NA, 0L, NA, NA),
    leaf_index = c(NA, 0L, NA, 1L, 2L),
    leaf_parent = c(NA, 0L, NA, 1L, 1L),
    threshold = c(10.0, NA, 5.0, NA, NA),
    decision_type = c("<=", NA, "<=", NA, NA),
    default_left = c("TRUE", NA, "FALSE", NA, NA),
    leaf_value = c(NA, 100.0, NA, 200.0, 300.0),
    stringsAsFactors = FALSE
  )

  tree_result <- tidypredict:::get_lgb_tree(tree_df)

  # leaf_0: path is just [x1 <= 10] (left of root)
  expect_equal(tree_result[[1]]$prediction, 100.0)
  expect_length(tree_result[[1]]$path, 1)
  expect_equal(tree_result[[1]]$path[[1]]$col, "x1")
  expect_equal(tree_result[[1]]$path[[1]]$op, "less-equal")
  expect_equal(tree_result[[1]]$path[[1]]$missing, TRUE) # default_left TRUE at root

  # leaf_1: path is [x1 > 10, x2 <= 5] (right of root, left of split_1)
  expect_equal(tree_result[[2]]$prediction, 200.0)
  expect_length(tree_result[[2]]$path, 2)
  # First condition: x1 > 10 (right of root)
  expect_equal(tree_result[[2]]$path[[1]]$col, "x1")
  expect_equal(tree_result[[2]]$path[[1]]$op, "more")
  expect_equal(tree_result[[2]]$path[[1]]$missing, FALSE) # went right, default_left TRUE
  # Second condition: x2 <= 5 (left of split_1)
  expect_equal(tree_result[[2]]$path[[2]]$col, "x2")
  expect_equal(tree_result[[2]]$path[[2]]$op, "less-equal")
  expect_equal(tree_result[[2]]$path[[2]]$missing, FALSE) # default_left FALSE at split_1

  # leaf_2: path is [x1 > 10, x2 > 5] (right of root, right of split_1)
  expect_equal(tree_result[[3]]$prediction, 300.0)
  expect_length(tree_result[[3]]$path, 2)
  expect_equal(tree_result[[3]]$path[[1]]$op, "more")
  expect_equal(tree_result[[3]]$path[[2]]$op, "more")
  expect_equal(tree_result[[3]]$path[[2]]$missing, TRUE) # went right, default_left FALSE
})

test_that("single leaf tree (stump) has empty path", {
  # Edge case: tree with only a root leaf (no splits)
  tree_df <- data.frame(
    tree_index = 0L,
    split_index = NA_integer_,
    split_feature = NA_character_,
    node_parent = NA_integer_,
    leaf_index = 0L,
    leaf_parent = NA_integer_,
    threshold = NA_real_,
    decision_type = NA_character_,
    default_left = NA_character_,
    leaf_value = 42.0,
    stringsAsFactors = FALSE
  )

  tree_result <- tidypredict:::get_lgb_tree(tree_df)

  expect_length(tree_result, 1)
  expect_equal(tree_result[[1]]$prediction, 42.0)
  expect_length(tree_result[[1]]$path, 0) # No conditions for root leaf
})

test_that("mixed default_left values in same tree are handled correctly", {
  # Tree where different splits have different default_left values
  #           split_0 (x1 <= 10, default_left=TRUE)
  #          /                 \
  #      split_1              leaf_2
  #  (x2 <= 5, default_left=FALSE)
  #     /        \
  #  leaf_0    leaf_1
  #
  # Pre-order: split_0, split_1, leaf_0, leaf_1, leaf_2
  tree_df <- data.frame(
    tree_index = c(0L, 0L, 0L, 0L, 0L),
    split_index = c(0L, 1L, NA, NA, NA),
    split_feature = c("x1", "x2", NA, NA, NA),
    node_parent = c(NA, 0L, NA, NA, NA),
    leaf_index = c(NA, NA, 0L, 1L, 2L),
    leaf_parent = c(NA, NA, 1L, 1L, 0L),
    threshold = c(10.0, 5.0, NA, NA, NA),
    decision_type = c("<=", "<=", NA, NA, NA),
    default_left = c("TRUE", "FALSE", NA, NA, NA),
    leaf_value = c(NA, NA, 100.0, 200.0, 300.0),
    stringsAsFactors = FALSE
  )

  tree_result <- tidypredict:::get_lgb_tree(tree_df)

  # leaf_0: path [x1 <= 10, x2 <= 5]
  # x1: left child, default_left=TRUE -> missing=TRUE
  # x2: left child, default_left=FALSE -> missing=FALSE
  expect_equal(tree_result[[1]]$path[[1]]$missing, TRUE)
  expect_equal(tree_result[[1]]$path[[2]]$missing, FALSE)

  # leaf_1: path [x1 <= 10, x2 > 5]
  # x1: left child, default_left=TRUE -> missing=TRUE
  # x2: right child, default_left=FALSE -> missing=TRUE
  expect_equal(tree_result[[2]]$path[[1]]$missing, TRUE)
  expect_equal(tree_result[[2]]$path[[2]]$missing, TRUE)

  # leaf_2: path [x1 > 10]
  # x1: right child, default_left=TRUE -> missing=FALSE
  expect_equal(tree_result[[3]]$path[[1]]$missing, FALSE)
})

test_that("model with missing values produces valid parse", {
  skip_if_not_installed("lightgbm")

  # Create data with missing values
  set.seed(456)
  X <- data.matrix(mtcars[, c("mpg", "cyl")])
  y <- mtcars$hp
  X_with_na <- X
  X_with_na[1:5, 1] <- NA
  X_with_na[10:15, 2] <- NA

  dtrain <- lightgbm::lgb.Dataset(
    params = list(num_threads = 1L),
    X_with_na,
    label = y,
    colnames = c("mpg", "cyl")
  )

  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      num_leaves = 4L,
      learning_rate = 1.0,
      objective = "regression",
      min_data_in_leaf = 1L,
      use_missing = TRUE
    ),
    data = dtrain,
    nrounds = 3L,
    verbose = -1L
  )

  pm <- parse_model(model)

  expect_s3_class(pm, "pm_lgb")
  expect_length(pm$trees, 3)

  # Verify all paths have valid missing flags (TRUE or FALSE, not NA)
  all_missing_flags <- unlist(lapply(pm$trees, function(tree) {
    lapply(tree, function(leaf) {
      vapply(leaf$path, function(cond) cond$missing, logical(1))
    })
  }))

  expect_type(all_missing_flags, "logical")
  expect_false(anyNA(all_missing_flags))
})

# Fit formula tests -----------------------------------------------------------

test_that("tidypredict_fit returns language object", {
  skip_if_not_installed("lightgbm")
  model <- make_lgb_model()

  fit_formula <- tidypredict_fit(model)

  expect_type(fit_formula, "language")
})

test_that("tidypredict_fit works on parsed model", {
  skip_if_not_installed("lightgbm")
  model <- make_lgb_model()
  pm <- parse_model(model)

  fit_formula <- tidypredict_fit(pm)

  expect_type(fit_formula, "language")
})

test_that("produced case_when uses .default", {
  skip_if_not_installed("lightgbm")
  model <- make_lgb_model()

  fit <- tidypredict_fit(model)
  fit_text <- rlang::expr_text(fit)

  expect_match(fit_text, "\\.default")
})

test_that("regression predictions match native predict", {
  skip_if_not_installed("lightgbm")

  set.seed(123)
  X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
  y <- mtcars$hp
  dtrain <- lightgbm::lgb.Dataset(
    params = list(num_threads = 1L),
    X,
    label = y,
    colnames = c("mpg", "cyl", "disp")
  )

  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      num_leaves = 4L,
      learning_rate = 0.3,
      objective = "regression",
      min_data_in_leaf = 1L
    ),
    data = dtrain,
    nrounds = 5L,
    verbose = -1L
  )

  fit_formula <- tidypredict_fit(model)
  native_preds <- predict(model, params = list(num_threads = 1L), X)
  tidy_preds <- dplyr::mutate(mtcars, pred = !!fit_formula)$pred

  expect_equal(unname(tidy_preds), unname(native_preds), tolerance = 1e-10)
})

test_that("binary classification predictions match native predict", {
  skip_if_not_installed("lightgbm")

  set.seed(123)
  X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
  y <- mtcars$am
  dtrain <- lightgbm::lgb.Dataset(
    params = list(num_threads = 1L),
    X,
    label = y,
    colnames = c("mpg", "cyl", "disp")
  )

  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      num_leaves = 4L,
      learning_rate = 0.3,
      objective = "binary",
      min_data_in_leaf = 1L
    ),
    data = dtrain,
    nrounds = 5L,
    verbose = -1L
  )

  fit_formula <- tidypredict_fit(model)
  native_preds <- predict(model, params = list(num_threads = 1L), X)
  tidy_preds <- dplyr::mutate(mtcars, pred = !!fit_formula)$pred

  expect_equal(unname(tidy_preds), unname(native_preds), tolerance = 1e-10)
})

test_that("poisson predictions match native predict", {
  skip_if_not_installed("lightgbm")

  set.seed(123)
  X <- data.matrix(mtcars[, c("mpg", "disp")])
  y <- mtcars$carb
  dtrain <- lightgbm::lgb.Dataset(
    params = list(num_threads = 1L),
    X,
    label = y,
    colnames = c("mpg", "disp")
  )

  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      num_leaves = 4L,
      learning_rate = 0.3,
      objective = "poisson",
      min_data_in_leaf = 1L
    ),
    data = dtrain,
    nrounds = 5L,
    verbose = -1L
  )

  fit_formula <- tidypredict_fit(model)
  native_preds <- predict(model, params = list(num_threads = 1L), X)
  tidy_preds <- dplyr::mutate(mtcars, pred = !!fit_formula)$pred

  expect_equal(unname(tidy_preds), unname(native_preds), tolerance = 1e-10)
})

test_that("regression_l1 predictions match native predict", {
  skip_if_not_installed("lightgbm")

  set.seed(123)
  X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
  y <- mtcars$hp
  dtrain <- lightgbm::lgb.Dataset(
    params = list(num_threads = 1L),
    X,
    label = y,
    colnames = c("mpg", "cyl", "disp")
  )

  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      num_leaves = 4L,
      learning_rate = 0.3,
      objective = "regression_l1",
      min_data_in_leaf = 1L
    ),
    data = dtrain,
    nrounds = 5L,
    verbose = -1L
  )

  fit_formula <- tidypredict_fit(model)
  native_preds <- predict(model, params = list(num_threads = 1L), X)
  tidy_preds <- dplyr::mutate(mtcars, pred = !!fit_formula)$pred

  expect_equal(unname(tidy_preds), unname(native_preds), tolerance = 1e-10)
})

test_that("regression_l2 predictions match native predict", {
  skip_if_not_installed("lightgbm")

  set.seed(123)
  X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
  y <- mtcars$hp
  dtrain <- lightgbm::lgb.Dataset(
    params = list(num_threads = 1L),
    X,
    label = y,
    colnames = c("mpg", "cyl", "disp")
  )

  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      num_leaves = 4L,
      learning_rate = 0.3,
      objective = "regression_l2",
      min_data_in_leaf = 1L
    ),
    data = dtrain,
    nrounds = 5L,
    verbose = -1L
  )

  fit_formula <- tidypredict_fit(model)
  native_preds <- predict(model, params = list(num_threads = 1L), X)
  tidy_preds <- dplyr::mutate(mtcars, pred = !!fit_formula)$pred

  expect_equal(unname(tidy_preds), unname(native_preds), tolerance = 1e-10)
})

test_that("mape predictions match native predict", {
  skip_if_not_installed("lightgbm")

  set.seed(123)
  X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
  y <- mtcars$hp
  dtrain <- lightgbm::lgb.Dataset(
    params = list(num_threads = 1L),
    X,
    label = y,
    colnames = c("mpg", "cyl", "disp")
  )

  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      num_leaves = 4L,
      learning_rate = 0.3,
      objective = "mape",
      min_data_in_leaf = 1L
    ),
    data = dtrain,
    nrounds = 5L,
    verbose = -1L
  )

  fit_formula <- tidypredict_fit(model)
  native_preds <- predict(model, params = list(num_threads = 1L), X)
  tidy_preds <- dplyr::mutate(mtcars, pred = !!fit_formula)$pred

  expect_equal(unname(tidy_preds), unname(native_preds), tolerance = 1e-10)
})

test_that("huber predictions match native predict", {
  skip_if_not_installed("lightgbm")

  set.seed(123)
  X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
  y <- mtcars$hp
  dtrain <- lightgbm::lgb.Dataset(
    params = list(num_threads = 1L),
    X,
    label = y,
    colnames = c("mpg", "cyl", "disp")
  )

  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      num_leaves = 4L,
      learning_rate = 0.3,
      objective = "huber",
      min_data_in_leaf = 1L
    ),
    data = dtrain,
    nrounds = 5L,
    verbose = -1L
  )

  fit_formula <- tidypredict_fit(model)
  native_preds <- predict(model, params = list(num_threads = 1L), X)
  tidy_preds <- dplyr::mutate(mtcars, pred = !!fit_formula)$pred

  expect_equal(unname(tidy_preds), unname(native_preds), tolerance = 1e-10)
})

test_that("fair predictions match native predict", {
  skip_if_not_installed("lightgbm")

  set.seed(123)
  X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
  y <- mtcars$hp
  dtrain <- lightgbm::lgb.Dataset(
    params = list(num_threads = 1L),
    X,
    label = y,
    colnames = c("mpg", "cyl", "disp")
  )

  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      num_leaves = 4L,
      learning_rate = 0.3,
      objective = "fair",
      min_data_in_leaf = 1L
    ),
    data = dtrain,
    nrounds = 5L,
    verbose = -1L
  )

  fit_formula <- tidypredict_fit(model)
  native_preds <- predict(model, params = list(num_threads = 1L), X)
  tidy_preds <- dplyr::mutate(mtcars, pred = !!fit_formula)$pred

  expect_equal(unname(tidy_preds), unname(native_preds), tolerance = 1e-10)
})

test_that("quantile predictions match native predict", {
  skip_if_not_installed("lightgbm")

  set.seed(123)
  X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
  y <- mtcars$hp
  dtrain <- lightgbm::lgb.Dataset(
    params = list(num_threads = 1L),
    X,
    label = y,
    colnames = c("mpg", "cyl", "disp")
  )

  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      num_leaves = 4L,
      learning_rate = 0.3,
      objective = "quantile",
      alpha = 0.5,
      min_data_in_leaf = 1L
    ),
    data = dtrain,
    nrounds = 5L,
    verbose = -1L
  )

  fit_formula <- tidypredict_fit(model)
  native_preds <- predict(model, params = list(num_threads = 1L), X)
  tidy_preds <- dplyr::mutate(mtcars, pred = !!fit_formula)$pred

  expect_equal(unname(tidy_preds), unname(native_preds), tolerance = 1e-10)
})

test_that("gamma predictions match native predict", {
  skip_if_not_installed("lightgbm")

  set.seed(123)
  X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
  y <- mtcars$hp
  dtrain <- lightgbm::lgb.Dataset(
    params = list(num_threads = 1L),
    X,
    label = y,
    colnames = c("mpg", "cyl", "disp")
  )

  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      num_leaves = 4L,
      learning_rate = 0.3,
      objective = "gamma",
      min_data_in_leaf = 1L
    ),
    data = dtrain,
    nrounds = 5L,
    verbose = -1L
  )

  fit_formula <- tidypredict_fit(model)
  native_preds <- predict(model, params = list(num_threads = 1L), X)
  tidy_preds <- dplyr::mutate(mtcars, pred = !!fit_formula)$pred

  expect_equal(unname(tidy_preds), unname(native_preds), tolerance = 1e-10)
})

test_that("tweedie predictions match native predict", {
  skip_if_not_installed("lightgbm")

  set.seed(123)
  X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
  y <- mtcars$hp
  dtrain <- lightgbm::lgb.Dataset(
    params = list(num_threads = 1L),
    X,
    label = y,
    colnames = c("mpg", "cyl", "disp")
  )

  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      num_leaves = 4L,
      learning_rate = 0.3,
      objective = "tweedie",
      tweedie_variance_power = 1.5,
      min_data_in_leaf = 1L
    ),
    data = dtrain,
    nrounds = 5L,
    verbose = -1L
  )

  fit_formula <- tidypredict_fit(model)
  native_preds <- predict(model, params = list(num_threads = 1L), X)
  tidy_preds <- dplyr::mutate(mtcars, pred = !!fit_formula)$pred

  expect_equal(unname(tidy_preds), unname(native_preds), tolerance = 1e-10)
})

test_that("cross_entropy predictions match native predict", {
  skip_if_not_installed("lightgbm")

  set.seed(123)
  X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
  y <- mtcars$am
  dtrain <- lightgbm::lgb.Dataset(
    params = list(num_threads = 1L),
    X,
    label = y,
    colnames = c("mpg", "cyl", "disp")
  )

  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      num_leaves = 4L,
      learning_rate = 0.3,
      objective = "cross_entropy",
      min_data_in_leaf = 1L
    ),
    data = dtrain,
    nrounds = 5L,
    verbose = -1L
  )

  fit_formula <- tidypredict_fit(model)
  native_preds <- predict(model, params = list(num_threads = 1L), X)
  tidy_preds <- dplyr::mutate(mtcars, pred = !!fit_formula)$pred

  expect_equal(unname(tidy_preds), unname(native_preds), tolerance = 1e-10)
})

test_that("RF boosting predictions match native predict (#185)", {
  skip_if_not_installed("lightgbm")

  set.seed(123)
  X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
  y <- mtcars$hp
  dtrain <- lightgbm::lgb.Dataset(
    params = list(num_threads = 1L),
    X,
    label = y,
    colnames = c("mpg", "cyl", "disp")
  )

  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      boosting = "rf",
      num_leaves = 4L,
      objective = "regression",
      min_data_in_leaf = 1L,
      bagging_freq = 1,
      bagging_fraction = 0.8
    ),
    data = dtrain,
    nrounds = 5L,
    verbose = -1L
  )

  fit_formula <- tidypredict_fit(model)
  native_preds <- predict(model, params = list(num_threads = 1L), X)
  tidy_preds <- dplyr::mutate(mtcars, pred = !!fit_formula)$pred

  expect_equal(unname(tidy_preds), unname(native_preds), tolerance = 1e-10)
})

test_that("predictions with missing values match", {
  skip_if_not_installed("lightgbm")

  # Training data with NAs
  set.seed(456)
  X <- data.matrix(mtcars[, c("mpg", "cyl")])
  y <- mtcars$hp
  X_train <- X
  X_train[1:3, 1] <- NA
  dtrain <- lightgbm::lgb.Dataset(
    params = list(num_threads = 1L),
    X_train,
    label = y,
    colnames = c("mpg", "cyl")
  )

  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      num_leaves = 4L,
      learning_rate = 0.5,
      objective = "regression",
      min_data_in_leaf = 1L,
      use_missing = TRUE
    ),
    data = dtrain,
    nrounds = 3L,
    verbose = -1L
  )

  X_pred <- X
  X_pred[5:7, 1] <- NA
  X_pred[10:12, 2] <- NA

  fit_formula <- tidypredict_fit(model)
  native_preds <- predict(model, params = list(num_threads = 1L), X_pred)

  pred_df <- as.data.frame(X_pred)
  tidy_preds <- dplyr::mutate(pred_df, pred = !!fit_formula)$pred

  expect_equal(unname(tidy_preds), unname(native_preds), tolerance = 1e-10)
})

test_that("unsupported objective throws error", {
  skip_if_not_installed("lightgbm")
  model <- make_lgb_model()
  pm <- parse_model(model)

  pm$general$params$objective <- "unsupported_objective"

  expect_snapshot(tidypredict_fit(pm), error = TRUE)
})

# SQL generation tests ------------------------------------------------------

test_that("tidypredict_sql returns SQL class", {
  skip_if_not_installed("lightgbm")
  skip_if_not_installed("dbplyr")
  model <- make_lgb_model()

  sql_result <- tidypredict_sql(model, dbplyr::simulate_dbi())

  expect_s3_class(sql_result, "sql")
})

test_that("tidypredict_sql works with parsed model", {
  skip_if_not_installed("lightgbm")
  skip_if_not_installed("dbplyr")
  model <- make_lgb_model()
  pm <- parse_model(model)

  sql_result <- tidypredict_sql(pm, dbplyr::simulate_dbi())

  expect_s3_class(sql_result, "sql")
})

test_that("SQL predictions match native predictions with SQLite", {
  skip_if_not_installed("lightgbm")
  skip_if_not_installed("DBI")
  skip_if_not_installed("RSQLite")
  skip_if_not_installed("dbplyr")

  model <- make_lgb_model()

  con <- DBI::dbConnect(RSQLite::SQLite(), ":memory:")
  on.exit(DBI::dbDisconnect(con), add = TRUE)

  test_data <- mtcars[, c("mpg", "cyl", "disp")]
  DBI::dbWriteTable(con, "test_data", test_data)

  sql_query <- tidypredict_sql(model, con)
  db_result <- DBI::dbGetQuery(
    con,
    paste0("SELECT ", sql_query, " AS pred FROM test_data")
  )

  X <- data.matrix(test_data)
  native_preds <- predict(model, params = list(num_threads = 1L), X)

  expect_equal(db_result$pred, unname(native_preds), tolerance = 1e-10)
})

test_that("SQL predictions match for binary classification with SQLite", {
  skip_if_not_installed("lightgbm")
  skip_if_not_installed("DBI")
  skip_if_not_installed("RSQLite")
  skip_if_not_installed("dbplyr")

  set.seed(456)
  X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
  y <- as.integer(mtcars$am)
  dtrain <- lightgbm::lgb.Dataset(params = list(num_threads = 1L), X, label = y)
  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      num_leaves = 4L,
      learning_rate = 0.5,
      objective = "binary",
      min_data_in_leaf = 1L
    ),
    data = dtrain,
    nrounds = 5L,
    verbose = -1L
  )

  con <- DBI::dbConnect(RSQLite::SQLite(), ":memory:")
  on.exit(DBI::dbDisconnect(con), add = TRUE)

  test_data <- mtcars[, c("mpg", "cyl", "disp")]
  DBI::dbWriteTable(con, "test_data", test_data)

  sql_query <- tidypredict_sql(model, con)
  db_result <- DBI::dbGetQuery(
    con,
    paste0("SELECT ", sql_query, " AS pred FROM test_data")
  )

  native_preds <- predict(model, params = list(num_threads = 1L), X)

  expect_equal(db_result$pred, unname(native_preds), tolerance = 1e-10)
})

# Multiclass tests ----------------------------------------------------------

test_that("parse_model extracts num_class for multiclass", {
  skip_if_not_installed("lightgbm")

  set.seed(123)
  X <- data.matrix(iris[, 1:4])
  colnames(X) <- c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width")
  y <- as.integer(iris$Species) - 1L

  dtrain <- lightgbm::lgb.Dataset(
    params = list(num_threads = 1L),
    X,
    label = y,
    colnames = colnames(X)
  )
  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      num_leaves = 4L,
      learning_rate = 0.5,
      objective = "multiclass",
      num_class = 3L,
      min_data_in_leaf = 1L
    ),
    data = dtrain,
    nrounds = 2L,
    verbose = -1L
  )

  pm <- parse_model(model)

  expect_equal(pm$general$num_class, 3)
  expect_equal(pm$general$num_tree_per_iteration, 3)
  # 2 rounds * 3 classes = 6 trees
  expect_equal(length(pm$trees), 6)
})

test_that("tidypredict_fit returns list for multiclass", {
  skip_if_not_installed("lightgbm")

  set.seed(123)
  X <- data.matrix(iris[, 1:4])
  colnames(X) <- c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width")
  y <- as.integer(iris$Species) - 1L

  dtrain <- lightgbm::lgb.Dataset(
    params = list(num_threads = 1L),
    X,
    label = y,
    colnames = colnames(X)
  )
  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      num_leaves = 4L,
      learning_rate = 0.5,
      objective = "multiclass",
      num_class = 3L,
      min_data_in_leaf = 1L
    ),
    data = dtrain,
    nrounds = 3L,
    verbose = -1L
  )

  fit_formulas <- tidypredict_fit(model)

  expect_type(fit_formulas, "list")
  expect_length(fit_formulas, 3)
  expect_named(fit_formulas, c("class_0", "class_1", "class_2"))

  types <- vapply(fit_formulas, typeof, character(1))
  expect_all_equal(types, "language")
})

test_that("multiclass predictions match native predictions", {
  skip_if_not_installed("lightgbm")

  set.seed(789)
  X <- data.matrix(iris[, 1:4])
  colnames(X) <- c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width")
  y <- as.integer(iris$Species) - 1L

  dtrain <- lightgbm::lgb.Dataset(
    params = list(num_threads = 1L),
    X,
    label = y,
    colnames = colnames(X)
  )
  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      num_leaves = 8L,
      learning_rate = 0.3,
      objective = "multiclass",
      num_class = 3L,
      min_data_in_leaf = 1L
    ),
    data = dtrain,
    nrounds = 5L,
    verbose = -1L
  )

  fit_formulas <- tidypredict_fit(model)
  native_preds <- predict(model, params = list(num_threads = 1L), X)

  test_df <- as.data.frame(X)
  tidy_preds <- dplyr::mutate(
    test_df,
    class_0 = !!fit_formulas$class_0,
    class_1 = !!fit_formulas$class_1,
    class_2 = !!fit_formulas$class_2
  )
  tidy_mat <- as.matrix(tidy_preds[, c("class_0", "class_1", "class_2")])

  expect_equal(unname(tidy_mat), unname(native_preds), tolerance = 1e-10)
})

test_that("multiclass probabilities sum to 1", {
  skip_if_not_installed("lightgbm")

  set.seed(321)
  X <- data.matrix(iris[, 1:4])
  colnames(X) <- c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width")
  y <- as.integer(iris$Species) - 1L

  dtrain <- lightgbm::lgb.Dataset(
    params = list(num_threads = 1L),
    X,
    label = y,
    colnames = colnames(X)
  )
  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      num_leaves = 4L,
      learning_rate = 0.5,
      objective = "multiclass",
      num_class = 3L,
      min_data_in_leaf = 1L
    ),
    data = dtrain,
    nrounds = 3L,
    verbose = -1L
  )

  fit_formulas <- tidypredict_fit(model)

  test_df <- as.data.frame(X)
  tidy_preds <- dplyr::mutate(
    test_df,
    class_0 = !!fit_formulas$class_0,
    class_1 = !!fit_formulas$class_1,
    class_2 = !!fit_formulas$class_2
  )

  row_sums <- tidy_preds$class_0 + tidy_preds$class_1 + tidy_preds$class_2
  expect_equal(row_sums, rep(1, nrow(X)), tolerance = 1e-10)
})

test_that("multiclassova predictions match native predictions", {
  skip_if_not_installed("lightgbm")

  set.seed(654)
  X <- data.matrix(iris[, 1:4])
  colnames(X) <- c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width")
  y <- as.integer(iris$Species) - 1L

  dtrain <- lightgbm::lgb.Dataset(
    params = list(num_threads = 1L),
    X,
    label = y,
    colnames = colnames(X)
  )
  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      num_leaves = 8L,
      learning_rate = 0.3,
      objective = "multiclassova",
      num_class = 3L,
      min_data_in_leaf = 1L
    ),
    data = dtrain,
    nrounds = 5L,
    verbose = -1L
  )

  fit_formulas <- tidypredict_fit(model)
  native_preds <- predict(model, params = list(num_threads = 1L), X)

  expect_named(fit_formulas, c("class_0", "class_1", "class_2"))

  test_df <- as.data.frame(X)
  tidy_preds <- dplyr::mutate(
    test_df,
    class_0 = !!fit_formulas$class_0,
    class_1 = !!fit_formulas$class_1,
    class_2 = !!fit_formulas$class_2
  )
  tidy_mat <- as.matrix(tidy_preds[, c("class_0", "class_1", "class_2")])

  expect_equal(unname(tidy_mat), unname(native_preds), tolerance = 1e-10)
})

test_that("multiclass SQL generation returns list of SQL", {
  skip_if_not_installed("lightgbm")
  skip_if_not_installed("dbplyr")

  set.seed(123)
  X <- data.matrix(iris[, 1:4])
  colnames(X) <- c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width")
  y <- as.integer(iris$Species) - 1L

  dtrain <- lightgbm::lgb.Dataset(
    params = list(num_threads = 1L),
    X,
    label = y,
    colnames = colnames(X)
  )
  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      num_leaves = 4L,
      learning_rate = 0.5,
      objective = "multiclass",
      num_class = 3L,
      min_data_in_leaf = 1L
    ),
    data = dtrain,
    nrounds = 2L,
    verbose = -1L
  )

  sql_result <- tidypredict_sql(model, dbplyr::simulate_dbi())

  expect_type(sql_result, "list")
  expect_length(sql_result, 3)
  is_sql <- vapply(sql_result, \(x) inherits(x, "sql"), logical(1))
  expect_all_equal(is_sql, TRUE)
})

# Edge case tests -----------------------------------------------------------

test_that("empty trees throws error", {
  pm <- list()
  pm$general$model <- "lgb.Booster"
  pm$general$type <- "lgb"
  pm$general$version <- 3
  pm$general$params <- list(objective = "regression")
  pm$trees <- list()
  class(pm) <- c("pm_lgb", "parsed_model", "list")

  expect_snapshot(tidypredict_fit(pm), error = TRUE)
})

test_that("multiclass with num_class < 2 throws error", {
  pm <- list()
  pm$general$model <- "lgb.Booster"
  pm$general$type <- "lgb"
  pm$general$version <- 3
  pm$general$params <- list(objective = "multiclass")
  pm$general$num_class <- 1
  pm$trees <- list(list(list(prediction = 1, path = list())))
  class(pm) <- c("pm_lgb", "parsed_model", "list")

  expect_snapshot(tidypredict_fit(pm), error = TRUE)
})

test_that("multiclass with NULL num_class throws error", {
  pm <- list()
  pm$general$model <- "lgb.Booster"
  pm$general$type <- "lgb"
  pm$general$version <- 3
  pm$general$params <- list(objective = "multiclass")
  pm$general$num_class <- NULL
  pm$trees <- list(list(list(prediction = 1, path = list())))
  class(pm) <- c("pm_lgb", "parsed_model", "list")

  expect_snapshot(tidypredict_fit(pm), error = TRUE)
})

test_that("RF boosting in from_parsed averages trees", {
  pm <- list()
  pm$general$model <- "lgb.Booster"
  pm$general$type <- "lgb"
  pm$general$version <- 3
  pm$general$params <- list(objective = "regression", boosting = "rf")
  pm$trees <- list(
    list(
      list(
        prediction = 10,
        path = list(list(
          type = "conditional",
          col = "x",
          val = 5,
          op = "less-equal",
          missing = FALSE
        ))
      ),
      list(
        prediction = 20,
        path = list(list(
          type = "conditional",
          col = "x",
          val = 5,
          op = "more",
          missing = FALSE
        ))
      )
    ),
    list(
      list(
        prediction = 30,
        path = list(list(
          type = "conditional",
          col = "x",
          val = 5,
          op = "less-equal",
          missing = FALSE
        ))
      ),
      list(
        prediction = 40,
        path = list(list(
          type = "conditional",
          col = "x",
          val = 5,
          op = "more",
          missing = FALSE
        ))
      )
    )
  )
  class(pm) <- c("pm_lgb", "parsed_model", "list")

  fit_formula <- tidypredict_fit(pm)

  formula_str <- paste(deparse(fit_formula), collapse = "")
  expect_match(formula_str, "/2")
})

test_that("from_parsed handles set type with missing", {
  pm <- list()
  pm$general$model <- "lgb.Booster"
  pm$general$type <- "lgb"
  pm$general$version <- 3
  pm$general$params <- list(objective = "regression")
  pm$trees <- list(
    list(
      list(
        prediction = 10,
        path = list(list(
          type = "set",
          col = "cat_feat",
          vals = c(0L, 1L),
          op = "in",
          missing = TRUE
        ))
      ),
      list(
        prediction = 20,
        path = list(list(
          type = "set",
          col = "cat_feat",
          vals = c(0L, 1L),
          op = "not-in",
          missing = FALSE
        ))
      )
    )
  )
  class(pm) <- c("pm_lgb", "parsed_model", "list")

  fit_formula <- tidypredict_fit(pm)

  formula_str <- deparse(fit_formula)
  expect_match(formula_str, "%in%")
  expect_match(formula_str, "is.na")
})

test_that("from_parsed handles set type without missing", {
  pm <- list()
  pm$general$model <- "lgb.Booster"
  pm$general$type <- "lgb"
  pm$general$version <- 3
  pm$general$params <- list(objective = "regression")
  pm$trees <- list(
    list(
      list(
        prediction = 10,
        path = list(list(
          type = "set",
          col = "cat_feat",
          vals = c(0L, 1L),
          op = "in",
          missing = FALSE
        ))
      ),
      list(
        prediction = 20,
        path = list(list(
          type = "set",
          col = "cat_feat",
          vals = c(0L, 1L),
          op = "not-in",
          missing = FALSE
        ))
      )
    )
  )
  class(pm) <- c("pm_lgb", "parsed_model", "list")

  fit_formula <- tidypredict_fit(pm)

  formula_str <- deparse(fit_formula)
  expect_match(formula_str, "%in%")
  expect_no_match(formula_str, "is.na")
})

test_that("from_parsed handles conditional without missing", {
  pm <- list()
  pm$general$model <- "lgb.Booster"
  pm$general$type <- "lgb"
  pm$general$version <- 3
  pm$general$params <- list(objective = "regression")
  pm$trees <- list(
    list(
      list(
        prediction = 10,
        path = list(list(
          type = "conditional",
          col = "x",
          val = 5,
          op = "less-equal",
          missing = FALSE
        ))
      ),
      list(
        prediction = 20,
        path = list(list(
          type = "conditional",
          col = "x",
          val = 5,
          op = "more",
          missing = FALSE
        ))
      )
    )
  )
  class(pm) <- c("pm_lgb", "parsed_model", "list")

  fit_formula <- tidypredict_fit(pm)

  formula_str <- deparse(fit_formula)
  expect_match(formula_str, "<=")
  expect_no_match(formula_str, "is.na")
})

test_that("build_lgb_nested_condition errors on unknown type", {
  condition <- list(
    type = "unknown_type",
    col = "x",
    val = 1,
    op = "less-equal",
    missing = FALSE
  )

  expect_snapshot(
    tidypredict:::build_lgb_nested_condition(condition),
    error = TRUE
  )
})

# Categorical feature tests -------------------------------------------------

test_that("parse_model handles categorical splits", {
  skip_if_not_installed("lightgbm")

  set.seed(123)
  n <- 200
  cat_int <- sample(0:3, n, replace = TRUE)
  y <- ifelse(cat_int %in% c(0, 1), 10, -10) + rnorm(n, sd = 0.3)

  X <- matrix(cat_int, ncol = 1)
  colnames(X) <- "cat_feat"

  dtrain <- lightgbm::lgb.Dataset(
    params = list(num_threads = 1L),
    X,
    label = y,
    categorical_feature = "cat_feat"
  )
  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      num_leaves = 4L,
      learning_rate = 1.0,
      objective = "regression",
      min_data_in_leaf = 1L
    ),
    data = dtrain,
    nrounds = 1L,
    verbose = -1L
  )

  pm <- parse_model(model)

  expect_s3_class(pm, "parsed_model")
  expect_gt(length(pm$trees), 0)

  first_leaf <- pm$trees[[1]][[1]]
  expect_gt(length(first_leaf$path), 0)
  expect_equal(first_leaf$path[[1]]$type, "set")
  expect_equal(first_leaf$path[[1]]$op, "in")
  expect_type(first_leaf$path[[1]]$vals, "integer")
})

test_that("categorical predictions match native predictions", {
  skip_if_not_installed("lightgbm")

  set.seed(457)
  n <- 200
  cat_int <- sample(0:3, n, replace = TRUE)
  y <- ifelse(cat_int %in% c(0, 1), 10, -10) + rnorm(n, sd = 2)

  X <- matrix(cat_int, ncol = 1)
  colnames(X) <- "cat_feat"

  dtrain <- lightgbm::lgb.Dataset(
    params = list(num_threads = 1L),
    X,
    label = y,
    categorical_feature = "cat_feat"
  )
  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      num_leaves = 4L,
      learning_rate = 1.0,
      objective = "regression",
      min_data_in_leaf = 1L
    ),
    data = dtrain,
    nrounds = 2L,
    verbose = -1L
  )

  fit_formula <- tidypredict_fit(model)
  native_preds <- predict(model, params = list(num_threads = 1L), X)

  test_df <- data.frame(cat_feat = cat_int)
  tidy_preds <- dplyr::mutate(test_df, pred = !!fit_formula)$pred

  expect_equal(unname(tidy_preds), unname(native_preds), tolerance = 1e-10)
})

test_that("mixed numerical + categorical predictions match", {
  skip_if_not_installed("lightgbm")

  set.seed(789)
  n <- 300
  cat_int <- sample(0:3, n, replace = TRUE)
  num_feat <- rnorm(n)
  y <- ifelse(cat_int %in% c(0, 1), 5, -5) + num_feat * 2 + rnorm(n, sd = 0.3)

  X <- cbind(num_feat = num_feat, cat_feat = cat_int)

  dtrain <- lightgbm::lgb.Dataset(
    params = list(num_threads = 1L),
    X,
    label = y,
    categorical_feature = "cat_feat"
  )
  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      num_leaves = 8L,
      learning_rate = 0.5,
      objective = "regression",
      min_data_in_leaf = 1L
    ),
    data = dtrain,
    nrounds = 3L,
    verbose = -1L
  )

  fit_formula <- tidypredict_fit(model)
  native_preds <- predict(model, params = list(num_threads = 1L), X)

  test_df <- as.data.frame(X)
  tidy_preds <- dplyr::mutate(test_df, pred = !!fit_formula)$pred

  expect_equal(unname(tidy_preds), unname(native_preds), tolerance = 1e-10)
})

test_that("categorical with missing values predictions match", {
  skip_if_not_installed("lightgbm")

  set.seed(321)
  n <- 300
  cat_int <- sample(0:3, n, replace = TRUE)
  y <- ifelse(cat_int %in% c(0, 1), 10, -10) + rnorm(n, sd = 0.3)

  # Add NAs
  na_idx <- sample(n, 30)
  cat_int[na_idx] <- NA
  y[na_idx] <- 10 + rnorm(30, sd = 0.3)

  X <- matrix(as.numeric(cat_int), ncol = 1)
  colnames(X) <- "cat_feat"

  dtrain <- lightgbm::lgb.Dataset(
    params = list(num_threads = 1L),
    X,
    label = y,
    categorical_feature = "cat_feat"
  )
  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      num_leaves = 4L,
      learning_rate = 1.0,
      objective = "regression",
      min_data_in_leaf = 1L,
      use_missing = TRUE
    ),
    data = dtrain,
    nrounds = 1L,
    verbose = -1L
  )

  fit_formula <- tidypredict_fit(model)

  test_X <- matrix(c(0, 1, 2, 3, NA), ncol = 1)
  colnames(test_X) <- "cat_feat"
  test_df <- data.frame(cat_feat = c(0, 1, 2, 3, NA))

  native_preds <- predict(model, params = list(num_threads = 1L), test_X)
  tidy_preds <- dplyr::mutate(test_df, pred = !!fit_formula)$pred

  expect_equal(unname(tidy_preds), unname(native_preds), tolerance = 1e-10)
})

test_that("categorical SQL generation works", {
  skip_if_not_installed("lightgbm")
  skip_if_not_installed("dbplyr")
  skip_if_not_installed("DBI")
  skip_if_not_installed("RSQLite")

  set.seed(123)
  n <- 200
  cat_int <- sample(0:3, n, replace = TRUE)
  y <- ifelse(cat_int %in% c(0, 1), 10, -10) + rnorm(n, sd = 2)

  X <- matrix(cat_int, ncol = 1)
  colnames(X) <- "cat_feat"

  dtrain <- lightgbm::lgb.Dataset(
    params = list(num_threads = 1L),
    X,
    label = y,
    categorical_feature = "cat_feat"
  )
  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      num_leaves = 4L,
      learning_rate = 1.0,
      objective = "regression",
      min_data_in_leaf = 1L
    ),
    data = dtrain,
    nrounds = 1L,
    verbose = -1L
  )

  sql_result <- tidypredict_sql(model, dbplyr::simulate_dbi())
  expect_s3_class(sql_result, "sql")

  con <- DBI::dbConnect(RSQLite::SQLite(), ":memory:")
  on.exit(DBI::dbDisconnect(con), add = TRUE)

  test_data <- data.frame(cat_feat = 0:3)
  DBI::dbWriteTable(con, "test_data", test_data)

  sql_query <- tidypredict_sql(model, con)
  db_result <- DBI::dbGetQuery(
    con,
    paste0("SELECT ", sql_query, " AS pred FROM test_data")
  )

  test_X <- matrix(0:3, ncol = 1)
  colnames(test_X) <- "cat_feat"
  native_preds <- predict(model, params = list(num_threads = 1L), test_X)

  expect_equal(db_result$pred, unname(native_preds), tolerance = 1e-10)
})

test_that("parse_lgb_categorical_threshold handles various formats", {
  expect_equal(
    parse_lgb_categorical_threshold("0||1"),
    c(0L, 1L)
  )

  expect_equal(
    parse_lgb_categorical_threshold("0||1||3"),
    c(0L, 1L, 3L)
  )

  expect_equal(
    parse_lgb_categorical_threshold("2"),
    2L
  )
})

test_that("categorical path contains both in and not-in operators", {
  skip_if_not_installed("lightgbm")

  set.seed(123)
  n <- 200
  cat_int <- sample(0:3, n, replace = TRUE)
  y <- ifelse(cat_int %in% c(0, 1), 10, -10) + rnorm(n, sd = 0.3)

  X <- matrix(cat_int, ncol = 1)
  colnames(X) <- "cat_feat"

  dtrain <- lightgbm::lgb.Dataset(
    params = list(num_threads = 1L),
    X,
    label = y,
    categorical_feature = "cat_feat"
  )
  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      num_leaves = 4L,
      learning_rate = 1.0,
      objective = "regression",
      min_data_in_leaf = 1L
    ),
    data = dtrain,
    nrounds = 1L,
    verbose = -1L
  )

  pm <- parse_model(model)

  # Collect all operators from paths
  all_ops <- unlist(lapply(pm$trees[[1]], function(leaf) {
    sapply(leaf$path, function(p) p$op)
  }))

  expect_contains(all_ops, "in")
  expect_contains(all_ops, "not-in")
})

test_that("categorical with many categories works", {
  skip_if_not_installed("lightgbm")

  set.seed(555)
  n <- 400
  cat_int <- sample(0:7, n, replace = TRUE)
  y <- ifelse(cat_int < 4, 10, -10) + rnorm(n, sd = 2)

  X <- matrix(cat_int, ncol = 1)
  colnames(X) <- "cat_feat"

  dtrain <- lightgbm::lgb.Dataset(
    params = list(num_threads = 1L),
    X,
    label = y,
    categorical_feature = "cat_feat"
  )
  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      num_leaves = 4L,
      learning_rate = 1.0,
      objective = "regression",
      min_data_in_leaf = 1L
    ),
    data = dtrain,
    nrounds = 2L,
    verbose = -1L
  )

  fit_formula <- tidypredict_fit(model)
  native_preds <- predict(model, params = list(num_threads = 1L), X)

  test_df <- data.frame(cat_feat = cat_int)
  tidy_preds <- dplyr::mutate(test_df, pred = !!fit_formula)$pred

  expect_equal(unname(tidy_preds), unname(native_preds), tolerance = 1e-10)
})

test_that("parsed model can be saved and loaded via YAML", {
  skip_if_not_installed("lightgbm")
  skip_if_not_installed("yaml")

  set.seed(123)
  X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
  y <- mtcars$hp

  dtrain <- lightgbm::lgb.Dataset(
    params = list(num_threads = 1L),
    X,
    label = y,
    colnames = c("mpg", "cyl", "disp")
  )
  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      num_leaves = 4L,
      learning_rate = 0.5,
      objective = "regression",
      min_data_in_leaf = 1L
    ),
    data = dtrain,
    nrounds = 5L,
    verbose = -1L
  )

  pm <- parse_model(model)
  mp <- tempfile(fileext = ".yml")
  yaml::write_yaml(pm, mp)
  l <- yaml::read_yaml(mp)
  pm_loaded <- as_parsed_model(l)

  fit_original <- tidypredict_fit(pm)
  fit_loaded <- tidypredict_fit(pm_loaded)

  test_df <- as.data.frame(X)
  preds_original <- dplyr::mutate(test_df, pred = !!fit_original)$pred
  preds_loaded <- dplyr::mutate(test_df, pred = !!fit_loaded)$pred

  expect_equal(preds_original, preds_loaded, tolerance = 1e-6)
})

test_that("parsed multiclass model can be saved and loaded via YAML", {
  skip_if_not_installed("lightgbm")
  skip_if_not_installed("yaml")

  set.seed(123)
  X <- data.matrix(iris[, 1:4])
  colnames(X) <- c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width")
  y <- as.integer(iris$Species) - 1L

  dtrain <- lightgbm::lgb.Dataset(
    params = list(num_threads = 1L),
    X,
    label = y,
    colnames = colnames(X)
  )
  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      num_leaves = 4L,
      learning_rate = 0.5,
      objective = "multiclass",
      num_class = 3L,
      min_data_in_leaf = 1L
    ),
    data = dtrain,
    nrounds = 3L,
    verbose = -1L
  )

  pm <- parse_model(model)
  mp <- tempfile(fileext = ".yml")
  yaml::write_yaml(pm, mp)
  l <- yaml::read_yaml(mp)
  pm_loaded <- as_parsed_model(l)

  fit_original <- tidypredict_fit(pm)
  fit_loaded <- tidypredict_fit(pm_loaded)

  test_df <- as.data.frame(X)
  preds_original <- lapply(fit_original, \(f) {
    dplyr::mutate(test_df, pred = !!f)$pred
  })
  preds_loaded <- lapply(fit_loaded, \(f) {
    dplyr::mutate(test_df, pred = !!f)$pred
  })

  expect_equal(preds_original, preds_loaded, tolerance = 1e-6)
})

test_that("tidypredict_test works for regression model", {
  skip_if_not_installed("lightgbm")

  set.seed(123)
  X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
  y <- mtcars$hp

  dtrain <- lightgbm::lgb.Dataset(
    params = list(num_threads = 1L),
    X,
    label = y,
    colnames = c("mpg", "cyl", "disp")
  )
  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      num_leaves = 4L,
      learning_rate = 0.5,
      objective = "regression",
      min_data_in_leaf = 1L
    ),
    data = dtrain,
    nrounds = 5L,
    verbose = -1L
  )

  result <- tidypredict_test(model, xg_df = X)

  expect_s3_class(result, "tidypredict_test")
  expect_false(result$alert)
  expect_equal(max(result$raw_results$fit_diff), 0, tolerance = 1e-10)
})

test_that("tidypredict_test works for binary classification model", {
  skip_if_not_installed("lightgbm")

  set.seed(456)
  X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
  y <- as.integer(mtcars$am)

  dtrain <- lightgbm::lgb.Dataset(params = list(num_threads = 1L), X, label = y)
  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      num_leaves = 4L,
      learning_rate = 0.5,
      objective = "binary",
      min_data_in_leaf = 1L
    ),
    data = dtrain,
    nrounds = 5L,
    verbose = -1L
  )

  result <- tidypredict_test(model, xg_df = X)

  expect_s3_class(result, "tidypredict_test")
  expect_false(result$alert)
})

test_that("tidypredict_test errors for multiclass model", {
  skip_if_not_installed("lightgbm")

  set.seed(123)
  X <- data.matrix(iris[, 1:4])
  colnames(X) <- c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width")
  y <- as.integer(iris$Species) - 1L

  dtrain <- lightgbm::lgb.Dataset(
    params = list(num_threads = 1L),
    X,
    label = y,
    colnames = colnames(X)
  )
  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      num_leaves = 4L,
      learning_rate = 0.5,
      objective = "multiclass",
      num_class = 3L,
      min_data_in_leaf = 1L
    ),
    data = dtrain,
    nrounds = 3L,
    verbose = -1L
  )

  expect_snapshot(tidypredict_test(model, xg_df = X), error = TRUE)
})

test_that("tidypredict_test errors when matrix not provided", {
  skip_if_not_installed("lightgbm")

  set.seed(123)
  X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
  y <- mtcars$hp

  dtrain <- lightgbm::lgb.Dataset(
    params = list(num_threads = 1L),
    X,
    label = y,
    colnames = c("mpg", "cyl", "disp")
  )
  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      num_leaves = 4L,
      learning_rate = 0.5,
      objective = "regression",
      min_data_in_leaf = 1L
    ),
    data = dtrain,
    nrounds = 5L,
    verbose = -1L
  )

  expect_snapshot(tidypredict_test(model), error = TRUE)
})

test_that("tidypredict_test respects max_rows parameter", {
  skip_if_not_installed("lightgbm")

  set.seed(123)
  X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
  y <- mtcars$hp

  dtrain <- lightgbm::lgb.Dataset(
    params = list(num_threads = 1L),
    X,
    label = y,
    colnames = c("mpg", "cyl", "disp")
  )
  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      num_leaves = 4L,
      learning_rate = 0.5,
      objective = "regression",
      min_data_in_leaf = 1L
    ),
    data = dtrain,
    nrounds = 5L,
    verbose = -1L
  )

  result <- tidypredict_test(model, xg_df = X, max_rows = 10)

  expect_equal(nrow(result$raw_results), 10)
})

test_that(".extract_lgb_trees returns list of tree expressions", {
  skip_if_not_installed("lightgbm")

  set.seed(123)
  X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
  y <- mtcars$hp

  dtrain <- lightgbm::lgb.Dataset(
    params = list(num_threads = 1L),
    X,
    label = y,
    colnames = c("mpg", "cyl", "disp")
  )
  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      num_leaves = 4L,
      learning_rate = 0.5,
      objective = "regression",
      min_data_in_leaf = 1L
    ),
    data = dtrain,
    nrounds = 5L,
    verbose = -1L
  )

  trees <- .extract_lgb_trees(model)

  expect_type(trees, "list")
  expect_length(trees, 5)

  types <- vapply(trees, typeof, character(1))
  expect_all_equal(types, "language")
})

test_that(".extract_lgb_trees combined results match tidypredict_fit", {
  skip_if_not_installed("lightgbm")
  model <- make_lgb_model()
  test_data <- mtcars[, c("mpg", "cyl", "disp")]

  trees <- .extract_lgb_trees(model)
  eval_env <- rlang::new_environment(
    data = as.list(test_data),
    parent = asNamespace("dplyr")
  )
  tree_preds <- lapply(trees, rlang::eval_tidy, env = eval_env)
  combined <- Reduce(`+`, tree_preds)

  fit_result <- rlang::eval_tidy(tidypredict_fit(model), test_data)

  expect_equal(combined, fit_result)
})

test_that(".extract_lgb_trees errors on non-lgb.Booster", {
  expect_snapshot(.extract_lgb_trees(list()), error = TRUE)
})

test_that("tidypredict works with parsnip/bonsai lightgbm model", {
  skip_if_not_installed("lightgbm")
  skip_if_not_installed("parsnip")
  skip_if_not_installed("bonsai")

  set.seed(123)
  train_data <- mtcars[, c("hp", "mpg", "cyl", "disp")]

  model_spec <- parsnip::boost_tree(
    trees = 5,
    tree_depth = 3,
    min_n = 1
  ) |>
    parsnip::set_engine("lightgbm") |>
    parsnip::set_mode("regression")

  model_fit <- parsnip::fit(
    model_spec,
    hp ~ mpg + cyl + disp,
    data = train_data
  )

  lgb_model <- model_fit$fit

  expect_s3_class(lgb_model, "lgb.Booster")

  pm <- parse_model(lgb_model)
  expect_s3_class(pm, "parsed_model")
  expect_s3_class(pm, "pm_lgb")
  expect_gt(length(pm$trees), 0)

  fit_formula <- tidypredict_fit(lgb_model)
  expect_type(fit_formula, "language")

  X <- data.matrix(train_data[, c("mpg", "cyl", "disp")])
  native_preds <- predict(lgb_model, params = list(num_threads = 1L), X)
  tidy_preds <- dplyr::mutate(train_data, pred = !!fit_formula)$pred

  expect_equal(unname(tidy_preds), unname(native_preds), tolerance = 1e-10)
})

test_that("tidypredict works with parsnip/bonsai binary classification", {
  skip_if_not_installed("lightgbm")
  skip_if_not_installed("parsnip")
  skip_if_not_installed("bonsai")

  set.seed(456)
  train_data <- mtcars[, c("am", "mpg", "cyl", "disp")]
  train_data$am <- factor(train_data$am)

  model_spec <- parsnip::boost_tree(
    trees = 5,
    tree_depth = 3,
    min_n = 1
  ) |>
    parsnip::set_engine("lightgbm") |>
    parsnip::set_mode("classification")

  model_fit <- parsnip::fit(
    model_spec,
    am ~ mpg + cyl + disp,
    data = train_data
  )

  lgb_model <- model_fit$fit

  expect_s3_class(lgb_model, "lgb.Booster")

  fit_formula <- tidypredict_fit(lgb_model)
  expect_type(fit_formula, "language")

  X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
  native_preds <- predict(lgb_model, params = list(num_threads = 1L), X)
  tidy_preds <- dplyr::mutate(mtcars, pred = !!fit_formula)$pred

  expect_equal(unname(tidy_preds), unname(native_preds), tolerance = 1e-10)
})

test_that("tidypredict_sql works with parsnip/bonsai model", {
  skip_if_not_installed("lightgbm")
  skip_if_not_installed("parsnip")
  skip_if_not_installed("bonsai")
  skip_if_not_installed("dbplyr")

  set.seed(123)
  train_data <- mtcars[, c("hp", "mpg", "cyl", "disp")]

  model_spec <- parsnip::boost_tree(
    trees = 3,
    tree_depth = 2,
    min_n = 1
  ) |>
    parsnip::set_engine("lightgbm") |>
    parsnip::set_mode("regression")

  model_fit <- parsnip::fit(
    model_spec,
    hp ~ mpg + cyl + disp,
    data = train_data
  )
  lgb_model <- model_fit$fit

  sql_result <- tidypredict_sql(lgb_model, dbplyr::simulate_dbi())

  expect_s3_class(sql_result, "sql")
})

test_that("tidypredict_test works with parsnip/bonsai model", {
  skip_if_not_installed("lightgbm")
  skip_if_not_installed("parsnip")
  skip_if_not_installed("bonsai")

  set.seed(123)
  train_data <- mtcars[, c("hp", "mpg", "cyl", "disp")]

  model_spec <- parsnip::boost_tree(
    trees = 5,
    tree_depth = 3,
    min_n = 1
  ) |>
    parsnip::set_engine("lightgbm") |>
    parsnip::set_mode("regression")

  model_fit <- parsnip::fit(
    model_spec,
    hp ~ mpg + cyl + disp,
    data = train_data
  )
  lgb_model <- model_fit$fit

  X <- data.matrix(train_data[, c("mpg", "cyl", "disp")])
  result <- tidypredict_test(lgb_model, xg_df = X)

  expect_s3_class(result, "tidypredict_test")
  expect_false(result$alert)
})

# Linear tree tests ----------------------------------------------------------

test_that("linear tree regression predictions match native predict (#186)", {
  skip_if_not_installed("lightgbm")

  set.seed(123)
  n <- 100
  X <- cbind(x1 = rnorm(n), x2 = rnorm(n))
  y <- 2 * X[, 1] + 3 * X[, 2] + rnorm(n, sd = 0.1)

  dtrain <- lightgbm::lgb.Dataset(
    params = list(num_threads = 1L),
    X,
    label = y,
    colnames = c("x1", "x2")
  )
  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      objective = "regression",
      linear_tree = TRUE,
      num_leaves = 4L,
      min_data_in_leaf = 10L
    ),
    data = dtrain,
    nrounds = 5L,
    verbose = -1L
  )

  fit_formula <- tidypredict_fit(model)
  native_preds <- predict(model, params = list(num_threads = 1L), X)

  test_df <- as.data.frame(X)
  tidy_preds <- dplyr::mutate(test_df, pred = !!fit_formula)$pred

  expect_equal(unname(tidy_preds), unname(native_preds), tolerance = 1e-10)
})

test_that("linear tree binary classification predictions match (#186)", {
  skip_if_not_installed("lightgbm")

  set.seed(456)
  n <- 200
  X <- cbind(x1 = rnorm(n), x2 = rnorm(n))
  y <- as.numeric((2 * X[, 1] + 3 * X[, 2] + rnorm(n)) > 0)

  dtrain <- lightgbm::lgb.Dataset(
    params = list(num_threads = 1L),
    X,
    label = y,
    colnames = c("x1", "x2")
  )
  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      objective = "binary",
      linear_tree = TRUE,
      num_leaves = 4L,
      min_data_in_leaf = 20L
    ),
    data = dtrain,
    nrounds = 5L,
    verbose = -1L
  )

  fit_formula <- tidypredict_fit(model)
  native_preds <- predict(model, params = list(num_threads = 1L), X)

  test_df <- as.data.frame(X)
  tidy_preds <- dplyr::mutate(test_df, pred = !!fit_formula)$pred

  expect_equal(unname(tidy_preds), unname(native_preds), tolerance = 1e-10)
})

test_that("linear tree multiclass predictions match (#186)", {
  skip_if_not_installed("lightgbm")

  set.seed(789)
  n <- 300
  X <- cbind(x1 = rnorm(n), x2 = rnorm(n))
  y <- as.integer(cut(X[, 1] + X[, 2] + rnorm(n, sd = 0.5), breaks = 3)) - 1L

  dtrain <- lightgbm::lgb.Dataset(
    params = list(num_threads = 1L),
    X,
    label = y,
    colnames = c("x1", "x2")
  )
  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      objective = "multiclass",
      num_class = 3L,
      linear_tree = TRUE,
      num_leaves = 4L,
      min_data_in_leaf = 20L
    ),
    data = dtrain,
    nrounds = 3L,
    verbose = -1L
  )

  fit_formulas <- tidypredict_fit(model)
  native_preds <- predict(model, params = list(num_threads = 1L), X)
  native_mat <- matrix(native_preds, ncol = 3, byrow = FALSE)

  test_df <- as.data.frame(X)
  tidy_preds <- dplyr::mutate(
    test_df,
    class_0 = !!fit_formulas$class_0,
    class_1 = !!fit_formulas$class_1,
    class_2 = !!fit_formulas$class_2
  )
  tidy_mat <- as.matrix(tidy_preds[, c("class_0", "class_1", "class_2")])

  expect_equal(unname(tidy_mat), unname(native_mat), tolerance = 1e-10)
})

test_that("linear tree with RF boosting predictions match (#186)", {
  skip_if_not_installed("lightgbm")

  set.seed(321)
  n <- 100
  X <- cbind(x1 = rnorm(n), x2 = rnorm(n))
  y <- 2 * X[, 1] + 3 * X[, 2] + rnorm(n, sd = 0.1)

  dtrain <- lightgbm::lgb.Dataset(
    params = list(num_threads = 1L),
    X,
    label = y,
    colnames = c("x1", "x2")
  )
  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      boosting = "rf",
      objective = "regression",
      linear_tree = TRUE,
      num_leaves = 4L,
      min_data_in_leaf = 10L,
      bagging_freq = 1,
      bagging_fraction = 0.8
    ),
    data = dtrain,
    nrounds = 5L,
    verbose = -1L
  )

  fit_formula <- tidypredict_fit(model)
  native_preds <- predict(model, params = list(num_threads = 1L), X)

  test_df <- as.data.frame(X)
  tidy_preds <- dplyr::mutate(test_df, pred = !!fit_formula)$pred

  expect_equal(unname(tidy_preds), unname(native_preds), tolerance = 1e-10)
})

test_that("linear tree parsed model has correct structure", {
  skip_if_not_installed("lightgbm")

  set.seed(123)
  n <- 100
  X <- cbind(x1 = rnorm(n), x2 = rnorm(n))
  y <- 2 * X[, 1] + 3 * X[, 2] + rnorm(n, sd = 0.1)

  dtrain <- lightgbm::lgb.Dataset(
    params = list(num_threads = 1L),
    X,
    label = y,
    colnames = c("x1", "x2")
  )
  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      objective = "regression",
      linear_tree = TRUE,
      num_leaves = 4L,
      min_data_in_leaf = 10L
    ),
    data = dtrain,
    nrounds = 3L,
    verbose = -1L
  )

  pm <- parse_model(model)

  expect_s3_class(pm, "parsed_model")
  expect_length(pm$trees, 3)

  # Check that leaves have linear info instead of constant predictions
  first_tree <- pm$trees[[1]]
  first_leaf <- first_tree[[1]]

  expect_contains(names(first_leaf), c("prediction", "linear", "path"))

  # Either prediction is NULL (linear) or linear is NULL (constant)
  has_linear <- !is.null(first_leaf$linear)
  if (has_linear) {
    expect_null(first_leaf$prediction)
    expect_contains(
      names(first_leaf$linear),
      c("intercept", "feature_names", "coefficients")
    )
    expect_type(first_leaf$linear$intercept, "double")
    expect_type(first_leaf$linear$feature_names, "character")
    expect_type(first_leaf$linear$coefficients, "double")
  }
})

test_that("linear tree handles NA values correctly when trained with NAs (#186)", {
  skip_if_not_installed("lightgbm")

  set.seed(123)
  n <- 100
  x1 <- rnorm(n)
  x2 <- rnorm(n)
  y <- 2 + 0.5 * x1 + 0.3 * x2 + rnorm(n, sd = 0.1)

  # Add NAs to training data so model learns NA handling

  x1[c(5, 15, 25)] <- NA
  x2[c(10, 20, 30)] <- NA

  X <- cbind(x1 = x1, x2 = x2)
  dtrain <- lightgbm::lgb.Dataset(
    params = list(num_threads = 1L),
    X,
    label = y,
    colnames = c("x1", "x2")
  )
  model <- lightgbm::lgb.train(
    params = list(
      num_threads = 1,
      objective = "regression",
      linear_tree = TRUE,
      num_leaves = 4L,
      min_data_in_leaf = 10L
    ),
    data = dtrain,
    nrounds = 3L,
    verbose = -1L
  )

  # Test data with various NA patterns
  test_X <- rbind(
    c(0.5, 0.5), # No NAs
    c(NA, 0.5), # NA in x1
    c(0.5, NA), # NA in x2
    c(NA, NA) # Both NA
  )
  colnames(test_X) <- c("x1", "x2")

  fit_formula <- tidypredict_fit(model)
  native_preds <- predict(model, params = list(num_threads = 1L), test_X)

  test_df <- as.data.frame(test_X)
  tidy_preds <- dplyr::mutate(test_df, pred = !!fit_formula)$pred

  expect_equal(unname(tidy_preds), unname(native_preds), tolerance = 1e-10)
})

#' Generate nested case_when trees
#'
#' These functions generate nested `case_when()` expressions for decision trees,
#' which are more efficient than flat `case_when()` for both R/dplyr and SQL
#' execution.
#'
#' The following tree:
#'
#'             +-----+
#'        +----|x > 0|----+
#'        |    +-----+    |
#'        v               v
#'    +------+        +--------+
#' +--|y < 20|--+  +--|z <= 10 |--+
#' |  +------+  |  |  +--------+  |
#' v            v  v              v
#' a            b  c              d
#'
#' will be turned into the following nested `case_when()` statement:
#'
#' ```r
#' case_when(
#'   x > 0 ~ case_when(
#'     y < 20 ~ "a",
#'     .default = "b"
#'   ),
#'   .default = case_when(
#'     z <= 10 ~ "c",
#'     .default = "d"
#'   )
#' )
#' ```
#'
#' @details
#' NA values in predictor columns are not handled by the generated expression.
#' Users should ensure that predictor columns do not contain NA values before
#' using the generated expression, or the results will be NA for those rows.
#'
#' @keywords internal

#' Generate nested case_when for a tree
#'
#' @param tree_info A tree info list from `rpart_tree_info_full()` or similar
#' @keywords internal
generate_nested_case_when_tree <- function(tree_info) {
  build_nested_node(0L, tree_info)
}

#' Build a nested case_when expression for a single node
#'
#' @param node_id The node ID to build (0-indexed)
#' @param tree_info Tree info list with nodeID, leftChild, rightChild,
#'   splitvarName, terminal, prediction, and node_splits
#' @keywords internal
build_nested_node <- function(node_id, tree_info) {
  node_idx <- which(tree_info$nodeID == node_id)

  # Leaf node: return prediction
  if (tree_info$terminal[node_idx]) {
    prediction <- tree_info$prediction[node_idx]
    if (is.factor(prediction)) {
      prediction <- as.character(prediction)
    }
    return(prediction)
  }

  # Internal node: build nested case_when
  left_id <- tree_info$leftChild[node_idx]
  right_id <- tree_info$rightChild[node_idx]

  # Get split info
  split_info <- tree_info$node_splits[[node_idx]]

  # Recursively build subtrees
  left_subtree <- build_nested_node(left_id, tree_info)
  right_subtree <- build_nested_node(right_id, tree_info)

  # Build condition
  condition <- build_nested_split_condition(split_info$primary)
  expr(case_when(!!condition ~ !!left_subtree, .default = !!right_subtree))
}

#' Build a split condition expression for nested trees (left branch)
#'
#' @param split A split info list with col, val/vals, is_categorical
#' @keywords internal
build_nested_split_condition <- function(split) {
  col <- rlang::sym(split$col)

  if (split$is_categorical) {
    # Categorical split: x %in% c("a", "b")
    vals <- unlist(split$vals)
    expr(!!col %in% !!vals)
  } else {
    # Continuous split: x <= threshold (left branch)
    expr(!!col <= !!split$val)
  }
}

# For {orbital}
#' Build nested case_when expression from tree info
#'
#' Shared helper for building nested tree expressions. This is the nested
#' equivalent of `.build_case_when_tree()`.
#'
#' @param tree_info A tree info list with nodeID, leftChild, rightChild,
#'   splitvarName, terminal, prediction, and node_splits
#' @keywords internal
#' @export
.build_nested_case_when_tree <- function(tree_info) {
  generate_nested_case_when_tree(tree_info)
}

# Build nested case_when from flat paths format
#
# Converts a flat list of leaf paths into a nested case_when expression.
# Used by parsed models (xgboost, lightgbm, catboost) when reconstructing
# trees from serialized format.
#
# @param leaves List of leaves, each with `prediction` and `path`. Each path
#   element must have an `op` field indicating the branch direction.
# @param build_condition_fn Function to build a condition expression from a
#   path element. Should only build the LEFT branch condition (the right
#   branch is handled by `.default`).
#
# ## Operator naming convention
#
# Different models use different operator names, but they follow a pattern:
# - **Left branch operators** (condition is TRUE): "less", "less-equal", "in",
#   "more-equal", "equal"
# - **Right branch operators** (condition is FALSE): "more", "not-in",
#   "not-equal"
#
# Model-specific conventions:
# - **xgboost**: "less" and "more-equal" for left, inverse for right
# - **lightgbm**: "less-equal" and "in" for left, "more" and "not-in" for right
# - **catboost**: "less-equal" and "equal" for left, "more" and "not-equal" for right
#
# The `build_condition_fn` is only called with left-branch path elements,
# so it only needs to handle left-branch operators.
build_nested_from_flat_paths <- function(leaves, build_condition_fn) {
  if (length(leaves) == 0) {
    cli::cli_abort("Empty tree.", .internal = TRUE)
  }

  # Single leaf (stump)
  if (length(leaves) == 1 && length(leaves[[1]]$path) == 0) {
    return(leaves[[1]]$prediction)
  }

  build_nested_from_paths_recursive(leaves, build_condition_fn, path_depth = 1)
}

build_nested_from_paths_recursive <- function(
  leaves,
  build_condition_fn,
  path_depth
) {
  if (length(leaves) == 1) {
    return(leaves[[1]]$prediction)
  }

  first_leaf <- leaves[[1]]
  if (path_depth > length(first_leaf$path)) {
    return(first_leaf$prediction)
  }

  # Get condition at this depth
  split_info <- first_leaf$path[[path_depth]]

  # Partition leaves by left vs right condition based on operator name.
  # See "Operator naming convention" in build_nested_from_flat_paths docs.
  is_left_condition <- function(leaf) {
    if (path_depth > length(leaf$path)) {
      return(TRUE)
    }
    op <- leaf$path[[path_depth]]$op
    op %in% c("less", "less-equal", "in", "more-equal", "equal")
  }

  left_leaves <- Filter(is_left_condition, leaves)
  right_leaves <- Filter(Negate(is_left_condition), leaves)

  if (length(left_leaves) == 0 || length(right_leaves) == 0) {
    return(build_nested_from_paths_recursive(
      leaves,
      build_condition_fn,
      path_depth + 1
    ))
  }

  condition <- build_condition_fn(split_info)

  left_subtree <- build_nested_from_paths_recursive(
    left_leaves,
    build_condition_fn,
    path_depth + 1
  )
  right_subtree <- build_nested_from_paths_recursive(
    right_leaves,
    build_condition_fn,
    path_depth + 1
  )

  expr(case_when(!!condition ~ !!left_subtree, .default = !!right_subtree))
}

test_that("generate_nested_case_when_tree works for regression", {
  skip_if_not_installed("rpart")

  model <- rpart::rpart(Sepal.Length ~ Sepal.Width + Petal.Length, data = iris)
  tree_info <- rpart_tree_info_full(model)
  nested_expr <- generate_nested_case_when_tree(tree_info)

  nested_pred <- dplyr::mutate(iris, pred = !!nested_expr)$pred
  original_pred <- predict(model, iris)

  expect_equal(nested_pred, unname(original_pred))
})

test_that("generate_nested_case_when_tree works for classification", {
  skip_if_not_installed("rpart")

  model <- rpart::rpart(
    Species ~ Sepal.Width + Petal.Length,
    data = iris,
    method = "class"
  )
  tree_info <- rpart_tree_info_full(model)
  nested_expr <- generate_nested_case_when_tree(tree_info)

  nested_pred <- dplyr::mutate(iris, pred = !!nested_expr)$pred
  original_pred <- as.character(predict(model, iris, type = "class"))

  expect_equal(nested_pred, original_pred)
})

test_that("generate_nested_case_when_tree handles categorical splits", {
  skip_if_not_installed("rpart")

  data <- mtcars
  data$cyl <- factor(data$cyl)
  model <- rpart::rpart(mpg ~ cyl + hp, data = data)
  tree_info <- rpart_tree_info_full(model)
  nested_expr <- generate_nested_case_when_tree(tree_info)

  nested_pred <- dplyr::mutate(data, pred = !!nested_expr)$pred
  original_pred <- predict(model, data)

  expect_equal(nested_pred, unname(original_pred))
})

test_that("generate_nested_case_when_tree handles stump trees", {
  skip_if_not_installed("rpart")

  model <- rpart::rpart(Sepal.Length ~ Sepal.Width, data = iris, cp = 1)
  tree_info <- rpart_tree_info_full(model)

  expect_equal(sum(tree_info$terminal), 1L)

  nested_expr <- generate_nested_case_when_tree(tree_info)

  expect_true(is.numeric(nested_expr))

  nested_pred <- dplyr::mutate(iris, pred = !!nested_expr)$pred
  original_pred <- predict(model, iris)

  expect_equal(nested_pred, unname(original_pred))
})

test_that("generate_nested_case_when_tree produces nested case_when structure", {
  skip_if_not_installed("rpart")

  model <- rpart::rpart(Sepal.Length ~ Petal.Length, data = iris)
  tree_info <- rpart_tree_info_full(model)
  nested_expr <- generate_nested_case_when_tree(tree_info)

  expr_str <- rlang::expr_deparse(nested_expr)
  expect_true(any(grepl("case_when", expr_str)))
  expect_true(any(grepl("\\.default", expr_str)))
})

test_that("build_nested_split_condition handles continuous splits", {
  split <- list(col = "x", val = 5, is_categorical = FALSE)
  cond <- build_nested_split_condition(split)

  expect_equal(rlang::expr_deparse(cond), "x <= 5")
})

test_that("build_nested_split_condition handles categorical splits", {
  split <- list(col = "x", vals = list("a", "b"), is_categorical = TRUE)
  cond <- build_nested_split_condition(split)

  data <- data.frame(x = c("a", "b", "c", "d"))
  result <- dplyr::mutate(data, match = !!cond)$match
  expect_equal(result, c(TRUE, TRUE, FALSE, FALSE))
})

test_that(".build_nested_case_when_tree is exported and works", {
  skip_if_not_installed("rpart")

  model <- rpart::rpart(Sepal.Length ~ Petal.Length, data = iris)
  tree_info <- rpart_tree_info_full(model)

  nested_expr <- .build_nested_case_when_tree(tree_info)

  nested_pred <- dplyr::mutate(iris, pred = !!nested_expr)$pred
  original_pred <- predict(model, iris)

  expect_equal(nested_pred, unname(original_pred))
})

# Deep tree tests

test_that("generate_nested_case_when_tree works for deep trees (depth 6+)", {
  skip_if_not_installed("rpart")

  ctrl <- rpart::rpart.control(cp = 0.001, minsplit = 2, maxdepth = 10)
  model <- rpart::rpart(
    Sepal.Length ~ Sepal.Width + Petal.Length + Petal.Width,
    data = iris,
    control = ctrl
  )

  tree_info <- rpart_tree_info_full(model)
  n_nodes <- length(tree_info$nodeID)
  expect_gt(n_nodes, 10)

  nested_expr <- generate_nested_case_when_tree(tree_info)
  nested_pred <- dplyr::mutate(iris, pred = !!nested_expr)$pred
  original_pred <- predict(model, iris)

  expect_equal(nested_pred, unname(original_pred))
})

# tidypredict_fit tests

test_that("tidypredict_fit produces correct predictions for rpart regression", {
  skip_if_not_installed("rpart")

  model <- rpart::rpart(mpg ~ ., data = mtcars)

  fit_expr <- tidypredict_fit(model)
  fit_pred <- dplyr::mutate(mtcars, pred = !!fit_expr)$pred
  original_pred <- predict(model, mtcars)

  expect_equal(fit_pred, unname(original_pred))
})

test_that("tidypredict_fit produces correct predictions for rpart classification", {
  skip_if_not_installed("rpart")

  model <- rpart::rpart(Species ~ ., data = iris, method = "class")

  fit_expr <- tidypredict_fit(model)
  fit_pred <- dplyr::mutate(iris, pred = !!fit_expr)$pred
  original_pred <- as.character(predict(model, iris, type = "class"))

  expect_equal(fit_pred, original_pred)
})

# Script to regenerate rf-v2-classification.rds
# Uses CRAN tidypredict 1.0.1 which produced v2 format
# Note: This model will error when used with tidypredict_fit()
# because classification is not supported.

tmp_lib <- tempfile("lib")
dir.create(tmp_lib)
install.packages(
  "tidypredict",
  lib = tmp_lib,
  repos = "https://cloud.r-project.org"
)
library(tidypredict, lib.loc = tmp_lib)

library(randomForest)
set.seed(123)
model <- randomForest(
  Species ~ Sepal.Length + Sepal.Width,
  data = iris,
  ntree = 3
)

pm <- parse_model(model)
saveRDS(pm, "tests/testthat/backwards-compat/rf-v2-classification.rds")

#' Prepares parsed model object
#'
#' @param x A parsed model object
#'
#' @export
as_parsed_model <- function(x) {
  UseMethod("as_parsed_model")
}

#' @export
as_parsed_model.list <- function(x) {
  t <- paste0("pm_", x$general$type)
  class(x) <- c("parsed_model", t, class(x))
  x
}

## ----setup, include = FALSE---------------------------------------------------
if (requireNamespace("lightgbm", quietly = TRUE)) {
  library(tidypredict)
  library(lightgbm)
  library(dplyr)
  eval_code <- TRUE
} else {
  eval_code <- FALSE
}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = eval_code
)
source("_threads.R")

## -----------------------------------------------------------------------------
library(lightgbm)

# Prepare data
X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
y <- mtcars$hp

dtrain <- lgb.Dataset(X, label = y, colnames = c("mpg", "cyl", "disp"))

model <- lgb.train(
  params = list(
    num_leaves = 4L,
    learning_rate = 0.5,
    objective = "regression",
    min_data_in_leaf = 1L
  ),
  data = dtrain,
  nrounds = 10L,
  verbose = -1L
)

## -----------------------------------------------------------------------------
tidypredict_fit(model)

## -----------------------------------------------------------------------------
library(dplyr)

mtcars %>%
  tidypredict_to_column(model) %>%
  glimpse()

## -----------------------------------------------------------------------------
tidypredict_test(model, xg_df = X)

## -----------------------------------------------------------------------------
X_bin <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
y_bin <- mtcars$am

dtrain_bin <- lgb.Dataset(X_bin, label = y_bin, colnames = c("mpg", "cyl", "disp"))

model_bin <- lgb.train(
  params = list(
    num_leaves = 4L,
    learning_rate = 0.5,
    objective = "binary",
    min_data_in_leaf = 1L
  ),
  data = dtrain_bin,
  nrounds = 10L,
  verbose = -1L
)

tidypredict_test(model_bin, xg_df = X_bin)

## -----------------------------------------------------------------------------
X_iris <- data.matrix(iris[, 1:4])
colnames(X_iris) <- c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width")
y_iris <- as.integer(iris$Species) - 1L

dtrain_iris <- lgb.Dataset(X_iris, label = y_iris, colnames = colnames(X_iris))

model_multi <- lgb.train(
  params = list(
    num_leaves = 4L,
    learning_rate = 0.5,
    objective = "multiclass",
    num_class = 3L,
    min_data_in_leaf = 1L
  ),
  data = dtrain_iris,
  nrounds = 5L,
  verbose = -1L
)

fit_formulas <- tidypredict_fit(model_multi)
names(fit_formulas)

## -----------------------------------------------------------------------------
iris %>%
  mutate(
    prob_setosa = !!fit_formulas$class_0,
    prob_versicolor = !!fit_formulas$class_1,
    prob_virginica = !!fit_formulas$class_2
  ) %>%
  select(Species, starts_with("prob_")) %>%
  head()

## -----------------------------------------------------------------------------
set.seed(123)
n <- 200
cat_data <- data.frame(
  cat_feat = sample(0:3, n, replace = TRUE),
  y = NA
)
cat_data$y <- ifelse(cat_data$cat_feat %in% c(0, 1), 10, -10) + rnorm(n, sd = 2)

X_cat <- matrix(cat_data$cat_feat, ncol = 1)
colnames(X_cat) <- "cat_feat"

dtrain_cat <- lgb.Dataset(
  X_cat,
  label = cat_data$y,
  categorical_feature = "cat_feat"
)

model_cat <- lgb.train(
  params = list(
    num_leaves = 4L,
    learning_rate = 1.0,
    objective = "regression",
    min_data_in_leaf = 1L
  ),
  data = dtrain_cat,
  nrounds = 2L,
  verbose = -1L
)

tidypredict_fit(model_cat)

## ----eval = requireNamespace("parsnip", quietly = TRUE) && requireNamespace("bonsai", quietly = TRUE)----
library(parsnip)
library(bonsai)

p_model <- boost_tree(
  trees = 10,
  tree_depth = 3,
  min_n = 1
) %>%
  set_engine("lightgbm") %>%
  set_mode("regression") %>%
  fit(hp ~ mpg + cyl + disp, data = mtcars)

# Extract the underlying lgb.Booster
lgb_model <- p_model$fit

tidypredict_test(lgb_model, xg_df = X)

## -----------------------------------------------------------------------------
pm <- parse_model(model)
str(pm, 2)

## -----------------------------------------------------------------------------
str(pm$trees[1])

## ----setup, include = FALSE---------------------------------------------------
if (FALSE) {
  library(tidypredict)
  library(dplyr)
  eval_code <- TRUE
} else {
  eval_code <- FALSE
}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = eval_code
)
source("_threads.R")

#' Adds the prediction columns to a piped command set.
#'
#' Adds a new column with the results from tidypredict_fit() to a piped command set.
#' If add_interval is set to TRUE, it will add two additional columns- one
#' for the lower and another for the upper prediction interval bounds.
#'
#' @param df A data.frame or tibble
#' @param model An R model or a parsed model inside a data frame
#' @param add_interval Switch that indicates if the prediction interval columns should be added. Defaults
#' to FALSE
#' @param interval The prediction interval, defaults to 0.95. Ignored if add_interval is set to
#' FALSE
#' @param vars The name of the variables that this function will produce. Defaults to "fit", "upper", and "lower".
#'
#' @export
tidypredict_to_column <- function(
  df,
  model,
  add_interval = FALSE,
  interval = 0.95,
  vars = c("fit", "upper", "lower")
) {
  fit_model <- tidypredict_fit(model)

  if (inherits(fit_model, "list")) {
    cli::cli_abort(
      "{.fn tidypredict_to_column} does not support tree based models."
    )
  }

  fit <- vars[1]
  upper <- vars[2]
  lower <- vars[3]

  df <- mutate(df, !!fit := !!fit_model)

  if (add_interval) {
    formulas <- c(
      as.name(fit),
      tidypredict_interval(model, interval = interval)
    )
    upper_formula <- reduce_addition(formulas)
    lower_formula <- reduce_subtraction(formulas)

    df <- mutate(
      df,
      !!upper := !!upper_formula,
      !!lower := !!lower_formula
    )
  }

  df
}

# Script to generate ranger-v2-classification.rds for backwards compatibility testing
#
# This model was created with tidypredict 1.0.1 (CRAN) which used ranger's
# v2 parsed model format with flat path structure in $trees.
#
# Classification models are parsed but error when calling tidypredict_fit().
# This file is used to test that error path.
#
# Run this script to regenerate the test fixture if needed.

# Install CRAN version of tidypredict (1.0.1)
tmp_lib <- tempfile("lib")
dir.create(tmp_lib)
install.packages(
  "tidypredict",
  lib = tmp_lib,
  repos = "https://cloud.r-project.org",
  quiet = TRUE
)

# Load CRAN version
library(tidypredict, lib.loc = tmp_lib)
cat("tidypredict version:", as.character(packageVersion("tidypredict")), "\n")

# ranger is also needed
library(ranger)
cat("ranger version:", as.character(packageVersion("ranger")), "\n")

# Create a classification model
set.seed(123)
model <- ranger::ranger(
  Species ~ Sepal.Length + Sepal.Width,
  data = iris,
  num.trees = 2,
  max.depth = 2,
  seed = 123,
  num.threads = 1
)

pm <- parse_model(model)
cat("\nClassification model:\n")
cat("Version:", pm$general$version, "\n")
cat("Has $trees:", !is.null(pm$trees), "\n")
cat("First prediction:", pm$trees[[1]][[1]]$prediction, "\n")

saveRDS(pm, "ranger-v2-classification.rds")
cat("Saved to ranger-v2-classification.rds\n")

# Cleanup
unlink(tmp_lib, recursive = TRUE)

# Predict ---------------------------------------

#' @export
tidypredict_fit.lm <- function(model) {
  parsedmodel <- parse_model(model)
  build_fit_formula(parsedmodel)
}

build_fit_formula <- function(parsedmodel) {
  parsed_f <- map(
    parsedmodel$terms,
    ~ {
      if (.x$is_intercept == 0) {
        cols <- map(.x$fields, lm_constructor)
        cols <- reduce_multiplication(cols)
        expr((!!cols * !!.x$coef))
      } else {
        if (.x$coef == 0) {
          NULL
        } else {
          expr(!!.x$coef)
        }
      }
    }
  )
  parsed_f <- purrr::discard(parsed_f, is.null)

  f <- reduce_addition(parsed_f)

  if (!is.null(parsedmodel$general$offset)) {
    f <- expr_addition(f, parsedmodel$general$offset)
  }

  if (parsedmodel$general$is_glm == 1) {
    link <- parsedmodel$general$link
    f <- apply_inverse_link(f, link)
  }
  f
}

apply_inverse_link <- function(f, link) {
  switch(
    link,
    "identity" = f,
    "logit" = expr(1 - 1 / (1 + exp(!!f))),
    "log" = expr(exp(!!f)),
    "inverse" = expr(1 / (!!f)),
    "1/mu^2" = expr(1 / sqrt(!!f)),
    # Bowling et al. approximation to normal CDF (max error ~0.014%)
    "probit" = expr(1 / (1 + exp(-0.07056 * (!!f)^3 - 1.5976 * (!!f)))),
    "cloglog" = expr(1 - exp(-exp(!!f))),
    "sqrt" = expr((!!f)^2),
    cli::cli_abort("Link {.val {link}} is not supported.")
  )
}

# Parse model --------------------------------------

#' @export
parse_model.lm <- function(model) parse_model_lm(model)

parse_model_lm <- function(model, call = rlang::caller_env()) {
  acceptable_formula(model)

  coefs <- as.numeric(model$coefficients)
  labels <- names(model$coefficients)
  vars <- names(attr(model$terms, "dataClasses"))
  qr <- NULL
  if (!is.null(model$qr)) {
    qr <- tryCatch(
      qr.solve(qr.R(model$qr)),
      error = function(cnd) {
        if (grepl("singular matrix", cnd$message)) {
          cli::cli_abort(
            c(
              x = "Unable to calculate inverse of QR decomposition.",
              i = "This is likely happening because the predictors contain a 
              linear combination of predictors. Please remove and try again."
            ),
            call = call
          )
        }
        stop(cnd)
      }
    )
  }

  pm <- list()
  pm$general$model <- class(model)[[1]]
  pm$general$version <- 2
  pm$general$type <- "regression"
  pm$general$residual <- model$df.residual

  if (length(summary(model)$sigma^2) > 0) {
    pm$general$sigma2 <- summary(model)$sigma^2
  }
  if (!is.null(model$family$family)) {
    pm$general$family <- model$family$family
  }
  if (!is.null(model$family$link)) {
    pm$general$link <- model$family$link
  }
  if (!is.null(model$call$offset)) {
    pm$general$offset <- model$call$offset
  }
  pm$general$is_glm <- 0
  if (class(model)[[1]] == "glm") {
    pm$general$is_glm <- 1
  }
  terms <- map(
    seq_len(length(labels)),
    ~ {
      list(
        label = labels[.x],
        coef = coefs[.x],
        is_intercept = ifelse(labels[.x] == "(Intercept)", 1, 0),
        fields = parse_label_lm(labels[.x], vars),
        qr = parse_qr_lm(labels[.x], qr)
      )
    }
  )
  pm$terms <- terms
  as_parsed_model(pm)
}

parse_label_lm <- function(label, vars) {
  all_items <- NULL
  items <- strsplit(label, "\\:")[[1]]
  for (i in seq_len(length(items))) {
    item <- list(
      type = "ordinary",
      col = items[i]
    )
    cat_match <- map_lgl(vars, ~ .x == substr(items[i], 1, nchar(.x)))
    if (
      any(cat_match) &&
        any(vars[cat_match] != items[i]) &&
        !(items[i] %in% vars)
    ) {
      cat_match_vars <- vars[cat_match]
      sole_cat_match <- cat_match_vars[rank(-nchar(cat_match_vars))][[1]]
      item <- list(
        type = "conditional",
        col = sole_cat_match,
        val = substr(items[i], nchar(sole_cat_match) + 1, nchar(items[i])),
        op = "equal"
      )
    }
    all_items <- c(all_items, list(item))
  }
  all_items
}

parse_qr_lm <- function(label, qr) {
  qrs <- qr[label == rownames(qr)]
  qrs <- set_names(
    as.list(qrs),
    paste0("qr_", 1:length(qrs))
  )
}

# Intervals -----------------------------------------------

#' @export
tidypredict_interval.lm <- function(model, interval = 0.95) {
  parsedmodel <- parse_model(model)
  te_interval_lm(parsedmodel, interval)
}

get_qr_lm <- function(qr_name, parsedmodel) {
  q <- map(
    parsedmodel$terms,
    ~ {
      cqr <- .x$qr[qr_name][[1]]

      if (.x$is_intercept == 0) {
        cols <- map(.x$fields, lm_constructor)
        cols <- reduce_multiplication(cols)
        if (cqr != 0) {
          expr_multiplication(cols, cqr)
        }
      } else {
        expr(!!cqr)
      }
    }
  )
  f <- reduce_addition(q[!map_lgl(q, is.null)])

  expr(((!!f)) * ((!!f)) * !!parsedmodel$general$sigma2)
}

te_interval_lm <- function(parsedmodel, interval = 0.95) {
  qr_names <- names(parsedmodel$terms[[1]]$qr)
  qrs_map <- map(
    qr_names,
    ~ get_qr_lm(.x, parsedmodel)
  )
  qrs <- reduce_addition(qrs_map)
  tfrac <- qt(1 - (1 - 0.95) / 2, parsedmodel$general$residual)
  expr(!!tfrac * sqrt((!!qrs) + (!!parsedmodel$general$sigma2)))
}

# Helpers -------------------------------------------------

lm_constructor <- function(x) {
  f <- NULL
  if (x$type == "ordinary") {
    f <- expr(!!as.name(x$col))
  }
  if (x$type == "conditional") {
    f <- expr(ifelse(!!as.name(x$col) == !!x$val, 1, 0))
  }
  if (x$type == "operation") {
    if (x$op == "morethan") {
      f <- expr(ifelse(
        !!as.name(x$col) > !!x$val,
        !!as.name(x$col) - !!x$val,
        0
      ))
    }
    if (x$op == "lessthan") {
      f <- expr(ifelse(
        !!as.name(x$col) < !!x$val,
        !!x$val - !!as.name(x$col),
        0
      ))
    }
  }
  f
}

---
title: "Float precision at split boundaries"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Float precision at split boundaries}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

## The issue

Tree-based models like XGBoost, LightGBM, and CatBoost internally convert data to 32-bit floats during training. This means that split thresholds are chosen based on 32-bit precision values. However, R uses 64-bit doubles by default, and most databases also use higher precision floating-point numbers.

This precision mismatch can cause predictions to differ when a data point falls exactly on or very close to a split boundary. The 32-bit and 64-bit representations of the same number may round differently, causing the data point to go left in one system and right in another.

## Which models are affected?

XGBoost and Cubist store everything as 32-bit floats, making them most susceptible to this issue. LightGBM and CatBoost use 64-bit doubles for leaf values, which reduces (but does not eliminate) the risk.

## Example

Here is a real example from a Cubist model. When we extract the split values used in the model's rules, we see values like:

```
variable     value
lstat    9.5299997
rm       6.2259998
rm       6.546
lstat    5.3899999
```

These split values should correspond to actual values in the training data. But when we check, only one of the four matches exactly:

```
# Exact matches
variable value
rm       6.546

# Non-matches
variable     value
lstat    9.5299997
rm       6.2259998
lstat    5.3899999
```

If we look for nearby values in the training data:

```
variable value_data value_split
rm            6.226   6.2259998
rm            6.546   6.546
lstat         5.39    5.3899999
lstat         9.53    9.5299997
```

The original training values were `6.226`, `5.39`, and `9.53`, but they were converted to 32-bit floats during model training, resulting in slightly different stored thresholds.

Why does this matter? Consider a model with two rules:

```
rule 1: rm > 6.2259998
rule 2: rm <= 6.2259998
```

If you pass in an observation where `rm` is `6.226`, you might expect rule 1 to apply since `6.226 > 6.2259998`. But the native model applies rule 2 because it internally converted `6.226` to `6.2259998` during training, making them equal.

## What tidypredict does

tidypredict extracts split thresholds from the model and uses them in R formulas or SQL queries. Since R and databases typically use 64-bit floats, the comparisons are done at 64-bit precision against thresholds that were originally determined at 32-bit precision.

In most cases, this works fine because data points rarely fall exactly on split boundaries. However, you should always verify predictions match using `tidypredict_test()`.

## Pros and cons

Pros of using tidypredict despite this issue:

- In-database scoring avoids moving large datasets out of the database
- SQL translation is portable across database systems
- For most rows, predictions will match exactly

Risks:

- A small fraction of predictions may differ unpredictably
- For classification, boundary cases could flip the predicted class
- It is hard to know in advance which specific rows will be affected

When are values likely to hit boundaries?

- Integer or rounded data (e.g., age in whole years, prices rounded to cents)
- Data with limited unique values or repeated measurements
- Scoring the same data used for training

Continuous or high-precision real-world measurements are less likely to land exactly on split boundaries.

Considerations:

- Use `tidypredict_test()` to quantify the discrepancy rate on your data
- The magnitude of difference is usually small (just a neighboring leaf value)
- High-stakes applications (medical, financial) may need stricter validation

## Recommendations

1. Accept small differences: For production use, consider that a tiny fraction of predictions may differ at exact boundaries. Decide if this is acceptable for your use case.

2. Use native predictions when possible: For applications where perfect agreement is critical, consider using the native model's predict function rather than SQL translation.

---
title: "How tidypredict generates tree formulas"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{How tidypredict generates tree formulas}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

library(dplyr)
library(tidypredict)
library(rpart)
set.seed(100)
```

This vignette explains how tidypredict converts decision tree models into dplyr formulas, and why we chose nested `case_when()` expressions over flat ones.

## Nested vs flat case_when

Consider a simple decision tree:

```
           +-------+
      +----|x <= 5 |----+
      |    +-------+    |
      v                 v
  +-------+           "high"
  |y <= 3 |
  +-------+
   |     |
   v     v
 "low" "med"
```

This tree has three leaves with predictions "low", "med", and "high".

### Flat case_when (old approach)

The flat approach lists every leaf path as a separate condition:

```r
case_when(
  x <= 5 & y <= 3 ~ "low",
  x <= 5 & y > 3  ~ "med",
  x > 5           ~ "high"
)
```

Each condition must encode the **entire path** from root to leaf. For a tree with depth `d`, each condition can have up to `d` comparisons joined by `&`.

### Nested case_when (current approach)

The nested approach mirrors the tree structure:

```r
case_when(
  x <= 5 ~ case_when(
    y <= 3 ~ "low",
    .default = "med"
  ),
  .default = "high"
)
```

Each node becomes its own `case_when()`, with the left branch as the condition and the right branch as `.default`.

### Why nested is better

**1. Fewer comparisons at runtime**

With flat `case_when`, R evaluates conditions sequentially until one matches. In the worst case (the last leaf), all conditions are checked. Each condition re-evaluates splits that were already decided higher in the tree.

With nested `case_when`, each split is evaluated exactly once. The `.default` clause handles the "else" branch without re-checking the condition.

**2. More efficient SQL**

SQL databases optimize nested `CASE WHEN` statements better than flat ones with compound `AND` conditions. The nested structure allows the query planner to short-circuit evaluation.

Flat SQL:
```sql
CASE
  WHEN x <= 5 AND y <= 3 THEN 'low'
  WHEN x <= 5 AND y > 3 THEN 'med'
  WHEN x > 5 THEN 'high'
END
```

Nested SQL:
```sql
CASE
  WHEN x <= 5 THEN
    CASE
      WHEN y <= 3 THEN 'low'
      ELSE 'med'
    END
  ELSE 'high'
END
```

**3. Smaller formula size**

For a balanced tree of depth `d` with `2^d` leaves:

- Flat: Each leaf condition has `d` terms, so total terms = `d * 2^d`
- Nested: Each split appears once, so total terms = `2^d - 1`

For a tree of depth 2 (4 leaves):

- Flat: 2 * 4 = 8 comparison terms
- Nested: 4 - 1 = 3 comparison terms

For a tree of depth 6 (64 leaves):

- Flat: 6 * 64 = 384 comparison terms
- Nested: 64 - 1 = 63 comparison terms

## Parsed model versions

tidypredict uses a version number in parsed models to track format changes:
- **Version 1-2**: Used flat `case_when()` for trees
- **Version 3**: Uses nested `case_when()` (current)

When loading a model saved with an older version, tidypredict automatically uses the appropriate formula builder for backwards compatibility. See `?parse_model` for details.

---
title: "Random Forest, using Ranger"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Random Forest, using Ranger}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

library(dplyr)
library(tidypredict)
library(parsnip)
library(ranger)
set.seed(100)
```

| Function                                                      |Works|
|---------------------------------------------------------------|-----|
|`tidypredict_fit()`, `tidypredict_sql()`, `parse_model()`      |  ✔  |
|`tidypredict_to_column()`                                      |  ✗  |
|`tidypredict_test()`                                           |  ✔  |
|`tidypredict_interval()`, `tidypredict_sql_interval()`         |  ✗  |
|`parsnip`                                                      |  ✔  |

## How it works

Here is a simple `ranger()` model using the `mtcars` dataset:
```{r}
library(dplyr)
library(tidypredict)
library(ranger)

model <- ranger(mpg ~ ., data = mtcars, num.trees = 5, max.depth = 2)
```

## Under the hood

The parser is based on the output from the `ranger::treeInfo()` function.  It will return as many decision paths as there are non-NA rows in the `prediction` field.
```{r}
treeInfo(model) %>%
  head()
```


The output from `parse_model()` is transformed into a `dplyr`, a.k.a Tidy Eval, formula. Each decision tree becomes one `dplyr::case_when()` statement, which are then combined.

```{r}
tidypredict_fit(model)
```

From there, the Tidy Eval formula can be used anywhere where it can be operated. `tidypredict` provides three paths:

  - Use directly inside `dplyr`,  `mutate(iris, !! tidypredict_fit(model))`
  - Use `tidypredict_to_column(model)` to a piped command set
  - Use `tidypredict_to_sql(model)` to retrieve the SQL statement


## parsnip

`tidypredict` also supports `ranger` model objects fitted via the `parsnip` package.

```{r}
library(parsnip)

parsnip_model <- rand_forest(mode = "regression", trees = 5) %>%
  set_engine("ranger", max.depth = 2) %>%
  fit(mpg ~ ., data = mtcars)

tidypredict_fit(parsnip_model)
```

---
title: "Create a regression spec - version 2"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Create a regression spec - version 2}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r, echo = FALSE}
spec_frame <- read.csv("regression.csv")
knitr::kable(spec_frame, row.names = FALSE)
```

---
title: "XGBoost models"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{XGBoost models}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
if (requireNamespace("xgboost", quietly = TRUE)) {
  library(tidypredict)
  library(xgboost)
  library(dplyr)
  eval_code <- TRUE
} else {
  eval_code <- FALSE
}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = eval_code
)
source("_threads.R")
```

| Function                                                      |Works|
|---------------------------------------------------------------|-----|
|`tidypredict_fit()`, `tidypredict_sql()`, `parse_model()`      |  ✔  |
|`tidypredict_to_column()`                                      |  ✔  |
|`tidypredict_test()`                                           |  ✔  |
|`tidypredict_interval()`, `tidypredict_sql_interval()`         |  ✗  |
|`parsnip`                                                      |  ✔  |

## `tidypredict_` functions

```{r}
library(xgboost)

logregobj <- function(preds, dtrain) {
  labels <- xgboost::getinfo(dtrain, "label")
  preds <- 1 / (1 + exp(-preds))
  grad <- preds - labels
  hess <- preds * (1 - preds)
  return(list(grad = grad, hess = hess))
}

xgb_bin_data <- xgboost::xgb.DMatrix(
  as.matrix(mtcars[, -9]),
  label = mtcars$am
)

model <- xgboost::xgb.train(
  params = list(max_depth = 2, objective = "binary:logistic", base_score = 0.5),
  data = xgb_bin_data,
  nrounds = 50
)
```

- Create the R formula
    ```{r}
tidypredict_fit(model)
    ```

- Add the prediction to the original table
    ```{r}
library(dplyr)

mtcars %>%
  tidypredict_to_column(model) %>%
  glimpse()
    ```

- Confirm that `tidypredict` results match to the model's `predict()` results. The `xg_df` argument expects the `xgb.DMatrix` data set.
    ```{r}
tidypredict_test(model, mtcars, xg_df = xgb_bin_data)
    ```

Please be aware that XGBoost converts data into 32-bit floats internally. This could lead to prediction discrepancies at exact split boundaries. Always verify that predictions match using `tidypredict_test()`. See the [float precision](float-precision.html) article for more details.

## parsnip

`parsnip` fitted models are also supported by `tidypredict`:
```{r}
library(parsnip)

p_model <- boost_tree(mode = "regression") %>%
  set_engine("xgboost") %>%
  fit(am ~ ., data = mtcars)
```


```{r}
tidypredict_test(p_model, mtcars, xg_df = xgb_bin_data)
```

## Parse model spec

Here is an example of the model spec:
```{r}
pm <- parse_model(model)
str(pm, 2)
```

```{r}
str(pm$trees[1])
```

---
title: "Generalized Linear Regression"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Generalized Linear Regression}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

library(dplyr)
library(tidypredict)
```

## Highlights & Limitations

- Defaults to 0-to-1 predictions for `binomial` family models. That is akin to running `predict(model, type = "response")`
- Only *treatment* contrast (`contr.treatment`) are supported.
- `offset` is supported
- Categorical variables are supported
- In-line functions in the formulas are **not supported**:  
     - OK - `wt ~ mpg + am` 
     - OK - `mutate(mtcars, newam = paste0(am))` and then `wt ~ mpg + newam`
     - Not OK - `wt ~ mpg + as.factor(am)`
     - Not OK - `wt ~ mpg + as.character(am)`
- Interval functions are not supported: `tidypredict_interval()` & `tidypredict_sql_interval()`

## How it works

```{r}
library(tidypredict)
library(dplyr)

df <- mtcars %>%
  mutate(char_cyl = paste0("cyl", cyl)) %>%
  select(wt, char_cyl, am)

model <- glm(am ~ wt + char_cyl, data = df, family = "binomial")
```

It returns a SQL query that contains the coefficients (`model$coefficients`) operated against the correct variable or categorical variable value.  In most cases the resulting SQL is one short `CASE WHEN` statement per coefficient.  It appends the `offset` field or value, if one is provided.

For `binomial` models, the [sigmoid](https://en.wikipedia.org/wiki/Sigmoid_function) equation is applied. This means that the target SQL database type will need to support the exponent function. 

```{r}
library(tidypredict)
tidypredict_sql(model, dbplyr::simulate_mssql())
```

Alternatively, use `tidypredict_to_column()` if the results are the be used or previewed in `dplyr`.

```{r}
df %>%
  tidypredict_to_column(model) %>%
  head(10)
```

## Under the hood

The parser reads several parts of the `glm` object to tabulate all of the needed variables.  One entry per coefficient is added to the final table. Other variables are added at the end. Some variables are not required for every parsed model.  For example, `offset` is listed because it's part of the formula (call) of the model, if there were no offset in a given model, that line would not exist.

```{r}
pm <- parse_model(model)
str(pm, 2)
```

The output from `parse_model()` is transformed into a `dplyr`, a.k.a Tidy Eval, formula.  All categorical variables are operated using `if_else()`.
```{r}
tidypredict_fit(model)
```

From there, the Tidy Eval formula can be used anywhere where it can be operated. `tidypredict` provides three paths:

  - Use directly inside `dplyr`,  `mutate(df, !! tidypredict_fit(model))`
  - Use `tidypredict_to_column(model)` to a piped command set
  - Use `tidypredict_to_sql(model)` to retrieve the SQL statement

The same applies to the prediction interval functions.


## How it performs


Testing the `tidypredict` results is easy.  The `tidypredict_test()` function automatically uses the `lm` model object's data frame, to compare `tidypredict_fit()`, and `tidypredict_interval()` to the results given by `predict()`

```{r}
tidypredict_test(model)
```

---
title: "glmnet models"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{glmnet models}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
if (requireNamespace("glmnet", quietly = TRUE)) {
  library(tidypredict)
  library(glmnet)
  library(dplyr)
  eval_code <- TRUE
} else {
  eval_code <- FALSE
}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = eval_code
)
source("_threads.R")
```

| Function                                                      |Works|
|---------------------------------------------------------------|-----|
|`tidypredict_fit()`, `tidypredict_sql()`, `parse_model()`      |  ✔  |
|`tidypredict_to_column()`                                      |  ✔  |
|`tidypredict_test()`                                           |  ✔  |
|`tidypredict_interval()`, `tidypredict_sql_interval()`         |  ✗  |
|`parsnip`                                                      |  ✔  |

## `tidypredict_` functions

```{r}
library(glmnet)

model <- glmnet::glmnet(mtcars[, -1], mtcars$mpg, lambda = 1)
```

- Create the R formula
    ```{r}
tidypredict_fit(model)
    ```

- Add the prediction to the original table
    ```{r}
library(dplyr)

mtcars %>%
  tidypredict_to_column(model) %>%
  glimpse()
    ```

- Confirm that `tidypredict` results match to the model's `predict()` results.
    ```{r}
tidypredict_test(model, mtcars[, -1])
    ```

## parsnip

`parsnip` fitted models are also supported by `tidypredict`:
```{r}
library(parsnip)

p_model <- linear_reg(penalty = 1) %>%
  set_engine("glmnet") %>%
  fit(mpg ~ ., data = mtcars)
```

```{r}
tidypredict_fit(p_model)
```

## Parse model spec

Here is an example of the model spec:
```{r}
pm <- parse_model(model)
str(pm, 2)
```

```{r}
str(pm$trees[1])
```

---
title: "Linear Regression"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Linear Regression}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

library(dplyr)
library(tidypredict)
library(parsnip)
```

## Highlights & Limitations

- **Supports prediction intervals**, it uses the `qr.solve()` function to parse the interval coefficient of each term.
- Supports categorical variables and interactions
- Only *treatment* contrast (`contr.treatment`) are supported.
- `offset` is supported
- Categorical variables are supported
- Interactions with `:` and `*` are supported
- In-line functions in the formulas are **not supported**:  
     - OK - `wt ~ mpg + am` 
     - OK - `wt ~ mpg * am + vs:disp` 
     - OK - `mutate(mtcars, newam = paste0(am))` and then `wt ~ mpg + newam`
     - Not OK - `wt ~ mpg + as.factor(am)`
     - Not OK - `wt ~ mpg + as.character(am)`

## How it works

```{r}
library(dplyr)
library(tidypredict)

df <- mtcars %>%
  mutate(char_cyl = paste0("cyl", cyl)) %>%
  select(mpg, wt, char_cyl, am)

model <- lm(mpg ~ wt + char_cyl, offset = am, data = df)
```

It returns a SQL query that contains the coefficients (`model$coefficients`) operated against the correct variable or categorical variable value.  In most cases the resulting SQL is one short `CASE WHEN` statement per coefficient.  It appends the `offset` field or value, if one is provided.
```{r}
library(tidypredict)
tidypredict_sql(model, dbplyr::simulate_mssql())
```

Alternatively, use `tidypredict_to_column()` if the results are the be used or previewed in `dplyr`.

```{r}
df %>%
  tidypredict_to_column(model) %>%
  head(10)
```

## Prediction intervals

Use `tidypredict_sql_interval()` to get the SQL query that operates the prediction interval.  The `interval` defaults to 0.95
```{r}
tidypredict_sql_interval(model, dbplyr::simulate_mssql())
```

Prediction intervals also works in the `tidypredict_to_column()`, just set the `add_interval` argument to `TRUE`.
```{r}
df %>%
  tidypredict_to_column(model, add_interval = TRUE) %>%
  head(10)
```

## Under the hood

The parser reads several parts of the `lm` object to tabulate all of the needed variables.  One entry per coefficient is added to the final table, those entries will have the results of `qr.solve()` already operated and placed in the correct column, they will have a `qr_` prefix.  There will be one `qr_` column per coefficient.  

Other variables are added at the end. Some variables are not required for every parsed model.  For example, `offset` is listed because it's part of the formula (call) of the model, if there were no offset in a given model, that line would not exist.

```{r}
pm <- parse_model(model)
str(pm, 2)
```

The output from `parse_model()` is transformed into a `dplyr`, a.k.a Tidy Eval, formula.  All categorical variables are operated using `if_else()`.
```{r}
tidypredict_fit(model)
```

A function to put together the Tidy Eval interval formula is also supported
```{r}
tidypredict_interval(model)
```

From there, the Tidy Eval formula can be used anywhere where it can be operated. `tidypredict` provides three paths:

  - Use directly inside `dplyr`,  `mutate(df, !! tidypredict_fit(model))`
  - Use `tidypredict_to_column(model)` to a piped command set
  - Use `tidypredict_to_sql(model)` to retrieve the SQL statement

The same applies to the prediction interval functions.

## How it performs

Testing the `tidypredict` results is easy.  The `tidypredict_test()` function automatically uses the `lm` model object's data frame, to compare `tidypredict_fit()`, and `tidypredict_interval()` to the results given by `predict()`

```{r}
tidypredict_test(model)
```

To run with prediction intervals set the `include_intervals` argument to `TRUE`

```{r}
tidypredict_test(model, include_intervals = TRUE)
```

## parsnip

`tidypredict` also supports `lm()` model objects fitted via the `parsnip` package.

```{r}
library(parsnip)

parsnip_model <- linear_reg() %>%
  set_engine("lm") %>%
  fit(mpg ~ wt + cyl, offset = am, data = mtcars)

tidypredict_fit(parsnip_model)
```

---
title: "Cubist models"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Cubist models}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
library(tidypredict)
library(Cubist)
library(dplyr)
```


| Function                                                      |Works|
|---------------------------------------------------------------|-----|
|`tidypredict_fit()`, `tidypredict_sql()`, `parse_model()`      |  ✔  |
|`tidypredict_to_column()`                                      |  ✔  |
|`tidypredict_test()`                                           |  ✗  |
|`tidypredict_interval()`, `tidypredict_sql_interval()`         |  ✗  |
|`parsnip`                                                      |  ✗  |


## `tidypredict_` functions

```{r}
library(Cubist)
data("BostonHousing", package = "mlbench")

model <- Cubist::cubist(
  x = BostonHousing[, -14],
  y = BostonHousing$medv,
  committees = 3
)
```

- Create the R formula
    ```{r}
tidypredict_fit(model)
    ```

- SQL output example
    ```{r}
tidypredict_sql(model, dbplyr::simulate_odbc())
    ```

- Add the prediction to the original table
    ```{r}
library(dplyr)

BostonHousing %>%
  tidypredict_to_column(model) %>%
  glimpse()
    ```

We are not able to give an exact match of the original predictions [due to a minor bug](https://github.com/topepo/Cubist/issues/62) in Cubist.

## Parse model spec

Here is an example of the model spec:
```{r}
pm <- parse_model(model)
str(pm, 2)
```

```{r}
str(pm$terms[1:2])
```

## Limitations

- `tidypredict_test()` is not supported
- Prediction intervals are not supported
- Cubist uses 32-bit floats internally, which may cause prediction discrepancies at exact split boundaries. See the [float precision](float-precision.html) article for details.

---
title: "catboost models"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{CatBoost models}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
if (FALSE) {
  library(tidypredict)
  library(dplyr)
  eval_code <- TRUE
} else {
  eval_code <- FALSE
}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = eval_code
)
```

| Function                                                      |Works|
|---------------------------------------------------------------|-----|
|`tidypredict_fit()`, `tidypredict_sql()`, `parse_model()`      |
|`tidypredict_to_column()`                                      |
|`tidypredict_test()`                                           |
|`tidypredict_interval()`, `tidypredict_sql_interval()`         |
|`parsnip`                                                      |

## `tidypredict_` functions

```r
library(catboost)
# Prepare data
X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
y <- mtcars$hp

pool <- catboost.load_pool(
  X,
  label = y,
  feature_names = as.list(c("mpg", "cyl", "disp"))
)

model <- catboost.train(
  pool,
  params = list(
    iterations = 10L,
    depth = 3L,
    learning_rate = 0.5,
    loss_function = "RMSE",
    logging_level = "Silent",
    allow_writing_files = FALSE
  )
)
```

- Create the R formula
    ```r
tidypredict_fit(model)
    ```

- Add the prediction to the original table
    ```r
library(dplyr)

mtcars %>%
  tidypredict_to_column(model) %>%
  glimpse()
    ```

- Confirm that `tidypredict` results match to the model's `predict()` results. The `xg_df` argument expects the matrix data set.
    ```r
tidypredict_test(model, xg_df = X)
    ```

## Supported objectives

CatBoost supports many objective functions. The following objectives are supported by `tidypredict`:

### Regression objectives (identity transform)

- `RMSE` (default)
- `MAE`
- `Quantile`
- `MAPE`
- `Poisson`

### Binary classification (sigmoid transform)

- `Logloss`
- `CrossEntropy`

### Multiclass classification

- `MultiClass` (softmax transform)
- `MultiClassOneVsAll` (sigmoid per class)

## Binary classification example

```r
X_bin <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
y_bin <- mtcars$am

pool_bin <- catboost.load_pool(
  X_bin,
  label = y_bin,
  feature_names = as.list(c("mpg", "cyl", "disp"))
)

model_bin <- catboost.train(
  pool_bin,
  params = list(
    iterations = 10L,
    depth = 3L,
    learning_rate = 0.5,
    loss_function = "Logloss",
    logging_level = "Silent",
    allow_writing_files = FALSE
  )
)

tidypredict_test(model_bin, xg_df = X_bin)
```

## Multiclass classification example

```r
X_multi <- data.matrix(iris[, 1:4])
y_multi <- as.integer(iris$Species) - 1L

pool_multi <- catboost.load_pool(
  X_multi,
  label = y_multi,
  feature_names = as.list(colnames(iris)[1:4])
)

model_multi <- catboost.train(
  pool_multi,
  params = list(
    iterations = 10L,
    depth = 3L,
    learning_rate = 0.5,
    loss_function = "MultiClass",
    logging_level = "Silent",
    allow_writing_files = FALSE
  )
)

# Multiclass returns a list of formulas, one per class
formulas <- tidypredict_fit(model_multi)
names(formulas)
```

Test multiclass predictions:

```r
tidypredict_test(model_multi, xg_df = X_multi)
```

## Categorical features

CatBoost models can use categorical features with one-hot encoding.

### With parsnip/bonsai (recommended)

When using parsnip/bonsai, categorical features are handled automatically:

```r
library(parsnip)
library(bonsai)

df_cat <- data.frame(
  num_feat = mtcars$mpg,
  cat_feat = factor(ifelse(mtcars$am == 1, "manual", "auto")),
  target = mtcars$hp
)

model_spec <- boost_tree(trees = 10, tree_depth = 3) |>
  set_engine("catboost", logging_level = "Silent", one_hot_max_size = 10) |>
  set_mode("regression")

model_fit <- fit(model_spec, target ~ num_feat + cat_feat, data = df_cat)

# Categorical features are handled automatically
tidypredict_fit(model_fit)
```

### With raw CatBoost

For raw CatBoost models, you need to manually establish the hash-to-category mapping:

```r
pool_cat <- catboost.load_pool(
  df_cat[, c("num_feat", "cat_feat")],
  label = df_cat$target
)

model_cat <- catboost.train(
  pool_cat,
  params = list(
    iterations = 10L,
    depth = 3L,
    learning_rate = 0.5,
    loss_function = "RMSE",
    logging_level = "Silent",
    allow_writing_files = FALSE,
    one_hot_max_size = 10
  )
)

# Parse and set category mapping manually
pm_cat <- parse_model(model_cat)
pm_cat <- set_catboost_categories(pm_cat, model_cat, df_cat)

# Now use the parsed model
tidypredict_fit(pm_cat)
```

## Parse model spec

Here is an example of the model spec:
```r
pm <- parse_model(model)
str(pm, 2)
```

```r
str(pm$trees[1])
```

## Limitations

- Prediction intervals are not supported
- CatBoost uses 32-bit floats for split thresholds, which may cause prediction discrepancies at exact split boundaries. See the [float precision](float-precision.html) article for details.

---
title: "XGBoost models"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{XGBoost models}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
if (requireNamespace("xgboost", quietly = TRUE)) {
  library(tidypredict)
  library(xgboost)
  library(dplyr)
  eval_code <- TRUE
} else {
  eval_code <- FALSE
}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = eval_code
)
```

| Function                                                      |Works|
|---------------------------------------------------------------|-----|
|`tidypredict_fit()`, `tidypredict_sql()`, `parse_model()`      |  ✔  |
|`tidypredict_to_column()`                                      |  ✔  |
|`tidypredict_test()`                                           |  ✔  |
|`tidypredict_interval()`, `tidypredict_sql_interval()`         |  ✗  |
|`parsnip`                                                      |  ✔  |

## `tidypredict_` functions

```{r}
library(xgboost)

logregobj <- function(preds, dtrain) {
  labels <- xgboost::getinfo(dtrain, "label")
  preds <- 1 / (1 + exp(-preds))
  grad <- preds - labels
  hess <- preds * (1 - preds)
  return(list(grad = grad, hess = hess))
}

xgb_bin_data <- xgboost::xgb.DMatrix(
  as.matrix(mtcars[, -9]),
  label = mtcars$am
)

model <- xgboost::xgb.train(
  params = list(max_depth = 2, objective = "binary:logistic", base_score = 0.5),
  data = xgb_bin_data,
  nrounds = 50
)
```

- Create the R formula
    ```{r}
tidypredict_fit(model)
    ```

- Add the prediction to the original table
    ```{r}
library(dplyr)

mtcars %>%
  tidypredict_to_column(model) %>%
  glimpse()
    ```

- Confirm that `tidypredict` results match to the model's `predict()` results. The `xg_df` argument expects the `xgb.DMatrix` data set.
    ```{r}
tidypredict_test(model, mtcars, xg_df = xgb_bin_data)
    ```

Please be aware that XGBoost converts data into 32-bit floats internally. This could lead to prediction discrepancies at exact split boundaries. Always verify that predictions match using `tidypredict_test()`. See the [float precision](float-precision.html) article for more details.

## parsnip

`parsnip` fitted models are also supported by `tidypredict`:
```{r}
library(parsnip)

p_model <- boost_tree(mode = "regression") %>%
  set_engine("xgboost") %>%
  fit(am ~ ., data = mtcars)
```


```{r}
tidypredict_test(p_model, mtcars, xg_df = xgb_bin_data)
```

## Parse model spec

Here is an example of the model spec:
```{r}
pm <- parse_model(model)
str(pm, 2)
```

```{r}
str(pm$trees[1])
```

---
title: "MARS models via the `earth` package"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{MARS models via the `earth` package}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
library(tidypredict)
library(earth)
library(dplyr)
```

| Function                                                      |Works|
|---------------------------------------------------------------|-----|
|`tidypredict_fit()`, `tidypredict_sql()`, `parse_model()`      |  ✔  |
|`tidypredict_to_column()`                                      |  ✔  |
|`tidypredict_test()`                                           |  ✔  |
|`tidypredict_interval()`, `tidypredict_sql_interval()`         |  ✗  |
|`parsnip`                                                      |  ✔  |


## `tidypredict_` functions

```{r}
library(earth)
data("etitanic", package = "earth")

model <- earth(age ~ sibsp + parch, data = etitanic, degree = 3)
```

- Create the R formula
    ```{r}
tidypredict_fit(model)
    ```

- SQL output example
    ```{r}
tidypredict_sql(model, dbplyr::simulate_odbc())
    ```

- Add the prediction to the original table
    ```{r}
library(dplyr)

etitanic %>%
  tidypredict_to_column(model) %>%
  glimpse()
    ```

- Confirm that `tidypredict` results match to the model's `predict()` results
    ```{r}
tidypredict_test(model, etitanic)
    ```

## GLM models

`tidypredict` supports the `glm` argument as well:
```{r}
model <- earth(survived ~ .,
  data = etitanic,
  glm = list(family = binomial), degree = 2
)

tidypredict_fit(model)
```

The spec sets the `is_glm` entry to 1, as well as the `family` and `link` entries. 
```{r}
str(parse_model(model), 2)
```

## parsnip

`parsnip` fitted models are also supported by `tidypredict`:
```{r}
library(parsnip)

p_model <- mars(mode = "regression", prod_degree = 3) %>%
  set_engine("earth") %>%
  fit(age ~ sibsp + parch, data = etitanic)

tidypredict_fit(p_model)
```

## Parse model spec

Here is an example of the model spec:
```{r}
pm <- parse_model(model)
str(pm, 2)
```

```{r}
str(pm$terms[1:2])
```

---
title: "Non-R Models"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Non-R Models}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

A model that is trained in any language are able to integrate with `tidypredict`, and thus with `broom`.  The requirement is that the model in that language is exported using the parse model spec.  The easiest file format would be YAML. 

## python example

A model that was fitted using `sklearn`'s `linear_model`. The model is based on diabetes data. Ten baseline variables, age, sex, body mass index, average blood pressure, and six blood serum measurements were obtained for each of n = 442 diabetes patients, as well as the response of interest, a quantitative measure of disease progression one year after baseline. The model's results were converted to YAML by the same python script, I copied and pasted the top part here:

```
general:
  is_glm: 0
  model: lm
  residual: 0
  sigma2: 0
  type: regression
  version: 2.0
terms:
- coef: 152.76430691633442
  fields:
  - col: (Intercept)
    type: ordinary
  is_intercept: 1
  label: (Intercept)
```

## Read in R

The YAML data can be read in R by using the `yaml` package.  In this example, we have copy-pasted most of the models inside a variable called `sklearn_model`. Because `yaml` requires local YAML variables to be split by line, we use `strsplit()`.

```{r}
library(yaml)

sklearn_model <- strsplit("general:
  is_glm: 0
  model: lm
  residual: 0
  sigma2: 0
  type: regression
  version: 2.0
terms:
- coef: 152.76430691633442
  fields:
  - col: (Intercept)
    type: ordinary
  is_intercept: 1
  label: (Intercept)
- coef: 0.3034995490660432
  fields:
  - col: age
    type: ordinary
  is_intercept: 0
  label: age
- coef: -237.63931533353403
  fields:
  - col: sex
    type: ordinary
  is_intercept: 0
  label: sex
- coef: 510.5306054362253
  fields:
  - col: bmi
    type: ordinary
  is_intercept: 0
  label: bmi
- coef: 327.7369804093466
  fields:
  - col: bp
    type: ordinary
  is_intercept: 0
  label: bp
- coef: -814.1317093725387
  fields:
  - col: s1
    type: ordinary
  is_intercept: 0
  label: s1
", split = "\n")[[1]]
```

Now the model is converted to an R `list` using `yaml.load`.
```{r}
sklearn_model <- yaml.load(sklearn_model)

str(sklearn_model, 2)
```

## `tidypredict`

The `list` object needs to be recognized as a `tidypredict` parsed model.  To do that, we use `as_parsed_model()`

```{r}
library(tidypredict)

spm <- as_parsed_model(sklearn_model)

class(spm)
```

The `spm` variable now works just as any parsed model inside R.  Use `tidypredict_fit()` to view the resulting formula.

```{r}
tidypredict_fit(spm)
```

Now, the model can run **inside a database**

```{r}
tidypredict_sql(spm, dbplyr::simulate_mssql())
```

## `broom`

Now that we have a `parsed_model` object, it is possible to use `broom`'s `tidy()` function.  This means that we are able to integrate a totally external model, with `broom`.

```{r}
tidy(spm)
```

---
title: "Save and re-load models"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Save and re-load models}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
library(tidypredict)
library(yaml)
```

`tidypredict` splits the translation process in two.  It first parses the model to extract the needed components to produce the prediction.  And second, it uses the object with the parsed information to produce the R formula. Thanks to this two step process, `tidypredict` does not need to parse the model every time.  `tidypredict`'s functions also accept R objects that contained already models that have been parsed already.  Additionally, because the parsed model object is made up of a list made up of basic variables, it is possible to save it in a file.  Currently, the best file format is YAML.

For this article, we will use the following model:

```{r}
model <- lm(mpg ~ (wt + disp) * cyl, data = mtcars)
```

## Parse model

The `parse_model()` function allows to run the first step manually.  It will return an R list object which contains all of the needed information to produce a prediction calculation.  The structure of the parsed model varies based on what kind of model is being processed.  In general, it is consistent in what kind of information it expects from each model type. For example, in the example the `lm()` model object will return variables such as `sigma2`,  which would not be used in other model types, such as decision trees.

```{r}
library(tidypredict)

parsed <- parse_model(model)
str(parsed, 2)
```

Usually, we pass an R model object to functions such as: `tidypredict_fit()`, and `tidypredict_sql()`.  These functions also accept a previously parsed model.

```{r}
tidypredict_fit(parsed)
```



```{r, include = FALSE}
library(yaml)
model_file <- tempfile(fileext = ".yml")
write_yaml(parsed, model_file)
loaded_model <- read_yaml(model_file)
loaded_model <- as_parsed_model(loaded_model)
```

## Saving the model

Saving the model is quite easy, use the package such as `yaml` to write the model object as a YAML file.  Any format that can persist a ragged list object should work as well. 

```{r, eval = FALSE}
library(yaml)

write_yaml(parsed, "my_model.yml")
```

## Re-load the model

In a new R session, we can read the YAML file into our environment. 

```{r, eval = FALSE}
library(tidypredict)
library(yaml)

loaded_model <- read_yaml("my_model")

loaded_model <- as_parsed_model(loaded_model)
```

The preview of the file looks exactly as the preview of the original parsed model. 

```{r}
str(loaded_model, 2)
```

`tidypredict` is able to read the new R variable and use it to create the formula.  

```{r}
tidypredict_fit(loaded_model)
```

The same variable can be used with other `tidypredict` functions, such as `tidypredict_sql()`

```{r}
tidypredict_sql(loaded_model, dbplyr::simulate_odbc())
```

## `broom`

The `parsed_model` object integrates with `tidy()` from `broom`.  

```{r}
tidy(loaded_model)
```

---
title: "LightGBM models"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{LightGBM models}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
if (requireNamespace("lightgbm", quietly = TRUE)) {
  library(tidypredict)
  library(lightgbm)
  library(dplyr)
  eval_code <- TRUE
} else {
  eval_code <- FALSE
}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = eval_code
)
```

| Function                                                      |Works|
|---------------------------------------------------------------|-----|
|`tidypredict_fit()`, `tidypredict_sql()`, `parse_model()`      |  ✔  |
|`tidypredict_to_column()`                                      |  ✔  |
|`tidypredict_test()`                                           |  ✔  |
|`tidypredict_interval()`, `tidypredict_sql_interval()`         |  ✗  |
|`parsnip`                                                      |  ✔  |

## `tidypredict_` functions

```{r}
library(lightgbm)

# Prepare data
X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
y <- mtcars$hp

dtrain <- lgb.Dataset(X, label = y, colnames = c("mpg", "cyl", "disp"))

model <- lgb.train(
  params = list(
    num_leaves = 4L,
    learning_rate = 0.5,
    objective = "regression",
    min_data_in_leaf = 1L
  ),
  data = dtrain,
  nrounds = 10L,
  verbose = -1L
)
```

- Create the R formula
    ```{r}
tidypredict_fit(model)
    ```

- Add the prediction to the original table
    ```{r}
library(dplyr)

mtcars %>%
  tidypredict_to_column(model) %>%
  glimpse()
    ```

- Confirm that `tidypredict` results match to the model's `predict()` results. The `xg_df` argument expects the matrix data set.
    ```{r}
tidypredict_test(model, xg_df = X)
    ```

## Supported objectives

LightGBM supports many objective functions. The following objectives are supported by `tidypredict`:

### Regression objectives (identity transform)

- `regression` / `regression_l2` (default)
- `regression_l1`
- `huber`
- `fair`
- `quantile`
- `mape`

### Regression objectives (exp transform)

- `poisson`
- `gamma`
- `tweedie`

### Binary classification (sigmoid transform)

- `binary`
- `cross_entropy`

### Multiclass classification

- `multiclass` (softmax transform)
- `multiclassova` (per-class sigmoid)

## Binary classification example

```{r}
X_bin <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
y_bin <- mtcars$am

dtrain_bin <- lgb.Dataset(X_bin, label = y_bin, colnames = c("mpg", "cyl", "disp"))

model_bin <- lgb.train(
  params = list(
    num_leaves = 4L,
    learning_rate = 0.5,
    objective = "binary",
    min_data_in_leaf = 1L
  ),
  data = dtrain_bin,
  nrounds = 10L,
  verbose = -1L
)

tidypredict_test(model_bin, xg_df = X_bin)
```

## Multiclass classification

For multiclass models, `tidypredict_fit()` returns a named list of formulas, one for each class:

```{r}
X_iris <- data.matrix(iris[, 1:4])
colnames(X_iris) <- c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width")
y_iris <- as.integer(iris$Species) - 1L

dtrain_iris <- lgb.Dataset(X_iris, label = y_iris, colnames = colnames(X_iris))

model_multi <- lgb.train(
  params = list(
    num_leaves = 4L,
    learning_rate = 0.5,
    objective = "multiclass",
    num_class = 3L,
    min_data_in_leaf = 1L
  ),
  data = dtrain_iris,
  nrounds = 5L,
  verbose = -1L
)

fit_formulas <- tidypredict_fit(model_multi)
names(fit_formulas)
```

Each formula produces the predicted probability for that class:

```{r}
iris %>%
  mutate(
    prob_setosa = !!fit_formulas$class_0,
    prob_versicolor = !!fit_formulas$class_1,
    prob_virginica = !!fit_formulas$class_2
  ) %>%
  select(Species, starts_with("prob_")) %>%
  head()
```

Note: `tidypredict_test()` does not support multiclass models. Use `tidypredict_fit()` directly.

## Categorical features

LightGBM supports native categorical features. When a feature is marked as categorical, `tidypredict` generates appropriate `%in%` conditions:

```{r}
set.seed(123)
n <- 200
cat_data <- data.frame(
  cat_feat = sample(0:3, n, replace = TRUE),
  y = NA
)
cat_data$y <- ifelse(cat_data$cat_feat %in% c(0, 1), 10, -10) + rnorm(n, sd = 2)

X_cat <- matrix(cat_data$cat_feat, ncol = 1)
colnames(X_cat) <- "cat_feat"

dtrain_cat <- lgb.Dataset(
  X_cat,
  label = cat_data$y,
  categorical_feature = "cat_feat"
)

model_cat <- lgb.train(
  params = list(
    num_leaves = 4L,
    learning_rate = 1.0,
    objective = "regression",
    min_data_in_leaf = 1L
  ),
  data = dtrain_cat,
  nrounds = 2L,
  verbose = -1L
)

tidypredict_fit(model_cat)
```

## parsnip

`parsnip` fitted models (via the `bonsai` package) are also supported by `tidypredict`:

```{r, eval = requireNamespace("parsnip", quietly = TRUE) && requireNamespace("bonsai", quietly = TRUE)}
library(parsnip)
library(bonsai)

p_model <- boost_tree(
  trees = 10,
  tree_depth = 3,
  min_n = 1
) %>%
  set_engine("lightgbm") %>%
  set_mode("regression") %>%
  fit(hp ~ mpg + cyl + disp, data = mtcars)

# Extract the underlying lgb.Booster
lgb_model <- p_model$fit

tidypredict_test(lgb_model, xg_df = X)
```

## Parse model spec

Here is an example of the model spec:
```{r}
pm <- parse_model(model)
str(pm, 2)
```

```{r}
str(pm$trees[1])
```

## Limitations

- Ranking objectives (`lambdarank`, `rank_xendcg`) are not supported
- Prediction intervals are not supported
- `tidypredict_test()` does not support multiclass models
- LightGBM uses 32-bit floats for split thresholds, which may cause prediction discrepancies at exact split boundaries. See the [float precision](float-precision.html) article for details.

---
title: "glmnet models"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{glmnet models}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
if (requireNamespace("glmnet", quietly = TRUE)) {
  library(tidypredict)
  library(glmnet)
  library(dplyr)
  eval_code <- TRUE
} else {
  eval_code <- FALSE
}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = eval_code
)
```

| Function                                                      |Works|
|---------------------------------------------------------------|-----|
|`tidypredict_fit()`, `tidypredict_sql()`, `parse_model()`      |  ✔  |
|`tidypredict_to_column()`                                      |  ✔  |
|`tidypredict_test()`                                           |  ✔  |
|`tidypredict_interval()`, `tidypredict_sql_interval()`         |  ✗  |
|`parsnip`                                                      |  ✔  |

## `tidypredict_` functions

```{r}
library(glmnet)

model <- glmnet::glmnet(mtcars[, -1], mtcars$mpg, lambda = 1)
```

- Create the R formula
    ```{r}
tidypredict_fit(model)
    ```

- Add the prediction to the original table
    ```{r}
library(dplyr)

mtcars %>%
  tidypredict_to_column(model) %>%
  glimpse()
    ```

- Confirm that `tidypredict` results match to the model's `predict()` results.
    ```{r}
tidypredict_test(model, mtcars[, -1])
    ```

## parsnip

`parsnip` fitted models are also supported by `tidypredict`:
```{r}
library(parsnip)

p_model <- linear_reg(penalty = 1) %>%
  set_engine("glmnet") %>%
  fit(mpg ~ ., data = mtcars)
```

```{r}
tidypredict_fit(p_model)
```

## Parse model spec

Here is an example of the model spec:
```{r}
pm <- parse_model(model)
str(pm, 2)
```

```{r}
str(pm$trees[1])
```

---
title: "LightGBM models"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{LightGBM models}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
if (requireNamespace("lightgbm", quietly = TRUE)) {
  library(tidypredict)
  library(lightgbm)
  library(dplyr)
  eval_code <- TRUE
} else {
  eval_code <- FALSE
}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = eval_code
)
source("_threads.R")
```

| Function                                                      |Works|
|---------------------------------------------------------------|-----|
|`tidypredict_fit()`, `tidypredict_sql()`, `parse_model()`      |  ✔  |
|`tidypredict_to_column()`                                      |  ✔  |
|`tidypredict_test()`                                           |  ✔  |
|`tidypredict_interval()`, `tidypredict_sql_interval()`         |  ✗  |
|`parsnip`                                                      |  ✔  |

## `tidypredict_` functions

```{r}
library(lightgbm)

# Prepare data
X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
y <- mtcars$hp

dtrain <- lgb.Dataset(X, label = y, colnames = c("mpg", "cyl", "disp"))

model <- lgb.train(
  params = list(
    num_leaves = 4L,
    learning_rate = 0.5,
    objective = "regression",
    min_data_in_leaf = 1L
  ),
  data = dtrain,
  nrounds = 10L,
  verbose = -1L
)
```

- Create the R formula
    ```{r}
tidypredict_fit(model)
    ```

- Add the prediction to the original table
    ```{r}
library(dplyr)

mtcars %>%
  tidypredict_to_column(model) %>%
  glimpse()
    ```

- Confirm that `tidypredict` results match to the model's `predict()` results. The `xg_df` argument expects the matrix data set.
    ```{r}
tidypredict_test(model, xg_df = X)
    ```

## Supported objectives

LightGBM supports many objective functions. The following objectives are supported by `tidypredict`:

### Regression objectives (identity transform)

- `regression` / `regression_l2` (default)
- `regression_l1`
- `huber`
- `fair`
- `quantile`
- `mape`

### Regression objectives (exp transform)

- `poisson`
- `gamma`
- `tweedie`

### Binary classification (sigmoid transform)

- `binary`
- `cross_entropy`

### Multiclass classification

- `multiclass` (softmax transform)
- `multiclassova` (per-class sigmoid)

## Binary classification example

```{r}
X_bin <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
y_bin <- mtcars$am

dtrain_bin <- lgb.Dataset(X_bin, label = y_bin, colnames = c("mpg", "cyl", "disp"))

model_bin <- lgb.train(
  params = list(
    num_leaves = 4L,
    learning_rate = 0.5,
    objective = "binary",
    min_data_in_leaf = 1L
  ),
  data = dtrain_bin,
  nrounds = 10L,
  verbose = -1L
)

tidypredict_test(model_bin, xg_df = X_bin)
```

## Multiclass classification

For multiclass models, `tidypredict_fit()` returns a named list of formulas, one for each class:

```{r}
X_iris <- data.matrix(iris[, 1:4])
colnames(X_iris) <- c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width")
y_iris <- as.integer(iris$Species) - 1L

dtrain_iris <- lgb.Dataset(X_iris, label = y_iris, colnames = colnames(X_iris))

model_multi <- lgb.train(
  params = list(
    num_leaves = 4L,
    learning_rate = 0.5,
    objective = "multiclass",
    num_class = 3L,
    min_data_in_leaf = 1L
  ),
  data = dtrain_iris,
  nrounds = 5L,
  verbose = -1L
)

fit_formulas <- tidypredict_fit(model_multi)
names(fit_formulas)
```

Each formula produces the predicted probability for that class:

```{r}
iris %>%
  mutate(
    prob_setosa = !!fit_formulas$class_0,
    prob_versicolor = !!fit_formulas$class_1,
    prob_virginica = !!fit_formulas$class_2
  ) %>%
  select(Species, starts_with("prob_")) %>%
  head()
```

Note: `tidypredict_test()` does not support multiclass models. Use `tidypredict_fit()` directly.

## Categorical features

LightGBM supports native categorical features. When a feature is marked as categorical, `tidypredict` generates appropriate `%in%` conditions:

```{r}
set.seed(123)
n <- 200
cat_data <- data.frame(
  cat_feat = sample(0:3, n, replace = TRUE),
  y = NA
)
cat_data$y <- ifelse(cat_data$cat_feat %in% c(0, 1), 10, -10) + rnorm(n, sd = 2)

X_cat <- matrix(cat_data$cat_feat, ncol = 1)
colnames(X_cat) <- "cat_feat"

dtrain_cat <- lgb.Dataset(
  X_cat,
  label = cat_data$y,
  categorical_feature = "cat_feat"
)

model_cat <- lgb.train(
  params = list(
    num_leaves = 4L,
    learning_rate = 1.0,
    objective = "regression",
    min_data_in_leaf = 1L
  ),
  data = dtrain_cat,
  nrounds = 2L,
  verbose = -1L
)

tidypredict_fit(model_cat)
```

## parsnip

`parsnip` fitted models (via the `bonsai` package) are also supported by `tidypredict`:

```{r, eval = requireNamespace("parsnip", quietly = TRUE) && requireNamespace("bonsai", quietly = TRUE)}
library(parsnip)
library(bonsai)

p_model <- boost_tree(
  trees = 10,
  tree_depth = 3,
  min_n = 1
) %>%
  set_engine("lightgbm") %>%
  set_mode("regression") %>%
  fit(hp ~ mpg + cyl + disp, data = mtcars)

# Extract the underlying lgb.Booster
lgb_model <- p_model$fit

tidypredict_test(lgb_model, xg_df = X)
```

## Parse model spec

Here is an example of the model spec:
```{r}
pm <- parse_model(model)
str(pm, 2)
```

```{r}
str(pm$trees[1])
```

## Limitations

- Ranking objectives (`lambdarank`, `rank_xendcg`) are not supported
- Prediction intervals are not supported
- `tidypredict_test()` does not support multiclass models
- LightGBM uses 32-bit floats for split thresholds, which may cause prediction discrepancies at exact split boundaries. See the [float precision](float-precision.html) article for details.

---
title: "Random Forest, using Ranger"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Random Forest, using Ranger}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

source("_threads.R")
library(dplyr)
library(tidypredict)
library(parsnip)
library(ranger)
set.seed(100)
```

| Function                                                      |Works|
|---------------------------------------------------------------|-----|
|`tidypredict_fit()`, `tidypredict_sql()`, `parse_model()`      |  ✔  |
|`tidypredict_to_column()`                                      |  ✗  |
|`tidypredict_test()`                                           |  ✔  |
|`tidypredict_interval()`, `tidypredict_sql_interval()`         |  ✗  |
|`parsnip`                                                      |  ✔  |

## How it works

Here is a simple `ranger()` model using the `mtcars` dataset:
```{r}
library(dplyr)
library(tidypredict)
library(ranger)

model <- ranger(mpg ~ ., data = mtcars, num.trees = 5, max.depth = 2)
```

## Under the hood

The parser is based on the output from the `ranger::treeInfo()` function.  It will return as many decision paths as there are non-NA rows in the `prediction` field.
```{r}
treeInfo(model) %>%
  head()
```


The output from `parse_model()` is transformed into a `dplyr`, a.k.a Tidy Eval, formula. Each decision tree becomes one `dplyr::case_when()` statement, which are then combined.

```{r}
tidypredict_fit(model)
```

From there, the Tidy Eval formula can be used anywhere where it can be operated. `tidypredict` provides three paths:

  - Use directly inside `dplyr`,  `mutate(iris, !! tidypredict_fit(model))`
  - Use `tidypredict_to_column(model)` to a piped command set
  - Use `tidypredict_to_sql(model)` to retrieve the SQL statement


## parsnip

`tidypredict` also supports `ranger` model objects fitted via the `parsnip` package.

```{r}
library(parsnip)

parsnip_model <- rand_forest(mode = "regression", trees = 5) %>%
  set_engine("ranger", max.depth = 2) %>%
  fit(mpg ~ ., data = mtcars)

tidypredict_fit(parsnip_model)
```

---
title: "How tidypredict generates tree formulas"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{How tidypredict generates tree formulas}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

source("_threads.R")
library(dplyr)
library(tidypredict)
library(rpart)
set.seed(100)
```

This vignette explains how tidypredict converts decision tree models into dplyr formulas, and why we chose nested `case_when()` expressions over flat ones.

## Nested vs flat case_when

Consider a simple decision tree:

```
           +-------+
      +----|x <= 5 |----+
      |    +-------+    |
      v                 v
  +-------+           "high"
  |y <= 3 |
  +-------+
   |     |
   v     v
 "low" "med"
```

This tree has three leaves with predictions "low", "med", and "high".

### Flat case_when (old approach)

The flat approach lists every leaf path as a separate condition:

```r
case_when(
  x <= 5 & y <= 3 ~ "low",
  x <= 5 & y > 3  ~ "med",
  x > 5           ~ "high"
)
```

Each condition must encode the **entire path** from root to leaf. For a tree with depth `d`, each condition can have up to `d` comparisons joined by `&`.

### Nested case_when (current approach)

The nested approach mirrors the tree structure:

```r
case_when(
  x <= 5 ~ case_when(
    y <= 3 ~ "low",
    .default = "med"
  ),
  .default = "high"
)
```

Each node becomes its own `case_when()`, with the left branch as the condition and the right branch as `.default`.

### Why nested is better

**1. Fewer comparisons at runtime**

With flat `case_when`, R evaluates conditions sequentially until one matches. In the worst case (the last leaf), all conditions are checked. Each condition re-evaluates splits that were already decided higher in the tree.

With nested `case_when`, each split is evaluated exactly once. The `.default` clause handles the "else" branch without re-checking the condition.

**2. More efficient SQL**

SQL databases optimize nested `CASE WHEN` statements better than flat ones with compound `AND` conditions. The nested structure allows the query planner to short-circuit evaluation.

Flat SQL:
```sql
CASE
  WHEN x <= 5 AND y <= 3 THEN 'low'
  WHEN x <= 5 AND y > 3 THEN 'med'
  WHEN x > 5 THEN 'high'
END
```

Nested SQL:
```sql
CASE
  WHEN x <= 5 THEN
    CASE
      WHEN y <= 3 THEN 'low'
      ELSE 'med'
    END
  ELSE 'high'
END
```

**3. Smaller formula size**

For a balanced tree of depth `d` with `2^d` leaves:

- Flat: Each leaf condition has `d` terms, so total terms = `d * 2^d`
- Nested: Each split appears once, so total terms = `2^d - 1`

For a tree of depth 2 (4 leaves):

- Flat: 2 * 4 = 8 comparison terms
- Nested: 4 - 1 = 3 comparison terms

For a tree of depth 6 (64 leaves):

- Flat: 6 * 64 = 384 comparison terms
- Nested: 64 - 1 = 63 comparison terms

## Parsed model versions

tidypredict uses a version number in parsed models to track format changes:
- **Version 1-2**: Used flat `case_when()` for trees
- **Version 3**: Uses nested `case_when()` (current)

When loading a model saved with an older version, tidypredict automatically uses the appropriate formula builder for backwards compatibility. See `?parse_model` for details.

---
title: "Generalized Linear Regression"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Generalized Linear Regression}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

source("_threads.R")
library(dplyr)
library(tidypredict)
```

## Highlights & Limitations

- Defaults to 0-to-1 predictions for `binomial` family models. That is akin to running `predict(model, type = "response")`
- Only *treatment* contrast (`contr.treatment`) are supported.
- `offset` is supported
- Categorical variables are supported
- In-line functions in the formulas are **not supported**:  
     - OK - `wt ~ mpg + am` 
     - OK - `mutate(mtcars, newam = paste0(am))` and then `wt ~ mpg + newam`
     - Not OK - `wt ~ mpg + as.factor(am)`
     - Not OK - `wt ~ mpg + as.character(am)`
- Interval functions are not supported: `tidypredict_interval()` & `tidypredict_sql_interval()`

## How it works

```{r}
library(tidypredict)
library(dplyr)

df <- mtcars %>%
  mutate(char_cyl = paste0("cyl", cyl)) %>%
  select(wt, char_cyl, am)

model <- glm(am ~ wt + char_cyl, data = df, family = "binomial")
```

It returns a SQL query that contains the coefficients (`model$coefficients`) operated against the correct variable or categorical variable value.  In most cases the resulting SQL is one short `CASE WHEN` statement per coefficient.  It appends the `offset` field or value, if one is provided.

For `binomial` models, the [sigmoid](https://en.wikipedia.org/wiki/Sigmoid_function) equation is applied. This means that the target SQL database type will need to support the exponent function. 

```{r}
library(tidypredict)
tidypredict_sql(model, dbplyr::simulate_mssql())
```

Alternatively, use `tidypredict_to_column()` if the results are the be used or previewed in `dplyr`.

```{r}
df %>%
  tidypredict_to_column(model) %>%
  head(10)
```

## Under the hood

The parser reads several parts of the `glm` object to tabulate all of the needed variables.  One entry per coefficient is added to the final table. Other variables are added at the end. Some variables are not required for every parsed model.  For example, `offset` is listed because it's part of the formula (call) of the model, if there were no offset in a given model, that line would not exist.

```{r}
pm <- parse_model(model)
str(pm, 2)
```

The output from `parse_model()` is transformed into a `dplyr`, a.k.a Tidy Eval, formula.  All categorical variables are operated using `if_else()`.
```{r}
tidypredict_fit(model)
```

From there, the Tidy Eval formula can be used anywhere where it can be operated. `tidypredict` provides three paths:

  - Use directly inside `dplyr`,  `mutate(df, !! tidypredict_fit(model))`
  - Use `tidypredict_to_column(model)` to a piped command set
  - Use `tidypredict_to_sql(model)` to retrieve the SQL statement

The same applies to the prediction interval functions.


## How it performs


Testing the `tidypredict` results is easy.  The `tidypredict_test()` function automatically uses the `lm` model object's data frame, to compare `tidypredict_fit()`, and `tidypredict_interval()` to the results given by `predict()`

```{r}
tidypredict_test(model)
```

---
title: "Linear Regression"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Linear Regression}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

source("_threads.R")
library(dplyr)
library(tidypredict)
library(parsnip)
```

## Highlights & Limitations

- **Supports prediction intervals**, it uses the `qr.solve()` function to parse the interval coefficient of each term.
- Supports categorical variables and interactions
- Only *treatment* contrast (`contr.treatment`) are supported.
- `offset` is supported
- Categorical variables are supported
- Interactions with `:` and `*` are supported
- In-line functions in the formulas are **not supported**:  
     - OK - `wt ~ mpg + am` 
     - OK - `wt ~ mpg * am + vs:disp` 
     - OK - `mutate(mtcars, newam = paste0(am))` and then `wt ~ mpg + newam`
     - Not OK - `wt ~ mpg + as.factor(am)`
     - Not OK - `wt ~ mpg + as.character(am)`

## How it works

```{r}
library(dplyr)
library(tidypredict)

df <- mtcars %>%
  mutate(char_cyl = paste0("cyl", cyl)) %>%
  select(mpg, wt, char_cyl, am)

model <- lm(mpg ~ wt + char_cyl, offset = am, data = df)
```

It returns a SQL query that contains the coefficients (`model$coefficients`) operated against the correct variable or categorical variable value.  In most cases the resulting SQL is one short `CASE WHEN` statement per coefficient.  It appends the `offset` field or value, if one is provided.
```{r}
library(tidypredict)
tidypredict_sql(model, dbplyr::simulate_mssql())
```

Alternatively, use `tidypredict_to_column()` if the results are the be used or previewed in `dplyr`.

```{r}
df %>%
  tidypredict_to_column(model) %>%
  head(10)
```

## Prediction intervals

Use `tidypredict_sql_interval()` to get the SQL query that operates the prediction interval.  The `interval` defaults to 0.95
```{r}
tidypredict_sql_interval(model, dbplyr::simulate_mssql())
```

Prediction intervals also works in the `tidypredict_to_column()`, just set the `add_interval` argument to `TRUE`.
```{r}
df %>%
  tidypredict_to_column(model, add_interval = TRUE) %>%
  head(10)
```

## Under the hood

The parser reads several parts of the `lm` object to tabulate all of the needed variables.  One entry per coefficient is added to the final table, those entries will have the results of `qr.solve()` already operated and placed in the correct column, they will have a `qr_` prefix.  There will be one `qr_` column per coefficient.  

Other variables are added at the end. Some variables are not required for every parsed model.  For example, `offset` is listed because it's part of the formula (call) of the model, if there were no offset in a given model, that line would not exist.

```{r}
pm <- parse_model(model)
str(pm, 2)
```

The output from `parse_model()` is transformed into a `dplyr`, a.k.a Tidy Eval, formula.  All categorical variables are operated using `if_else()`.
```{r}
tidypredict_fit(model)
```

A function to put together the Tidy Eval interval formula is also supported
```{r}
tidypredict_interval(model)
```

From there, the Tidy Eval formula can be used anywhere where it can be operated. `tidypredict` provides three paths:

  - Use directly inside `dplyr`,  `mutate(df, !! tidypredict_fit(model))`
  - Use `tidypredict_to_column(model)` to a piped command set
  - Use `tidypredict_to_sql(model)` to retrieve the SQL statement

The same applies to the prediction interval functions.

## How it performs

Testing the `tidypredict` results is easy.  The `tidypredict_test()` function automatically uses the `lm` model object's data frame, to compare `tidypredict_fit()`, and `tidypredict_interval()` to the results given by `predict()`

```{r}
tidypredict_test(model)
```

To run with prediction intervals set the `include_intervals` argument to `TRUE`

```{r}
tidypredict_test(model, include_intervals = TRUE)
```

## parsnip

`tidypredict` also supports `lm()` model objects fitted via the `parsnip` package.

```{r}
library(parsnip)

parsnip_model <- linear_reg() %>%
  set_engine("lm") %>%
  fit(mpg ~ wt + cyl, offset = am, data = mtcars)

tidypredict_fit(parsnip_model)
```

---
title: "Database write-back"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Database write-back}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r pre, include = FALSE}
if (!rlang::is_installed("randomForest")) {
  knitr::opts_chunk$set(
    eval = FALSE
  )
}
```

```{r setup, include=FALSE}
source("_threads.R")
library(dplyr)
library(tidypredict)
library(randomForest)
library(dbplyr)
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
set.seed(100)
```

This article reviews two possible scenarios for "writing back" predictions to the database, and without importing the data into memory first. Both scenarios share a common model preparation method.

## Example setup

To keep the example reproducible, a SQLite database will be used to simulate a larger scale deployment. 

First the data is prepared in memory.  The article will use the `nycflights13::flights` data, with a couple of modifications
```{r}
library(dplyr)
library(tidypredict)
library(randomForest)
library(dbplyr)

flights_table <- nycflights13::flights %>%
  mutate(
    current_score = 0,
    flight_id = row_number()
  )
```

A new database is created using `RSQLite`. 
```{r}
library(DBI)

con <- dbConnect(RSQLite::SQLite(), path = ":memory:")
db_fligths <- copy_to(con, flights_table)
```

## Model preparation

A sample is downloaded from the database for modeling. This example already selects the needed variables.
```{r}
df <- db_fligths %>%
  select(dep_delay, hour, distance) %>%
  head(1000) %>%
  collect()
```

A linear model is fitted using `lm()`
```{r}
model <- lm(dep_delay ~ ., data = df)
```

It is highly recommendable to always run `tidypredict_test()` to make sure that the predictions are in line with what the `predict()` command returns.

The `lm` and `glm` models contain the data that they were fitted with, so it's easy to run a test.  `tidypredict_test()` uses the model's internal data set by default
```{r}
tidypredict_test(model)
```

In cases where the model re-fitting is automated, the `tidypredict_test()` function returns an `alert` in case the threshold is exceeded, which can be used to fail the automated script. 
```{r, eval = FALSE}
if (tidypredict_test(model)$alert) stop("Threshold exceeded!")
```

## Scenario 1 - Update scores

In this scenario, The table that supplies the term values is also the recipient of the new score.  This is done by updating a specific field in the table.  For this example, the field is called `current_score`.  The following SQL UPDATE statement should work in most databases:
```{r}
library(dbplyr)

update_statement <- build_sql("UPDATE flights_table SET current_score  = ", tidypredict_sql(model, con = con), con = con)

update_statement
```

This statement can be then passed on to the database team, via documentation or an automated process.  In cases where the analyst has the responsibility to run the new SQL statement, or if R is being used to automate the scoring, the next line can be used:
```{r}
dbSendQuery(con, update_statement)
```

Here is a sample of the newly populated field:
```{r}
db_fligths %>%
  select(current_score) %>%
  head(10)
```

## Scenario 2- Append new scores

There may be the need to retain not only the new score, but when it was determined and its history.  This is usually possible because the source of record possesses a unique key identifier per entity, such as transaction ID or customer ID.  In the example, `flights_id` is the unique identifier.

In this example, the new scores will be stored in a new table called `daily_scores`.  The following code is part of the example preparation, it creates the table and seeds it with a single row.  This is not the best way to create an empty table, but it'll do for the purposes of this example.
```{r}
dbWriteTable(
  con, "daily_scores",
  tibble(
    flight_id = 0,
    score = 0,
    date = ""
  )
)
```

The plan is to use a SQL statement that most vendors support, is called INSERT INTO SELECT.  The idea is to use `dplyr` laziness to prepare the data transformation and predictions, but it's not going to be executed until is parsed into SQL and sent as part of a another statement.  The INSERT INTO SELECT statement allows for the results of a query to be saved in a table, and without leaving the database. 

In this example, predictions are going to be executed for just the records in the month of December. The data is filtered, and then `tidypredict_to_column()` is used to create the new fit field.  The results are then transformed to match to the structure of the new `daily_scores` table. 
```{r}
new_predictions <- db_fligths %>%
  filter(month == 12) %>%
  tidypredict_to_column(model, vars = "score") %>%
  select(
    flight_id,
    score
  ) %>%
  mutate(date = "01/01/2018")
```



```{r}
insert_scores <- build_sql("INSERT INTO daily_scores ", sql_render(new_predictions, con = con), con = con)
insert_scores
```

As in the first scenario, this statement can be then passed on to the database team, via documentation or an automated process.  In cases where the analyst has the responsibility to run the new SQL statement, or if R is being used to automate the scoring, the next line can be used:
```{r}
dbSendQuery(con, insert_scores)
```

A simple table join can be used to confirm that the new update worked.  For real life scenarios, a more sophisticated query should be performed in order to only get the latest score.  For this example, we simple filter on the same date we inserted
```{r}
tbl(con, "daily_scores") %>%
  inner_join(tbl(con, "flights_table"), by = "flight_id") %>%
  filter(date == "01/01/2018") %>%
  select(dep_delay, hour, distance, score, date)
```


```{r, include = FALSE}
dbDisconnect(con)
```

---
title: "Decision trees, using rpart"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Decision trees, using rpart}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

source("_threads.R")
library(dplyr)
library(tidypredict)
library(rpart)
set.seed(100)
```

| Function                                                      |Works|
|---------------------------------------------------------------|-----|
|`tidypredict_fit()`, `tidypredict_sql()`, `parse_model()`      |
|`tidypredict_to_column()`                                      |
|`tidypredict_test()`                                           |
|`tidypredict_interval()`, `tidypredict_sql_interval()`         |
|`parsnip`                                                      |

## How it works

Here is a simple `rpart()` model using the `mtcars` dataset:

```{r}
library(dplyr)
library(tidypredict)
library(rpart)

model <- rpart(mpg ~ ., data = mtcars)
```

## Under the hood

The parser extracts the tree structure from the model's `frame` and `splits` components. It handles both numeric and categorical splits, as well as surrogate splits for missing value handling.

```{r}
model$frame |>
  head()
```

The output from `parse_model()` is transformed into a `dplyr`, a.k.a Tidy Eval, formula. The decision tree becomes a `dplyr::case_when()` statement.

```{r}
tidypredict_fit(model)
```

From there, the Tidy Eval formula can be used anywhere where it can be operated. `tidypredict` provides three paths:

- Use directly inside `dplyr`,
  `mutate(mtcars, !! tidypredict_fit(model))`
- Use `tidypredict_to_column(model)` to a piped command set
- Use `tidypredict_to_sql(model)` to retrieve the SQL statement

## Classification

`rpart` classification models are also supported:

```{r}
model_class <- rpart(Species ~ ., data = iris)
tidypredict_fit(model_class)
```

## parsnip

`tidypredict` also supports `rpart` model objects fitted via the `parsnip` package.

```{r}
library(parsnip)

parsnip_model <- decision_tree(mode = "regression") |>
  set_engine("rpart") |>
  fit(mpg ~ ., data = mtcars)

tidypredict_fit(parsnip_model)
```

## Categorical predictors

`rpart` handles categorical predictors natively. The generated formula uses `%in%` for categorical splits:
```{r}
mtcars2 <- mtcars
mtcars2$cyl <- factor(mtcars2$cyl)

model_cat <- rpart(mpg ~ cyl + wt + hp, data = mtcars2)
tidypredict_fit(model_cat)
```

## Surrogate splits

`rpart` uses surrogate splits to handle missing values during prediction. When the primary split variable is missing, the model uses surrogate variables (other variables that produce similar splits) to route the observation. This behavior is controlled by the `usesurrogate` parameter in `rpart.control()`.

---
title: "MARS models via the `earth` package"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{MARS models via the `earth` package}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
source("_threads.R")
library(tidypredict)
library(earth)
library(dplyr)
```

| Function                                                      |Works|
|---------------------------------------------------------------|-----|
|`tidypredict_fit()`, `tidypredict_sql()`, `parse_model()`      |  ✔  |
|`tidypredict_to_column()`                                      |  ✔  |
|`tidypredict_test()`                                           |  ✔  |
|`tidypredict_interval()`, `tidypredict_sql_interval()`         |  ✗  |
|`parsnip`                                                      |  ✔  |


## `tidypredict_` functions

```{r}
library(earth)
data("etitanic", package = "earth")

model <- earth(age ~ sibsp + parch, data = etitanic, degree = 3)
```

- Create the R formula
    ```{r}
tidypredict_fit(model)
    ```

- SQL output example
    ```{r}
tidypredict_sql(model, dbplyr::simulate_odbc())
    ```

- Add the prediction to the original table
    ```{r}
library(dplyr)

etitanic %>%
  tidypredict_to_column(model) %>%
  glimpse()
    ```

- Confirm that `tidypredict` results match to the model's `predict()` results
    ```{r}
tidypredict_test(model, etitanic)
    ```

## GLM models

`tidypredict` supports the `glm` argument as well:
```{r}
model <- earth(survived ~ .,
  data = etitanic,
  glm = list(family = binomial), degree = 2
)

tidypredict_fit(model)
```

The spec sets the `is_glm` entry to 1, as well as the `family` and `link` entries. 
```{r}
str(parse_model(model), 2)
```

## parsnip

`parsnip` fitted models are also supported by `tidypredict`:
```{r}
library(parsnip)

p_model <- mars(mode = "regression", prod_degree = 3) %>%
  set_engine("earth") %>%
  fit(age ~ sibsp + parch, data = etitanic)

tidypredict_fit(p_model)
```

## Parse model spec

Here is an example of the model spec:
```{r}
pm <- parse_model(model)
str(pm, 2)
```

```{r}
str(pm$terms[1:2])
```

---
title: "Cubist models"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Cubist models}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
source("_threads.R")
library(tidypredict)
library(Cubist)
library(dplyr)
```


| Function                                                      |Works|
|---------------------------------------------------------------|-----|
|`tidypredict_fit()`, `tidypredict_sql()`, `parse_model()`      |  ✔  |
|`tidypredict_to_column()`                                      |  ✔  |
|`tidypredict_test()`                                           |  ✗  |
|`tidypredict_interval()`, `tidypredict_sql_interval()`         |  ✗  |
|`parsnip`                                                      |  ✗  |


## `tidypredict_` functions

```{r}
library(Cubist)
data("BostonHousing", package = "mlbench")

model <- Cubist::cubist(
  x = BostonHousing[, -14],
  y = BostonHousing$medv,
  committees = 3
)
```

- Create the R formula
    ```{r}
tidypredict_fit(model)
    ```

- SQL output example
    ```{r}
tidypredict_sql(model, dbplyr::simulate_odbc())
    ```

- Add the prediction to the original table
    ```{r}
library(dplyr)

BostonHousing %>%
  tidypredict_to_column(model) %>%
  glimpse()
    ```

We are not able to give an exact match of the original predictions [due to a minor bug](https://github.com/topepo/Cubist/issues/62) in Cubist.

## Parse model spec

Here is an example of the model spec:
```{r}
pm <- parse_model(model)
str(pm, 2)
```

```{r}
str(pm$terms[1:2])
```

## Limitations

- `tidypredict_test()` is not supported
- Prediction intervals are not supported
- Cubist uses 32-bit floats internally, which may cause prediction discrepancies at exact split boundaries. See the [float precision](float-precision.html) article for details.

---
title: "catboost models"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{CatBoost models}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
if (FALSE) {
  library(tidypredict)
  library(dplyr)
  eval_code <- TRUE
} else {
  eval_code <- FALSE
}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = eval_code
)
source("_threads.R")
```

| Function                                                      |Works|
|---------------------------------------------------------------|-----|
|`tidypredict_fit()`, `tidypredict_sql()`, `parse_model()`      |
|`tidypredict_to_column()`                                      |
|`tidypredict_test()`                                           |
|`tidypredict_interval()`, `tidypredict_sql_interval()`         |
|`parsnip`                                                      |

## `tidypredict_` functions

```r
library(catboost)
# Prepare data
X <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
y <- mtcars$hp

pool <- catboost.load_pool(
  X,
  label = y,
  feature_names = as.list(c("mpg", "cyl", "disp"))
)

model <- catboost.train(
  pool,
  params = list(
    iterations = 10L,
    depth = 3L,
    learning_rate = 0.5,
    loss_function = "RMSE",
    logging_level = "Silent",
    allow_writing_files = FALSE
  )
)
```

- Create the R formula
    ```r
tidypredict_fit(model)
    ```

- Add the prediction to the original table
    ```r
library(dplyr)

mtcars %>%
  tidypredict_to_column(model) %>%
  glimpse()
    ```

- Confirm that `tidypredict` results match to the model's `predict()` results. The `xg_df` argument expects the matrix data set.
    ```r
tidypredict_test(model, xg_df = X)
    ```

## Supported objectives

CatBoost supports many objective functions. The following objectives are supported by `tidypredict`:

### Regression objectives (identity transform)

- `RMSE` (default)
- `MAE`
- `Quantile`
- `MAPE`
- `Poisson`

### Binary classification (sigmoid transform)

- `Logloss`
- `CrossEntropy`

### Multiclass classification

- `MultiClass` (softmax transform)
- `MultiClassOneVsAll` (sigmoid per class)

## Binary classification example

```r
X_bin <- data.matrix(mtcars[, c("mpg", "cyl", "disp")])
y_bin <- mtcars$am

pool_bin <- catboost.load_pool(
  X_bin,
  label = y_bin,
  feature_names = as.list(c("mpg", "cyl", "disp"))
)

model_bin <- catboost.train(
  pool_bin,
  params = list(
    iterations = 10L,
    depth = 3L,
    learning_rate = 0.5,
    loss_function = "Logloss",
    logging_level = "Silent",
    allow_writing_files = FALSE
  )
)

tidypredict_test(model_bin, xg_df = X_bin)
```

## Multiclass classification example

```r
X_multi <- data.matrix(iris[, 1:4])
y_multi <- as.integer(iris$Species) - 1L

pool_multi <- catboost.load_pool(
  X_multi,
  label = y_multi,
  feature_names = as.list(colnames(iris)[1:4])
)

model_multi <- catboost.train(
  pool_multi,
  params = list(
    iterations = 10L,
    depth = 3L,
    learning_rate = 0.5,
    loss_function = "MultiClass",
    logging_level = "Silent",
    allow_writing_files = FALSE
  )
)

# Multiclass returns a list of formulas, one per class
formulas <- tidypredict_fit(model_multi)
names(formulas)
```

Test multiclass predictions:

```r
tidypredict_test(model_multi, xg_df = X_multi)
```

## Categorical features

CatBoost models can use categorical features with one-hot encoding.

### With parsnip/bonsai (recommended)

When using parsnip/bonsai, categorical features are handled automatically:

```r
library(parsnip)
library(bonsai)

df_cat <- data.frame(
  num_feat = mtcars$mpg,
  cat_feat = factor(ifelse(mtcars$am == 1, "manual", "auto")),
  target = mtcars$hp
)

model_spec <- boost_tree(trees = 10, tree_depth = 3) |>
  set_engine("catboost", logging_level = "Silent", one_hot_max_size = 10) |>
  set_mode("regression")

model_fit <- fit(model_spec, target ~ num_feat + cat_feat, data = df_cat)

# Categorical features are handled automatically
tidypredict_fit(model_fit)
```

### With raw CatBoost

For raw CatBoost models, you need to manually establish the hash-to-category mapping:

```r
pool_cat <- catboost.load_pool(
  df_cat[, c("num_feat", "cat_feat")],
  label = df_cat$target
)

model_cat <- catboost.train(
  pool_cat,
  params = list(
    iterations = 10L,
    depth = 3L,
    learning_rate = 0.5,
    loss_function = "RMSE",
    logging_level = "Silent",
    allow_writing_files = FALSE,
    one_hot_max_size = 10
  )
)

# Parse and set category mapping manually
pm_cat <- parse_model(model_cat)
pm_cat <- set_catboost_categories(pm_cat, model_cat, df_cat)

# Now use the parsed model
tidypredict_fit(pm_cat)
```

## Parse model spec

Here is an example of the model spec:
```r
pm <- parse_model(model)
str(pm, 2)
```

```r
str(pm$trees[1])
```

## Limitations

- Prediction intervals are not supported
- CatBoost uses 32-bit floats for split thresholds, which may cause prediction discrepancies at exact split boundaries. See the [float precision](float-precision.html) article for details.

# alert triggered with threshold = 0 (mocked)

    Code
      cat(t$message)
    Output
      tidypredict test results
      Difference threshold: 0
      
      Fitted records above the threshold: 32
      
      Fit max  difference:
      Lower max difference:
      Upper max difference:-<number>

# alert with intervals (mocked)

    Code
      cat(t$message)
    Output
      tidypredict test results
      Difference threshold: 0
      
      Fitted records above the threshold: 32
      Lower interval records above the threshold: 32
      Upper interval records above the threshold: 32
      
      Fit max  difference:<number>
      Lower max difference:<number>
      Upper max difference:-<number>

# xgboost alert branch

    Code
      cat(t$message)
    Output
      tidypredict test results
      Difference threshold: 0
      
      Fitted records above the threshold: 18
      
      Max difference: <number>

# glmnet alert branch (mocked)

    Code
      cat(t$message)
    Output
      tidypredict test results
      Difference threshold: 0
      
      Fitted records above the threshold: 2368
      
      Max difference: -<number>

# lightgbm alert branch (mocked)

    Code
      cat(t$message)
    Output
      tidypredict test results
      Difference threshold: 0
      
      Fitted records above the threshold: 32
      
      Max difference: <number>

# catboost alert branch (mocked)

    Code
      cat(t$message)
    Output
      tidypredict test results
      Difference threshold: 0
      
      Fitted records above the threshold: 32
      
      Max difference: <number>

# unsupported objective throws error

    Code
      tidypredict_fit(pm)
    Condition
      Error in `build_fit_formula_catboost_nested()`:
      ! Unsupported objective: "UnsupportedObjective".
      i Supported objectives: "RMSE", "MAE", "Quantile", "MAPE", "Poisson", "Huber", "LogCosh", "Expectile", "Tweedie", "Logloss", "CrossEntropy", "MultiClass", and "MultiClassOneVsAll".

# empty trees throws error

    Code
      tidypredict_fit(pm)
    Condition
      Error in `build_fit_formula_catboost_nested()`:
      ! Model has no trees.

# tidypredict_test requires matrix

    Code
      tidypredict_test(model)
    Condition
      Error in `catboost_model()`:
      ! CatBoost models require a matrix for predictions.
      i Pass the prediction matrix via the `xg_df` argument.

# .extract_catboost_trees errors on non-catboost model

    Code
      .extract_catboost_trees(lm(mpg ~ wt, data = mtcars))
    Condition
      Error in `.extract_catboost_trees()`:
      ! `model` must be <catboost.Model>, not a <lm> object.

# multiclass model requires num_class >= 2

    Code
      tidypredict_fit(pm)
    Condition
      Error in `build_fit_formula_catboost_multiclass_nested()`:
      ! Multiclass model must have num_class >= 2.

# set_catboost_categories validates parsed_model argument

    Code
      set_catboost_categories("not a parsed model", model, data.frame())
    Condition
      Error in `set_catboost_categories()`:
      ! `parsed_model` must be a parsed CatBoost model.

# set_catboost_categories validates model argument

    Code
      set_catboost_categories(pm, "not a model", data.frame())
    Condition
      Error in `set_catboost_categories()`:
      ! `model` must be a <catboost.Model>, not a string.

# set_catboost_categories errors when column not found in data

    Code
      set_catboost_categories(pm, model, wrong_data)
    Condition
      Error in `set_catboost_categories()`:
      ! Column "cat_feat" not found in `data`.

# set_catboost_categories errors when column is not a factor

    Code
      set_catboost_categories(pm, model, wrong_data)
    Condition
      Error in `set_catboost_categories()`:
      ! Column "cat_feat" must be a factor.

# categorical features without mapping throws error

    Code
      tidypredict_fit(pm)
    Condition
      Error in `map()`:
      i In index: 1.
      Caused by error in `build_nested_catboost_categorical()`:
      ! No category mapping found for hash -254607792.
      i For raw CatBoost models, use `set_catboost_categories()`.

# parsnip model without xlevels throws error

    Code
      tidypredict_fit(model_fit)
    Condition
      Error in `setup_catboost_parsnip_categories()`:
      ! Model has categorical features but no factor level information.
      i Ensure the model was fit with factor columns, not character columns.

# unsupported objective throws error

    Code
      tidypredict_fit(pm)
    Condition
      Error in `apply_xgb_objective()`:
      ! Objective "unsupported_objective" is not supported.
      i Supported objectives: "binary:hinge", "binary:logistic", "binary:logitraw", "count:poisson", "reg:absoluteerror", "reg:gamma", "reg:logistic", "reg:pseudohubererror", "reg:squarederror", "reg:squaredlogerror", "reg:tweedie".
      i Multiclass objectives ("multi:softmax", "multi:softprob") are not supported.

# NULL objective warns user

    Code
      tidypredict_fit(pm)
    Condition
      Warning:
      If the objective is a custom function, please explicitly apply it to the output.
    Output
      case_when(TRUE ~ 5)

# tidypredict_test xg_df argument is required

    Code
      tidypredict_test(model, mtcars)
    Condition
      Error in `xgb.DMatrix()`:
      ! xgb.DMatrix does not support construction from NULL

# .extract_xgb_trees errors on non-xgb.Booster

    Code
      .extract_xgb_trees(list())
    Condition
      Error in `.extract_xgb_trees()`:
      ! `model` must be <xgb.Booster>, not an empty list.

# NULL objective with non-zero base_score warns user

    Code
      fit <- tidypredict_fit(pm)
    Condition
      Warning:
      If the objective is a custom function, please explicitly apply it to the output.

# returns the right output

    Code
      rlang::expr_text(tf)
    Output
      [1] "(37.2 + hp * -0.0318 + wt * -3.88 + (ifelse(disp > 95.099998, \n    14.89 + hp * -0.0406 + drat * 2.4, 0) + ifelse(disp <= 95.099998, \n    33.06, 0))/((disp > 95.099998) + (disp <= 95.099998)) + (37.26 + \n    wt * -5.28))/3"

# formulas produces correct predictions

    Code
      tidypredict_test(model, non_split_data, threshold = 1e-05)
    Output
      tidypredict test results
      Difference threshold: 1e-05
      
       All results are within the difference threshold

# works with parsnip model specification

    Code
      tidypredict_test(model, df = etitanic_fac)
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

---

    Code
      tidypredict_test(model, df = etitanic_fac)
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

# returns the right output

    Code
      rlang::expr_text(tf)
    Output
      [1] "20.534817535821 + (ifelse(disp < 145, 145 - disp, 0) * 0.148589866311) + \n    (ifelse(disp > 145, disp - 145, 0) * -0.025012854678)"

# formulas produces correct predictions

    Code
      tidypredict_test(earth::earth(age ~ sibsp + parch, data = earth::etitanic),
      earth::etitanic)
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

---

    Code
      tidypredict_test(earth::earth(age ~ sibsp + parch, data = earth::etitanic,
      degree = 2), earth::etitanic)
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

---

    Code
      tidypredict_test(earth::earth(age ~ sibsp + parch, data = earth::etitanic,
      degree = 3), earth::etitanic)
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

---

    Code
      tidypredict_test(earth::earth(age ~ ., data = earth::etitanic), earth::etitanic)
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

---

    Code
      tidypredict_test(earth::earth(age ~ ., data = earth::etitanic, pmethod = "backward"),
      earth::etitanic)
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

---

    Code
      tidypredict_test(earth::earth(age ~ ., data = earth::etitanic, pmethod = "none"),
      earth::etitanic)
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

---

    Code
      tidypredict_test(earth::earth(age ~ ., data = earth::etitanic, pmethod = "exhaustive"),
      earth::etitanic)
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

---

    Code
      tidypredict_test(earth::earth(age ~ ., data = earth::etitanic, pmethod = "forward"),
      earth::etitanic)
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

---

    Code
      tidypredict_test(earth::earth(age ~ ., data = earth::etitanic, pmethod = "seqrep"),
      earth::etitanic)
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

---

    Code
      tidypredict_test(earth::earth(survived ~ age + sibsp, data = earth::etitanic,
      glm = list(family = binomial)), earth::etitanic)
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

---

    Code
      tidypredict_test(earth::earth(survived ~ age + sibsp, data = earth::etitanic,
      glm = list(family = binomial), degree = 2), earth::etitanic)
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

---

    Code
      tidypredict_test(earth::earth(survived ~ ., data = earth::etitanic, glm = list(
        family = binomial), pmethod = "backward"), earth::etitanic)
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

---

    Code
      tidypredict_test(earth::earth(survived ~ ., data = earth::etitanic, glm = list(
        family = binomial), pmethod = "none"), earth::etitanic)
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

---

    Code
      tidypredict_test(earth::earth(survived ~ ., data = earth::etitanic, glm = list(
        family = binomial), pmethod = "exhaustive"), earth::etitanic)
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

---

    Code
      tidypredict_test(earth::earth(survived ~ ., data = earth::etitanic, glm = list(
        family = binomial), pmethod = "forward"), earth::etitanic)
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

---

    Code
      tidypredict_test(earth::earth(survived ~ ., data = earth::etitanic, glm = list(
        family = binomial), pmethod = "seqrep"), earth::etitanic)
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

---

    Code
      tidypredict_test(earth::earth(Sepal.Length ~ ., data = iris), iris)
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

---

    Code
      tidypredict_test(earth::earth(x = iris[, -1], y = iris$Sepal.Length), iris)
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

---

    Code
      tidypredict_test(earth::earth(Sepal.Length ~ ., data = iris, degree = 2,
      pmethod = "none"), iris)
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

---

    Code
      tidypredict_test(earth::earth(x = iris[, -1], y = iris$Sepal.Length, degree = 2,
      pmethod = "none"), iris)
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

# .extract_earth_multiclass errors on non-earth model

    Code
      .extract_earth_multiclass(model)
    Condition
      Error in `.extract_earth_multiclass()`:
      ! `model` must be <earth>, not a <lm> object.

# .extract_earth_multiclass errors on binary model

    Code
      .extract_earth_multiclass(model)
    Condition
      Error in `.extract_earth_multiclass()`:
      ! Model does not contain multiclass information.
      i Fit the earth model with `glm = TRUE` for classification.

# .extract_earth_multiclass errors on regression model

    Code
      .extract_earth_multiclass(model)
    Condition
      Error in `.extract_earth_multiclass()`:
      ! Model does not contain multiclass information.
      i Fit the earth model with `glm = TRUE` for classification.

# unsupported objective throws error

    Code
      tidypredict_fit(pm)
    Condition
      Error in `build_fit_formula_lgb_from_parsed()`:
      ! Unsupported objective: "unsupported_objective".
      i Supported objectives: "regression", "regression_l2", "regression_l1", "huber", "fair", "quantile", "mape", "poisson", "gamma", "tweedie", "binary", "cross_entropy", "multiclass", and "multiclassova".

# empty trees throws error

    Code
      tidypredict_fit(pm)
    Condition
      Error in `build_fit_formula_lgb_from_parsed()`:
      ! Model has no trees.

# multiclass with num_class < 2 throws error

    Code
      tidypredict_fit(pm)
    Condition
      Error in `build_fit_formula_lgb_multiclass_from_parsed()`:
      ! Multiclass model must have num_class >= 2.

# multiclass with NULL num_class throws error

    Code
      tidypredict_fit(pm)
    Condition
      Error in `build_fit_formula_lgb_multiclass_from_parsed()`:
      ! Multiclass model must have num_class >= 2.

# build_lgb_nested_condition errors on unknown type

    Code
      tidypredict:::build_lgb_nested_condition(condition)
    Condition
      Error in `tidypredict:::build_lgb_nested_condition()`:
      ! Unknown path element type: "unknown_type"

# tidypredict_test errors for multiclass model

    Code
      tidypredict_test(model, xg_df = X)
    Condition
      Error in `lgb_booster()`:
      ! tidypredict_test does not support multiclass LightGBM models.
      i Use tidypredict_fit() directly for multiclass predictions.

# tidypredict_test errors when matrix not provided

    Code
      tidypredict_test(model)
    Condition
      Error in `lgb_booster()`:
      ! LightGBM models require a matrix for predictions.
      i Pass the prediction matrix via the `xg_df` argument.

# .extract_lgb_trees errors on non-lgb.Booster

    Code
      .extract_lgb_trees(list())
    Condition
      Error in `.extract_lgb_trees()`:
      ! `model` must be <lgb.Booster>, not an empty list.

# returns the right output

    Code
      rlang::expr_text(tf)
    Output
      [1] "1.520331147866 + (wt * -0.372988616484) + (cyl * 0.013885491477)"

# formulas produces correct predictions

    Code
      tidypredict_test(glm(am ~ wt + cyl + disp, data = mtcars, family = "gaussian"),
      mtcars)
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

---

    Code
      tidypredict_test(glm(am ~ wt + cyl + disp, data = mtcars, family = "binomial"),
      mtcars)
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

---

    Code
      tidypredict_test(glm(am ~ wt * cyl + disp, data = mtcars, family = "gaussian"),
      mtcars)
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

---

    Code
      tidypredict_test(glm(am ~ wt * cyl + disp, data = mtcars, family = "binomial"),
      mtcars)
    Condition
      Warning:
      glm.fit: fitted probabilities numerically 0 or 1 occurred
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

---

    Code
      tidypredict_test(glm(am ~ wt:cyl + disp, data = mtcars, family = "gaussian"),
      mtcars)
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

---

    Code
      tidypredict_test(glm(am ~ wt:cyl + disp, data = mtcars, family = "binomial"),
      mtcars)
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

# tidypredict works when variable names are subset of other variables

    Code
      tidypredict_test(model, mtcars)
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

# tidypredict_interval errors for non-gaussian glm

    Code
      tidypredict_interval(model)
    Condition
      Error in `te_interval_glm()`:
      ! Combination of family and link are not supported for prediction intervals.

# returns the right output

    Code
      round_print(tf)
    Output
      [1] "35.31378 + (cyl * -0.8714512) + (hp * -0.0101174) + (wt * -2.594437)"

# formulas produces correct predictions

    Code
      tidypredict_test(glmnet::glmnet(mtcars[, -1], mtcars$mpg, family = "gaussian",
      lambda = 1), mtcars[, -1])
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

---

    Code
      tidypredict_test(glmnet::glmnet(mtcars[, -8], mtcars$vs, family = "binomial",
      lambda = 1), mtcars[, -1])
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

---

    Code
      tidypredict_test(glmnet::glmnet(mtcars[, -8], mtcars$vs, family = "poisson",
      lambda = 1), mtcars[, -1])
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

# errors if more than 1 penalty is selected

    Code
      tidypredict_fit(model)
    Condition
      Error in `parse_model()`:
      ! `tidypredict_fit()` requires that there are only 1 penalty selected, 79 were provided.

---

    Code
      tidypredict_fit(model)
    Condition
      Error in `parse_model()`:
      ! `tidypredict_fit()` requires that there are only 1 penalty selected, 2 were provided.

# glmnet are handeld neatly with parsnip

    Code
      round_print(tf)
    Output
      [1] "35.31405 + (cyl * -0.8716234) + (hp * -0.01011579) + (wt * -2.594265)"

# multinomial family errors with helpful message (#198)

    Code
      tidypredict_fit(model)
    Condition
      Error in `tidypredict_fit()`:
      ! Multinomial glmnet models are not supported.
      i Models fit with `family = "multinomial"` have multiple outcome columns which is not supported.

# mgaussian family errors with helpful message (#199)

    Code
      tidypredict_fit(model)
    Condition
      Error in `tidypredict_fit()`:
      ! Multivariate gaussian glmnet models are not supported.
      i Models fit with `family = "mgaussian"` have multiple outcome columns which is not supported.

# .extract_glmnet_multiclass errors on non-multnet model

    Code
      .extract_glmnet_multiclass(model)
    Condition
      Error in `.extract_glmnet_multiclass()`:
      ! `model` must be <multnet>, not an <elnet> object.

# .extract_glmnet_multiclass errors with multiple penalties

    Code
      .extract_glmnet_multiclass(model)
    Condition
      Error in `.extract_glmnet_multiclass()`:
      ! glmnet model has multiple penalty values.
      i Specify a single `penalty` value.

# returns the right output

    Code
      rlang::expr_text(tf)
    Output
      [1] "(case_when(wt <= 2.3325 ~ case_when(disp <= 78.85 ~ 33.525, .default = 27.425), \n    .default = case_when(disp <= 281 ~ case_when(cyl <= 5 ~ case_when(drat <= \n        3.695 ~ 24.4, .default = 22.3666666666667), .default = case_when(wt <= \n        3.3275 ~ case_when(disp <= 152.5 ~ 19.7, .default = 21.1), \n        .default = case_when(disp <= 196.3 ~ 18.92, .default = 18.1))), \n        .default = 14.1)) + case_when(disp <= 266.9 ~ case_when(hp <= \n    80.5 ~ 28.5333333333333, .default = case_when(hp <= 118 ~ \n    case_when(wt <= 1.989 ~ 30.4, .default = case_when(hp <= \n        96 ~ 22.8, .default = case_when(qsec <= 19.725 ~ 21.4, \n        .default = 21.5))), .default = case_when(qsec <= 18.6 ~ \n    19.6, .default = 17.8))), .default = case_when(drat <= 3.035 ~ \n    10.4, .default = case_when(carb <= 2.5 ~ 18.7, .default = case_when(drat <= \n    3.635 ~ case_when(wt <= 4.5625 ~ 15.04, .default = 14.7), \n    .default = 13.3)))) + case_when(drat <= 3.75 ~ case_when(wt <= \n    3.4725 ~ 20.8333333333333, .default = case_when(qsec <= 16.23 ~ \n    13.3, .default = case_when(disp <= 456 ~ case_when(qsec <= \n    17.225 ~ 19.2, .default = case_when(wt <= 4.7075 ~ 16.34, \n    .default = 14.7)), .default = 10.4))), .default = case_when(disp <= \n    130.55 ~ case_when(disp <= 78.85 ~ 32.2333333333333, .default = 27.75), \n    .default = case_when(cyl <= 5 ~ 22.8, .default = case_when(qsec <= \n        17.66 ~ 21, .default = 19.2)))))/3"

# formulas produces correct predictions

    Code
      tidypredict_test(randomForest::randomForest(mpg ~ ., data = mtcars, ntree = 3),
      mtcars, )
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

# classification models error with clear message (#193)

    Code
      tidypredict_fit(model)
    Condition
      Error in `tidypredict_fit_rf_nested()`:
      ! Classification models are not supported for randomForest.
      i Only regression models can be converted to tidy formulas.
      i Classification requires a voting mechanism that cannot be expressed as a single formula.

# parse_model errors on classification model

    Code
      parse_model(model)
    Condition
      Error in `parse_model()`:
      ! Classification models are not supported for randomForest.
      i Only regression models can be converted to tidy formulas.
      i Classification requires a voting mechanism that cannot be expressed as a single formula.

# v2 parsed classification model errors

    Code
      tidypredict_fit(pm)
    Condition
      Error in `tidypredict_fit_randomForest()`:
      ! Classification models are not supported for randomForest.
      i Only regression models can be converted to tidy formulas.
      i Classification requires a voting mechanism that cannot be expressed as a single formula.

# .extract_rf_classprob errors on non-randomForest model

    Code
      .extract_rf_classprob(model)
    Condition
      Error in `.extract_rf_classprob()`:
      ! `model` must be <randomForest>, not a <lm> object.

# .extract_rf_classprob errors on regression model

    Code
      .extract_rf_classprob(model)
    Condition
      Error in `.extract_rf_classprob()`:
      ! Model is not a classification model.
      i Use `tidypredict_fit()` for regression models.

# .extract_rf_trees errors on non-randomForest model

    Code
      .extract_rf_trees(model)
    Condition
      Error in `.extract_rf_trees()`:
      ! `model` must be <randomForest>, not a <lm> object.

# .extract_rf_trees errors on classification model

    Code
      .extract_rf_trees(model)
    Condition
      Error in `.extract_rf_trees()`:
      ! Classification models are not supported.
      i Use `.extract_rf_classprob()` for classification models.

# returns the right output

    Code
      rlang::expr_text(tf)
    Output
      [1] "case_when(cyl <= 4 ~ 26.6636363636364, .default = case_when(cyl <= \n    6 ~ 19.7428571428571, .default = 15.1))"

# formulas produces correct predictions

    Code
      tidypredict_test(partykit::ctree(mpg ~ am + cyl, data = mtcars), mtcars)
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

---

    Code
      tidypredict_test(partykit::ctree(mpg ~ wt, offset = am1, data = mtcars), mtcars)
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

---

    Code
      tidypredict_test(partykit::ctree(mpg ~ wt + disp * cyl, data = mtcars), mtcars)
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

---

    Code
      tidypredict_test(partykit::ctree(mpg ~ (wt + disp) * cyl, data = mtcars),
      mtcars)
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

# .extract_partykit_classprob errors on non-party model

    Code
      .extract_partykit_classprob(list())
    Condition
      Error in `.extract_partykit_classprob()`:
      ! `model` must be <party>, not an empty list.

# returns the right output

    Code
      rlang::expr_text(tf)
    Output
      [1] "(case_when(hp <= 118 ~ case_when(cyl <= 5 ~ 26.7, .default = 21.1333333333333), \n    .default = case_when(hp <= 205 ~ 17.48, .default = 14.5416666666667)) + \n    case_when(hp <= 80.5 ~ 32.2333333333333, .default = case_when(hp <= \n        118 ~ 21.9545454545455, .default = 16.5722222222222)) + \n    case_when(disp <= 101.55 ~ 31.9, .default = case_when(cyl <= \n        7 ~ 20.8384615384615, .default = 14.8785714285714)))/3"

# formulas produces correct predictions

    Code
      tidypredict_test(ranger::ranger(mpg ~ ., data = mtcars, num.trees = 3,
      max.depth = 2, seed = 100, num.threads = 1), mtcars)
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

# classification models error with clear message (#191)

    Code
      tidypredict_fit(model)
    Condition
      Error in `tidypredict_fit_ranger_nested()`:
      ! Classification models are not supported for ranger.
      i Only regression models can be converted to tidy formulas.
      i Classification requires a voting mechanism that cannot be expressed as a single formula.

# .extract_ranger_classprob errors on non-ranger model

    Code
      .extract_ranger_classprob(model)
    Condition
      Error in `.extract_ranger_classprob()`:
      ! `model` must be <ranger>, not a <lm> object.

# .extract_ranger_classprob errors without probability = TRUE

    Code
      .extract_ranger_classprob(model)
    Condition
      Error in `.extract_ranger_classprob()`:
      ! Model does not contain probability information.
      i Fit the ranger model with `probability = TRUE`.

# .extract_ranger_trees errors on non-ranger model

    Code
      .extract_ranger_trees(model)
    Condition
      Error in `.extract_ranger_trees()`:
      ! `model` must be <ranger>, not a <lm> object.

# .extract_ranger_trees errors on classification model

    Code
      .extract_ranger_trees(model)
    Condition
      Error in `.extract_ranger_trees()`:
      ! Classification models are not supported.
      i Use `.extract_ranger_classprob()` for classification models.

# v2 parsed classification model errors

    Code
      tidypredict_fit(pm)
    Condition
      Error in `tidypredict_fit_ranger()`:
      ! Classification models are not supported for ranger.
      i Only regression models can be converted to tidy formulas.
      i Classification requires a voting mechanism that cannot be expressed as a single formula.

# parse_model.ranger errors on classification

    Code
      parse_model(model)
    Condition
      Error in `parse_model()`:
      ! Classification models are not supported for ranger.
      i Only regression models can be converted to tidy formulas.
      i Classification requires a voting mechanism that cannot be expressed as a single formula.

# returns the right output

    Code
      rlang::expr_text(tf)
    Output
      [1] "1.520331147866 + (wt * -0.372988616484) + (cyl * 0.013885491477)"

# formulas produces correct predictions

    Code
      tidypredict_test(lm(mpg ~ wt + am + cyl, data = mtcars), mtcars)
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

---

    Code
      tidypredict_test(lm(mpg ~ wt, offset = am, data = mtcars), mtcars)
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

---

    Code
      tidypredict_test(lm(mpg ~ wt + disp * cyl, data = mtcars), mtcars)
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

---

    Code
      tidypredict_test(lm(mpg ~ wt + disp:cyl, data = mtcars), mtcars)
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

---

    Code
      tidypredict_test(lm(mpg ~ (wt + disp) * cyl, data = mtcars), mtcars)
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

# tidypredict works when variable names are subset of other variables

    Code
      tidypredict_test(model, mtcars)
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

# we get better error from QR decomposition issues (#124)

    Code
      tidypredict::tidypredict_fit(lm_fit)
    Condition
      Error in `parse_model()`:
      x Unable to calculate inverse of QR decomposition.
      i This is likely happening because the predictors contain a linear combination of predictors. Please remove and try again.

# returns the right output

    Code
      rlang::expr_text(tf)
    Output
      [1] "case_when(cyl <= 5 ~ 26.6636363636364, .default = case_when(cyl <= \n    7 ~ 19.7428571428571, .default = 15.1))"

# formulas produce correct predictions - regression

    Code
      tidypredict_test(rpart::rpart(mpg ~ am + cyl + wt, data = mtcars), mtcars)
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

# formulas produce correct predictions - classification

    Code
      tidypredict_test(rpart::rpart(Species ~ ., data = iris), iris)
    Output
      tidypredict test results
      Difference threshold: 0
      
       All results are within the difference threshold

# categorical predictors work correctly

    Code
      tidypredict_test(rpart::rpart(mpg ~ cyl + wt, data = mtcars2), mtcars2)
    Output
      tidypredict test results
      Difference threshold: 1e-12
      
       All results are within the difference threshold

# .extract_rpart_classprob errors on non-rpart model

    Code
      .extract_rpart_classprob(list())
    Condition
      Error in `.extract_rpart_classprob()`:
      ! `model` must be <rpart>, not an empty list.

# .extract_rpart_classprob errors on regression model

    Code
      .extract_rpart_classprob(model)
    Condition
      Error in `.extract_rpart_classprob()`:
      ! `model` must be a classification model (method = 'class').

# path_formula() errors with unsupported values

    Code
      path_formula(list(type = "unknown", op = "less", col = "x", val = 0))
    Condition
      Error in `path_formula()`:
      ! type has unsupported value of unknown.
      i This is an internal error that was detected in the tidypredict package.
        Please report it at <https://github.com/tidymodels/tidypredict/issues> with a reprex (<https://tidyverse.org/help/>) and the full backtrace.

---

    Code
      path_formula(list(type = "conditional", op = "unknown", col = "x", val = 0))
    Condition
      Error in `build_comparison_expr()`:
      ! op has unsupported value of unknown.
      i This is an internal error that was detected in the tidypredict package.
        Please report it at <https://github.com/tidymodels/tidypredict/issues> with a reprex (<https://tidyverse.org/help/>) and the full backtrace.

---

    Code
      path_formula(list(type = "set", op = "unknown", col = "x", vals = 0))
    Condition
      Error in `build_set_expr()`:
      ! op has unsupported value of unknown.
      i This is an internal error that was detected in the tidypredict package.
        Please report it at <https://github.com/tidymodels/tidypredict/issues> with a reprex (<https://tidyverse.org/help/>) and the full backtrace.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model-catboost.R
\name{.extract_catboost_trees}
\alias{.extract_catboost_trees}
\title{Extract processed CatBoost trees}
\usage{
.extract_catboost_trees(model)
}
\arguments{
\item{model}{A CatBoost model object}
}
\description{
For use in orbital package.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict-interval.R
\name{tidypredict_interval}
\alias{tidypredict_interval}
\title{Returns a Tidy Eval formula to calculate prediction interval.}
\usage{
tidypredict_interval(model, interval = 0.95)
}
\arguments{
\item{model}{An R model or a list with a parsed model}

\item{interval}{The prediction interval, defaults to 0.95}
}
\description{
It parses a model or uses an already parsed model to return a
Tidy Eval formula that can then be used inside a dplyr command.
}
\details{
The result still has to be added to and subtracted from the fit to obtain the upper and
lower bound respectively.
}
\examples{

model <- lm(mpg ~ wt + cyl * disp, offset = am, data = mtcars)
tidypredict_interval(model)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as-parsed-model.R
\name{as_parsed_model}
\alias{as_parsed_model}
\title{Prepares parsed model object}
\usage{
as_parsed_model(x)
}
\arguments{
\item{x}{A parsed model object}
}
\description{
Prepares parsed model object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model-ranger.R
\name{.extract_ranger_classprob}
\alias{.extract_ranger_classprob}
\title{Extract classification probability trees for ranger models}
\usage{
.extract_ranger_classprob(model)
}
\arguments{
\item{model}{A ranger model object fitted with \code{probability = TRUE}}
}
\description{
For use in orbital package.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tidypredict_test.R
\name{tidypredict_test}
\alias{tidypredict_test}
\title{Tests base predict function against tidypredict}
\usage{
tidypredict_test(
  model,
  df = model$model,
  threshold = 1e-12,
  include_intervals = FALSE,
  max_rows = NULL,
  xg_df = NULL
)
}
\arguments{
\item{model}{An R model or a list with a parsed model. It currently supports
lm(), glm() and randomForest() models.}

\item{df}{A data frame that contains all of the needed fields to run the prediction.
It defaults to the "model" data frame object inside the model object.}

\item{threshold}{The number that a given result difference, between predict() and
tidypredict_to_column() should not exceed. For continuous predictions, the default
value is 0.000000000001 (1e-12), and for categorical predictions, the default value is
0.}

\item{include_intervals}{Switch to indicate if the prediction intervals should be
included in the test. It defaults to FALSE.}

\item{max_rows}{The number of rows in the object passed in the df argument. Highly
recommended for large data sets.}

\item{xg_df}{A xgb.DMatrix object, required only for XGBoost models. It defaults to
NULL
recommended for large data sets.}
}
\description{
Compares the results of predict() and tidypredict_to_column()
functions.
}
\examples{

model <- lm(mpg ~ wt + cyl * disp, offset = am, data = mtcars)
tidypredict_test(model)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model-earth.R
\name{.extract_earth_multiclass}
\alias{.extract_earth_multiclass}
\title{Extract multiclass linear predictors for earth models}
\usage{
.extract_earth_multiclass(model)
}
\arguments{
\item{model}{An earth model object with multiple classes (glm.list with >1 elements)}
}
\description{
For use in orbital package.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tidymodels.R
\name{tidy.pm_regression}
\alias{tidy.pm_regression}
\title{Tidy the parsed model results}
\usage{
\method{tidy}{pm_regression}(x, ...)
}
\arguments{
\item{x}{A parsed_model object}

\item{...}{Reserved for future use}
}
\description{
Tidy the parsed model results
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model-partykit.R
\name{.extract_partykit_classprob}
\alias{.extract_partykit_classprob}
\title{Extract classprob trees for partykit models}
\usage{
.extract_partykit_classprob(model)
}
\arguments{
\item{model}{A partykit model object}
}
\description{
For use in orbital package.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree-nested.R
\name{.build_nested_case_when_tree}
\alias{.build_nested_case_when_tree}
\title{Build nested case_when expression from tree info}
\usage{
.build_nested_case_when_tree(tree_info)
}
\arguments{
\item{tree_info}{A tree info list with nodeID, leftChild, rightChild,
splitvarName, terminal, prediction, and node_splits}
}
\description{
Shared helper for building nested tree expressions. This is the nested
equivalent of \code{.build_case_when_tree()}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree.R
\name{path_formulas}
\alias{path_formulas}
\title{Turn a path object into a combined expression}
\usage{
path_formulas(path)
}
\arguments{
\item{path}{a list of lists.

This list can contain 0 or more elemements. The elements but each be of the
following format:
\itemize{
\item \code{type} character, must be \code{"conditional"}, \code{"set"}, or \code{"all"}.
\item \code{op} character.
if \code{type == "conditional"} must be \code{"more"}, \code{"more-equal"}, \code{"less"}, or
\code{"less-equal"}.
if \code{type == "set"} must be \code{"in"} on \verb{not-in}.
\item \code{col} character.
\item \code{val} if \code{type == "conditional"} and \code{vals} if \code{type == "set"}.
Can be character or numeric.
@keywords internal
}}
}
\description{
Turn a path object into a combined expression
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model-rpart.R
\name{.extract_rpart_classprob}
\alias{.extract_rpart_classprob}
\title{Extract classprob trees for rpart models}
\usage{
.extract_rpart_classprob(model)
}
\arguments{
\item{model}{An rpart model object}
}
\description{
For use in orbital package.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model-partykit.R
\name{.partykit_tree_info_full}
\alias{.partykit_tree_info_full}
\title{Extract comprehensive tree info for partykit models}
\usage{
.partykit_tree_info_full(model)
}
\arguments{
\item{model}{A partykit model object}
}
\description{
Returns tree structure in format needed by nested case_when generator.
For use in orbital package.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree.R
\name{path_formula}
\alias{path_formula}
\title{Turn a path object into an expression}
\usage{
path_formula(x)
}
\arguments{
\item{x}{a list.

The input of this function is a list with 4 values.
\itemize{
\item \code{type} character, must be \code{"conditional"} or \code{"set"}.
\item \code{op} character.
if \code{type == "conditional"} must be \code{"more"}, \code{"more-equal"}, \code{"less"}, or
\code{"less-equal"}.
if \code{type == "set"} must be \code{"in"} on \verb{not-in}.
\item \code{col} character.
\item \code{val} if \code{type == "conditional"} and \code{vals} if \code{type == "set"}.
Can be character or numeric.
@keywords internal
}}
}
\description{
Turn a path object into an expression
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree.R
\name{generate_tree_node}
\alias{generate_tree_node}
\title{Construct a single node of a tree}
\usage{
generate_tree_node(node, calc_mode = "")
}
\arguments{
\item{node}{a list with named elements \code{path} and \code{prediction}. See details
for more.}

\item{calc_mode}{character, takes values \code{""} and \code{"calc_mode"}.

The \code{node} list should contain the two lists \code{path} and \code{prediction}.

The \code{path} element has the following structure:

This list can contain 0 or more elemements. The elements but each be of the
following format:
\itemize{
\item \code{type} character, must be \code{"conditional"}, \code{"set"}, or \code{"all"}.
\item \code{op} character.
if \code{type == "conditional"} must be \code{"more"}, \code{"more-equal"}, \code{"less"}, or
\code{"less-equal"}.
if \code{type == "set"} must be \code{"in"} on \verb{not-in}.
\item \code{col} character.
\item \code{val} if \code{type == "conditional"} and \code{vals} if \code{type == "set"}.
Can be character or numeric.
}

The \code{prediction} list has the following structure:

It can either be a singular value or a list.
If it is a list it will have the following 4 named elements \code{col}, \code{val},
\code{op}, and \code{is_intercept}.
\itemize{
\item \code{col} character, name of column
\item \code{val} val, numeric of character
\item \code{op} character, known values are \code{"none"} and \code{"multiply"}. \code{"none"} is
used then \code{is_intercept == 1}.
\item \code{is_intercept}integer, takes values \code{0} and \code{1}.`
}

@keywords internal}
}
\description{
Construct a single node of a tree
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree-nested.R
\name{generate_nested_case_when_tree}
\alias{generate_nested_case_when_tree}
\title{Generate nested case_when trees}
\usage{
generate_nested_case_when_tree(tree_info)
}
\arguments{
\item{tree_info}{A tree info list from \code{rpart_tree_info_full()} or similar}
}
\description{
These functions generate nested \code{case_when()} expressions for decision trees,
which are more efficient than flat \code{case_when()} for both R/dplyr and SQL
execution.
}
\details{
The following tree:

\if{html}{\out{<div class="sourceCode">}}\preformatted{        +-----+
   +----|x > 0|----+
   |    +-----+    |
   v               v
}\if{html}{\out{</div>}}

+------+        +--------+
+--|y < 20|--+  +--|z <= 10 |--+
|  +------+  |  |  +--------+  |
v            v  v              v
a            b  c              d

will be turned into the following nested \code{case_when()} statement:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{case_when(
  x > 0 ~ case_when(
    y < 20 ~ "a",
    .default = "b"
  ),
  .default = case_when(
    z <= 10 ~ "c",
    .default = "d"
  )
)
}\if{html}{\out{</div>}}

NA values in predictor columns are not handled by the generated expression.
Users should ensure that predictor columns do not contain NA values before
using the generated expression, or the results will be NA for those rows.
}
\keyword{Generate}
\keyword{a}
\keyword{case_when}
\keyword{for}
\keyword{internal}
\keyword{nested}
\keyword{tree}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model-glmnet.R
\name{.extract_glmnet_multiclass}
\alias{.extract_glmnet_multiclass}
\title{Extract multiclass linear predictors for glmnet models}
\usage{
.extract_glmnet_multiclass(model, penalty = NULL)
}
\arguments{
\item{model}{A glmnet model object with class "multnet"}

\item{penalty}{The penalty value to use for coefficient extraction}
}
\description{
For use in orbital package.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model-catboost.R
\name{set_catboost_categories}
\alias{set_catboost_categories}
\title{Set categorical feature mappings for CatBoost model}
\usage{
set_catboost_categories(parsed_model, model, data)
}
\arguments{
\item{parsed_model}{A parsed CatBoost model from \code{parse_model()}}

\item{model}{The original CatBoost model object}

\item{data}{A data frame containing factor columns matching the categorical
features used in the model. The factor levels must match those from
training.}
}
\value{
The parsed model with category mappings added
}
\description{
CatBoost stores categorical features as hash values internally. This function
establishes the mapping between hash values and category names by examining
a data frame with the same factor columns used during training.
}
\details{
This function is only needed when using raw CatBoost models (trained with
\code{catboost.train()}). When using parsnip/bonsai, categorical features are
handled automatically and this function is not required.
}
\examples{
\dontrun{
# For raw CatBoost models with categorical features:
pm <- parse_model(catboost_model)
pm <- set_catboost_categories(pm, catboost_model, training_data)
tidypredict_fit(pm)

# For parsnip/bonsai models, this is not needed:
# tidypredict_fit(parsnip_model_fit)  # works automatically
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict-column.R
\name{tidypredict_to_column}
\alias{tidypredict_to_column}
\title{Adds the prediction columns to a piped command set.}
\usage{
tidypredict_to_column(
  df,
  model,
  add_interval = FALSE,
  interval = 0.95,
  vars = c("fit", "upper", "lower")
)
}
\arguments{
\item{df}{A data.frame or tibble}

\item{model}{An R model or a parsed model inside a data frame}

\item{add_interval}{Switch that indicates if the prediction interval columns should be added. Defaults
to FALSE}

\item{interval}{The prediction interval, defaults to 0.95. Ignored if add_interval is set to
FALSE}

\item{vars}{The name of the variables that this function will produce. Defaults to "fit", "upper", and "lower".}
}
\description{
Adds a new column with the results from tidypredict_fit() to a piped command set.
If add_interval is set to TRUE, it will add two additional columns- one
for the lower and another for the upper prediction interval bounds.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree-nested.R
\name{build_nested_node}
\alias{build_nested_node}
\title{Build a nested case_when expression for a single node}
\usage{
build_nested_node(node_id, tree_info)
}
\arguments{
\item{node_id}{The node ID to build (0-indexed)}

\item{tree_info}{Tree info list with nodeID, leftChild, rightChild,
splitvarName, terminal, prediction, and node_splits}
}
\description{
Build a nested case_when expression for a single node
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tidypredict-package.R
\docType{package}
\name{tidypredict-package}
\alias{tidypredict}
\alias{tidypredict-package}
\title{tidypredict: Run Predictions Inside the Database}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

It parses a fitted 'R' model object, and returns a formula in 'Tidy Eval' code that calculates the predictions. It works with several databases back-ends because it leverages 'dplyr' and 'dbplyr' for the final 'SQL' translation of the algorithm. It currently supports lm(), glm(), randomForest(), ranger(), rpart(), earth(), xgb.Booster.complete(), lgb.Booster(), catboost.Model(), cubist(), and ctree() models.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://tidypredict.tidymodels.org}
  \item \url{https://github.com/tidymodels/tidypredict}
  \item Report bugs at \url{https://github.com/tidymodels/tidypredict/issues}
}

}
\author{
\strong{Maintainer}: Emil Hvitfeldt \email{emil.hvitfeldt@posit.co}

Authors:
\itemize{
  \item Edgar Ruiz \email{edgar@posit.co}
  \item Max Kuhn \email{max@posit.co}
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model-lightgbm.R
\name{.extract_lgb_trees}
\alias{.extract_lgb_trees}
\title{Extract processed LightGBM trees}
\usage{
.extract_lgb_trees(model)
}
\arguments{
\item{model}{A LightGBM model object}
}
\description{
For use in orbital package.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model-ranger.R
\name{.extract_ranger_trees}
\alias{.extract_ranger_trees}
\title{Extract regression trees for ranger models}
\usage{
.extract_ranger_trees(model)
}
\arguments{
\item{model}{A ranger model object (regression)}
}
\description{
For use in orbital package.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parsemodel.R
\name{parse_model}
\alias{parse_model}
\title{Converts an R model object into a parsed model}
\usage{
parse_model(model)
}
\arguments{
\item{model}{An R model object.}
}
\value{
A parsed model object with class \code{parsed_model} and a model-specific
subclass (e.g., \code{pm_xgb}, \code{pm_tree}, \code{pm_regression}). The object contains:
\itemize{
\item \verb{$general}: List with model metadata including \code{model} (model type),
\code{type} (used for S3 dispatch), \code{version} (parsed model format version),
and model-specific parameters.
\item Model-specific fields containing coefficients, tree structures, etc.
}
}
\description{
Parses a fitted R model's structure and extracts the components needed to
create a dplyr formula for prediction. The parsed model can be serialized
(e.g., saved to YAML) and later used to generate predictions without the
original model object.
}
\section{Parsed model versions}{


The \verb{$general$version} field indicates the parsed model format:
\itemize{
\item \strong{Version 1}: Original format. Linear models store coefficients in a
data frame. Tree models use flat \code{case_when()} expressions where all leaf
conditions are at the same level.
\item \strong{Version 2}: Improved coefficient storage for linear models (lm, earth).
Tree models still use flat \code{case_when()}.
\item \strong{Version 3}: Current format. Tree models (rpart, ranger, randomForest,
xgboost, lightgbm, catboost, partykit, cubist) use nested \code{case_when()}
expressions that mirror the tree structure. This produces more efficient
SQL and R code because conditions are evaluated hierarchically rather than
checking all leaf paths.
}

When loading a parsed model saved with an older version, tidypredict
automatically uses the appropriate formula builder for backwards
compatibility.
}

\section{Model types}{


Each parsed model has a type that determines the S3 class used for dispatch:
\itemize{
\item \code{pm_regression}: Linear models (lm, glm, earth, glmnet)
\item \code{pm_tree}: Single trees and forests (rpart, partykit, ranger, randomForest,
cubist)
\item \code{pm_xgb}: XGBoost gradient boosting models
\item \code{pm_lgb}: LightGBM gradient boosting models
\item \code{pm_catboost}: CatBoost gradient boosting models
}
}

\examples{
library(dplyr)
df <- mutate(mtcars, cyl = paste0("cyl", cyl))
model <- lm(mpg ~ wt + cyl * disp, offset = am, data = df)
parse_model(model)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tidypredict_test.R
\name{print.tidypredict_test}
\alias{print.tidypredict_test}
\title{print method for test predictions results}
\usage{
\method{print}{tidypredict_test}(x, ...)
}
\description{
print method for test predictions results
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/acceptable.R
\name{acceptable_formula}
\alias{acceptable_formula}
\title{Checks that the formula can be parsed}
\usage{
acceptable_formula(model)
}
\arguments{
\item{model}{An R model object}
}
\description{
Uses an S3 method to check that a given formula can be parsed based on its class.
It currently scans for contrasts that are not supported and in-line functions.
(e.g: lm(wt ~ as.factor(am))). Since this function is meant for function interaction,
as opposed to human interaction, a successful check is silent.
}
\examples{

model <- lm(mpg ~ wt, mtcars)
acceptable_formula(model)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree.R
\name{generate_case_when_trees}
\alias{generate_case_when_trees}
\title{Generate trees}
\usage{
generate_case_when_trees(parsedmodel, default = TRUE)
}
\description{
Each tree is generated as a flat tree with each node being a seperate part of
the case when.
This means that the following tree:
}
\details{
\if{html}{\out{<div class="sourceCode">}}\preformatted{        +-----+
   +----|x > 0|----+
   |    +-----+    |
   v               v
}\if{html}{\out{</div>}}

+------+        +--------+
+--|y < 20|--+  +--|z <= 10 |--+
|  +------+  |  |  +--------+  |
v            v  v              v
a            b  c              d

will be turned into the following \code{case_when()} statement.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{case_when(
  x >  0 & y <  20 ~ "a",
  x >  0 & y >= 20 ~ "b",
  x <= 0 & z <= 10 ~ "c",
  x <= 0 & z >  10 ~ "d"
)
}\if{html}{\out{</div>}}

instead of a nested \code{case_when()}s` like this

\if{html}{\out{<div class="sourceCode r">}}\preformatted{case_when(
  x >  0 ~ case_when(
             y <  20 ~ "a",
             y >= 10 ~ "b"
           ),
  x <= 0 ~ case_when(
             z <= 10 ~ "c",
             z >  10 ~ "d"
           )
)
}\if{html}{\out{</div>}}

The functions in this file generates these tree.
\code{generate_case_when_tree()} generates a single tree with
\code{generate_case_when_trees()} being a convinience wrapper for multiple trees.

\code{generate_tree_node()} generates the expressions for each a single ndoe in
the tree, where \code{generate_tree_nodes()} is a convinience wrapper for
calculating all notes.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree-nested.R
\name{build_nested_split_condition}
\alias{build_nested_split_condition}
\title{Build a split condition expression for nested trees (left branch)}
\usage{
build_nested_split_condition(split)
}
\arguments{
\item{split}{A split info list with col, val/vals, is_categorical}
}
\description{
Build a split condition expression for nested trees (left branch)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree.R
\name{.build_case_when_tree}
\alias{.build_case_when_tree}
\title{Build case_when expression from nodes with predictions and paths}
\usage{
.build_case_when_tree(nodes)
}
\arguments{
\item{nodes}{A list of lists, each with \code{prediction} (numeric) and \code{path}}
}
\description{
Shared helper for building tree expressions used by ranger and randomForest
classification extractors.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model-glmnet.R
\name{.build_linear_pred}
\alias{.build_linear_pred}
\title{Build linear predictor expression from coefficient names and values}
\usage{
.build_linear_pred(coef_names, coef_values)
}
\arguments{
\item{coef_names}{Character vector of coefficient names (including "(Intercept)")}

\item{coef_values}{Numeric vector of coefficient values}
}
\description{
Shared helper for building linear predictor strings from coefficients.
Used by orbital package for glmnet models.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/to_sql.R
\name{tidypredict_sql}
\alias{tidypredict_sql}
\title{Returns a SQL query with formula to calculate fitted values}
\usage{
tidypredict_sql(model, con)
}
\arguments{
\item{model}{An R model or a list with a parsed model}

\item{con}{Database connection object. It is used to select
the correct SQL translation syntax.}
}
\description{
Returns a SQL query with formula to calculate fitted values
}
\examples{
library(dbplyr)

model <- lm(mpg ~ wt + am + cyl, data = mtcars)
tidypredict_sql(model, simulate_dbi())
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict-fit.R
\name{tidypredict_fit}
\alias{tidypredict_fit}
\title{Returns a Tidy Eval formula to calculate fitted values}
\usage{
tidypredict_fit(model)
}
\arguments{
\item{model}{An R model or a list with a parsed model.}
}
\description{
It parses a model or uses an already parsed model to return a
Tidy Eval formula that can then be used inside a dplyr command.
}
\examples{

model <- lm(mpg ~ wt + cyl * disp, offset = am, data = mtcars)
tidypredict_fit(model)
}
